/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.animation;

import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.lang.reflect.Method;
import java.util.List;

import net.java.games.jogl.GLCanvas;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import fhm.edu.fem.app.AnimationStartup;
import fhm.edu.fem.calc.FEMGaussQuadrature;
import fhm.edu.fem.calc.FEMManager;
import fhm.edu.fem.calc.objects.MathObject;
import fhm.edu.fem.calc.objects.MembraneFunctions;
import fhm.edu.fem.calc.objects.Node;
import fhm.edu.fem.calc.objects.NodeList;
import fhm.edu.fem.calc.objects.Vector;
import fhm.edu.fem.calc.objects.matrix.Matrix;
import fhm.edu.fem.calc.objects.matrix.MatrixStandard;
import fhm.edu.fem.constants.ApplicationConstants;
import fhm.edu.fem.constants.GaussConstants;
import fhm.edu.fem.util.logging.LoggingManager;

/**
 * Class Description:
 *		This class is being used when we have a FEM4 files for Membran
 *		calculations. Any node designated as border node in the FEMIN File
 *		is taken out of the calculations (all other get Alias IDs)
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class CalcAnimation implements KeyListener {
	
	/* Our local logger object */
	protected static Logger localLogger = LoggingManager.getLogger(CalcAnimation.class.getName());
	
	/* Inner and Outer Points ("Stuetzstellen des inneren und aeusseren Integrals") */
	protected static int INNER_POINTS = 2;
	protected static int OUTER_POINTS = 2;
	
	// The "Auslenkungs" point
	protected double pX = -1;
	protected double pY = -1;
	protected double pZ = -1;
	
	// The required delta T
	protected double deltaT = -1;
	
	// The initial values (remember them for the reset case)
	protected double initialDeltaT = -1;
	protected double initialPZ = -1;
	
	// The current frame count in the animation
	protected int frameCount = 0;
	
	// The calculated size of membran
	protected double dimensionA = -1;
	protected double dimensionB = -1;
	
	// Where we store the calculated omega
	protected Vector omega = null;
	
	// Where we store the calculated vector w0
	protected Vector w0Vector = null;
	
	// Where we store the calculated vector a(i)
	protected Vector aiVector = null;
	
	// The starting time
	protected long startTime = 0;
	
	// To temporarily remember the time
	protected long tmpTime = 0;
	
	// To remember the FEMManager
	protected FEMManager femManager = FEMManager.getInstance();
	
	/**
	 * Basic Constructor needed for other animation classes extending from this one
	 */
	protected CalcAnimation() {}
	
    /**
     * Create a new Calc Animation Object
     * @param args, the command-line arguments
     */
    public CalcAnimation(String[] args) {
    	
	    if(args.length < 5) {
	        
			if(localLogger.isEnabledFor(Level.FATAL)) {
				localLogger.log(Level.FATAL, "Not enough arguments given !");
			}
			
	        AnimationStartup.abort(true);
	        
	    } else if(args.length > 6) {
	        
			if(localLogger.isEnabledFor(Level.FATAL)) {
				localLogger.log(Level.FATAL, "Too many arguments given !");
			}
			
	        AnimationStartup.abort(true);
	    }

        // Grab the second command-line argument - X-value
        String xString = args[1];
        
        // Convert the argument to a double value
        double x = Double.parseDouble(xString);
        
        if(x <= 0) {
        	
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "X-Value must be > 0 !");
			}
			
			// Abort the program
			AnimationStartup.abort(false);
        }
        
        this.pX = x;
        
        // Grab the third command-line argument - Y-value
        String yString = args[2];
        
        // Convert the argument to a double value
        double y = Double.parseDouble(yString);
        
        if(y <= 0) {
        	
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Y-Value must be > 0 !");
			}
			
			// Abort the program
			AnimationStartup.abort(false);
        }
        
        this.pY = y;
        
        // Grab the third command-line argument - Z-value
        String zString = args[3];
        
        // Convert the argument to a double value
        double z = Double.parseDouble(zString);
        
        this.pZ = z;
        
        // Grab the fourth command-line argument - delta-T
        String deltaTString = args[4];
        
        // Convert the argument to a double value
        double deltaT = Double.parseDouble(deltaTString);
    	
    	this.deltaT = deltaT;
    	
    	// Set additional parameters
    	this.initialDeltaT = deltaT;
    	this.initialPZ = z;
    }
    
    /**
     * Call this method to start the animation
     * @return true if all went well, false otherwise
     */
    public boolean start() {
    	
        // Remember the starting time
        this.startTime = System.currentTimeMillis();
        
        // Clear the result list of the FEMManager
        femManager.clearResultList();
        
        /**
         * CALCULATE THE DIMENSION OF THE MEMBRAN (A, B)
         */
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Calculating Membran Dimensions a, b ...");
        }
        
        // Calculate the dimensions
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.calculateMembranDimensions()) {
        	
        	if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error during the calculation of the Membran Dimension !");
            }
            
            return false;
        }
        
        // Calc the needed time
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Calculation of the Membran Dimensions successfully completed: " + this.tmpTime + " ms");
        }
        
        /**
         * CALCULATE THE VECTOR OMEGA
         */
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Calculating Vector Omega ...");
        }
        
        // Calculate the dimensions
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.calculateOmega()) {
        	
        	if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error during the calculation of Vector Omega !");
            }
            
            return false;
        }
        
        // Calc the needed time
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Calculation of Vector Omega successfully completed: " + this.tmpTime + " ms");
        }
        
        /**
         * CALCULATE THE STARTING AMPLITUDE w0
         */
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Calculating the Starting Amplitude of the Membran ...");
        }
        
        // Calculate the dimensions
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.calculateStartingAmplitude()) {
        	
        	if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error during the calculation of the starting amplitude !");
            }
            
            return false;
        }
        
        // Calc the needed time
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Calculation of the Starting Amplitude successfully completed: " + this.tmpTime + " ms");
        }
        
        /**
         * CALCULATE B-NORM OF ALL EIGENVECTORS
         */
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Calculating the b-normalization of all Eigenvectors ...");
        }
        
        // Calculate the dimensions
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.normB()) {
        	
        	if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error during the calculation of b-normalization !");
            }
            
            return false;
        }
        
        // Calc the needed time
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Calculation of the b-normalization successfully completed: " + this.tmpTime + " ms");
        }
        
        /**
         * CALCULATE a(i)
         */
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Calculating the Vector a(i) ...");
        }
        
        // Calculate the dimensions
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.calculateAI()) {
        	
        	if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error during the calculation of Vector a(i) !");
            }
            
            return false;
        }
        
        // Calc the needed time
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Calculation of the Vector a(i) successfully completed: " + this.tmpTime + " ms");
        }
        
        /**
         * CALCULATE THE EXACT a(m, n)
         */
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Calculating the exact Matrix a(m, n) ...");
        }
        
        // Calculate the dimensions
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.calculateExactA()) {
        	
        	if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error during the calculation of Matrix a(m,n) !");
            }
            
            return false;
        }
        
        // Calc the needed time
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Calculation of the a(m, n) successfully completed: " + this.tmpTime + " ms");
        }
        
        /**
         * CALCULATE THE ANIMATION
         */
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Calculating the Animation ...");
        }
        
        // Calculate the dimensions
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.calculateAnimation()) {
        	
        	if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error during the calculation of the animation !");
            }
            
            return false;
        }
        
        // Calc the needed time
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Calculation of the animation successfully completed: " + this.tmpTime + " ms");
        }
    	
    	return true;
    }
    
    /**
     * Calculate the Membran Dimensions
     * @return true if all went well, false otherwise
     */
    protected boolean calculateMembranDimensions() {
    	
    	// Reset the dimensions
    	dimensionA = 0;
    	dimensionB = 0;
    	
    	// Grab the list of all nodes
    	NodeList completeNodeList = femManager.getNodeList();
    	
    	// Find the largest X- and Y-values of any node (these are the membran sizes)
    	for(int i = 1; i <= completeNodeList.size(); i++) {
    		
    		// Check the x- and y-values of the current node
    		Node currentNode = completeNodeList.getNodeInCorrectOrder(i);
    		
    		if(currentNode.getX() > dimensionA) {
    			dimensionA = currentNode.getX();
    		}
    		
    		if(currentNode.getY() > dimensionB) {
    			dimensionB = currentNode.getY();
    		}
    	}
    	
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Membran Dimensions (a/b): " + dimensionA + " / " + dimensionB);
        }
		
    	return true;
    }
    
    /**
     * Calculate the Vector Omega
     * @return true if all went well, false otherwise
     */
    protected boolean calculateOmega() {
    	
    	// Grab the vector of eigenvalues
    	Vector eigenvalueVector = femManager.getEigenvalueVector();
    	
    	// Create the vector Omega
    	this.omega = new Vector("Omega", eigenvalueVector.size());
    	
    	// Run through all eigenvalues
    	for(int i = 1; i <= eigenvalueVector.size(); i++) {
    		
    		// Calculate : omega(i) = lambda(i) * sigma / g
    		double result = eigenvalueVector.getValue(i) * ApplicationConstants.ANIMATION_MEMBRAN_MATERIAL_CONSTANT_SIGMA / ApplicationConstants.ANIMATION_MEMBRAN_MATERIAL_CONSTANT_G;
    		
    		result = Math.sqrt(result);
    		this.omega.setValue(i, result);
    	}
		
    	return true;
    }
    
    /**
     * Calculate the Starting Amplitude
     * @return true if all went well, false otherwise
     */
    protected boolean calculateStartingAmplitude() {
    	    	
    	// The list of all nodes with an alias ID
    	NodeList aliasIDNodeList = femManager.getAliasIDNodeList();
    	
    	// Vector holding the w0 for all nodes (all nodes which have an alias ID)
    	Vector tmpW0Vector = new Vector("w0", aliasIDNodeList.size());
    	
    	// Store the values in the static methods class so we can use the w0 function there
    	// It is handeled that way because we need the same function for w0 later on in the double integral
		MembraneFunctions.storeMembraneValues(pX, pY, pZ, dimensionA, dimensionB);
    	
    	// Run through all alias ID Node and calculate each w0 Value
    	for(int i = 1; i <= aliasIDNodeList.size(); i++) {
    		
    		Node currentNode = aliasIDNodeList.getNodeInCorrectOrder(i);
    		double nodeX = currentNode.getX();
    		double nodeY = currentNode.getY();
    	
    		// Calculate w0 of the current point by calling the proper function
    		double w0Value = MembraneFunctions.w0(new Double(nodeX), new Double(nodeY), new Integer(0));
    		
    		// Save the calculated value in the w0 vector
    		tmpW0Vector.setValue(i, w0Value);
    	}
    	
    	// Set the new w0Vector
    	this.w0Vector = tmpW0Vector;
    	
    	// Display the maximum value in w0
    	double maxValue = Double.MIN_VALUE;
    	
    	for(int i = 1; i <= this.w0Vector.size(); i++) {
    		
    		if(w0Vector.getValue(i) > maxValue) {
    			maxValue = w0Vector.getValue(i);
    		}
    	}
    	
    	if(localLogger.isEnabledFor(Level.INFO)) {
			localLogger.log(Level.INFO, "Maximum Value in w0 (Starting Amplitude) is " + maxValue);
		}
    	
    	return true;
    }
    
    /**
     * B-normalize all eigenvectors
     * @return true if all went well, false otherwise
     */
    protected boolean normB() {
    	
    	// B-Normalize all eigenvectors
    	List<Vector> resultListSmall = femManager.getResultListSmall();
    	
    	// Run through all results
    	for(int resultIndex = 0; resultIndex < resultListSmall.size(); resultIndex++) {
    		
    		// Grab the current result
    		Vector currentResult = resultListSmall.get(resultIndex);
    		
//    		double sum = 0;
//    		
//    		for(int i = 1; i <= currentResult.size(); i++) {
//    			
//    			sum += Math.pow(currentResult.getValue(i),2 );
//    		}
//    		
//    		sum = Math.sqrt(sum);
//    		
//    		for(int i = 1; i <= currentResult.size(); i++) {
//    			
//    			currentResult.setValue(i, (currentResult.getValue(i) / sum));
//    		}
//    		
//        	// Calculate the length of the current eigenvector
//    		double lengthCheck = 0;
//    		for(int i = 1; i <= currentResult.size(); i++) {
//    			
//    			lengthCheck += Math.pow(currentResult.getValue(i), 2);
//    		}
//    		lengthCheck = Math.sqrt(lengthCheck);
//    		
//            if(localLogger.isEnabledFor(Level.INFO)) {
//                localLogger.log(Level.INFO, "Eigenvector #" + resultIndex + " b-ortonormalized. New vector length: " + lengthCheck);
//            }
    	}
    	
    	/**
    	 * Norm u(i)^T * M * u(i)  so that the result = 1
    	 */
    	
    	// Grab the Complete Mass Matrix M
    	Matrix completeMassMatrix = femManager.getCompleteMassMatrix();
    	
    	// Run through all results u
    	for(int resultIndex = 0; resultIndex < resultListSmall.size(); resultIndex++) {
    		
    		// Grab the current result u
    		Vector currentResult = resultListSmall.get(resultIndex);
    		
    		// Calculate M * u
    		Vector partOfResult = completeMassMatrix.mult(currentResult);
    		
    		// Calculate u^T * partOfResult
			double result = 0;
    		
			for(int row = 1; row <= currentResult.size(); row++) {
				
				result += currentResult.getValue(row) * partOfResult.getValue(row);
			}

			// The result is the factor which needs to be applied to the vector to norm it
			currentResult = currentResult.mult(1/Math.sqrt(result));
			
			// Recalculate to show the user the length
			partOfResult = completeMassMatrix.mult(currentResult);
			
			result = 0;
    		
			for(int row = 1; row <= currentResult.size(); row++) {
				
				result += currentResult.getValue(row) * partOfResult.getValue(row);
			}
		
    		// Give an output
			if(localLogger.isEnabledFor(Level.INFO)) {
				localLogger.log(Level.INFO, "Eigenvector #" + resultIndex + " b-ortonormalized. Result (needs to be 1): " + result);
			}
    	}
    	
    	/**
    	 * Test for u(k)^T * M * u(i) = 0 for i != k and 1 for i == k
    	 */
    	
//    	// Run through all results u
//    	for(int resultIndex1 = 0; resultIndex1 < resultListSmall.size(); resultIndex1++) {
//    		
//    		// Grab the current result u 1
//    		Vector currentResult1 = resultListSmall.get(resultIndex1);
//    		
//    		for(int resultIndex2 = 0; resultIndex2 < resultListSmall.size(); resultIndex2++) {
//    			
//        		// Grab the current result u
//        		Vector currentResult2 = resultListSmall.get(resultIndex2);
//    		
//				double result = 0;
//	    		
//	    		for(int row = 1; row <= completeMassMatrix.size(); row++) {    			
//	    			for(int col = 1; col <= completeMassMatrix.size(); col++) {
//	    				
//	    				result += currentResult1.getValue(col) * completeMassMatrix.getValue(row, col) * currentResult2.getValue(row);
//	    			}
//	    		}
//    		
//	    		// Give an output
//	    		System.err.println("Test: For (i/k): " + resultIndex1 + " / " + resultIndex2 + " the Result is: " + result);
//    		}
//    	}
    	
    	/**
    	 * End of Test
    	 */
    	
    	return true;
    }
    
    /**
     * Calculate the Vector a(i)
     * @return true if all went well, false otherwise
     */
    protected boolean calculateAI() {
    	
    	// Grab the small u-Vectors (without zeros for non-alias-ID-nodes) from the FEMManager
    	List<Vector> resultListSmall = femManager.getResultListSmall();
    	
    	Vector tmpAiVector = new Vector("a(i)", resultListSmall.size());
    	
    	// Grab the complete-Mass-Matrix from the FEMManager
    	Matrix completeMassMatrix = femManager.getCompleteMassMatrix();

    	// Calculate M * w0Vector once
    	Vector partOfResult = completeMassMatrix.mult(w0Vector);
    	
    	// Run through all results u
    	for(int resultIndex = 0; resultIndex < resultListSmall.size(); resultIndex++) {
    		
    		// Grab the current result u
    		Vector currentResult = resultListSmall.get(resultIndex);
    		
    		// Calculate u^T * partOfResult
			double result = 0;
    		
			for(int row = 1; row <= currentResult.size(); row++) {
				
				result += currentResult.getValue(row) * partOfResult.getValue(row);
			}
    		
    		tmpAiVector.setValue(resultIndex+1, result);
    	}  	
    	
    	// Set the new a(i) vector
    	this.aiVector = tmpAiVector;
    	
    	return true;
    }
    
    /**
     * Calculate the exact Matrix a(m,n) via formula: 4/(ab) double-integral { w(x,y,0) sin(m pi x/a) sin(n pi y/b) dx dy
     * @return true if all went well, false otherwise
     */
    protected boolean calculateExactA() {
    	
    	try {
    		
    		// Store the values needed to calculate some functions in the static function class
			MembraneFunctions.storeMembraneValues(pX, pY, pZ, dimensionA, dimensionB);
			
			/**
			 * SINUS PARTS OF THE INTERVAL
			 */
			
			// The integral input variables
			Method h1 = MembraneFunctions.getMethod(MembraneFunctions.SIN_M_X_A);
			Method h2 = MembraneFunctions.getMethod(MembraneFunctions.SIN_N_Y_B);
			Method q = MembraneFunctions.getMethod(MembraneFunctions.W0_X_Y);
			int m = 4;
			int n = 4;
			Vector ax = GaussConstants.get(INNER_POINTS, GaussConstants.VALUE_A);
			Vector hx = GaussConstants.get(INNER_POINTS, GaussConstants.VALUE_H);
			Vector ay = GaussConstants.get(OUTER_POINTS, GaussConstants.VALUE_A);
			Vector hy = GaussConstants.get(OUTER_POINTS, GaussConstants.VALUE_H);
			
			// The full result matrix
			Matrix fullResult = new MatrixStandard("Full Result", m);
			
			// Separate the integration distance for both integrals in smaller steps
			int amountOfIntegralSteps = 16;
			
			for (int nStepIndex = 0; nStepIndex < amountOfIntegralSteps; nStepIndex++) {

				// Calculate the current lower and upper border of the inner integral
				double bLower = nStepIndex * dimensionB / amountOfIntegralSteps;
				double bUpper = (nStepIndex + 1) * dimensionB / amountOfIntegralSteps;

				for (int mStepIndex = 0; mStepIndex < amountOfIntegralSteps; mStepIndex++) {

					// Calculate the current lower and upper border of the outer integral
					double aLower = mStepIndex * dimensionA / amountOfIntegralSteps;
					double aUpper = (mStepIndex + 1) * dimensionA / amountOfIntegralSteps;
					
					// The result object
					MathObject integralResultObject = null;

					// Calculate the double integral
					integralResultObject = FEMGaussQuadrature.calcDoubleIntegration(bLower, bUpper, aLower,	aUpper, h1, h2, q, m, n, ax, hx, ay, hy);

					// The result object should be a matrix, cast it
					Matrix result = null;
					if (integralResultObject instanceof Matrix) {

						result = (Matrix) integralResultObject;
						fullResult = fullResult.add(result);

					} else {
						
				        if(localLogger.isEnabledFor(Level.ERROR)) {
				            localLogger.log(Level.ERROR, "Eror while calculating a(m,n) - Integral Result Object is not a Matrix !");
				        }
					}
				}
			}
			
			// The factor applied to the result
			double factor = (4 / (dimensionA * dimensionB));
			fullResult = fullResult.mult(factor);
			
	        if(localLogger.isEnabledFor(Level.DEBUG)) {
	            localLogger.log(Level.DEBUG, "Vector a(i): " + this.aiVector.toString());
	        }
			
	        // Check if we have a quadratic membrane
	        if(this.dimensionA == this.dimensionB) {
	        	
		        /**
		         * QUADRATIC MEMBRANE
		         */
	        	
				// Display the absolute difference to the a(i) vector we calculated before
		        if(localLogger.isEnabledFor(Level.INFO)) {
		            localLogger.log(Level.INFO, "Quadratic Membrane absolute difference to Vector a(i):");
		            localLogger.log(Level.INFO, "	a(1) to A(1,1) => " + this.aiVector.getValue(1) + " to " + fullResult.getValue(1,1) + " = " + Math.abs(this.aiVector.getValue(1) - fullResult.getValue(1,1)));
		            localLogger.log(Level.INFO, "	a(2) to A(1,2) => " + this.aiVector.getValue(2) + " to " + fullResult.getValue(1,2) + " = " + Math.abs(this.aiVector.getValue(2) - fullResult.getValue(1,2)));
		            localLogger.log(Level.INFO, "	a(3) to A(2,1) => " + this.aiVector.getValue(3) + " to " + fullResult.getValue(2,1) + " = " + Math.abs(this.aiVector.getValue(3) - fullResult.getValue(2,1)));
		            localLogger.log(Level.INFO, "	a(4) to A(2,2) => " + this.aiVector.getValue(4) + " to " + fullResult.getValue(2,2) + " = " + Math.abs(this.aiVector.getValue(4) - fullResult.getValue(2,2)));
		            localLogger.log(Level.INFO, "	a(5) to A(1,3) => " + this.aiVector.getValue(5) + " to " + fullResult.getValue(1,3) + " = " + Math.abs(this.aiVector.getValue(5) - fullResult.getValue(1,3)));
		            localLogger.log(Level.INFO, "	a(6) to A(3,1) => " + this.aiVector.getValue(6) + " to " + fullResult.getValue(3,1) + " = " + Math.abs(this.aiVector.getValue(6) - fullResult.getValue(3,1)));
		            localLogger.log(Level.INFO, "	a(7) to A(2,3) => " + this.aiVector.getValue(7) + " to " + fullResult.getValue(2,3) + " = " + Math.abs(this.aiVector.getValue(7) - fullResult.getValue(2,3)));
		            localLogger.log(Level.INFO, "	a(8) to A(3,2) => " + this.aiVector.getValue(8) + " to " + fullResult.getValue(3,2) + " = " + Math.abs(this.aiVector.getValue(8) - fullResult.getValue(3,2)));
		        }
		        
		        if(localLogger.isEnabledFor(Level.INFO)) {
		            localLogger.log(Level.INFO, "Relation of errors:");
		            localLogger.log(Level.INFO, "	a(1) / A(1,1) => " + Math.abs(this.aiVector.getValue(1) / fullResult.getValue(1,1)));
		            localLogger.log(Level.INFO, "	a(2) / A(1,2) => " + Math.abs(this.aiVector.getValue(2) / fullResult.getValue(1,2)));
		            localLogger.log(Level.INFO, "	a(3) / A(2,1) => " + Math.abs(this.aiVector.getValue(3) / fullResult.getValue(2,1)));
		            localLogger.log(Level.INFO, "	a(4) / A(2,2) => " + Math.abs(this.aiVector.getValue(4) / fullResult.getValue(2,2)));
		            localLogger.log(Level.INFO, "	a(5) / A(1,3) => " + Math.abs(this.aiVector.getValue(5) / fullResult.getValue(1,3)));
		            localLogger.log(Level.INFO, "	a(6) / A(3,1) => " + Math.abs(this.aiVector.getValue(6) / fullResult.getValue(3,1)));
		            localLogger.log(Level.INFO, "	a(7) / A(2,3) => " + Math.abs(this.aiVector.getValue(7) / fullResult.getValue(2,3)));
		            localLogger.log(Level.INFO, "	a(8) / A(3,2) => " + Math.abs(this.aiVector.getValue(8) / fullResult.getValue(3,2)));
		        }
		        
				// REPLACE A(I) with the CALCULATED VALUES HERE
//				this.aiVector.setValue(1, fullResult.getValue(1,1));
//				this.aiVector.setValue(2, fullResult.getValue(1,2));
//				this.aiVector.setValue(3, fullResult.getValue(2,1));
//				this.aiVector.setValue(4, fullResult.getValue(2,2));
//				this.aiVector.setValue(5, fullResult.getValue(1,3));
//				this.aiVector.setValue(6, fullResult.getValue(3,1));
//				this.aiVector.setValue(7, fullResult.getValue(2,3));
//				this.aiVector.setValue(8, fullResult.getValue(3,2));
	        	
	        } else {
	        
		        /**
		         * RECTANGULAR MEMBRANE
		         */
		        
				// Display the absolute difference to the a(i) vector we calculated before
		        if(localLogger.isEnabledFor(Level.INFO)) {
		            localLogger.log(Level.INFO, "Rectangular Membrane absolute difference to Vector a(i):");
		            localLogger.log(Level.INFO, "	a(1) / A(1,1) => " + this.aiVector.getValue(1) + " to " + fullResult.getValue(1,1) + " = " + Math.abs(this.aiVector.getValue(1) - fullResult.getValue(1,1)));
		            localLogger.log(Level.INFO, "	a(2) / A(1,2) => " + this.aiVector.getValue(2) + " to " + fullResult.getValue(1,2) + " = " + Math.abs(this.aiVector.getValue(2) - fullResult.getValue(1,2)));
		            localLogger.log(Level.INFO, "	a(3) / A(2,1) => " + this.aiVector.getValue(3) + " to " + fullResult.getValue(2,1) + " = " + Math.abs(this.aiVector.getValue(3) - fullResult.getValue(2,1)));
		            localLogger.log(Level.INFO, "	a(4) / A(1,3) => " + this.aiVector.getValue(4) + " to " + fullResult.getValue(1,3) + " = " + Math.abs(this.aiVector.getValue(4) - fullResult.getValue(1,3)));
		            localLogger.log(Level.INFO, "	a(5) / A(2,2) => " + this.aiVector.getValue(5) + " to " + fullResult.getValue(2,2) + " = " + Math.abs(this.aiVector.getValue(5) - fullResult.getValue(2,2)));
		            localLogger.log(Level.INFO, "	a(6) / A(2,3) => " + this.aiVector.getValue(6) + " to " + fullResult.getValue(2,3) + " = " + Math.abs(this.aiVector.getValue(6) - fullResult.getValue(2,3)));
		            localLogger.log(Level.INFO, "	a(7) / A(1,4) => " + this.aiVector.getValue(7) + " to " + fullResult.getValue(1,4) + " = " + Math.abs(this.aiVector.getValue(7) - fullResult.getValue(1,4)));
		            localLogger.log(Level.INFO, "	a(8) / A(3,1) => " + this.aiVector.getValue(8) + " to " + fullResult.getValue(3,1) + " = " + Math.abs(this.aiVector.getValue(8) - fullResult.getValue(3,1)));
		        }
		        
		        if(localLogger.isEnabledFor(Level.INFO)) {
		            localLogger.log(Level.INFO, "Relation of errors:");
		            localLogger.log(Level.INFO, "	a(1) to A(1,1) => " + Math.abs(this.aiVector.getValue(1) / fullResult.getValue(1,1)));
		            localLogger.log(Level.INFO, "	a(2) to A(1,2) => " + Math.abs(this.aiVector.getValue(2) / fullResult.getValue(1,2)));
		            localLogger.log(Level.INFO, "	a(3) to A(2,1) => " + Math.abs(this.aiVector.getValue(3) / fullResult.getValue(2,1)));
		            localLogger.log(Level.INFO, "	a(4) to A(1,3) => " + Math.abs(this.aiVector.getValue(4) / fullResult.getValue(1,3)));
		            localLogger.log(Level.INFO, "	a(5) to A(2,2) => " + Math.abs(this.aiVector.getValue(5) / fullResult.getValue(2,2)));
		            localLogger.log(Level.INFO, "	a(6) to A(2,3) => " + Math.abs(this.aiVector.getValue(6) / fullResult.getValue(2,3)));
		            localLogger.log(Level.INFO, "	a(7) to A(1,4) => " + Math.abs(this.aiVector.getValue(7) / fullResult.getValue(1,4)));
		            localLogger.log(Level.INFO, "	a(8) to A(3,1) => " + Math.abs(this.aiVector.getValue(8) / fullResult.getValue(3,1)));
		        }
		        
				// REPLACE A(I) with the CALCULATED VALUES HERE
//				this.aiVector.setValue(1, fullResult.getValue(1,1));
//				this.aiVector.setValue(2, fullResult.getValue(1,2));
//				this.aiVector.setValue(3, fullResult.getValue(2,1));
//				this.aiVector.setValue(4, fullResult.getValue(1,3));
//				this.aiVector.setValue(5, fullResult.getValue(2,2));
//				this.aiVector.setValue(6, fullResult.getValue(2,3));
//				this.aiVector.setValue(7, fullResult.getValue(1,4));
//				this.aiVector.setValue(8, fullResult.getValue(3,1));
	        
	        }

			return true;

		} catch (NoSuchMethodException e) {
			
	        if(localLogger.isEnabledFor(Level.ERROR)) {
	            localLogger.log(Level.ERROR, "Eror while calculating a(m,n)", e);
	        }
	        
	        return false;
		}
    }
    
    /**
     * Calculate the Animation
     * @return true if all went well, false otherwise
     */
    protected boolean calculateAnimation() {
    	
    	/**
    	 * OPEN THE FRAME WITH THE W0
    	 */       
        // Create the Membrane 3D Frame
        GLMembraneFrame frame = new GLMembraneFrame("FEM - Membrane 3D - Java OpenGL", this.dimensionA, this.dimensionB, this.w0Vector);
        
        // Grab the GLCanvas from the frame and add us as keylistener to it
        GLCanvas canvas = (GLCanvas)(frame.getComponent(0));
        canvas.addKeyListener(this);
        
        // Also add us to the frame as keylistener (needed due to a bug which switches the key events from the frame to the GLCanvas after doing a rotation with the mouse)
        frame.addKeyListener(this);
    	
		/**
		 * CALCULATE THE CURRENT W(T) AND DISPLAY IT
		 */		
		
    	// Grab the small u-Vectors (without zeros for non-alias-ID-nodes) from the FEMManager
    	List<Vector> resultListSmall = femManager.getResultListSmall();
    	
    	// The list of all nodes with an alias ID
    	NodeList aliasIDNodeList = femManager.getAliasIDNodeList();

    	// Grab the vector of all eigenvalues
    	Vector eigenvalueVector = femManager.getEigenvalueVector();

    	// Run forever
    	for(frameCount = 0; frameCount <= Integer.MAX_VALUE; frameCount++) {
    		
        	// The vector w(t)
        	Vector wT = new Vector("w(t)", aliasIDNodeList.size());
    		
    		// Run through all aliasID nodes
    		for(int j = 1; j <= aliasIDNodeList.size(); j++) {
    			
    			double sum = 0;
    			
    			// Calculate: w(t) = sum[i=1 to p(Amount of Eigenvalues)](a(i) * u(i) * cos(omega(i) * t)
    			for(int i = 1; i <= eigenvalueVector.size(); i++) {
    				
    				// Grab the eigenvector for the current eigenvalue
    				Vector eigenvector = resultListSmall.get(i-1);
    				
    				sum += aiVector.getValue(i) * eigenvector.getValue(j) * Math.cos(this.omega.getValue(i) * (frameCount * deltaT));
    			}
    			
    			// Set the calculated w(t) value
    			wT.setValue(j, sum);
    		}
    		
    		// TEST: Check if the new wT Vector holds any value which is bigger than the starting amplitude
    		for(int i = 1; i <= wT.size(); i++) {

    			if(pZ < wT.getValue(i)) {
    				if(localLogger.isEnabledFor(Level.ERROR)) {
			            localLogger.log(Level.ERROR, "Error while calculating the animation. New calculated z is bigger than the starting amplitude: " + wT.getValue(i));
			        }
    			}
    		}
    		
    		// We calculated the new Vector w(t), tell the membrane frame about it
    		frame.setNewVectorW(wT);
    		
    		try {
				// Wait a little bit for the renderer to display the changes
				Thread.sleep(40);

			} catch (InterruptedException e) {}
    	}
    	
    	return true;
    }
    
    /**
     * METHODS NEEDED BY THE KEYLISTENER INTERFACE
     */
    
	public void keyPressed(KeyEvent e) {}
	public void keyReleased(KeyEvent e) {}
	
	/**
	 * Automatically called when the user presses a key
	 * @param e, the key event
	 */
	public void keyTyped(KeyEvent e) {
		
		// Grab the pressed key from the event
		char pressedKey = e.getKeyChar();
		
		// Boolean to remember if we need to recalculate a(i) and therefore also w(0)
		boolean recalculate = false;
		
		// Check if the user pressed a key we support
		switch(pressedKey) {
		
			case 'r':
				// Reset everything to the initial values
		        if(localLogger.isEnabledFor(Level.INFO)) {
		            localLogger.log(Level.INFO, "UserInput - Resetting Membran Amplitude and Delta-T to initial value");
		        }
		        
		        this.deltaT = this.initialDeltaT;
		        this.pZ = this.initialPZ;
		        recalculate = true;
		        
			break;
			
			case 'q':
				// Increase the membran height
		        if(localLogger.isEnabledFor(Level.INFO)) {
		            localLogger.log(Level.INFO, "UserInput - Increasing Membran Amplitude");
		        }
		        
		        this.pZ = this.pZ * 2;
		        recalculate = true;
		        
			break;
			
			case 'a':
				// Decrease the membran height
		        if(localLogger.isEnabledFor(Level.INFO)) {
		            localLogger.log(Level.INFO, "UserInput - Decreasing Membran Amplitude");
		        }
		        
		        this.pZ = this.pZ / 2;
		        recalculate = true;
		        
			break;
			
			case '-':
				// Use a smaller delta T
		        if(localLogger.isEnabledFor(Level.INFO)) {
		            localLogger.log(Level.INFO, "UserInput - Decreasing delta-T for smaller Steps");
		        }
		        
		        this.deltaT = this.deltaT/2;
		        this.frameCount = this.frameCount * 2;
		        
			break;
			
			case '+':
				// Use a greater delta T
		        if(localLogger.isEnabledFor(Level.INFO)) {
		            localLogger.log(Level.INFO, "UserInput - Increasing delta-T for bigger Steps");
		        }
		        
		        this.deltaT = this.deltaT * 2;
		        this.frameCount = this.frameCount / 2;
		        
			break;
		}

		// Check if we need to recalculate a(i) and w(0)
		if(recalculate) {
			
	        if(localLogger.isEnabledFor(Level.INFO)) {
	            localLogger.log(Level.INFO, "Recalculating a(i) and w(0) due to User Input");
	        }
			
	        long recalcStartTime = System.currentTimeMillis();
	        
			this.calculateStartingAmplitude();
			this.calculateAI();
			
	        if(localLogger.isEnabledFor(Level.INFO)) {
	            localLogger.log(Level.INFO, "Recalculation done. Needed time: " + (System.currentTimeMillis() - recalcStartTime) + " ms");
	        }
		}
	}
}