/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.animation;

import java.awt.event.KeyEvent;

import net.java.games.jogl.GLCanvas;

import org.apache.log4j.Level;

import fhm.edu.fem.app.AnimationStartup;
import fhm.edu.fem.calc.FEMCalcSpanner;
import fhm.edu.fem.calc.objects.Force;
import fhm.edu.fem.calc.objects.ForceList;
import fhm.edu.fem.calc.objects.Node;
import fhm.edu.fem.calc.objects.NodeList;

/**
 * Class Description:
 *		This class is being used when we have FEM6 files for Spanner
 *		calculations. It animates the Spanner and offers the user the ability
 *		to increase or decrease the used force.
 * 
 * @author Thomas Woellert (thomas.woellert@t-online.de)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class CalcAnimationSpanner extends CalcAnimation {

	/* The calculation object to use */
	protected FEMCalcSpanner femCalc = null;
	
	/* The frame we use */
	protected GLSpannerFrame femFrame = null;
	
	/* The steps used when increasing or decreasing the forces due to user input 
	 * Needs to be negative for better usability. If negative x-y-vector forces are applied the
	 * spanner bows downwards. So with + we push down and with - up. */
	protected double forceUserStep = -1.5;
	
    /**
     * Create a new Calc Animation Object
     * @param femCalc, the calculation object to use
     * @param args, the given command line arguments (none needed)
     */
    public CalcAnimationSpanner(FEMCalcSpanner femCalc, String[] args) {
    	
    	// Call the super constructor
    	super();
    	
	    if(args.length < 1) {
	        
			if(localLogger.isEnabledFor(Level.FATAL)) {
				localLogger.log(Level.FATAL, "Not enough arguments given !");
			}
			
	        AnimationStartup.abort(true);
	        
	    } else if(args.length > 1) {
	        
			if(localLogger.isEnabledFor(Level.FATAL)) {
				localLogger.log(Level.FATAL, "Too many arguments given !");
			}
			
	        AnimationStartup.abort(true);
	    }
    	
    	this.femCalc = femCalc;
    }
    
    /**
     * Call this method to start the animation
     * @return true if all went well, false otherwise
     */
    public boolean start() {
    	
    	/**
    	 * INITIALIZE THE DISPLAY
    	 */
    	
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Initializing the Display ...");
        }
        
        // Initialize the Display
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.initDisplay()) {
        	
        	if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error during the initialization of the Display !");
            }
            
            return false;
        }
        
        /**
         * MAKE THE FIRST CALCULATION
         */
        
	    // Start the calculation
	    if(!femCalc.start()) {
	        
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error while running the calculation !");
            }
            
            AnimationStartup.abort(false);
	    }
    	
    	return true;
    }
    
    /**
     * Initialize the display
     * @return
     */
    protected boolean initDisplay() {
    	
        // Create the Spanner 3D Frame
        femFrame = new GLSpannerFrame("FEM - Spanner 3D - Java OpenGL", femCalc);
        
        // Grab the GLCanvas from the frame and add us as keylistener to it
        GLCanvas canvas = (GLCanvas)(femFrame.getComponent(0));
        canvas.addKeyListener(this);
        
        // Also add us to the frame as keylistener (needed due to a bug which switches the key events from the frame to the GLCanvas after doing a rotation with the mouse)
        femFrame.addKeyListener(this);
        
        return true;
    }
    
	/**
	 * Automatically called when the user presses a key
	 * @param e, the key event
	 */
	public void keyTyped(KeyEvent e) {
		
		// Grab the pressed key from the event
		char pressedKey = e.getKeyChar();
		
		// Boolean to remember if we need to recalculate
		boolean recalculate = false;
		
		// Check if the user pressed a key we support
		switch(pressedKey) {
		
		case '-':
			// Use a lesser force

	        // Adjust the changed forces in the FEMManager so that the FEMCalcSpanner gets the new ones	        
	        if(localLogger.isEnabledFor(Level.INFO)) {
	            localLogger.log(Level.INFO, "UserInput - Decreasing Forces. Adding variable: " + this.forceUserStep);
	        }
	        
	        addVariableToForces(-1.0 * this.forceUserStep);
	        
	        // Reset the x- and y-coordinates of the real basic figure to the old values so that the changed forces take correct effect
	        resetNodePositions();
	        
	        recalculate = true;
	        
		break;
		
		case '+':
			// Use a greater force
	        
	        // Adjust the changed forces in the FEMManager so that the FEMCalcSpanner gets the new ones	        
	        if(localLogger.isEnabledFor(Level.INFO)) {
	            localLogger.log(Level.INFO, "UserInput - Increasing Forces. Adding variable: " + this.forceUserStep);
	        }
	        
	        addVariableToForces(this.forceUserStep);
	        
	        // Reset the x- and y-coordinates of the real basic figure to the old values so that the changed forces take correct effect
	        resetNodePositions();
	        
	        recalculate = true;
	        
		break;
		}
		
		// Check if we need to recalculate
		if(recalculate) {
			
	        if(localLogger.isEnabledFor(Level.INFO)) {
	            localLogger.log(Level.INFO, "Re-calculating due to User Input");
	        }  
	        
	        // We do not need to start the whole calculation over because the Complete-Stiffness-Matrix won't change when the Forces are changed
	        if(!femCalc.restartAfterForceChange()) {
	        	
	        	if(localLogger.isEnabledFor(Level.ERROR)) {
	                localLogger.log(Level.ERROR, "Error during Re-calculation !");
	            }
	            
	            return;
	        }
		}
	}
	
	/**
	 * Adjust the forces stored in the FEMManager by the given factor
	 * @param factor, the factor to add to the forces (OLD_FORCE+FACTOR = NEW_FORCE)
	 */
	protected void addVariableToForces(double factor) {

		// Grab the force list from the FEMManager
		ForceList forceList = femManager.getSpannerForceList();
		
		// Run through all forces in the list
		for(int i = 1; i <= forceList.size(); i++) {
			
			Force currentForce = forceList.getForce(i);
			
			currentForce.setX(currentForce.getX() + factor);
			currentForce.setY(currentForce.getY() + factor);
			
	        if(localLogger.isEnabledFor(Level.INFO)) {
	        	localLogger.log(Level.INFO, "Changed Force #" + i + " to (x/y): " + currentForce.getX() + " / " + currentForce.getY());
	        }  
		}
		
		// Set the new force list
		femManager.setSpannerForceList(forceList);
	}
	
	/**
	 * Reset the node positions to the initial old figure, so that the changed forces
	 * can take correct effect.
	 */
	protected void resetNodePositions() {
		
		// Run through the node list
		NodeList nodeList = femManager.getNodeList();
		
		for(int i = 1; i <= nodeList.size(); i++) {
			
			Node currentNode = nodeList.getNodeInCorrectOrder(i);
			
			// Reset the position of the new node to the original coordinates
			currentNode.setX(currentNode.getXOriginal());
			currentNode.setY(currentNode.getYOriginal());
		}
		
		femManager.setNodeList(nodeList);
	}
}
