/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.animation;

import java.awt.Dimension;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;

import fhm.edu.fem.calc.FEMManager;
import fhm.edu.fem.calc.objects.Node;
import fhm.edu.fem.calc.objects.NodeList;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLDrawable;
import net.java.games.jogl.GLEventListener;

/**
 * Class Description:
 *		This class is the base class for all renderers. It offers an abstract method
 *		which is automatically called by this class during rendering, where the extended classes
 *		can put their render-stuff.
 * 
 * @author Thomas Woellert (thomas.woellert@t-online.de)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
abstract public class GLFEMRenderer implements GLEventListener, MouseListener, MouseMotionListener, KeyListener {

	// Store the rotation variables
	protected float viewRotX = 0.0f;
	protected float viewRotY = 0.0f;
	protected float viewRotZ = 0.0f;

	// The current range of the camera from the object
	protected float nRange = 15.0f;

	// Remember the previous x-/y-values of the event when we pressed the mouse (used for movement and rotation)
	protected int prevMouseX;
	protected int prevMouseY;
	
	// The calculated x-/y-Translation to make sure that the membrane is drawn in the middle of the coordinate system
	protected double transX = 0.0;
	protected double transY = 0.0;

	// Remember when the user holds the right/left mouse button (used for movement)
	protected boolean mouseRButtonDown = false;
	protected boolean mouseLButtonDown = false;
	
	// To remember if we need to update the view due to user input
	protected boolean viewChangedByUser = false;
	
	// To remember the view's x/y/width/height values
	protected int viewX;
	protected int viewY;
	protected int viewWidth;
	protected int viewHeight;
	
	// To remember if we should show only points or lines
	protected boolean showPointsOnly = false;
	
	// To remember if we should display in color mode
	protected boolean useColorMode = true;
	
	// To remember if we should draw polygons
	protected boolean showPolygons = false;
	
	// To make the FEMManager holding the Node and Element data more readily available
	protected FEMManager femManager = FEMManager.getInstance();
	
	/**
	 * Constructor of the FEM Renderer, the mid point is automatically set at the mid of the
	 * whole figure shown
	 * @param xMid, the x-value which should be moved to the middle of the screen
	 * @param yMid, the y-value which should be moved to the middle of the screen
	 */
	public GLFEMRenderer() {
		
		// Set the mid point of the graphic based on all nodes in the displayed figure
		setMidPointFromNodes();
	}
	
	/**
	 * Constructor of the FEM Renderer, manually setting the mid point of the drawing
	 * @param xMid, the x-value which should be moved to the middle of the screen
	 * @param yMid, the y-value which should be moved to the middle of the screen
	 */
	public GLFEMRenderer(double xMid, double yMid) {
		
		transX = xMid;
		transY = yMid;
	}
	
	/**
	 * The abstract draw method which needs to be implemented in all extending renderers
	 * @param drawable, the main GLDrawable given by OpenGL
	 */
	abstract protected void drawContents(GLDrawable drawable);
	
	/**
	 * Calculate and set the mid point of the graphic based on all nodes in the figure
	 */
	protected void setMidPointFromNodes() {
		
		// Find the min and max X- and Y-value of all available nodes
		double maxX = Double.MIN_VALUE;
		double minX = Double.MAX_VALUE;
		
		double maxY = Double.MIN_VALUE;
		double minY = Double.MAX_VALUE;
		
		// Grab the full node list from the femmanager
		NodeList nodeList = femManager.getNodeList();
		
		for(int i = 1; i <= nodeList.size(); i++) {
			
			Node currentNode = nodeList.getNodeInCorrectOrder(i);
			
			if(currentNode.getX() > maxX) {
				maxX = currentNode.getX();
			}
			
			if(currentNode.getX() < minX) {
				minX = currentNode.getX();
			}
			
			if(currentNode.getY() > maxY) {
				maxY = currentNode.getY();
			}
			
			if(currentNode.getY() < minY) {
				minY = currentNode.getY();
			}
		}
		
		// Set the new mid point
		double xWidth = maxX - minX;
		double yWidth = maxY - minY;
		
		transX = -(xWidth/2); 
		transY = -(yWidth/2);
	}
	
	/**
	 * Initialize
	 * This method is automatically called by OpenGL
	 * @param drawable, the main GLDrawable given by OpenGL
	 */
	public void init(GLDrawable drawable) {
		
		// Grab the main GL object we need to draw on
		GL gl = drawable.getGL();
			
		// Clear the screen
		if(useColorMode) {
			gl.glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
		} else {
			gl.glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
		}
		
		// Enable smooth shading
		gl.glShadeModel(GL.GL_SMOOTH);
		
		// Reset the projection matrix
		gl.glMatrixMode(GL.GL_PROJECTION);
		gl.glLoadIdentity();
		
		// Switch to the modelview state
		gl.glMatrixMode(GL.GL_MODELVIEW);

		// Add this renderer as mouse listener to the drawable
		drawable.addMouseListener(this);
		drawable.addMouseMotionListener(this);
	}
	
	/**
	 * This method is called when the user resizes the window
	 * @param drawable, the main GLDrawable given by OpenGL
	 * @param x, the x-value of the view
	 * @param y, the y-value of the view
	 * @param w, the width of the view
	 * @param h, the height of the view
	 */
	public void reshape(GLDrawable drawable, int x, int y, int w, int h) {
		
		// Remember these values
		this.viewX = x;
		this.viewY = y;
		this.viewWidth = w;
		this.viewHeight = h;
		
		// Grab the main GL object to draw on
		GL gl = drawable.getGL();
		
		// Clear the COLOR and the DEPTH buffer
		gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);

		// Setze den Viewport gemaess der Window-Groesse
	    gl.glViewport(0, 0, w, h);

		// Ruecksetzung des Projection matrix stack
		gl.glMatrixMode(GL.GL_PROJECTION);
		gl.glLoadIdentity();

		// Definiere das viewing volume (left, right, bottom, top, near, far)
	    if (w <= h) {
	    	
			gl.glOrtho (-nRange, nRange, -nRange*h/w, nRange*h/w, -3*nRange, 3*nRange);
	    } else {
	    	
			gl.glOrtho (-nRange*w/h, nRange*w/h, -nRange, nRange, -3*nRange, 3*nRange);
	    }
	    
	    // Reset the boolean variable
	    this.viewChangedByUser = false;
	}
	
	/**
	 * This method is automatically called in a loop to keep drawing the window
	 * @param drawable, the main GLDrawable given by OpenGL
	 */
	public void display(GLDrawable drawable) {

		// Grab the main GL object to draw on
		GL gl = drawable.getGL();
		
		// Clear the screen
		if(useColorMode) {
			gl.glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
		} else {
			gl.glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
		}
		
		// Check if we need to update the viewport due to movement done by user input
		if(viewChangedByUser) {
			
			// Setze den Viewport gemaess der Window-Groesse
		    gl.glViewport(0, 0, viewWidth, viewHeight);

			// Ruecksetzung des Projection matrix stack
			gl.glMatrixMode(GL.GL_PROJECTION);
			gl.glLoadIdentity();

			// Definiere das viewing volume (left, right, bottom, top, near, far)
		    if (viewWidth <= viewHeight) {
		    	
				gl.glOrtho (-nRange, nRange, -nRange*viewHeight/viewWidth, nRange*viewHeight/viewWidth, -nRange, nRange);
		    } else {
		    	
				gl.glOrtho (-nRange*viewWidth/viewHeight, nRange*viewWidth/viewHeight, -nRange, nRange, -nRange, nRange);
		    }
		    
		    // Reset the boolean variable
		    this.viewChangedByUser = false;
		}
		
		// Switch to model view
	    gl.glMatrixMode(GL.GL_MODELVIEW);
		
	    // Clear the color and depth buffer bit
		gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);

		// Make sure all rotations done by the user take effect
		gl.glPushMatrix();
			gl.glRotatef(viewRotX, 1.0f, 0.0f, 0.0f);
			gl.glRotatef(viewRotY, 0.0f, 1.0f, 0.0f);
			gl.glRotatef(viewRotZ, 0.0f, 0.0f, 1.0f);

			// Place and draw the coordinate system
			gl.glPushMatrix();
				gl.glTranslatef(-6.0f, -6.0f, 0.0f);
				gl.glScalef(1.5f, 1.5f, 1.5f);
				this.drawCoordSystem(drawable);
				gl.glPopMatrix();
		
				// Place and draw the membrane based on the current w(t)
				gl.glPushMatrix();
					gl.glScalef(1.5f, 1.5f, 1.5f);
					this.drawContents(drawable);
				gl.glPopMatrix();

		// Make the last pop
		gl.glPopMatrix();
	}
	

	/**
	 * Draw the coordinate system
	 * @param drawable, the main GLDrawable given by OpenGL
	 */
	protected void drawCoordSystem(GLDrawable drawable) {
		
		// Grab the main GL object to draw on
		GL gl = drawable.getGL();
		
		// Set the color
		gl.glColor3f(0.0f, 1.0f, 0.0f);
		
		// Draw the x-axis
		gl.glBegin(GL.GL_LINE_LOOP);
			gl.glVertex3f(0.0f, 0.0f, 0.0f);
			gl.glVertex3f(1.0f, 0.0f, 0.0f);
		gl.glEnd();
		
		// Draw the y-axis
		gl.glBegin(GL.GL_LINE_LOOP);
			gl.glVertex3f(0.0f, 0.0f, 0.0f);
			gl.glVertex3f(0.0f, 1.0f, 0.0f);
		gl.glEnd();
		
		// Draw the z-axis
		gl.glBegin(GL.GL_LINE_LOOP);
			gl.glVertex3f(0.0f, 0.0f, 0.0f);
			gl.glVertex3f(0.0f, 0.0f, 1.0f);
		gl.glEnd();
	}
	
	/**
	 * Method required by the GLEventListener Interface
	 * @param drawable
	 * @param modeChanged
	 * @param deviceChanged
	 */
	public void displayChanged(GLDrawable drawable, boolean modeChanged, boolean deviceChanged) {}

	/**
	 * METHODS REQUIRED BY THE MOUSE LISTENERS
	 */
	
	/**
	 * Called when the user presses any mouse button
	 * @param e, the mouse event
	 */
	public void mousePressed(MouseEvent e) {
		
		// Remember the mouse coordinates
		prevMouseX = e.getX();
		prevMouseY = e.getY();
		
		// Check if the user pressed the Mouse Button 3 (normally the right mouse button)
		if ((e.getModifiers() & InputEvent.BUTTON3_MASK) != 0) {
			mouseRButtonDown = true;
			
		} else if ((e.getModifiers() & InputEvent.BUTTON1_MASK) != 0) {
			mouseLButtonDown = true;
		}
	}

	/**
	 * Called when the user releases the mouse button
	 * @param e, the mouse event
	 */
	public void mouseReleased(MouseEvent e) {
		
		// Check if the user released the Mouse Button 3 (normally the right mouse button)
		if ((e.getModifiers() & InputEvent.BUTTON3_MASK) != 0) {
			mouseRButtonDown = false;
			
		} else if ((e.getModifiers() & InputEvent.BUTTON1_MASK) != 0) {
			mouseLButtonDown = false;
		}
	}

	/**
	 * Called when the user drags the mouse
	 * @param e, the mouse event
	 */
	public void mouseDragged(MouseEvent e) {
		
		// Grab the current location of the mouse cursor
		int x = e.getX();
		int y = e.getY();
		
		// Grab the size of the component the user drags on (normally the OpenGL View)
		Dimension size = e.getComponent().getSize();

		// Check if the user currently holds the right mouse button
		if(mouseRButtonDown) {
			
			// Change the distance of the camera to the objects based on the y-value change when the user drags
			nRange = nRange + (y - prevMouseY);
			this.viewChangedByUser = true;

		} else if(mouseLButtonDown) {
			
			// Rotate the objects based on the user input
			float thetaY = 360.0f * ((float) (x - prevMouseX) / (float) size.width);
			float thetaX = 360.0f * ((float) (y - prevMouseY) / (float) size.height);

			viewRotX += thetaX;
			viewRotY += thetaY;
		}
		
		// Remember the mouse position for later use
		prevMouseX = x;
		prevMouseY = y;
	}

	/**
	 * METHODS REQUIRED BY THE MOUSE LISTENERS - NOT USED
	 */
	public void mouseClicked(MouseEvent e) {}
	public void mouseMoved(MouseEvent e) {}
	public void mouseEntered(MouseEvent e) {}
	public void mouseExited(MouseEvent e) {}
	
    /**
     * METHODS NEEDED BY THE KEYLISTENER INTERFACE
     */
    
	public void keyPressed(KeyEvent e) {}
	public void keyReleased(KeyEvent e) {}
	
	/**
	 * Automatically called when the user presses a key
	 * @param e, the key event
	 */
	public void keyTyped(KeyEvent e) {
		
		// Grab the pressed key from the event
		char pressedKey = e.getKeyChar();
		
		// Check if the user pressed a key we support
		switch(pressedKey) {
		
			case 'p':
				
				// Switch from lines to points only and back
				showPointsOnly = !showPointsOnly;
			break;
			
			case 'c':
				
				// Switch the color mode
				useColorMode = !useColorMode;
			break;
			
			case 'd':
				
				// Switch the draw-polygons mode
				showPolygons = !showPolygons;
			break;
		}
	}
}
