/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.animation;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLDrawable;
import fhm.edu.fem.calc.objects.Element;
import fhm.edu.fem.calc.objects.ElementList;
import fhm.edu.fem.calc.objects.Node;
import fhm.edu.fem.calc.objects.NodeList;
import fhm.edu.fem.calc.objects.Vector;

/**
 * Class Description:
 *		This class takes care of actually drawing the membrane
 * 
 * @author Thomas Woellert (thomas.woellert@t-online.de)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class GLMembraneRenderer extends GLFEMRenderer {

	// Vector holding the initial z-values
	protected Vector w0Vector = null;

	// The current w(t) vector we are drawing
	protected Vector currentWt = null;
	
	// The dimensions of the membrane
	protected double dimensionA = 0.0;
	protected double dimensionB = 0.0;

	/**
	 * Create a new OpenGL Membrane Renderer
	 * @param dimensionA, length of membrane side A
	 * @param dimensionB, length of membrane side B
	 * @param w0, the initial vector holding the z-values of all aliasID nodes
	 */
	public GLMembraneRenderer(double dimensionA, double dimensionB, Vector w0) {

		// Call the super constructor
		super();
		
		// Remember the variables
		this.w0Vector = w0;
		this.dimensionA = dimensionA;
		this.dimensionB = dimensionB;
		
		// Set w0 as the current w(t) vector
		currentWt = w0;
	}
	
	/**
	 * Set a new vector w(t) while running
	 * @param wT, the new Vector wT
	 */
	public void setNewVectorW(Vector wT) {
		this.currentWt = wT;
	}
	
	/**
	 * Draw the contents
	 * @param drawable, the main GLDrawable given by OpenGL
	 */
	protected void drawContents(GLDrawable drawable) {
		
		// Grab the main GL object to draw on
		GL gl = drawable.getGL();
		
		// Enable the DEPTH TEST
		gl.glEnable(GL.GL_DEPTH_TEST);
		
		// Check if we should show full polygons
		if(showPolygons) {
			
			// Draw the whole figure in filled polygons
			this.drawMembranePolygons(drawable, this.currentWt);
			
		} else {
		
			// Check if we should show only the points or the full elements
			if(showPointsOnly) {
				
				// Draw all nodes
				this.drawMembraneNodes(drawable, this.currentWt);
			} else {
				
				// Draw all elements
				this.drawMembraneElements(drawable, this.currentWt);
				
			}
		}
	}
	
	/**
	 * Draw the membrane's nodes
	 * @param drawable, the main GLDrawable given by OpenGL
	 * @param w, the vector w(t) to use
	 */
	protected void drawMembraneNodes(GLDrawable drawable, Vector w) {
		
		// Grab the main GL object to draw on
		GL gl = drawable.getGL();
		
		// Grab the node list
		NodeList fullNodeList = femManager.getNodeList();
		
		gl.glBegin(GL.GL_POINTS);
			
			// Run through all nodes
			for(int i = 1; i <= fullNodeList.size(); i++) {
				
				// Grab the current node
				Node currentNode = fullNodeList.getNodeInCorrectOrder(i);
				
				// Check if the current Node has an alias ID
				if(currentNode.getAliasId() == 0) {
					
					// This is a border node
					// Clear the screen
					gl.glColor3f(1.0f, 0.0f, 0.0f);
					gl.glVertex3d(transX + currentNode.getX(), transY + currentNode.getY(), 0.0d);
					
				} else {
					
					// This is an alias node with a specific initial z-value taken from the vector
					if(useColorMode) {
						gl.glColor3f(1.0f, 1.0f, 0.0f);
					} else {
						gl.glColor3f(0.0f, 0.0f, 0.0f);
					}
					gl.glVertex3d(transX + currentNode.getX(), transY + currentNode.getY(), w.getValue(currentNode.getAliasId()));
				}
			}
			
		gl.glEnd();
	}
	
	/**
	 * Draw the membrane's elements
	 * @param drawable, the main GLDrawable given by OpenGL
	 * @param w, the vector w(t) to use
	 */
	protected void drawMembraneElements(GLDrawable drawable, Vector w) {
		
		// Grab the main GL object to draw on
		GL gl = drawable.getGL();
		
		// Grab the element list
		ElementList fullElementList = femManager.getElementList();
		
		// Run through all elements
		for(int i = 1; i <= fullElementList.size(); i++) {
			
			// Grab the current element
			Element currentElement = fullElementList.getElement(i);
			NodeList elementNodeList = currentElement.getNodeList();
			
			// Run through all corner nodes in the element
			int elementNodeCount = -1;
			
			// To remember if there are more nodes ("krummlinige Elemente")
			boolean moreNodes = false;
			
			switch(elementNodeList.size()) {
				
				case 6:
					elementNodeCount = 3;
					moreNodes = true;
				break;
				
				case 8:
					elementNodeCount = 4;
					moreNodes = true;
				break;
				
				default:
					// The element has 3 or 4 nodes
					elementNodeCount = elementNodeList.size();
					moreNodes = false;
			}
			
			gl.glBegin(GL.GL_LINE_STRIP);

			// Remember the first node
			Node firstNode = elementNodeList.getNodeInCorrectOrder(1);
			
			for(int j = 1; j <= elementNodeCount; j++) {
				
				// Grab the current node in the element
				Node currentNode = elementNodeList.getNodeInCorrectOrder(j);
				int aliasID = currentNode.getAliasId();
				
				// Check if the current Node has an alias ID
				if(aliasID == 0) {
					
					// This is a border node
					gl.glColor3f(1.0f, 0.0f, 0.0f);
					gl.glVertex3d(transX + currentNode.getX(), transY + currentNode.getY(), 0.0d);
					
				} else {
					
					// This is an alias node with a specific initial z-value taken from the vector
					if(useColorMode) {
						gl.glColor3f(1.0f, 1.0f, 0.0f);
					} else {
						gl.glColor3f(0.0f, 0.0f, 0.0f);
					}
					gl.glVertex3d(transX + currentNode.getX(), transY + currentNode.getY(), w.getValue(aliasID));
				}
				
				// Check if there are more nodes (in case of "krummlinige Elemente")
				if(moreNodes) {
					
					// Grab the current node in the element
					currentNode = elementNodeList.getNodeInCorrectOrder(j+elementNodeCount);
					aliasID = currentNode.getAliasId();
					
					// Check if the current Node has an alias ID
					if(aliasID == 0) {
						
						// This is a border node
						gl.glColor3f(1.0f, 0.0f, 0.0f);
						gl.glVertex3d(transX + currentNode.getX(), transY + currentNode.getY(), 0.0d);
						
					} else {
						
						// This is an alias node with a specific initial z-value taken from the vector
						if(useColorMode) {
							gl.glColor3f(1.0f, 1.0f, 0.0f);
						} else {
							gl.glColor3f(0.0f, 0.0f, 0.0f);
						}
						gl.glVertex3d(transX + currentNode.getX(), transY + currentNode.getY(), w.getValue(aliasID));
					}
				}
			}

			// Make the connection from the last node in the element's line array to the first (to close the element)
			// Check if the current Node has an alias ID
			int aliasID = firstNode.getAliasId();
			if(aliasID == 0) {
				
				// This is a border node
				gl.glColor3f(1.0f, 0.0f, 0.0f);
				gl.glVertex3d(transX + firstNode.getX(), transY + firstNode.getY(), 0.0d);
				
			} else {
				
				// This is an alias node with a specific initial z-value taken from the vector
				if(useColorMode) {
					gl.glColor3f(1.0f, 1.0f, 0.0f);
				} else {
					gl.glColor3f(0.0f, 0.0f, 0.0f);
				}
				gl.glVertex3d(transX + firstNode.getX(), transY + firstNode.getY(), w.getValue(aliasID));
			}
			
			gl.glEnd();
		}
	}
	
	/**
	 * Draw the membrane's polygons
	 * @param drawable, the main GLDrawable given by OpenGL
	 * @param w, the vector w(t) to use
	 */
	protected void drawMembranePolygons(GLDrawable drawable, Vector w) {
		
		// Grab the main GL object to draw on
		GL gl = drawable.getGL();
		
		// Grab the element list
		ElementList fullElementList = femManager.getElementList();
		
		// Run through all elements
		for(int i = 1; i <= fullElementList.size(); i++) {
			
			// Grab the current element
			Element currentElement = fullElementList.getElement(i);
			NodeList elementNodeList = currentElement.getNodeList();
			
			// Run through all corner nodes in the element
			int elementNodeCount = -1;
			
			// To remember if there are more nodes ("krummlinige Elemente")
			boolean moreNodes = false;
			
			switch(elementNodeList.size()) {
				
				case 6:
					elementNodeCount = 3;
					moreNodes = true;
				break;
				
				case 8:
					elementNodeCount = 4;
					moreNodes = true;
				break;
				
				default:
					// The element has 3 or 4 nodes
					elementNodeCount = elementNodeList.size();
					moreNodes = false;
			}
			
			gl.glBegin(GL.GL_POLYGON);

			// Remember the first node to close the drawing at the end of the element
			Node firstNode = elementNodeList.getNodeInCorrectOrder(1);
			
			for(int j = 1; j <= elementNodeCount; j++) {
				
				// Grab the first Node
				Node currentNode = elementNodeList.getNodeInCorrectOrder(j);
				int aliasID = currentNode.getAliasId();
				
				// Check if the current Node has an alias ID
				if(aliasID == 0) {
					
					// This is a border node
					gl.glColor3f(1.0f, 0.0f, 0.0f);
					gl.glVertex3d(transX + currentNode.getX(), transY + currentNode.getY(), 0.0d);
					
				} else {
					
					// This is an alias node with a specific initial z-value taken from the vector
					if(useColorMode) {
						gl.glColor3f(1.0f, 1.0f, 0.0f);
					} else {
						gl.glColor3f(0.0f, 0.0f, 0.0f);
					}
					gl.glVertex3d(transX + currentNode.getX(), transY + currentNode.getY(), w.getValue(aliasID));
				}
				
				// Additional drawing for "krummlinige" elements
				if(moreNodes) {
					
					// Grab the first "krummlinig" Node
					currentNode = elementNodeList.getNodeInCorrectOrder(j+elementNodeCount);
					aliasID = currentNode.getAliasId();
					
					// Check if the current Node has an alias ID
					if(aliasID == 0) {
						
						// This is a border node
						gl.glColor3f(1.0f, 0.0f, 0.0f);
						gl.glVertex3d(transX + currentNode.getX(), transY + currentNode.getY(), 0.0d);
						
					} else {
						
						// This is an alias node with a specific initial z-value taken from the vector
						if(useColorMode) {
							gl.glColor3f(1.0f, 1.0f, 0.0f);
						} else {
							gl.glColor3f(0.0f, 0.0f, 0.0f);
						}
						gl.glVertex3d(transX + currentNode.getX(), transY + currentNode.getY(), w.getValue(aliasID));
					}
				}
			}

			// Make the connection from the last node in the element's line array to the first (to close the element)
			// Check if the current Node has an alias ID
			int aliasID = firstNode.getAliasId();
			if(aliasID == 0) {
				
				// This is a border node
				gl.glColor3f(1.0f, 0.0f, 0.0f);
				gl.glVertex3d(transX + firstNode.getX(), transY + firstNode.getY(), 0.0d);
				
			} else {
				
				// This is an alias node with a specific initial z-value taken from the vector
				if(useColorMode) {
					gl.glColor3f(1.0f, 1.0f, 0.0f);
				} else {
					gl.glColor3f(0.0f, 0.0f, 0.0f);
				}
				gl.glVertex3d(transX + firstNode.getX(), transY + firstNode.getY(), w.getValue(aliasID));
			}
			
			gl.glEnd();
		}
	}
}
