/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.animation;

import java.awt.Color;
import java.awt.event.KeyEvent;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLDrawable;
import fhm.edu.fem.calc.FEMCalcSpanner;
import fhm.edu.fem.calc.objects.BorderList;
import fhm.edu.fem.calc.objects.Element;
import fhm.edu.fem.calc.objects.ElementList;
import fhm.edu.fem.calc.objects.Node;
import fhm.edu.fem.calc.objects.NodeList;

/**
 * Class Description:
 *		This class takes care of actually drawing the spanner
 * 
 * @author Thomas Woellert (thomas.woellert@t-online.de)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class GLSpannerRenderer extends GLFEMRenderer {

	/* The FEMCalc object we are using to calculate the renderer look */
	protected FEMCalcSpanner femCalc = null;
	
	/* The Alias ID Node List - only for drawing purposes */
	protected NodeList aliasIDNodeList = new NodeList();
	
	/* Enable or disable drawing of the back-figure based on the thickness "h" of the spanner */
	protected boolean drawBackFigure = false;
	
	/* Enable or disable drawing of the new figure */
	protected boolean drawNewFigure = true;
	
	/* Enable or disable drawing of the "Spannungsbetrag" colors */
	protected boolean drawSpannungsbetragColors = false;
	
	/* The list of all colors for the elements' "Spannungsbetrag" */
	protected GLColor[] spannungsbetragColors;
	
	/**
	 * Create a new OpenGL Spanner Renderer
	 * @param femCalc, the calculator we use
	 */
	public GLSpannerRenderer(FEMCalcSpanner femCalc) {
		
		// Call super constructor
		super();
		
		// Remember the variables
		this.femCalc = femCalc;
		
		// Create the alias ID NodeList for drawing purposes
		createAliasList();
	}
	
	/**
	 * Draw the contents
	 * @param drawable, the main GLDrawable given by OpenGL
	 */
	protected void drawContents(GLDrawable drawable) {

		// Grab the main GL object to draw on
		GL gl = drawable.getGL();
		
		// The new actual node and element lists to draw
		NodeList newNodeList = femManager.getNodeList();
		ElementList newElementList = femManager.getElementList();
		
		// Build the color list for all elements, displaying the "Spannungsbetrag" colors
		buildSpannungsbetragsColors(newElementList);
		
		// Check if we should show full polygons
		if(showPolygons) {
			
			// Enable the DEPTH TEST
			gl.glEnable(GL.GL_DEPTH_TEST);
			
			// Draw the whole figure in filled polygons
			this.drawSpannerPolygons(drawable, newElementList, 0.0, true);
			
			if(drawNewFigure) {
				// Draw the front polygons of the new figure (need to move the z-value a little bit to avoid polygons from flickering because
				// they overlap each other)
				this.drawSpannerPolygons(drawable, newElementList, 0.0+0.01, false);
			}
			
			if(drawBackFigure) {
				
				// Draw the back polygons of the initial figure
				this.drawSpannerPolygons(drawable, newElementList, (-1.0*femManager.getSpannerMaterialConstantH()), true);
				
				if(drawNewFigure) {
					// Draw the back polygons of the new figure (need to move the z-value a little bit to avoid polygons from flickering because
					// they overlap each other)
					this.drawSpannerPolygons(drawable, newElementList, (-1.0*femManager.getSpannerMaterialConstantH())-0.01, false);
				}
			}
			
			// Disable the DEPTH TEST
			gl.glDisable(GL.GL_DEPTH_TEST);
			
		} else {
			
			// Check if we should show only the points or the full elements
			if(showPointsOnly) {
				
				// Draw all nodes of the initial figure
				this.drawSpannerNodes(drawable, newNodeList, 0.0, true);
				
				if(drawNewFigure) {
					// Draw all elements of the new figure
					this.drawSpannerNodes(drawable, newNodeList, 0.0, false);
				}
				
				if(drawBackFigure) {
					
					// Draw the back nodes  of the initial figure
					this.drawSpannerNodes(drawable, newNodeList, (-1.0*femManager.getSpannerMaterialConstantH()), true);
					
					if(drawNewFigure) {
						// Draw the back nodes of the new figure
						this.drawSpannerNodes(drawable, newNodeList, (-1.0*femManager.getSpannerMaterialConstantH()), false);
					}
				}
				
			} else {
				
				// Draw all elements of the initial figure
				this.drawSpannerElements(drawable, newElementList, 0.0, true);
				
				if(drawNewFigure) {
					// Draw all elements of the new figure
					this.drawSpannerElements(drawable, newElementList, 0.0, false);
				}
				
				if(drawBackFigure) {
					
					// Draw the back elements  of the initial figure
					this.drawSpannerElements(drawable, newElementList, (-1.0*femManager.getSpannerMaterialConstantH()), true);
					
					if(drawNewFigure) {
						// Draw the back elements of the new figure
						this.drawSpannerElements(drawable, newElementList, (-1.0*femManager.getSpannerMaterialConstantH()), false);
					}
				}
			}
		}
	}
	
	/**
	 * Draw the spanners current nodes
	 * @param drawable, the main GLDrawable given by OpenGL
	 * @param nodeList, the nodelist to draw
	 * @param zValue, the z-value of the polygon nodes
	 * @param useOriginalCoordinates, draw the nodes using their original coordinates
	 */
	protected void drawSpannerNodes(GLDrawable drawable, NodeList nodeList, double zValue, boolean useOriginalCoordinates) {
		
		// Grab the main GL object to draw on
		GL gl = drawable.getGL();

		gl.glBegin(GL.GL_POINTS);
			
		// Run through all nodes
		for(int i = 1; i <= nodeList.size(); i++) {
			
			// Grab the current node
			Node currentNode = nodeList.getNodeInCorrectOrder(i);
			
			// Check if the current Node has an alias ID (the Spanner has no real Alias IDs during calculation, they are solely used for display purposes)
			if(!femManager.getAliasIDNodeList().contains(currentNode)) {
				
				// This is a border node
				gl.glColor3f(1.0f, 0.0f, 0.0f);
				
				if(useOriginalCoordinates) {
					gl.glColor3f(0.5f, 0.5f, 0.5f);
					gl.glVertex3d(transX + currentNode.getXOriginal(), transY + currentNode.getYOriginal(), zValue);
				} else {
					gl.glVertex3d(transX + currentNode.getX(), transY + currentNode.getY(), zValue);
				}
				
			} else {
				
				// This is a node which has a border condition defined
				if(useColorMode) {
					gl.glColor3f(1.0f, 1.0f, 0.0f);
				} else {
					gl.glColor3f(0.0f, 0.0f, 0.0f);
				}
				
				if(useOriginalCoordinates) {
					gl.glColor3f(0.5f, 0.5f, 0.5f);
					gl.glVertex3d(transX + currentNode.getXOriginal(), transY + currentNode.getYOriginal(), zValue);
				} else {
					gl.glVertex3d(transX + currentNode.getX(), transY + currentNode.getY(), zValue);
				}
			}
		}
			
		gl.glEnd();
	}
	
	/**
	 * Draw the spanners current elements
	 * @param drawable, the main GLDrawable given by OpenGL
	 * @param elementList, the elementList to draw
	 * @param zValue, the z-value of the polygon nodes
	 * @param useOriginalCoordinates, draw the nodes using their original coordinates
	 */
	protected void drawSpannerElements(GLDrawable drawable, ElementList elementList, double zValue, boolean useOriginalCoordinates) {
		
		// Grab the main GL object to draw on
		GL gl = drawable.getGL();
		
		// Grab the index of the elements with the highest "Hauptspannung"
		int maxSpannungsBetragElementIndex = this.femCalc.getMaxSigmaElementIndex();
		
		// Run through all elements
		for(int i = 1; i <= elementList.size(); i++) {
			
			// Grab the current element
			Element currentElement = elementList.getElement(i);
			NodeList elementNodeList = currentElement.getNodeList();
			
			// Run through all corner nodes in the element
			int elementNodeCount = -1;
			
			// To remember if there are more nodes ("krummlinige Elemente")
			boolean moreNodes = false;
			
			switch(elementNodeList.size()) {
				
				case 6:
					elementNodeCount = 3;
					moreNodes = true;
				break;
				
				case 8:
					elementNodeCount = 4;
					moreNodes = true;
				break;
				
				default:
					// The element has 3 or 4 nodes
					elementNodeCount = elementNodeList.size();
					moreNodes = false;
			}
			
			/**
			 * DRAW THE ELEMENTS
			 */
			
			gl.glBegin(GL.GL_LINE_STRIP);

			// Remember the first node
			Node firstNode = elementNodeList.getNodeInCorrectOrder(1);
			
			// Draw the front elements
			for(int j = 1; j <= elementNodeCount; j++) {
				
				// Grab the current node in the element
				Node currentNode = elementNodeList.getNodeInCorrectOrder(j);

				// Check if the current Node has an alias ID
				if(!femManager.getAliasIDNodeList().contains(currentNode)) {
					
					// This is a border node
					if(drawSpannungsbetragColors) {
						gl.glColor3f(spannungsbetragColors[i-1].getRedGL(), spannungsbetragColors[i-1].getGreenGL(), spannungsbetragColors[i-1].getBlueGL());	
					} else if(i == maxSpannungsBetragElementIndex) {
						gl.glColor3f(0.0f, 1.0f, 0.0f);					
					} else {
						gl.glColor3f(1.0f, 0.0f, 0.0f);
					}
					
					if(useOriginalCoordinates) {
						gl.glColor3f(0.5f, 0.5f, 0.5f);
						gl.glVertex3d(transX + currentNode.getXOriginal(), transY + currentNode.getYOriginal(), zValue);
					} else {
						gl.glVertex3d(transX + currentNode.getX(), transY + currentNode.getY(), zValue);
					}
					
				} else {
					
					// This is an alias node with a specific initial z-value taken from the vector
					if(useColorMode) {
						if(drawSpannungsbetragColors) {
							gl.glColor3f(spannungsbetragColors[i-1].getRedGL(), spannungsbetragColors[i-1].getGreenGL(), spannungsbetragColors[i-1].getBlueGL());	
						} else if(i == maxSpannungsBetragElementIndex) {
							gl.glColor3f(0.0f, 1.0f, 0.0f);
						} else {
							gl.glColor3f(1.0f, 1.0f, 0.0f);
						}
					} else {						
						if(drawSpannungsbetragColors) {
							gl.glColor3f(spannungsbetragColors[i-1].getRedGL(), spannungsbetragColors[i-1].getGreenGL(), spannungsbetragColors[i-1].getBlueGL());	
						} else if(i == maxSpannungsBetragElementIndex) {
							gl.glColor3f(0.0f, 1.0f, 0.0f);
						} else {
							gl.glColor3f(0.0f, 0.0f, 0.0f);
						}
					}
					if(useOriginalCoordinates) {
						gl.glColor3f(0.5f, 0.5f, 0.5f);
						gl.glVertex3d(transX + currentNode.getXOriginal(), transY + currentNode.getYOriginal(), zValue);
					} else {
						gl.glVertex3d(transX + currentNode.getX(), transY + currentNode.getY(), zValue);
					}
				}
				
				// Check if there are more nodes (in case of "krummlinige Elemente")
				if(moreNodes) {
					
					// Grab the current node in the element
					currentNode = elementNodeList.getNodeInCorrectOrder(j+elementNodeCount);
					
					// Check if the current Node has an alias ID
					if(!femManager.getAliasIDNodeList().contains(currentNode)) {
						
						// This is a border node
						if(drawSpannungsbetragColors) {
							gl.glColor3f(spannungsbetragColors[i-1].getRedGL(), spannungsbetragColors[i-1].getGreenGL(), spannungsbetragColors[i-1].getBlueGL());	
						} else if(i == maxSpannungsBetragElementIndex) {
							gl.glColor3f(0.0f, 1.0f, 0.0f);
						} else {
							gl.glColor3f(1.0f, 0.0f, 0.0f);
						}
						if(useOriginalCoordinates) {
							gl.glColor3f(0.5f, 0.5f, 0.5f);
							gl.glVertex3d(transX + currentNode.getXOriginal(), transY + currentNode.getYOriginal(), zValue);
						} else {
							gl.glVertex3d(transX + currentNode.getX(), transY + currentNode.getY(), zValue);
						}
						
					} else {
						
						// This is an alias node with a specific initial z-value taken from the vector
						if(useColorMode) {
							
							if(drawSpannungsbetragColors) {
								gl.glColor3f(spannungsbetragColors[i-1].getRedGL(), spannungsbetragColors[i-1].getGreenGL(), spannungsbetragColors[i-1].getBlueGL());	
							} else if(i == maxSpannungsBetragElementIndex) {
								gl.glColor3f(0.0f, 1.0f, 0.0f);
							} else {
								gl.glColor3f(1.0f, 1.0f, 0.0f);
							}
							
						} else {
							
							if(drawSpannungsbetragColors) {
								gl.glColor3f(spannungsbetragColors[i-1].getRedGL(), spannungsbetragColors[i-1].getGreenGL(), spannungsbetragColors[i-1].getBlueGL());	
							} else if(i == maxSpannungsBetragElementIndex) {
								gl.glColor3f(0.0f, 1.0f, 0.0f);
							} else {
								gl.glColor3f(0.0f, 0.0f, 0.0f);
							}
						}
						if(useOriginalCoordinates) {
							gl.glColor3f(0.5f, 0.5f, 0.5f);
							gl.glVertex3d(transX + currentNode.getXOriginal(), transY + currentNode.getYOriginal(), zValue);
						} else {
							gl.glVertex3d(transX + currentNode.getX(), transY + currentNode.getY(), zValue);
						}
					}
				}
			}
			
			// Make the connection from the last node in the element's line array to the first (to close the element)
			// Check if the current Node has an alias ID
			if(!femManager.getAliasIDNodeList().contains(firstNode)) {
				
				// This is a border node
				if(drawSpannungsbetragColors) {
					gl.glColor3f(spannungsbetragColors[i-1].getRedGL(), spannungsbetragColors[i-1].getGreenGL(), spannungsbetragColors[i-1].getBlueGL());	
				} else if(i == maxSpannungsBetragElementIndex) {
					gl.glColor3f(0.0f, 1.0f, 0.0f);
				} else {
					gl.glColor3f(1.0f, 0.0f, 0.0f);
				}
				if(useOriginalCoordinates) {
					gl.glColor3f(0.5f, 0.5f, 0.5f);
					gl.glVertex3d(transX + firstNode.getXOriginal(), transY + firstNode.getYOriginal(), zValue);
				} else {
					gl.glVertex3d(transX + firstNode.getX(), transY + firstNode.getY(), zValue);
				}
				
			} else {
				
				// This is an alias node with a specific initial z-value taken from the vector
				if(useColorMode) {
					if(drawSpannungsbetragColors) {
						gl.glColor3f(spannungsbetragColors[i-1].getRedGL(), spannungsbetragColors[i-1].getGreenGL(), spannungsbetragColors[i-1].getBlueGL());	
					} else if(i == maxSpannungsBetragElementIndex) {
						gl.glColor3f(0.0f, 1.0f, 0.0f);
					} else {
						gl.glColor3f(1.0f, 1.0f, 0.0f);
					}
				} else {
					if(drawSpannungsbetragColors) {
						gl.glColor3f(spannungsbetragColors[i-1].getRedGL(), spannungsbetragColors[i-1].getGreenGL(), spannungsbetragColors[i-1].getBlueGL());	
					} else if(i == maxSpannungsBetragElementIndex) {
						gl.glColor3f(0.0f, 1.0f, 0.0f);
					} else {
						gl.glColor3f(0.0f, 0.0f, 0.0f);
					}
				}
				if(useOriginalCoordinates) {
					gl.glColor3f(0.5f, 0.5f, 0.5f);
					gl.glVertex3d(transX + firstNode.getXOriginal(), transY + firstNode.getYOriginal(), zValue);
				} else {
					gl.glVertex3d(transX + firstNode.getX(), transY + firstNode.getY(), zValue);
				}
			}
			
			gl.glEnd();
		}
	}
    
	/**
	 * Draw the spanner's front real polygons based on the node coordinates
	 * @param drawable, the main GLDrawable given by OpenGL
	 * @param fullElementList, the elementList to draw
	 * @param zValue, the z-value of the polygon nodes
	 * @param useOriginalCoordinates, draw the nodes using their original coordinates
	 */
	protected void drawSpannerPolygons(GLDrawable drawable, ElementList fullElementList, double zValue, boolean useOriginalCoordinates) {
		
		// Grab the main GL object to draw on
		GL gl = drawable.getGL();
		
		// Grab the index of the elements with the highest "Hauptspannung"
		int maxSpannungsBetragElementIndex = this.femCalc.getMaxSigmaElementIndex();
		
		// Run through all elements
		for(int i = 1; i <= fullElementList.size(); i++) {
			
			// Grab the current element
			Element currentElement = fullElementList.getElement(i);
			NodeList elementNodeList = currentElement.getNodeList();
			
			// Run through all corner nodes in the element
			int elementNodeCount = -1;
			
			// To remember if there are more nodes ("krummlinige Elemente")
			boolean moreNodes = false;
			
			switch(elementNodeList.size()) {
				
				case 6:
					elementNodeCount = 3;
					moreNodes = true;
				break;
				
				case 8:
					elementNodeCount = 4;
					moreNodes = true;
				break;
				
				default:
					// The element has 3 or 4 nodes
					elementNodeCount = elementNodeList.size();
					moreNodes = false;
			}
			
			gl.glBegin(GL.GL_POLYGON);

			// Remember the first node to close the drawing at the end of the element
			Node firstNode = elementNodeList.getNodeInCorrectOrder(1);
			
			for(int j = 1; j <= elementNodeCount; j++) {
				
				// Grab the first Node
				Node currentNode = elementNodeList.getNodeInCorrectOrder(j);
				
				// Check if the current Node has an alias ID
				if(!femManager.getAliasIDNodeList().contains(currentNode)) {
					
					// This is a border node
					if(drawSpannungsbetragColors) {
						gl.glColor3f(spannungsbetragColors[i-1].getRedGL(), spannungsbetragColors[i-1].getGreenGL(), spannungsbetragColors[i-1].getBlueGL());	
					} else if(i == maxSpannungsBetragElementIndex) {
						gl.glColor3f(0.0f, 1.0f, 0.0f);
					} else {
						gl.glColor3f(1.0f, 0.0f, 0.0f);
					}
					if(useOriginalCoordinates) {
						gl.glColor3f(0.5f, 0.5f, 0.5f);
						gl.glVertex3d(transX + currentNode.getXOriginal(), transY + currentNode.getYOriginal(), zValue);
					} else {
						gl.glVertex3d(transX + currentNode.getX(), transY + currentNode.getY(), zValue);
					}
					
				} else {
					
					// This is an alias node with a specific initial z-value taken from the vector
					if(useColorMode) {
						if(drawSpannungsbetragColors) {
							gl.glColor3f(spannungsbetragColors[i-1].getRedGL(), spannungsbetragColors[i-1].getGreenGL(), spannungsbetragColors[i-1].getBlueGL());	
						} else if(i == maxSpannungsBetragElementIndex) {
							gl.glColor3f(0.0f, 1.0f, 0.0f);
						} else {
							gl.glColor3f(1.0f, 1.0f, 0.0f);
						}
					} else {
						if(drawSpannungsbetragColors) {
							gl.glColor3f(spannungsbetragColors[i-1].getRedGL(), spannungsbetragColors[i-1].getGreenGL(), spannungsbetragColors[i-1].getBlueGL());	
						} else if(i == maxSpannungsBetragElementIndex) {
							gl.glColor3f(0.0f, 1.0f, 0.0f);
						} else {
							gl.glColor3f(0.0f, 0.0f, 0.0f);
						}
					}
					if(useOriginalCoordinates) {
						gl.glColor3f(0.5f, 0.5f, 0.5f);
						gl.glVertex3d(transX + currentNode.getXOriginal(), transY + currentNode.getYOriginal(), zValue);
					} else {
						gl.glVertex3d(transX + currentNode.getX(), transY + currentNode.getY(), zValue);
					}
				}
				
				// Additional drawing for "krummlinige" elements
				if(moreNodes) {
					
					// Grab the first "krummlinig" Node
					currentNode = elementNodeList.getNodeInCorrectOrder(j+elementNodeCount);
					
					// Check if the current Node has an alias ID
					if(!femManager.getAliasIDNodeList().contains(currentNode)) {
						
						// This is a border node
						if(drawSpannungsbetragColors) {
							gl.glColor3f(spannungsbetragColors[i-1].getRedGL(), spannungsbetragColors[i-1].getGreenGL(), spannungsbetragColors[i-1].getBlueGL());	
						} else if(i == maxSpannungsBetragElementIndex) {
							gl.glColor3f(0.0f, 1.0f, 0.0f);
						} else {
							gl.glColor3f(1.0f, 0.0f, 0.0f);
						}
						if(useOriginalCoordinates) {
							gl.glColor3f(1.0f, 0.0f, 0.0f);
							gl.glVertex3d(transX + currentNode.getXOriginal(), transY + currentNode.getYOriginal(), zValue);
						} else {
							gl.glVertex3d(transX + currentNode.getX(), transY + currentNode.getY(), zValue);
						}
						
					} else {
						
						// This is an alias node with a specific initial z-value taken from the vector
						if(useColorMode) {
							if(drawSpannungsbetragColors) {
								gl.glColor3f(spannungsbetragColors[i-1].getRedGL(), spannungsbetragColors[i-1].getGreenGL(), spannungsbetragColors[i-1].getBlueGL());	
							} else if(i == maxSpannungsBetragElementIndex) {
								gl.glColor3f(0.0f, 1.0f, 0.0f);
							} else {
								gl.glColor3f(1.0f, 1.0f, 0.0f);
							}
						} else {
							if(drawSpannungsbetragColors) {
								gl.glColor3f(spannungsbetragColors[i-1].getRedGL(), spannungsbetragColors[i-1].getGreenGL(), spannungsbetragColors[i-1].getBlueGL());	
							} else if(i == maxSpannungsBetragElementIndex) {
								gl.glColor3f(0.0f, 1.0f, 0.0f);
							} else {
								gl.glColor3f(0.0f, 0.0f, 0.0f);
							}
						}
						if(useOriginalCoordinates) {
							gl.glColor3f(0.5f, 0.5f, 0.5f);
							gl.glVertex3d(transX + currentNode.getXOriginal(), transY + currentNode.getYOriginal(), zValue);
						} else {
							gl.glVertex3d(transX + currentNode.getX(), transY + currentNode.getY(), zValue);
						}
					}
				}
			}

			// Make the connection from the last node in the element's line array to the first (to close the element)
			// Check if the current Node has an alias ID
			if(!femManager.getAliasIDNodeList().contains(firstNode)) {
				
				// This is a border node
				if(drawSpannungsbetragColors) {
					gl.glColor3f(spannungsbetragColors[i-1].getRedGL(), spannungsbetragColors[i-1].getGreenGL(), spannungsbetragColors[i-1].getBlueGL());	
				} else if(i == maxSpannungsBetragElementIndex) {
					gl.glColor3f(0.0f, 1.0f, 0.0f);
				} else {
					gl.glColor3f(1.0f, 0.0f, 0.0f);
				}
				if(useOriginalCoordinates) {
					gl.glColor3f(0.5f, 0.5f, 0.5f);
					gl.glVertex3d(transX + firstNode.getXOriginal(), transY + firstNode.getYOriginal(), zValue);
				} else {
					gl.glVertex3d(transX + firstNode.getX(), transY + firstNode.getY(), zValue);
				}
				
			} else {
				
				// This is an alias node with a specific initial z-value taken from the vector
				if(useColorMode) {
					if(drawSpannungsbetragColors) {
						gl.glColor3f(spannungsbetragColors[i-1].getRedGL(), spannungsbetragColors[i-1].getGreenGL(), spannungsbetragColors[i-1].getBlueGL());	
					} else if(i == maxSpannungsBetragElementIndex) {
						gl.glColor3f(0.0f, 1.0f, 0.0f);
					} else {
						gl.glColor3f(1.0f, 1.0f, 0.0f);
					}
				} else {
					if(drawSpannungsbetragColors) {
						gl.glColor3f(spannungsbetragColors[i-1].getRedGL(), spannungsbetragColors[i-1].getGreenGL(), spannungsbetragColors[i-1].getBlueGL());	
					} else if(i == maxSpannungsBetragElementIndex) {
						gl.glColor3f(0.0f, 1.0f, 0.0f);
					} else {
						gl.glColor3f(0.0f, 0.0f, 0.0f);
					}
				}
				if(useOriginalCoordinates) {
					gl.glColor3f(0.5f, 0.5f, 0.5f);
					gl.glVertex3d(transX + firstNode.getXOriginal(), transY + firstNode.getYOriginal(), zValue);
				} else {
					gl.glVertex3d(transX + firstNode.getX(), transY + firstNode.getY(), zValue);
				}
			}
			
			gl.glEnd();
		}
	}
	
    /**
     * Create the alias List
     * @return true if all went well, false otherwise
     */
    protected boolean createAliasList() {
    	
    	// Grab the list of all nodes
    	NodeList completeNodeList = femManager.getNodeList();
    	
    	// Grab the border list
    	BorderList borderList = femManager.getBorderList();
    	
    	// Make sure the list is empty
    	aliasIDNodeList.clear();
    	
    	// List of all border nodes
    	NodeList borderNodeList = new NodeList();
    	
    	// Run through all borders
    	for(int i = 1; i <= borderList.size(); i++) {

    		borderNodeList.addAll(borderList.getBorder(i).getNodes());
    	}
    	
		// Create the alias IDs based on the current border
		NodeList nodeList = completeNodeList.createAliasIDs(1, borderNodeList);
		
		// Add the node list we are getting back to the aliasIDNodeList
		aliasIDNodeList.addAll(nodeList);	
        
        // Save the alias ID list in the FEMManager for later use
        femManager.setAliasIDNodeList(this.aliasIDNodeList);

    	return true;
    }
    
    /**
     * Build the list of all element colors depending on the elements' spannungsbetrag
     * @param elementList, the element list to use
     */
    protected void buildSpannungsbetragsColors(ElementList elementList) {
    	
    	// Initialize the array
    	spannungsbetragColors = new GLColor[elementList.size()];
    	
    	// Get the max allowed "Hchstspannung"
    	double maxSpannung = femManager.getSpannerMaterialConstantHS();
    	maxSpannung = Math.sqrt(3*Math.pow(maxSpannung, 2));
    	
    	// Get the max Spannung we calculated via Sigma
    	int maxSpannungsBetragElementIndex = this.femCalc.getMaxSigmaElementIndex();
    	
    	// Check if we got a max spannungs betrag element index
    	if(maxSpannungsBetragElementIndex == -1) {
    		
    		return;
    	}
    	
    	double calcMaxSpannung = elementList.getElement(maxSpannungsBetragElementIndex).getSpannungsBetrag();
    	
    	// Calculate the color-steps ... we have 255 steps, from Color(0, 255, 0) till Color(255, 0, 0) (Red/Green/Blue)
    	double stepSpannung = maxSpannung / 255;
    	
    	// Run through all color steps
    	for(int i = 0; i < 256; i++) {
    		
    		// Build the current color
    		GLColor currentColor = new GLColor(i, (255-i), 0);
    		
    		// Run through all elements and determine if they are in the current color step
    		for(int elementIndex = 1; elementIndex <= elementList.size(); elementIndex++) {

    			Element currentElement = elementList.getElement(elementIndex);
    			double elementSpannung = currentElement.getSpannungsBetrag();
    			
    			if(elementSpannung >= (stepSpannung * i) && elementSpannung <= (stepSpannung * (i+1))) {
    				
    				// The element is in the current "Spannung"s-Step
    				spannungsbetragColors[elementIndex-1] = currentColor;
    			}
    		}
    	}
    	
    	// Check for elements which have no color yet (most likely they have a "Spannung" > "Hoechstspannung")
    	// No element can have a "Spannung" < 0
    	for(int i = 1; i <= elementList.size(); i++) {
    		
    		Element currentElement = elementList.getElement(i);
			double elementSpannung = currentElement.getSpannungsBetrag();
			
			if(elementSpannung > maxSpannung) {

					// Set the elements color to ORANGE to indicate that the element would have broken
					spannungsbetragColors[i-1] = new GLColor(255, 255, 255);
			}
    	}
    }
	
	/**
	 * Automatically called when the user presses a key
	 * @param e, the key event
	 */
	public void keyTyped(KeyEvent e) {
		
		// Call the super method, else all previous control methods won't work anymore
		super.keyTyped(e);
		
		// Grab the pressed key from the event
		char pressedKey = e.getKeyChar();
		
		// Check if the user pressed a key we support
		switch(pressedKey) {
		
			case 'b':
				
				// Switch the back-figure on and off
				drawBackFigure = !drawBackFigure;
			break;
			
			case 'n':
				
				// Switch the new-figure on and off
				drawNewFigure = !drawNewFigure;
			break;
			
			case 's':
				
				// Switch the "Spannungsbetrag" colors on and off
				drawSpannungsbetragColors = !drawSpannungsbetragColors;
			break;
		}
	}
}
