/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.app;

import java.io.File;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import fhm.edu.fem.animation.CalcAnimation;
import fhm.edu.fem.animation.CalcAnimationSpanner;
import fhm.edu.fem.calc.FEMCalc;
import fhm.edu.fem.calc.FEMCalcCircleMembrane;
import fhm.edu.fem.calc.FEMCalcMembrane;
import fhm.edu.fem.calc.FEMCalcSpanner;
import fhm.edu.fem.calc.FEMManager;
import fhm.edu.fem.constants.ApplicationConstants;
import fhm.edu.fem.constants.PropertyConstants;
import fhm.edu.fem.util.logging.LoggingManager;
import fhm.edu.fem.util.properties.PropertyCache;

/**
 * Class Description:
 * 		This class is starting the animation
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class AnimationStartup {
	
	/* Our local logger object: Null at the beginning, because we need to initialize the LoggingManager first */
	protected static Logger localLogger = null;
	
	/**
	 * Construct a new startup object
	 * @param inFile, FEMIN File
	 * @param args, the given arguments
	 */
	public AnimationStartup(File inFile, String[] args) {
		
		if(localLogger.isEnabledFor(Level.INFO)) {
			localLogger.log(Level.INFO, "Using FEMIN File: " + inFile.getAbsolutePath());
		}
		
	    // Load the file
	    FEMManager.getInstance().load(inFile);
	    
	    // Create the Calculation object
	    FEMCalc calc = null;
	    
	    // Check which way of compilation we need to use
	    if(FEMManager.getInstance().getFEMApproach() == 4) {
	    	
			// FEM4 means we are to calculate Membran (check 4.Exercise)
			calc = new FEMCalcMembrane();

	    } else if(FEMManager.getInstance().getFEMApproach() == 5) {
	    	
	    	// FEM5 means we are to calculate the Circle Membran (check 6.Exercise)
	    	calc = new FEMCalcCircleMembrane();
	    	
		} else if(FEMManager.getInstance().getFEMApproach() == 6) {
			
			// FEM6 - (check 6.Exercise) - Spanner
			calc = new FEMCalcSpanner();
			
		} else {

			// Unknown required compilation
			if (localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR,	"Unknown required way of compilation: " + FEMManager.getInstance().getFEMApproach());
			}

			AnimationStartup.abort(false);
	    }
	    
	    if(FEMManager.getInstance().getFEMApproach() != 6) {
	    
		    // Start the calculation
		    if(!calc.start()) {
		        
	            if(localLogger.isEnabledFor(Level.ERROR)) {
	                localLogger.log(Level.ERROR, "Error while running the calculation !");
	            }
	            
	            AnimationStartup.abort(false);
		    }
	    }
	    
		// Create the calculation object based on the needed FEM Approach
	    CalcAnimation calcAnim = null;
	    
	    if(FEMManager.getInstance().getFEMApproach() == 4 || FEMManager.getInstance().getFEMApproach() == 5) {
	    	
			// FEM4 and 5 means we are to calculate Membran (check 4.Exercise)
			calcAnim = new CalcAnimation(args);

	    } else if(FEMManager.getInstance().getFEMApproach() == 6) {
			
			// FEM6 - (check 6.Exercise) - Spanner
			calcAnim = new CalcAnimationSpanner((FEMCalcSpanner)calc, args);
			
		} else {

			// Unknown required compilation
			if (localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR,	"Unknown required way of animation: " + FEMManager.getInstance().getFEMApproach());
			}

			AnimationStartup.abort(false);
	    }
	    
	    // Start the calculation
	    if(!calcAnim.start()) {
	        
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error while running the animation !");
            }
            
            EigenvaluesStartup.abort(false);
	    }
	    
	    // End of Program execution
	    // System.exit(0);
	}

	/**
	 * Call this method to start the program
	 * @param args, command line arguments
	 */
	public static void main(String[] args) {
		try {
			// Initialize the logging manager
			LoggingManager.setProperties(PropertyCache.getProperties(PropertyConstants.GLOBAL_PROPS));
			
			// Now we can initialize our local logger
			localLogger = LoggingManager.getLogger(AnimationStartup.class.getName());
			
			// Check if we properly received a logger
			if(localLogger != null) {
				
				if(localLogger.isEnabledFor(Level.INFO)) {					
					// Print startup text
					System.out.println(ApplicationConstants.getStartupText());
					
					localLogger.log(Level.INFO, "Starting AnimationStartup ...");
				}				
				
				// Separate the command-line arguments and start the application
				AnimationStartup.start(args);
				
			} else {
				System.err.println("[FATAL] - LoggingManager and/or Logger initialization failed - Aborting !");
			}
		} catch (Exception e) {
			if(localLogger != null) {
				if(localLogger.isEnabledFor(Level.ERROR)) {
					localLogger.log(Level.ERROR, "Running Startup failed!", e);
				}
			} else {
				System.err.println("[FATAL] - Running Startup failed !");
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * Separate the given command-line arguments and start the application
	 * @param args, the command-line arguments
	 */
	protected static void start(String[] args) {
	    
	    // Check if we got all command-line arguments
	    if(args.length < 1) {
	        
	        System.err.println("[FATAL] - Not enough arguments given !");
	        AnimationStartup.abort(true);
	        
	    } else {
	    	
	        // Grab the first command-line argument which should point us at the FEMIN file
	        String inputFileName = args[0];
	        
	        // Create the input file object
	        File inputFile = new File(inputFileName);
	        
	        // Check if the file exists
	        if(!inputFile.exists()) {
	            
	            System.err.println("[FATAL] - Input File: " + inputFileName + " does not exist !");
	            AnimationStartup.abort(true);
	        }
	        
	        // The file exists, create the AnimationStartup object
	        AnimationStartup startup = new AnimationStartup(inputFile, args);
	    }	    
	}
	
	/**
	 * Method called when aborting the program
	 * @param true if help-text should be printed, false otherwise
	 */
	public static void abort(boolean printHelpText) {
	    
	    // Check if the help text should be printed or not
	    if(printHelpText) {
	        
	        // Print the help text
	        System.out.println(ApplicationConstants.getAnimationHelpText());
	    }
	    
	    // Exit the application
	    System.exit(0);
	}
}


