/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.app;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import fhm.edu.fem.calc.CalcEigenvalues;
import fhm.edu.fem.constants.ApplicationConstants;
import fhm.edu.fem.constants.PropertyConstants;
import fhm.edu.fem.util.logging.LoggingManager;
import fhm.edu.fem.util.properties.PropertyCache;

/**
 * Class Description:
 * 		This class is calculating the exact eigenvalues lambda(i,j)
 * 		of quadratic or rectangular membrans depending on the given membran size
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class EigenvaluesStartup {
	
	/* Our local logger object: Null at the beginning, because we need to initialize the LoggingManager first */
	protected static Logger localLogger = null;
	
	/**
	 * Construct a new startup object
	 * @param double a, the length of the membran side left-to-right/right-to-left
	 * @param double b, the length of the membran side down-to-up/up-to-down
	 */
	public EigenvaluesStartup(double a, double b) {

		// Create the calculation object
		CalcEigenvalues calc = new CalcEigenvalues(a, b);
	    
	    // Start the calculation
	    if(!calc.start()) {
	        
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error while running the calculation !");
            }
            
            EigenvaluesStartup.abort(false);
	    }
	    
	    // End of Program execution
	    // System.exit(0);
	}

	/**
	 * Call this method to start the program
	 * @param args, command line arguments
	 */
	public static void main(String[] args) {
		try {
			// Initialize the logging manager
			LoggingManager.setProperties(PropertyCache.getProperties(PropertyConstants.GLOBAL_PROPS));
			
			// Now we can initialize our local logger
			localLogger = LoggingManager.getLogger(EigenvaluesStartup.class.getName());
			
			// Check if we properly received a logger
			if(localLogger != null) {
				
				if(localLogger.isEnabledFor(Level.INFO)) {					
					// Print startup text
					System.out.println(ApplicationConstants.getStartupText());
					
					localLogger.log(Level.INFO, "Starting EigenvaluesStartup ...");
				}				
				
				// Separate the command-line arguments and start the application
				EigenvaluesStartup.start(args);
				
			} else {
				System.err.println("[FATAL] - LoggingManager and/or Logger initialization failed - Aborting !");
			}
		} catch (Exception e) {
			if(localLogger != null) {
				if(localLogger.isEnabledFor(Level.ERROR)) {
					localLogger.log(Level.ERROR, "Running Startup failed!", e);
				}
			} else {
				System.err.println("[FATAL] - Running Startup failed !");
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * Separate the given command-line arguments and start the application
	 * @param args, the command-line arguments
	 */
	protected static void start(String[] args) {
	    
	    // Check if we got all command-line arguments
	    if(args.length < 2) {
	        
	        System.err.println("[FATAL] - Not enough arguments given !");
	        EigenvaluesStartup.abort(true);
	        
	    } else if(args.length > 3) {
	        
	        System.err.println("[FATAL] - Too many arguments given !");
	        EigenvaluesStartup.abort(true);
	        
	    } else {
	        
	        // Grab the first command-line argument which should be the length of side A
	        String lengthSideAString = args[0];
	        
	        // Convert the argument to a double value and check if it is greater than zero
	        double lengthSideA = Double.parseDouble(lengthSideAString);
	        
	        if(lengthSideA <= 0) {
	        	
				if(localLogger.isEnabledFor(Level.ERROR)) {
					localLogger.log(Level.ERROR, "Length of Membran Side A must be > 0 !");
				}
				
				// Abort the program
				EigenvaluesStartup.abort(false);
	        }
	        
	        // Grab the second command-line argument which should be the length of side B
	        String lengthSideBString = args[1];
	        
	        // Convert the argument to a double value and check if it is greater than zero
	        double lengthSideB = Double.parseDouble(lengthSideBString);
	        
	        if(lengthSideB <= 0) {
	        	
				if(localLogger.isEnabledFor(Level.ERROR)) {
					localLogger.log(Level.ERROR, "Length of Membran Side B must be > 0 !");
				}
				
				// Abort the program
				EigenvaluesStartup.abort(false);
	        }
	        
	        // The file exists, create the EigenvaluesStartup object
	        EigenvaluesStartup startup = new EigenvaluesStartup(lengthSideA, lengthSideB);
	    }	    
	}
	
	/**
	 * Method called when aborting the program
	 * @param true if help-text should be printed, false otherwise
	 */
	public static void abort(boolean printHelpText) {
	    
	    // Check if the help text should be printed or not
	    if(printHelpText) {
	        
	        // Print the help text
	        System.out.println(ApplicationConstants.getEigenvaluesHelpText());
	    }
	    
	    // Exit the application
	    System.exit(0);
	}
}

