/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.app;

import java.io.File;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import fhm.edu.fem.calc.FEMManager;
import fhm.edu.fem.constants.ApplicationConstants;
import fhm.edu.fem.constants.PropertyConstants;
import fhm.edu.fem.convert.FEMINConverterFEM2toFEM3;
import fhm.edu.fem.util.logging.LoggingManager;
import fhm.edu.fem.util.properties.PropertyCache;

/**
 * Class Description:
 *		This class can be used when starting the FEMIN FEM2 to FEM3 converter program
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class FEMINConverterStartup {
    
	/* Our local logger object: Null at the beginning, because we need to initialize the LoggingManager first */
	protected static Logger localLogger = null;
	
	/* Tells us if the new file should be saved FEM_PLOTTER compatible */
	protected static boolean FEM_PLOTTER_COMPATIBLE_OUTPUT = true;
	
	/**
	 * Construct a new startup object
	 * @param femInputFile, the File object representing the input file
	 * @param femPlotterCompatible, true if the converted file should be saved compatible to FEM_PLOTTER
	 */
	public FEMINConverterStartup(File femInputFile, boolean femPlotterCompatible) {

	    // Load the file
	    FEMManager.getInstance().load(femInputFile);
	    
	    // Call the converter
	    FEMINConverterFEM2toFEM3 converter = new FEMINConverterFEM2toFEM3();
	    
	    // Convert the file
	    if(!converter.convertFEM2toFEM3()) {
	        
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Conversion failed !");
			}
			
			FEMINConverterStartup.abort(false);
	    }
	    
	    // Write the file
	    File outFile = new File(femInputFile.getAbsoluteFile() + ApplicationConstants.FEM3_FILE_EXTENSION);
	    
	    if(!FEMManager.getInstance().saveFEMIN(outFile, femPlotterCompatible)) {
	        
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Saving the file failed: " + outFile);
			}
			
			FEMINConverterStartup.abort(false);
	    }
	    
	    // Conversion complete
		if(localLogger.isEnabledFor(Level.INFO)) {
			localLogger.log(Level.INFO, "Conversion complete !");
		}
	}

	/**
	 * Call this method to start the FEM program
	 * @param args, command line arguments, none needed
	 */
	public static void main(String[] args) {
		try {
			// Initialize the logging manager
			LoggingManager.setProperties(PropertyCache.getProperties(PropertyConstants.GLOBAL_PROPS));
			
			// Now we can initialize our local logger
			localLogger = LoggingManager.getLogger(FEMINConverterStartup.class.getName());
			
			// Check if we properly received a logger
			if(localLogger != null) {
				
				if(localLogger.isEnabledFor(Level.INFO)) {					
					// Print startup text
					System.out.println(ApplicationConstants.getStartupText());
					
					localLogger.log(Level.INFO, "Starting FEM Converter ...");
				}
				
				// Separate the command-line arguments and start the application
				FEMINConverterStartup.start(args);
				
			} else {
				System.err.println("[FATAL] - LoggingManager and/or Logger initialization failed - Aborting !");
			}
		} catch (Exception e) {
			if(localLogger != null) {
				if(localLogger.isEnabledFor(Level.ERROR)) {
					localLogger.log(Level.ERROR, "Running Startup failed!", e);
				}
			} else {
				System.err.println("[FATAL] - Running Startup failed !");
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * Separate the given command-line arguments and start the application
	 * @param args, the command-line arguments
	 */
	protected static void start(String[] args) {
	    
	    // Check if we got all command-line arguments
	    if(args.length < 1) {
	        
	        System.err.println("[FATAL] - Not enough arguments given !");
	        FEMINConverterStartup.abort(true);
	        
	    } else if(args.length > 1) {
	        
	        System.err.println("[FATAL] - Too many arguments given !");
	        FEMINConverterStartup.abort(true);
	        
	    } else {
	        
	        // Grab the first command-line argument which should point us at the FEMIN file
	        String inputFileName = args[0];
	        
	        // Create the input file object
	        File inputFile = new File(inputFileName);
	        
	        // Check if the file exists
	        if(!inputFile.exists()) {
	            
	            System.err.println("[FATAL] - Input File: " + inputFileName + " does not exist !");
	            FEMINConverterStartup.abort(true);
	        }
	        
	        // The file exists, create the FEMINConverterStart object
	        FEMINConverterStartup startup = new FEMINConverterStartup(inputFile, FEMINConverterStartup.FEM_PLOTTER_COMPATIBLE_OUTPUT);
	    }	    
	}
	
	/**
	 * Method called when aborting the program
	 * @param true if help-text should be printed, false otherwise
	 */
	public static void abort(boolean printHelpText) {
	    
	    // Check if the help text should be printed or not
	    if(printHelpText) {
	        
	        // Print the help text
	        System.out.println(ApplicationConstants.getConverterHelpText());
	    }
	    
	    // Exit the application
	    System.exit(0);
	}
}
