/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.app;

import java.io.File;
import java.io.IOException;

import javax.swing.JFrame;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import fhm.edu.fem.calc.FEMCalc;
import fhm.edu.fem.calc.FEMCalcCircleMembrane;
import fhm.edu.fem.calc.FEMCalcGauss;
import fhm.edu.fem.calc.FEMCalcMembrane;
import fhm.edu.fem.calc.FEMCalcSpanner;
import fhm.edu.fem.calc.FEMCalcTemperature;
import fhm.edu.fem.calc.FEMManager;
import fhm.edu.fem.calc.objects.matrix.Matrix;
import fhm.edu.fem.calc.objects.matrix.MatrixPanel;
import fhm.edu.fem.constants.ApplicationConstants;
import fhm.edu.fem.constants.PropertyConstants;
import fhm.edu.fem.util.logging.LoggingManager;
import fhm.edu.fem.util.properties.PropertyCache;

/**
 * Class Description:
 * 		This class can be used when starting the FEM program
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class FEMStartup {
	
	/* Our local logger object: Null at the beginning, because we need to initialize the LoggingManager first */
	protected static Logger localLogger = null;
	
	/* Tells us if the new file should be saved FEM_PLOTTER compatible */
	protected static boolean FEM_PLOTTER_COMPATIBLE_OUTPUT = true;
	
	/**
	 * Construct a new startup object
	 * @param femInputFile, the File object representing the input file
	 * @param femOutputFile, the File object representing the output file
	 * @param showCompleteStiffnessMatrix, true if the complete-stiffness-matrix should be displayed, false otherwise
	 */
	public FEMStartup(File femInputFile, File femOutputFile, boolean showCompleteStiffnessMatrix) {

		if(localLogger.isEnabledFor(Level.INFO)) {
			localLogger.log(Level.INFO, "Using FEMIN File: " + femInputFile.getAbsolutePath());
			localLogger.log(Level.INFO, "Using FEMOUT File: " + femOutputFile.getAbsolutePath());
		}
		
	    // Load the file
	    FEMManager.getInstance().load(femInputFile);
	    
	    // Create the Calculation object
	    FEMCalc calc = null;
	    
	    // Check which way of compilation we need to use
	    if(FEMManager.getInstance().getFEMApproach() == 2) {
	    	
            // Check if we have any CTabs and need to calculate temperature-stuff
            if(FEMManager.getInstance().getCTabList().size() > 0) {
                
                if(localLogger.isEnabledFor(Level.INFO)) {
                    localLogger.log(Level.INFO, "CTabs Found ! Temperature !");
                    localLogger.log(Level.INFO, "Forcing Compilation: B-VECTOR via GAUSS");
                }
                
    	    	// Temperature calculations
    	    	calc = new FEMCalcTemperature();
                
            } else {
                
                if(localLogger.isEnabledFor(Level.INFO)) {
                    localLogger.log(Level.INFO, "No CTabs found !");
                    localLogger.log(Level.INFO, "Using Compilation: STANDARD");
                }
                
    	    	// Standard way of compilation using S-Matrices
    	    	calc = new FEMCalc();
            }
	    	
	    } else if(FEMManager.getInstance().getFEMApproach() == 3) {
	    	
            // Check if we have any CTabs and need to calculate temperature-stuff
            if(FEMManager.getInstance().getCTabList().size() > 0) {
                
                if(localLogger.isEnabledFor(Level.INFO)) {
                    localLogger.log(Level.INFO, "CTabs Found ! Temperature !");
                    localLogger.log(Level.INFO, "Forcing Compilation: B-VECTOR via GAUSS");
                }
                
    	    	// Temperature calculations
    	    	calc = new FEMCalcTemperature();
                
            } else {
                
                if (localLogger.isEnabledFor(Level.INFO)) {
					localLogger.log(Level.INFO, "No CTabs found !");
					localLogger.log(Level.INFO,	"Using Compilation: GAUSS_QUADRATURE");
				}

				// Compilation via Gauss
				calc = new FEMCalcGauss();
			}

		} else if (FEMManager.getInstance().getFEMApproach() == 4) {
			
			// FEM4 means we are to calculate Membran (check 4.Exercise)
			calc = new FEMCalcMembrane();

		} else if(FEMManager.getInstance().getFEMApproach() == 5) {

			//	FEM5 - (check 6.Exercise) - Circle-Ring- and Circle-Sector-Elements
			calc = new FEMCalcCircleMembrane();
			
		} else if(FEMManager.getInstance().getFEMApproach() == 6) {
			
			// FEM6 - (check 6.Exercise) - Spanner
			calc = new FEMCalcSpanner();
			
		} else {

			// Unknown required compilation
			if (localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR,	"Unknown required way of compilation: ");
			}

			FEMStartup.abort(false);
		}
	    
	    // Start the calculation
	    if(!calc.start()) {
	        
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error while running the calculation !");
            }
            
            FEMStartup.abort(false);
	    }
	    
	    // Save the results
	    if(!FEMManager.getInstance().saveFEMOUT(femOutputFile)) {
	        
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error while saving the file: " + femOutputFile.getName());
            }
            
            FEMStartup.abort(false);
	    }
	    
	    // Save a transformated FEMIN File for usage in either the FEM_PLOTTER or our other Calculation methods
	    if(FEMManager.getInstance().getFEMApproach() == 5) {
	    	
			// Save the tranformed FEMIN File
			String fullPath = femInputFile.getAbsolutePath();
			fullPath += ".conv";
			
			File outputFile = new File(fullPath);
	        
	        try {
                // Create the file
                outputFile.createNewFile();
                
            } catch (IOException e) {
                
                System.err.println("[FATAL] - IOException at FEM5 FEMIN CONVERT file: " + outputFile.getName());
            }
            
            // Call the FEMINSaver through the FEMManager
            FEMManager.getInstance().saveFEMIN(outputFile, FEMStartup.FEM_PLOTTER_COMPATIBLE_OUTPUT);
	    }
	    
	    // In case of FEM6 save a FEMIN File for use in FEM_PLOTTER with the new point coordinates of the spanner
	    if(FEMManager.getInstance().getFEMApproach() == 6) {
	    	
			// Save the tranformed FEMIN File
			String fullPath = femInputFile.getAbsolutePath();
			fullPath += ".result";
			
			File outputFile = new File(fullPath);
	        
	        try {
                // Create the file
                outputFile.createNewFile();
                
            } catch (IOException e) {
                
                System.err.println("[FATAL] - IOException at FEM6 FEMIN CONVERT file: " + outputFile.getName());
            }
            
            // Call the FEMINSaver through the FEMManager
            FEMManager.getInstance().saveFEMIN(outputFile, FEMStartup.FEM_PLOTTER_COMPATIBLE_OUTPUT);
	    }
	    
	    // Show the complete-stiffness-matrix if boolean is true
	    if(showCompleteStiffnessMatrix) {
	        
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Displaying the Complete-Stiffness-Matrix");
            }
            
            Matrix completeStiffnessMatrix = calc.getCompleteStiffnessMatrix();
            
            JFrame frame = new JFrame("Complete-Stiffness-Matrix");
            MatrixPanel mPanel = new MatrixPanel(completeStiffnessMatrix);
            frame.add(mPanel);
            frame.setSize(400, 400);
            frame.setVisible(true);  
	    }
	    
	    // End of Program execution
	    // System.exit(0);
	}

	/**
	 * Call this method to start the FEM program
	 * @param args, command line arguments, none needed
	 */
	public static void main(String[] args) {
		try {
			// Initialize the logging manager
			LoggingManager.setProperties(PropertyCache.getProperties(PropertyConstants.GLOBAL_PROPS));
			
			// Now we can initialize our local logger
			localLogger = LoggingManager.getLogger(FEMStartup.class.getName());
			
			// Check if we properly received a logger
			if(localLogger != null) {
				
				if(localLogger.isEnabledFor(Level.INFO)) {					
					// Print startup text
					System.out.println(ApplicationConstants.getStartupText());
					
					localLogger.log(Level.INFO, "Starting FEM ...");
				}				
				
				// Separate the command-line arguments and start the application
				FEMStartup.start(args);
				
			} else {
				System.err.println("[FATAL] - LoggingManager and/or Logger initialization failed - Aborting !");
			}
		} catch (Exception e) {
			if(localLogger != null) {
				if(localLogger.isEnabledFor(Level.ERROR)) {
					localLogger.log(Level.ERROR, "Running Startup failed!", e);
				}
			} else {
				System.err.println("[FATAL] - Running Startup failed !");
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * Separate the given command-line arguments and start the application
	 * @param args, the command-line arguments
	 */
	protected static void start(String[] args) {
	    
	    // Check if we got all command-line arguments
	    if(args.length < 2) {
	        
	        System.err.println("[FATAL] - Not enough arguments given !");
	        FEMStartup.abort(true);
	        
	    } else if(args.length > 3) {
	        
	        System.err.println("[FATAL] - Too many arguments given !");
	        FEMStartup.abort(true);
	        
	    } else {
	        
	        // Grab the first command-line argument which should point us at the FEMIN file
	        String inputFileName = args[0];
	        
	        // Create the input file object
	        File inputFile = new File(inputFileName);
	        
	        // Check if the file exists
	        if(!inputFile.exists()) {
	            
	            System.err.println("[FATAL] - Input File: " + inputFileName + " does not exist !");
	            FEMStartup.abort(true);
	        }
	        
	        // Grab the second command-line argument which should point us at the FEMOUT file
	        String outputFileName = args[1];
	        
	        // Create the input file object
	        File outputFile = new File(outputFileName);
	        
	        try {
                // Create the file
                outputFile.createNewFile();
                
            } catch (IOException e) {
                
                System.err.println("[FATAL] - IOException at FEMOUT file: " + outputFileName);
            }
            
            // Grab the third command-line argument telling us if the CompleteStiffnessMatrix should be displayed or not (OPTIONAL)
            boolean showCompleteStiffnessMatrix = false;
            
            if(args.length > 2) {
                
                // Any command-line argument is sufficient in that case, no matter what it was
                showCompleteStiffnessMatrix = true;
            }
	        
	        // The file exists, create the FEMStartup object
	        FEMStartup startup = new FEMStartup(inputFile, outputFile, showCompleteStiffnessMatrix);
	    }	    
	}
	
	/**
	 * Method called when aborting the program
	 * @param true if help-text should be printed, false otherwise
	 */
	public static void abort(boolean printHelpText) {
	    
	    // Check if the help text should be printed or not
	    if(printHelpText) {
	        
	        // Print the help text
	        System.out.println(ApplicationConstants.getHelpText());
	    }
	    
	    // Exit the application
	    System.exit(0);
	}
	
	/**
	 * Display the given Matrix in a panel
	 * @param m, the matrix to display
	 */
	public static void displayMatrix(Matrix m) {
        JFrame frame = new JFrame("Matrix - Name: " + m.getName() + " Size: " + m.size() + "x" + m.size());
        MatrixPanel mPanel = new MatrixPanel(m);
        
        frame.setSize(300, 300);
        frame.add(mPanel);
        
        frame.setVisible(true);  
	}
}
