/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.calc;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import fhm.edu.fem.app.EigenvaluesStartup;
import fhm.edu.fem.calc.objects.matrix.MatrixStandard;
import fhm.edu.fem.constants.ApplicationConstants;
import fhm.edu.fem.util.logging.LoggingManager;

/**
 * Class Description:
 *		This class executes the calculation of the exact eigenvalues
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class CalcEigenvalues {

	/* Our local logger object */
	protected static Logger localLogger = LoggingManager.getLogger(CalcEigenvalues.class.getName());
	
	// The membran dimension
	protected double lengthOfSideA = -1;
	protected double lengthOfSideB = -1;
	
	// To remember the start time
	protected long startTime = -1;
	
	// Constant needed for eigenvalue calculation (speeds things up a little)
	protected double piSquared = Math.PI*Math.PI;
	protected double sideASquared = -1;
	protected double sideBSquared = -1;
	
	// Where we store all found eigenvalues
	ArrayList<Double> eigenvalueList = new ArrayList<Double>();
	
    /**
     * Create a new Calculator to get the exact Eigenvalues
     * @param double, the positive length of side A
     * @param double, the positive length of side B
     */
    public CalcEigenvalues(double a, double b) {
        
    	// Check the values
        if(a <= 0) {
        	
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Length of Membran Side A must be > 0 !");
			}
			
			// Abort the program
			EigenvaluesStartup.abort(false);
        }
        
    	// Check the values
        if(b <= 0) {
        	
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Length of Membran Side B must be > 0 !");
			}
			
			// Abort the program
			EigenvaluesStartup.abort(false);
        }
        
        // Remeber the values
        this.lengthOfSideA = a;
        this.lengthOfSideB = b;
        
        // Precalculate some values
        this.sideASquared = a*a;
        this.sideBSquared = b*b;
    }
    
    /**
     * Call this method to start the calculation
     * @return true if all went well, false otherwise
     */
    public boolean start() {
    	
        // Remember the starting time
        this.startTime = System.currentTimeMillis();
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Starting to calculate the exact Eigenvalues ...");
            localLogger.log(Level.INFO, "###############################################");
            localLogger.log(Level.INFO, "# ATTENTION																													#");
            localLogger.log(Level.INFO, "###############################################");
            localLogger.log(Level.INFO, "Please note that this Program is not calculating all Eigenvalues but");
            localLogger.log(Level.INFO, "only 'NEEDED_NUMBER_OF_EIGENVALUES'^2");
            localLogger.log(Level.INFO, "Due to that in some cases it might fail to find some Eigenvalues because");
            localLogger.log(Level.INFO, "normally you would need to calculate 1 to INFINITY Eigenvalues.");
            localLogger.log(Level.INFO, "###############################################");
            localLogger.log(Level.INFO, "Size of Membran (A / B): " + lengthOfSideA + " / " + lengthOfSideB);
        }
        
        // Allocate a Matrix to store the calculated eigenvalues
        MatrixStandard eigenvalueMatrix = new MatrixStandard("Exact Eigenvalues", ApplicationConstants.EXACT_EIGENVALUES_NUMBER);
        
        // Fill the array with eigenvalues
        for(int i = 1; i <= eigenvalueMatrix.size(); i++) {
        	for(int j = 1; j <= eigenvalueMatrix.size(); j++) {
        		
        		// Call the formula for the current eigenvalue
        		eigenvalueMatrix.setValue(i, j, this.calcEigenvalueAt(i, j, this.lengthOfSideA, this.lengthOfSideB));
        	}
        }       
        
        // We calculated the eigenvalues, now sort them properly from small-to-big
       this.eigenvalueList.clear();
        
        double currentSmallestEigenvalue = Double.MAX_VALUE;
        int positionI = -1;
        int positionJ = -1;
        int eigenvalueCounter = 0;
        
        // Run till we found enough eigenvalues
        while(eigenvalueCounter < ApplicationConstants.EXACT_EIGENVALUES_NUMBER) {
        	
        	// Search through the array and find the current smallest eigenvalue
            for(int i = 1; i <= eigenvalueMatrix.size(); i++) {
            	for(int j = 1; j <= eigenvalueMatrix.size(); j++) {
            		
            		// Check if the current eigenvalue is smaller than the one we already found
            		if(eigenvalueMatrix.getValue(i, j) < currentSmallestEigenvalue) {
            			
            			// Remeber it and also it's position in the array
            			currentSmallestEigenvalue = eigenvalueMatrix.getValue(i, j);
            			positionI = i;
            			positionJ = j;
            		}
            	}
            }
            
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Found Eigenvalue #" + eigenvalueCounter + " at position (i/j) : ( " + positionI + " / " + positionJ + " ) : " + currentSmallestEigenvalue);
            }
            
            // Add the eigenvalue to the list
            eigenvalueList.add(new Double(currentSmallestEigenvalue));
            
            // Clear the eigenvalue in the array
            eigenvalueMatrix.setValue(positionI, positionJ, Double.MAX_VALUE);
            
            // Clear the found eigenvalue
            currentSmallestEigenvalue = Double.MAX_VALUE;
            
            eigenvalueCounter++;
        }
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Calculation complete");
        }
        
        // All went well
        return true;
    }
    
    /**
     * Retrieve the list of all found eigenvalues
     * @return List, the eigenvalue list
     */
    public List getEigenvalueList() {
    	return this.eigenvalueList;
    }
    
    /**
     * Calculate the eigenvalue using the formula (using the values stored in this class) : lambda(i,j) = pi^2 * [ (i^2/a^2) + (j^2/b^2) ]
     * @param i, the current i index
     * @param j, the current j index
     * @return double, the calculated exact eigenvalue lambda(i,j)
     */
    protected double calcEigenvalueAt(int i, int j, double a, double b) {
    	
    	return this.piSquared * (((i*i)/(this.sideASquared)) + ((j*j)/(this.sideBSquared)));
    }
}
