/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.calc;

import java.util.List;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import fhm.edu.fem.calc.objects.Border;
import fhm.edu.fem.calc.objects.Calculation;
import fhm.edu.fem.calc.objects.Condition;
import fhm.edu.fem.calc.objects.Element;
import fhm.edu.fem.calc.objects.ElementList;
import fhm.edu.fem.calc.objects.Node;
import fhm.edu.fem.calc.objects.NodeList;
import fhm.edu.fem.calc.objects.Vector;
import fhm.edu.fem.calc.objects.matrix.Matrix;
import fhm.edu.fem.calc.objects.matrix.MatrixFactory;
import fhm.edu.fem.constants.ApplicationConstants;
import fhm.edu.fem.constants.SMatrixConstants;
import fhm.edu.fem.exception.CholeskyFailedException;
import fhm.edu.fem.util.logging.LoggingManager;

/**
 * Class Description:
 *		This class executes the calculation itself
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class FEMCalc 
{
	/* Our local logger object */
	protected static Logger localLogger = LoggingManager.getLogger(FEMCalc.class.getName());
    
	// The FEM Manager where we got our data from
    protected static FEMManager femManager = FEMManager.getInstance();
    
    // The approach constants
    public static final int APPROACH_LINEAR = SMatrixConstants.APPROACH_LINEAR;
    public static final int APPROACH_QUADRATIC = SMatrixConstants.APPROACH_QUADRATIC;
    
    // The element stiffness matrix we are creating
    protected Matrix completeStiffnessMatrix = null;
    
    // Used for time measurements
    protected long startTime = 0;
    protected long tmpTime = 0;
    protected long totalTime = 0;
    
    /**
     * Create a new FEM Calculator with the given approach
     */
    public FEMCalc() {
        
        // Create the complete stiffness matrix
        this.completeStiffnessMatrix = MatrixFactory.createUserDefinedMatrix("Complete Stiffness Matrix" , femManager.getNodeList().size()); 
    }
    
    /**
     * Call this method to start the calculation
     * @return true if all went well, false otherwise
     */
    public boolean start() {
        
        // Remember the starting time
        this.startTime = System.currentTimeMillis();
        
        // Clear the result list of the FEMManager
        femManager.clearResultList();
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Doing the compilation ...");
        }
        
        // Do the compilation
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.compilation()) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error during compilation !");
            }
            
            return false;
        }
        
        // Calc the time needed for the compilation
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Compilation successfully completed: " + this.tmpTime + " ms");
            localLogger.log(Level.INFO, "Size of the Complete-Stiffness-Matrix: " + this.completeStiffnessMatrix.size() + "x" + this.completeStiffnessMatrix.size());
            localLogger.log(Level.INFO, "Required Memory of the Complete-Stiffness-Matrix: " + this.completeStiffnessMatrix.getRequiredMemory() + " bytes");
        }
        
        this.tmpTime = System.currentTimeMillis();
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Testing if the Complete-Stiffness-Matrix is symmetric ...");
        }
        
        // Compilation done, test if the complete stiffness matrix is symmetric
        if(!this.completeStiffnessMatrix.isSymmetric()) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Complete Stiffness Matrix is not symmetric !");
            }
            
            return false;
        }
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Testing if all elements on the main diagonal in the Complete-Stiffness-Matrix are not zero ...");
        }
        
        // Check if the main diagonal of the complete stiffness matrix is not zero
        if(!this.completeStiffnessMatrix.isMainDiagonalNotZero()) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: At least one element on the main diagonal of the complete stiffness matrix is zero !");
            }
            
            return false;
        }
        
        // Calc the time needed for the tests
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Time needed for the tests: " + this.tmpTime + " ms");
        }
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Doing the calculations ...");
        }
        
        // Run through all calculations        
        List<Calculation> calculationList = femManager.getCalculationList();
        
        for(int calcIndex = 0; calcIndex < calculationList.size(); calcIndex++) {
            
            // Remember the time at the start of this calculation
            this.tmpTime = System.currentTimeMillis();
            
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Starting calculation: " + calcIndex);
            }
            
            // Grab calculation
            Calculation calc = calculationList.get(calcIndex);
            
            // Do the calculation
            if(!this.calculate(calc)) {
                
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error during calculation: " + calcIndex);
                }
                
                return false;
            }
            
            // Calculate the time needed to complete the current calculation
            this.tmpTime = System.currentTimeMillis() - this.tmpTime;
            
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Calculation " + calcIndex + " successfully completed: " + this.tmpTime + " ms");
            }
        }
        
        // Calculate the total time
        this.totalTime = System.currentTimeMillis() - this.startTime;

        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "All calculations successfully completed - Total time: " + this.totalTime + " ms");
        }
        
        return true;
    }
    
    /**
     * Do the compilation
     * @return true if all went well, false otherwise
     */
    protected boolean compilation()
    {
        try {
            // Grab the approach from the manager
            int femApproach = femManager.getFEMApproach();
            
            // Grab the list of all elements
            ElementList elementlist = femManager.getElementList();
            
            // Calculate how many percentage one element makes up
            double oneElementPercentage = 100.0 / elementlist.size();
            
            // Calculate how many elements make up one percent
            double onePercentElements = elementlist.size() / 100.0;
            
            // Where we store the current actual percentage
            double currentPercentage = 0.0;
            
            // Run through all elements
            for(int elementIndex = 1; elementIndex <= elementlist.size(); elementIndex++)
            {                
                // Calculate the current percentage of elements already done in the compilation
                currentPercentage = (elementIndex-1) * oneElementPercentage;

                // Make a percentage output about every 5% of completed elements
                if(elementIndex%(onePercentElements*5.0) < 1) {
                    
                    if(localLogger.isEnabledFor(Level.INFO)) {
                        localLogger.log(Level.INFO, "Compilation at " + currentPercentage + " %");
                    }
                }

                // Get the current element
                Element tmpElement =  elementlist.getElement(elementIndex);

                NodeList nodeList =  tmpElement.getNodeList();
                int elementType = -1;
                
                // Get the correct nodes from the element
                Node node1 = nodeList.getNodeInCorrectOrder(1);
                Node node2 = nodeList.getNodeInCorrectOrder(2);
                Node nodeE = null;
                
                // Check which node we need to grab in order to get nodeE
		        if(nodeList.size() == 3 && femApproach == SMatrixConstants.APPROACH_LINEAR) {
		              
		            // Triangle with APPROACH_LINEAR
		            elementType = SMatrixConstants.FIGURE_TRIANGLE;
		            nodeE = nodeList.getNodeInCorrectOrder(3);
		              
		        } else if(nodeList.size() == 4 && femApproach == SMatrixConstants.APPROACH_LINEAR) {
		              
		            // Parallelogram with APPROACH_LINEAR
		            elementType = SMatrixConstants.FIGURE_QUADRAT;
		            nodeE = nodeList.getNodeInCorrectOrder(4);
		              
		        } else if(nodeList.size() == 6 && femApproach == SMatrixConstants.APPROACH_QUADRATIC) {
		              
		            // Triangle with APPROACH_QUADRATIC
		            elementType = SMatrixConstants.FIGURE_TRIANGLE;
		            nodeE = nodeList.getNodeInCorrectOrder(3);
		              
		        } else if(nodeList.size() == 8 && femApproach == SMatrixConstants.APPROACH_QUADRATIC) {
		              
		            // Parallelogram with APPROACH_QUADRATIC
		            elementType = SMatrixConstants.FIGURE_QUADRAT;
		            nodeE = nodeList.getNodeInCorrectOrder(4);
		            
		        } else {

	                if(localLogger.isEnabledFor(Level.ERROR)) {
	                    localLogger.log(Level.ERROR, "Node1 for Element " + elementIndex + " = " + node1+ " Number of Nodes in this Element = " + nodeList.size());
	                }
		        }

                // DEBUG Stuff
                if(localLogger.isEnabledFor(Level.DEBUG)) {
                    localLogger.log(Level.DEBUG, "Node1 for Element " + elementIndex + " = " + node1);
                    localLogger.log(Level.DEBUG, "Node2 for Element " + elementIndex + " = " + node2);
                    localLogger.log(Level.DEBUG, "NodeE for Element " + elementIndex + " = " + nodeE);
                }
                
                // DEBUG Stuff
                if(localLogger.isEnabledFor(Level.DEBUG)) {
                    localLogger.log(Level.DEBUG, "det(J) Calculation = " + (node2.getX() - node1.getX()) + " * " + (nodeE.getY() - node1.getY()) + " - " + (nodeE.getX() - node1.getX()) + " * " + (node2.getY() - node1.getY()));
                }
                
                // Calculate det(J)
                double detJ = (node2.getX() - node1.getX()) * 
                						  (nodeE.getY() - node1.getY()) - 
                						  (nodeE.getX() - node1.getX()) * 
                						  (node2.getY() - node1.getY());
                
                if(localLogger.isEnabledFor(Level.DEBUG)) {
                    localLogger.log(Level.DEBUG, "det(J) for Element " + elementIndex + " = " + detJ);
                }
                
                // Check if det(J) is smaller than 0
                if(detJ < 0) {
                    
                    if(localLogger.isEnabledFor(Level.ERROR)) {
                        localLogger.log(Level.ERROR, "det(J) for Element " + elementIndex + " is negative: " + detJ);
                    }
                    
                    return false;
                }
                
                // Calculate a, b and c of the integral
                double a = (Math.pow(nodeE.getX() - node1.getX(), 2) + Math.pow(nodeE.getY() - node1.getY(), 2)) / detJ; 
                
                double b = (  (nodeE.getX() - node1.getX()) *  (node2.getX() - node1.getX()) ) + ( (nodeE.getY() - node1.getY()) * (node2.getY() - node1.getY()) );      
                b = (b *-1)/detJ;
                
                double c = (Math.pow(node2.getY() - node1.getY(), 2) + Math.pow(node2.getX() - node1.getX(), 2)) /detJ;
                
                if(localLogger.isEnabledFor(Level.DEBUG)) {
                    localLogger.log(Level.DEBUG, "a/b/c for Element " + elementIndex + ": [a=" + a + "] [b=" + b + "] [c=" + c + "]");
                }
                
                // Calculate the element stiffness matrix
                Matrix S1 = SMatrixConstants.getSMatrix(femManager.getFEMApproach(), elementType, 1);
                Matrix S2 = SMatrixConstants.getSMatrix(femManager.getFEMApproach(), elementType, 2);
                Matrix S3 = SMatrixConstants.getSMatrix(femManager.getFEMApproach(), elementType, 3);

                // Calculate: S1*a + S2*b + S3*c
                S1 = S1.mult(a);
                S2 = S2.mult(b);
                S3 = S3.mult(c);
                Matrix elementStiffnessMatrix = (S1.add(S2)).add(S3);
                elementStiffnessMatrix.setName("ElementStiffnessMatrix: " + elementIndex);
                            
                if(localLogger.isEnabledFor(Level.DEBUG)) {
                    localLogger.log(Level.DEBUG, "ElementStiffnessMatrix for Element " + elementIndex + ": " + elementStiffnessMatrix);
                }
                          
                // Run through all nodes in the element to get the index i of the matrix
                for(int indexA = 1; indexA <= elementStiffnessMatrix.size(); indexA++) {
                   
                    // Grab the nodeID from the first node
                    Node tmpNodeA = nodeList.getNodeInCorrectOrder(indexA);
                    int i = tmpNodeA.getId();
                    
                    // Determine the starting valueof indexB for the next lines, depending if we use a MATRIX_STANDARD or MATRIX_HULL
                    int startOfIndexB = -1;
                    
                    if(ApplicationConstants.MATRIX_USE == ApplicationConstants.MATRIX_STANDARD) {
                        
                        // We are using a standard matrix
                        startOfIndexB = 1;
                        
                    } else if(ApplicationConstants.MATRIX_USE == ApplicationConstants.MATRIX_HULL) {
                        
                        // We are using a hull matrix
                        startOfIndexB = indexA;
                    }
                    
                   // Run through all nodes in the element to get the index j of the matrix
                   for(int indexB = startOfIndexB; indexB <= elementStiffnessMatrix.size(); indexB++) {
                       
                       Node tmpNodeB = nodeList.getNodeInCorrectOrder(indexB);
                       int j = tmpNodeB.getId();
                           
                       // Grab the current value from the element stiffness matrix and add it to the value in the complete stiffness matrix
                       double elementValue = elementStiffnessMatrix.getValue(indexA, indexB);
                       double newValue = this.completeStiffnessMatrix.getValue(i, j) + elementValue;                       
                       this.completeStiffnessMatrix.setValue(i, j, newValue);
                   }             
               }
            }
            
            if(localLogger.isEnabledFor(Level.DEBUG)) {
                localLogger.log(Level.DEBUG, "Compilation complete !");
                localLogger.log(Level.DEBUG, "Matrix size: " + this.completeStiffnessMatrix.size());
                localLogger.log(Level.DEBUG, "CompleteStiffnessMatrix: " + this.completeStiffnessMatrix);
            }
            
            return true;
            
        } catch (Exception e) {
            
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Failure during compilation!", e);
			}
			
			return false;
        }
    }
    
    /**
     * Do the calculation
     * @param calc, the calculation object
     * @return true if all went well, false otherwise
     */
    protected boolean calculate(Calculation calc) {
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Starting to clone the complete-stiffness-Matrix ...");
        }

        // Clone the complete stiffness matrix for us to change
        Matrix stiffnessMatrixClone = (Matrix)(this.completeStiffnessMatrix.clone());

        // Create the vectors we need in the calculation
        Vector b  = new Vector("b", this.completeStiffnessMatrix.size());
        
        /**
         * IMPLEMENT THE BORDER CONDITIONS
         * */
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Starting to implement the border conditions ...");
        }
        
        // Grab the conditions from the actual calculation
        List<Condition> conditionList = calc.getConditions();
        
        // Run through all conditions
        for(int i = 0; i < conditionList.size(); i++)
        {
            // Grab the current condition
            Condition condition =  conditionList.get(i);
            
            // Retrieve the value of the condition
            double conditionValue = condition.getValue();
            
            // Get the border affected by this condition and grab all nodes from this border
            Border conditionBorder = condition.getBorder();
            NodeList borderNodeList = conditionBorder.getNodes();

            // Run through all nodes in the border
            for(int j = 1; j <= borderNodeList.size(); j++){
                
                // Grab the current node from the border node list
                Node tmpNode = borderNodeList.getNodeInCorrectOrder(j);

                // Retrieve the ID number of the current node
                int nodeId = tmpNode.getId();
                
                // Run through the vector b
                for(int k = 1; k <= this.completeStiffnessMatrix.size(); k++){
                    
                    // Calculate the new value of b[k] made up of b[k] - ( K[k][nodeID] * conditionValue )
                    double newValue = b.getValue(k) - (stiffnessMatrixClone.getValue(k, nodeId) * conditionValue);    

                    // Set the new value for b
                    b.setValue(k, newValue);     
                }

                // Set the condition value to the proper line in the vector b
                b.setValue(nodeId, conditionValue);
               
                // Fill the column and the line identified by the node ID with zeros
                stiffnessMatrixClone.setLineAndColumn(nodeId, 0);
                
                // Set the value "1" at the main diagonal element identified by K[nodeID][nodeID]
                stiffnessMatrixClone.setValue(nodeId, nodeId, 1);
           }
        }
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Implementing border conditions successfully completed");
            localLogger.log(Level.INFO, "New required Memory of the MODIFIED Complete-Stiffness-Matrix: " + stiffnessMatrixClone.getRequiredMemory() + " bytes");
        }

        /**
         * Calculate the CHOLESKY
         * */
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Starting to calculate the CHOLESKY ...");
        }
        
        Matrix L = null;
        
        try {
            // Compute the CHOLESKY
            L = stiffnessMatrixClone.cholesky();

            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Computed CHOLESKY for Matrix: " + stiffnessMatrixClone.getName());
            }
            
            if(localLogger.isEnabledFor(Level.DEBUG)) {
                localLogger.log(Level.DEBUG, "CHOLESKY: " + L);
            }
            
        } catch (CholeskyFailedException e) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error while computing the CHOLESKY for Matrix: " + stiffnessMatrixClone.getName());
            }
            
            return false;
        }

        /**
         * Calculate the results by solving K*u = b
         * */
        
        // Calculate Vector y from L*y=b
        Vector y = new Vector("y", this.completeStiffnessMatrix.size());
        
        // Run through all element positions in the vector y
        for(int i = 1; i <= this.completeStiffnessMatrix.size(); i++) {
            
            double sum = 0;
            
            // Calculate the sum l[i][k]*y[k] from k=1 to [i-1]
            for(int k = 1; k <= (i-1); k++) {
            
                sum = sum + (L.getValue(i, k) * y.getValue(k));
            }
            
            double newValue = (b.getValue(i) - sum) / L.getValue(i, i);
            y.setValue(i, newValue);
        }
        
        // Solve "L(transponated) * u = y" by calculating Vector u
        Vector u = new Vector("u", this.completeStiffnessMatrix.size());
        
        // Run through all element positions in the vector u
        for(int i = this.completeStiffnessMatrix.size(); i >= 1; i--) {
            
            double sum = 0;
            
            // Calculate the sum l[k][i] * u[k] from k=i+1 to size of the matrix
            for(int k = (i+1); k <= this.completeStiffnessMatrix.size(); k++) {
            
                sum = sum + (L.getValue(k, i) * u.getValue(k));
            }
            
            double newValue = (y.getValue(i) - sum) / L.getValue(i, i);
            u.setValue(i, newValue);
        }    
        
        // Calculation for Vector u has been completed, save the results in the FEMManager
        femManager.addToResultList(u);
        
        if(localLogger.isEnabledFor(Level.DEBUG)) {
            localLogger.log(Level.DEBUG, "Vector u calculation successfull: " + u);
        }

        return true;
    }
    
    /**
     * Retrieve the complete stiffness matrix at the current point of the calculation
     * @return Returns the completeStiffnessMatrix.
     */
    public Matrix getCompleteStiffnessMatrix() {
        return completeStiffnessMatrix;
    }
    
    /**
     * Retrieve the time needed to complete all calculations
     * @return Returns the totalTime.
     */
    public long getTotalTime() {
        return totalTime;
    }
}
