/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.calc;

import java.lang.reflect.Method;
import java.util.Arrays;

import org.apache.log4j.Level;

import fhm.edu.fem.calc.objects.CompilationFunctions;
import fhm.edu.fem.calc.objects.Element;
import fhm.edu.fem.calc.objects.ElementList;
import fhm.edu.fem.calc.objects.MathObject;
import fhm.edu.fem.calc.objects.MembraneCircle;
import fhm.edu.fem.calc.objects.Node;
import fhm.edu.fem.calc.objects.NodeList;
import fhm.edu.fem.calc.objects.Vector;
import fhm.edu.fem.calc.objects.matrix.Matrix;
import fhm.edu.fem.calc.objects.matrix.MatrixFactory;
import fhm.edu.fem.constants.ApplicationConstants;
import fhm.edu.fem.constants.GaussConstants;
import fhm.edu.fem.constants.SMatrixConstants;

/**
 * Class Description:
 *		This class is being used when we have a FEM5 files for Membran
 *		calculations. Any node designated as border node in the FEMIN File
 *		is taken out of the calculations (all other get Alias IDs).
 *
 *		This class is especially used for Circle Membranes in case we want to use
 *		the circle-ring- and circle-sector-element calculation
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class FEMCalcCircleMembrane extends FEMCalcMembrane {
	
	/* The membrane circle object */
	protected MembraneCircle membraneCircle = null;
	
	/* Where we store the bessel values */
	protected double[][] besselValues;
	protected boolean besselValuesInitialized = false;
	
    /**
     * Create a new FEM Calculator for Circle Membrane special Calculations
     */
    public FEMCalcCircleMembrane() {
        
    	// Call the super constructor
        super();
    }
    
    /**
     * Call this method to start the calculation
     * @return true if all went well, false otherwise
     */
    public boolean start() {
    	
        // Remember the starting time
        this.startTime = System.currentTimeMillis();
        
        // Clear the result list of the FEMManager
        femManager.clearResultList();
        
        // Create the membrane circle object
        membraneCircle = new MembraneCircle(femManager);
        membraneCircle.calculateAll();
        
        /**
         * ALIAS ID LIST CREATION
         */
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Creating the Alias List ...");
        }
        
        // Build the alias ID list
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.createAliasList()) {
        	
        	if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error during the creation of the alias list !");
            }
            
            return false;
        }
        
        // Calc the time needed to create the alias list
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Creation of the Alias List successfully completed: " + this.tmpTime + " ms");
        }
        
        /**
         * COMPILATION OF THE COMPLETE STIFFNESS MATRIX
         */
        
        // Do the compilation for the normal Complete-Stiffness-Matrix
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.compileStiffnessMatrix()) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error during compilation of the complete-stiffness-matrix !");
            }
            
            return false;
        }
        
        // Calc the time needed for the compilation
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Compilation of the Complete-Stiffness-Matrix successfully completed: " + this.tmpTime + " ms");
            localLogger.log(Level.INFO, "Size of the Complete-Stiffness-Matrix: " + this.completeStiffnessMatrix.size() + "x" + this.completeStiffnessMatrix.size());
            localLogger.log(Level.INFO, "Required Memory of the Complete-Stiffness-Matrix: " + this.completeStiffnessMatrix.getRequiredMemory() + " bytes");
        }
        
        /**
         * COMPILATION OF THE COMPLETE MASS MATRIX
         */
        
        // Do the compilation for the MassMatrix
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.compileMassMatrix()) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error during compilation of the complete-mass-matrix !");
            }
            
            return false;
        }
        
        // Check if the user wants to shift the point of origin (in case eigenvalue = 0 fails)
        if(ApplicationConstants.MEMBRAN_SHIFT_ORIGIN > 0) {
        	
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Shifting point of Origin for Eigenvalues by: " + ApplicationConstants.MEMBRAN_SHIFT_ORIGIN);
            }
        	
        	// Shift the complete stiffness matrix
        	for(int i = 1; i <= this.completeStiffnessMatrix.size(); i++) {
        		
        		this.completeStiffnessMatrix.setValue(i, i, this.completeStiffnessMatrix.getValue(i, i) + ApplicationConstants.MEMBRAN_SHIFT_ORIGIN);
        	}
        }
        
        // Calc the time needed for the compilation
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Compilation of the Complete-Mass-Matrix successfully completed: " + this.tmpTime + " ms");
            localLogger.log(Level.INFO, "Size of the Complete-Mass-Matrix: " + this.completeMassMatrix.size() + "x" + this.completeMassMatrix.size());
            localLogger.log(Level.INFO, "Required Memory of the Complete-Mass-Matrix: " + this.completeMassMatrix.getRequiredMemory() + " bytes");
        }
        
        /**
         * ALGORITHM DECISION - Jacobi, Rayleigh, RayleighS, SVI1, SVI2, SVI3
         */
        
        // Run the Membran Algorithm (either Jacobi, Rayleigh or SVI)        
        if(ApplicationConstants.MEMBRAN_USE == ApplicationConstants.MEMBRAN_RAYLEIGH || 
           ApplicationConstants.MEMBRAN_USE == ApplicationConstants.MEMBRAN_RAYLEIGH_S &&
           ApplicationConstants.MATRIX_USE == ApplicationConstants.MATRIX_HULL)	{
        	
        	// RAYLEIGH or RAYLEIG_S - ALGORITHM
        	this.tmpTime = System.currentTimeMillis();
        	
        	if(!this.calculateRAY()){
        		
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error during calculation of RAYLEIGH !");
                    
                }    
                return false;
        	}
        	
            // Calc the time needed for the rayleigh calculation
            this.tmpTime = System.currentTimeMillis() - this.tmpTime;
            
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Calculation of Rayleigh successfully completed: " + this.tmpTime + " ms");
            }
            
        } else if(ApplicationConstants.MEMBRAN_USE == ApplicationConstants.MEMBRAN_SVI &&
                  ApplicationConstants.MATRIX_USE == ApplicationConstants.MATRIX_HULL) {
        	
        	// SVI - ALGORITHM
        	this.tmpTime = System.currentTimeMillis();
        	
        	if(!this.calculateSVI()){
        		
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error during calculation of SVI !");
                    
                }    
                return false;
        	}
        	
            // Calc the time needed for the SVI calculation
            this.tmpTime = System.currentTimeMillis() - this.tmpTime;
            
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Calculation of SVI successfully completed: " + this.tmpTime + " ms");
            }
        	
        } else {
        	
        	// JACOBI - ALGORITHM
        	
            // Calculate C
            this.tmpTime = System.currentTimeMillis();
            
            if(!this.calculateC()) {
                
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error during calculation of c !");
                }
                
                return false;
            }
            
            // Calc the time needed for the calculation of C
            this.tmpTime = System.currentTimeMillis() - this.tmpTime;
            
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Calculation of c successfully completed: " + this.tmpTime + " ms");
            }
            
            // Calculate the Jacobi
            this.tmpTime = System.currentTimeMillis();
            
            if(!this.calculateJacobi()) {
                
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error during calculation of Jacobi !");
                }
                
                return false;
            }
            
            // Calc the time needed for the calculation Jacobi
            this.tmpTime = System.currentTimeMillis() - this.tmpTime;
            
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Calculation of Jacobi successfully completed: " + this.tmpTime + " ms");
            }
        }
        
        // Calculate the total time
        this.totalTime = System.currentTimeMillis() - this.startTime;

        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Calculation successfully completed - Total time: " + this.totalTime + " ms");
        }
        
        /**
         * EXACT EIGENVALUES
         */
        
        // Calculate the exact eigenvalues for the circle membrane and display them
        this.calculateExactEigenvalues();
        
        return true;
    }
    
    /**
     * Do the compilation with the Gauss-Quadrature-Integration
     * with it's size depending on the number of inner nodes we still have after creating
     * the alias IDs
     * @return true if all went well, false otherwise
     */
    protected boolean compileStiffnessMatrix() {
    	
    	try {
        	this.completeStiffnessMatrix = MatrixFactory.createUserDefinedMatrix("Complete Stiffness Matrix" , this.aliasIDNodeList.size());
            
        	// Grab the pseudo section/Element list
        	ElementList elementList = femManager.getElementList();

            // The variables to use in the Gauss Quadrature Method call
            double a = 0;
            double b = 1;
            double factor_1 = 1;
            double factor_2 = 1;
            Method h1_1 = null;
            Method h1_2 = null;
            Method h2_1 = null;
            Method h2_2 = null;
            Method q_1 = null;
            Method q_2 = null;
            int z1 = -1;
            int z2 = -1;
    		Vector ax = GaussConstants.get(INNER_POINTS, GaussConstants.VALUE_A);
    		Vector hx = GaussConstants.get(INNER_POINTS, GaussConstants.VALUE_H);
    		Vector ay = GaussConstants.get(OUTER_POINTS, GaussConstants.VALUE_A);
    		Vector hy = GaussConstants.get(OUTER_POINTS, GaussConstants.VALUE_H);
            MathObject firstIntegralResultObject = null;
            MathObject secondIntegralResultObject = null;
                        
            // Run through all elements
            for(int elementIndex = 1; elementIndex <= elementList.size(); elementIndex++)
            {                
            	
            	// Grab the node list from the current section(element)
            	Element element = elementList.getElement(elementIndex);
            	NodeList nodeList = element.getNodeList();
            	
                // The element stiffness matrix to fill
                Matrix elementStiffnessMatrix = null;
                
                // Get the radius and angles
            	int currentSectionIndex = elementIndex%(membraneCircle.getSections()+1);
            	
            	if(elementIndex > membraneCircle.getSections()) {
            		currentSectionIndex++;
            	}
            	
            	double angle1 = membraneCircle.getAngle(currentSectionIndex);
            	double angle2 = Double.MAX_VALUE;
            	
            	// Check if we are at the last section and correct the angle if necessary
            	if(currentSectionIndex == membraneCircle.getSections()) {
            		
            		angle2 = 2*Math.PI;
            		
            	} else {
            		
            		angle2 = membraneCircle.getAngle(currentSectionIndex+1);
            	}

                // Check if the Element is a Triangle or Quadrat
                if(element.getElementType() == SMatrixConstants.FIGURE_QUADRAT) {
                	
                	//
                    // Calculate the "KREIS-RING-ELEMENT"
                    //
                	
                	double radius1 = membraneCircle.getRadius1();
                	double radius2 = membraneCircle.getRadius2();
                	
                    CompilationFunctions.storeRadiusAndAngle(radius1, radius2, angle1, angle2);
                    
            		factor_1 = (angle2 - angle1) / (radius2 - radius1);
            		factor_2 = 1/factor_1;
            		
                	// Quadrat
                	h1_1 = CompilationFunctions.getMethod(CompilationFunctions.QUADRAT_QUADRATIC_N_XI);
            		h2_1 = CompilationFunctions.getMethod(CompilationFunctions.QUADRAT_QUADRATIC_N_XI);
            		q_1 = CompilationFunctions.getMethod(CompilationFunctions.MEMBRANE_CIRCLE_RING_Q_1);
            		
            		h1_2 = CompilationFunctions.getMethod(CompilationFunctions.QUADRAT_QUADRATIC_N_ETA);
            		h2_2 = CompilationFunctions.getMethod(CompilationFunctions.QUADRAT_QUADRATIC_N_ETA);
            		q_2 = CompilationFunctions.getMethod(CompilationFunctions.MEMBRANE_CIRCLE_RING_Q_2);
            		
            		z1 = 8;
            		z2 = 8;
            		
                    // Calculate the first integral
                    firstIntegralResultObject = FEMGaussQuadrature.calcDoubleIntegration(a, b, a, b ,h1_1, h2_1, q_1, z1, z2, ax, hx, ay, hy);

                    // Calculate the second integral
                    secondIntegralResultObject = FEMGaussQuadrature.calcDoubleIntegration(a, b, a, b, h1_2, h2_2, q_2, z1, z2, ax, hx, ay, hy);
                    
                    // Check if both result objects of the first and second integral match our requirements
                    if(!(firstIntegralResultObject instanceof Matrix) || !(secondIntegralResultObject instanceof Matrix)) {
                    	
                    	// At least one of the result objects is not a Matrix => ERROR
                        if(localLogger.isEnabledFor(Level.ERROR)) {
                            localLogger.log(Level.ERROR, "One of the integral result objects is not a Matrix !");
                        }
                        
                        return false;
                    }
                    
                    // Bring in both factors
                    firstIntegralResultObject = ((Matrix)(firstIntegralResultObject)).mult(factor_1);
                    secondIntegralResultObject = ((Matrix)(secondIntegralResultObject)).mult(factor_2);
                    
                    // Combine the first and second integral result objects to one resulting element stiffness matrix
                    elementStiffnessMatrix = ((Matrix)(firstIntegralResultObject)).add((Matrix)(secondIntegralResultObject));
                    elementStiffnessMatrix.setName("ElementStiffnessMatrix: " + elementIndex);                    
                    
                    if(localLogger.isEnabledFor(Level.DEBUG)) {
                        localLogger.log(Level.DEBUG, "ElementStiffnessMatrix for Section " + elementIndex + ": " + elementStiffnessMatrix);
                    }
                	
                } else if(element.getElementType() == SMatrixConstants.FIGURE_TRIANGLE) {
                	
                	//
                    // CALCULATE THE RELATED "KREIS-SEKTOR-ELEMENT"
                    //
                	
                	double radius1 = 0;
                	double radius2 = membraneCircle.getRadius1();
                	
                	if(ApplicationConstants.MEMBRAN_CIRCLE_OVERLAPPING_ELEMENTS) {
                		
                		radius1 = 0;
                		radius2 = membraneCircle.getRadius2();
                	}

                    CompilationFunctions.storeRadiusAndAngle(radius1, radius2, angle1, angle2);
                    
                	// It is in the inner region - "KREIS-SEKTOR-ELEMENT"
            		factor_1 = (angle2 - angle1);
            		factor_2 = 1/factor_1;
                	
            		h1_1 = CompilationFunctions.getMethod(CompilationFunctions.MEMBRANE_CIRCLE_SECTOR_TRIANGLE_QUADRATIC_NXI);
            		h2_1 = CompilationFunctions.getMethod(CompilationFunctions.MEMBRANE_CIRCLE_SECTOR_TRIANGLE_QUADRATIC_NXI);
            		q_1 = CompilationFunctions.getMethod(CompilationFunctions.XI);
            		
            		h1_2 = CompilationFunctions.getMethod(CompilationFunctions.MEMBRANE_CIRCLE_SECTOR_TRIANGLE_QUADRATIC_NETA);
            		h2_2 = CompilationFunctions.getMethod(CompilationFunctions.MEMBRANE_CIRCLE_SECTOR_TRIANGLE_QUADRATIC_NETA_THROUGH_XI);
            		q_2 = CompilationFunctions.getMethod(CompilationFunctions.ONE);
            		
            		z1 = 6;
            		z2 = 6;

            		// Calculate the first integral
                    firstIntegralResultObject = FEMGaussQuadrature.calcDoubleIntegration(a, b, a, b ,h1_1, h2_1, q_1, z1, z2, ax, hx, ay, hy);
                    
                    // Calculate the second integral
                    secondIntegralResultObject = FEMGaussQuadrature.calcDoubleIntegration(a, b, a, b, h1_2, h2_2, q_2, z1, z2, ax, hx, ay, hy);
                    
                    // Check if both result objects of the first and second integral match our requirements
                    if(!(firstIntegralResultObject instanceof Matrix) || !(secondIntegralResultObject instanceof Matrix)) {
                    	
                    	// At least one of the result objects is not a Matrix => ERROR
                        if(localLogger.isEnabledFor(Level.ERROR)) {
                            localLogger.log(Level.ERROR, "One of the integral result objects is not a Matrix !");
                        }
                        
                        return false;
                    }
                    
                    // Bring in both factors
                    firstIntegralResultObject = ((Matrix)(firstIntegralResultObject)).mult(factor_1);
                    secondIntegralResultObject = ((Matrix)(secondIntegralResultObject)).mult(factor_2);
                    
                    // Combine the first and second integral result objects to one resulting element stiffness matrix
                    elementStiffnessMatrix = ((Matrix)(firstIntegralResultObject)).add((Matrix)(secondIntegralResultObject));
                    elementStiffnessMatrix.setName("ElementStiffnessMatrix: " + elementIndex);
                    
                    if(localLogger.isEnabledFor(Level.DEBUG)) {
                        localLogger.log(Level.DEBUG, "ElementStiffnessMatrix for Section " + elementIndex + ": " + elementStiffnessMatrix);
                    }
                	
                } else {
                	
                    if(localLogger.isEnabledFor(Level.ERROR)) {
                        localLogger.log(Level.ERROR, "Unknown Element Type. Node List size: " + nodeList.size() + " Element Type: " + element.getElementType());
                    }
                    
                    return false;
                }
                
                // Run through all nodes in the element to get the index i of the matrix
                for(int indexA = 1; indexA <= elementStiffnessMatrix.size(); indexA++) {
                   
                	// Grab the nodeID from the first node
                    Node tmpNodeA = nodeList.getNodeInCorrectOrder(indexA);
                    int i = tmpNodeA.getAliasId();
                    
                    if(i > 0) {
                    
                        // Determine the starting valueof indexB for the next lines, depending if we use a MATRIX_STANDARD or MATRIX_HULL
                        int startOfIndexB = -1;
                        
                        if(ApplicationConstants.MATRIX_USE == ApplicationConstants.MATRIX_STANDARD) {
                            
                            // We are using a standard matrix
                            startOfIndexB = 1;
                            
                        } else if(ApplicationConstants.MATRIX_USE == ApplicationConstants.MATRIX_HULL) {
                            
                            // We are using a hull matrix
                            startOfIndexB = indexA;
                        }
                        
                       // Run through all nodes in the element to get the index j of the matrix
                       for(int indexB = startOfIndexB; indexB <= elementStiffnessMatrix.size(); indexB++) {
                           
	                        Node tmpNodeB = nodeList.getNodeInCorrectOrder(indexB);
	                        int j = tmpNodeB.getAliasId();
	                        
	                        if(j > 0) {
	                        	
	                         // Grab the current value from the element stiffness matrix and add it to the value in the complete stiffness matrix
	                         double elementValue = elementStiffnessMatrix.getValue(indexA, indexB);
	                         double newValue = this.completeStiffnessMatrix.getValue(i, j) + elementValue;                       
	                         this.completeStiffnessMatrix.setValue(i, j, newValue);
	                        }
                       }
                    }   
                }
            }
            
            if(localLogger.isEnabledFor(Level.DEBUG)) {
                localLogger.log(Level.DEBUG, "Compilation complete !");
                localLogger.log(Level.DEBUG, "Matrix size: " + this.completeStiffnessMatrix.size());
                localLogger.log(Level.DEBUG, "CompleteStiffnessMatrix: " + this.completeStiffnessMatrix);
            }
            
            // Store the stiffness-matrix in the FEMManager for later use
            femManager.setCompleteStiffnessMatrix(this.completeStiffnessMatrix);
            return true;
            
        } catch (Exception e) {
            
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Failure during compilation!", e);
			}
			
			return false;
        }
    }
    
    
    /**
     * Compile the Complete-Mass-Matrix
     * @return true if all went well, false otherwise
     */
    protected boolean compileMassMatrix() {
    	
    	try {
    		// Create the complete mass matrix based on the length of the alias node ID list
            this.completeMassMatrix = MatrixFactory.createUserDefinedMatrix("Complete Mass Matrix" , aliasIDNodeList.size()); 
            
        	// Grab the pseudo section/Element list
        	ElementList elementList = femManager.getElementList();
        	
            // The variables to use in the Gauss Quadrature Method call
            double a = 0;
            double b = 1;
            double factor = 1;
            Method h1 = null;
            Method h2 = null;
            Method q = null;
            int z1 = -1;
            int z2 = -1;
    		Vector ax = GaussConstants.get(INNER_POINTS, GaussConstants.VALUE_A);
    		Vector hx = GaussConstants.get(INNER_POINTS, GaussConstants.VALUE_H);
    		Vector ay = GaussConstants.get(OUTER_POINTS, GaussConstants.VALUE_A);
    		Vector hy = GaussConstants.get(OUTER_POINTS, GaussConstants.VALUE_H);
            MathObject firstIntegralResultObject = null;
                        
            // Run through all elements
            for(int elementIndex = 1; elementIndex <= elementList.size(); elementIndex++)
            {                
            	
            	// Grab the node list from the current section(element)
            	Element element = elementList.getElement(elementIndex);
            	NodeList nodeList = element.getNodeList();
            	
                // The element mass matrix to fill
                Matrix elementMassMatrix = null;
                
                // Get the radius and angles
            	int currentSectionIndex = elementIndex%(membraneCircle.getSections()+1);
            	
            	if(elementIndex > membraneCircle.getSections()) {
            		currentSectionIndex++;
            	}
            	
            	double angle1 = membraneCircle.getAngle(currentSectionIndex);
            	double angle2 = Double.MAX_VALUE;
            	
            	// Check if we are at the last section and correct the angle if necessary
            	if(currentSectionIndex == membraneCircle.getSections()) {
            		
            		angle2 = 2*Math.PI;
            		
            	} else {
            		
            		angle2 = membraneCircle.getAngle(currentSectionIndex+1);
            	}

                // Check if the Element is a Triangle or Quadrat
                if(element.getElementType() == SMatrixConstants.FIGURE_QUADRAT) {
                	
                	//
                    // Calculate the "KREIS-RING-ELEMENT"
                    //
                	
                	double radius1 = membraneCircle.getRadius1();
                	double radius2 = membraneCircle.getRadius2();
                    CompilationFunctions.storeRadiusAndAngle(radius1, radius2, angle1, angle2);

                	factor =(radius2 - radius1) *  (angle2 - angle1);
            		
                	// Quadrat
                	h1 = CompilationFunctions.getMethod(CompilationFunctions.QUADRAT_QUADRATIC_N);
            		h2 = CompilationFunctions.getMethod(CompilationFunctions.QUADRAT_QUADRATIC_N);
            		q = CompilationFunctions.getMethod(CompilationFunctions.MEMBRANE_CIRCLE_RING_Q_1);
            		
            		z1 = 8;
            		z2 = 8;
            		
                    // Calculate the first integral
                    firstIntegralResultObject = FEMGaussQuadrature.calcDoubleIntegration(a, b, a, b ,h1, h2, q, z1, z2, ax, hx, ay, hy);
                    
                    // Check if both result objects of the first and second integral match our requirements
                    if(!(firstIntegralResultObject instanceof Matrix)) {
                    	
                    	// At least one of the result objects is not a Matrix => ERROR
                        if(localLogger.isEnabledFor(Level.ERROR)) {
                            localLogger.log(Level.ERROR, "One of the integral result objects is not a Matrix !");
                        }
                        
                        return false;
                    }
                    
                    // Bring in both factors
                    firstIntegralResultObject = ((Matrix)(firstIntegralResultObject)).mult(factor);
                    
                    // Combine the first and second integral result objects to one resulting element stiffness matrix
                    elementMassMatrix = ((Matrix)(firstIntegralResultObject));
                    elementMassMatrix.setName("ElementMassMatrix: " + elementIndex);

                    if(localLogger.isEnabledFor(Level.DEBUG)) {
                        localLogger.log(Level.DEBUG, "ElementMassMatrix for Section " + elementIndex + ": " + elementMassMatrix);
                    } 
                	
                } else if(element.getElementType() == SMatrixConstants.FIGURE_TRIANGLE) {
                	
                	//
                    // CALCULATE THE RELATED "KREIS-SEKTOR-ELEMENT"
                    //
                	
                	double radius1 = 0;
                	double radius2 = membraneCircle.getRadius1();
                	
                	if(ApplicationConstants.MEMBRAN_CIRCLE_OVERLAPPING_ELEMENTS) {
                		
                		radius1 = 0;
                		radius2 = membraneCircle.getRadius2();
                	}

                    CompilationFunctions.storeRadiusAndAngle(radius1, radius2, angle1, angle2);
                    
                	// It is in the inner region - "KREIS-SEKTOR-ELEMENT"
                    factor = Math.pow(radius2, 2) * (angle2 - angle1);
                	
                	// Triangle
            		h1 = CompilationFunctions.getMethod(CompilationFunctions.MEMBRANE_CIRCLE_SECTOR_TRIANGLE_QUADRATIC_N);
            		h2 = CompilationFunctions.getMethod(CompilationFunctions.MEMBRANE_CIRCLE_SECTOR_TRIANGLE_QUADRATIC_N);
            		q = CompilationFunctions.getMethod(CompilationFunctions.XI);
            		
            		z1 = 6;
            		z2 = 6;

            		// Calculate the first integral
                    firstIntegralResultObject = FEMGaussQuadrature.calcDoubleIntegration(a, b, a, b ,h1, h2, q, z1, z2, ax, hx, ay, hy);
                    
                    // Check if both result objects of the first and second integral match our requirements
                    if(!(firstIntegralResultObject instanceof Matrix)) {
                    	
                    	// At least one of the result objects is not a Matrix => ERROR
                        if(localLogger.isEnabledFor(Level.ERROR)) {
                            localLogger.log(Level.ERROR, "One of the integral result objects is not a Matrix !");
                        }
                        
                        return false;
                    }
                    
                    // Bring in both factors
                    firstIntegralResultObject = ((Matrix)(firstIntegralResultObject)).mult(factor);

                    // Combine the first and second integral result objects to one resulting element stiffness matrix
                    elementMassMatrix = ((Matrix)(firstIntegralResultObject));
                    elementMassMatrix.setName("ElementMassMatrix: " + elementIndex);

                    if(localLogger.isEnabledFor(Level.DEBUG)) {
                        localLogger.log(Level.DEBUG, "ElementMassMatrix for Section " + elementIndex + ": " + elementMassMatrix);
                    }
                } else {
                	
                    if(localLogger.isEnabledFor(Level.ERROR)) {
                        localLogger.log(Level.ERROR, "Unknown Element Type !");
                    }
                    
                    return false;
                }
                
                // Run through all nodes in the element to get the index i of the matrix
                for(int indexA = 1; indexA <= elementMassMatrix.size(); indexA++) {
                   
                	// Grab the nodeID from the first node
                    Node tmpNodeA = nodeList.getNodeInCorrectOrder(indexA);
                    int i = tmpNodeA.getAliasId();
                    
                    if(i > 0) {
                        
                        // Determine the starting valueof indexB for the next lines, depending if we use a MATRIX_STANDARD or MATRIX_HULL
                        int startOfIndexB = -1;
                        
                        if(ApplicationConstants.MATRIX_USE == ApplicationConstants.MATRIX_STANDARD) {
                            
                            // We are using a standard matrix
                            startOfIndexB = 1;
                            
                        } else if(ApplicationConstants.MATRIX_USE == ApplicationConstants.MATRIX_HULL) {
                            
                            // We are using a hull matrix
                            startOfIndexB = indexA;
                        }
                        
                       // Run through all nodes in the element to get the index j of the matrix
                       for(int indexB = startOfIndexB; indexB <= elementMassMatrix.size(); indexB++) {

                            Node tmpNodeB = nodeList.getNodeInCorrectOrder(indexB);
                            int j = tmpNodeB.getAliasId();
                                
                            if(j > 0) {
                              
                                // Grab the current value from the element stiffness matrix and add it to the value in the complete stiffness matrix
                                double elementValue = elementMassMatrix.getValue(indexA, indexB);
                                double newValue = this.completeMassMatrix.getValue(i, j) + elementValue;    
                                this.completeMassMatrix.setValue(i, j, newValue);
                            }
                       }   
                    }
                }
            }
            
            if(localLogger.isEnabledFor(Level.DEBUG)) {
                localLogger.log(Level.DEBUG, "Compilation complete !");
                localLogger.log(Level.DEBUG, "Matrix size: " + this.completeMassMatrix.size());
                localLogger.log(Level.DEBUG, "CompleteMassMatrix: " + this.completeMassMatrix);
            }

            // Store the stiffness-matrix in the FEMManager for later use
            femManager.setCompleteMassMatrix(this.completeMassMatrix);
            return true;
            
        } catch (Exception e) {
            
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Failure during compilation!", e);
			}
			
			return false;
        }
    }
    
    /**
     * Calculate the exact eigenvalues of this circle membrane and
     * display as many as needed, sorted from small to big
     */
    protected void calculateExactEigenvalues() {
    	
    	// Run through all known bessel values and calculate the exact eigenvalue
    	double[] eigenvalues = new double[54];
    	int eigenvalueIndex = 0;
		
    	for(int m = 1; m <= 9; m++) {
    		
    		for(int n = 0; n <= 5; n++) {
    			
    			// Get the bessel value
    			double besselValue = this.getBesselValue(m,n);
    			
    			// Calculate the eigenvalue = (besselValue/r)^2
    			double eigenvalue = Math.pow((besselValue/this.membraneCircle.getRadius2()), 2);
 
    			// Store the calculated eigenvalue
    			eigenvalues[eigenvalueIndex] = eigenvalue;
    			eigenvalueIndex++;
    		}
    	}
    	
    	// Sort the Array
    	Arrays.sort(eigenvalues);
    	
    	// Display the eigenvalues ascending
    	for(int i = 0; i < ApplicationConstants.MEMBRAN_NUMBER_OF_EIGENVALUES; i++) {
    		
    		if(localLogger.isEnabledFor(Level.INFO)) {
    			localLogger.log(Level.INFO, "Exact Eigenvalue #" + (i+1) + " : " + eigenvalues[i]);
    		}
    	}
    }
    
    /**
     * Get the bessel value at the given index
     * as shown in Schwarz Buch Page 140
     * @param, m, the row
     * @param n, the column
     * @return double
     */
    protected double getBesselValue(int m, int n) {
    	
    	// Check if we need to init the bessel values
    	if(!besselValuesInitialized) {
    		
    		this.besselValues = new double[10][6];
    		
    		besselValues[0][0] = Double.NaN;
    		besselValues[0][1] = 0.0;
    		besselValues[0][2] = 0.0;
    		besselValues[0][3] = 0.0;
    		besselValues[0][4] = 0.0;
    		besselValues[0][5] = 0.0;
    		
    		besselValues[1][0] = 2.405;
    		besselValues[1][1] = 3.832;
    		besselValues[1][2] = 5.135;
    		besselValues[1][3] = 6.379;
    		besselValues[1][4] = 7.588;
    		besselValues[1][5] = 8.711;
    		
    		besselValues[2][0] = 5.520;
    		besselValues[2][1] = 7.016;
    		besselValues[2][2] = 8.417;
    		besselValues[2][3] = 9.760;
    		besselValues[2][4] = 11.064;
    		besselValues[2][5] = 12.339;
    		
    		besselValues[3][0] = 8.654;
    		besselValues[3][1] = 10.173;
    		besselValues[3][2] = 11.620;
    		besselValues[3][3] = 13.015;
    		besselValues[3][4] = 14.373;
    		besselValues[3][5] = 15.700;
    		
    		besselValues[4][0] = 11.792;
    		besselValues[4][1] = 13.323;
    		besselValues[4][2] = 14.796;
    		besselValues[4][3] = 16.224;
    		besselValues[4][4] = 17.616;
    		besselValues[4][5] = 18.980;
    		
    		besselValues[5][0] = 14.931;
    		besselValues[5][1] = 16.470;
    		besselValues[5][2] = 17.960;
    		besselValues[5][3] = 19.410;
    		besselValues[5][4] = 20.827;
    		besselValues[5][5] = 22.218;
    		
    		besselValues[6][0] = 18.071;
    		besselValues[6][1] = 19.616;
    		besselValues[6][2] = 21.117;
    		besselValues[6][3] = 22.583;
    		besselValues[6][4] = 24.018;
    		besselValues[6][5] = 25.430;
    		
    		besselValues[7][0] = 21.212;
    		besselValues[7][1] = 22.760;
    		besselValues[7][2] = 34.270;
    		besselValues[7][3] = 25.749;
    		besselValues[7][4] = 27.200;
    		besselValues[7][5] = 28.627;
    		
    		besselValues[8][0] = 24.353;
    		besselValues[8][1] = 25.903;
    		besselValues[8][2] = 27.421;
    		besselValues[8][3] = 28.909;
    		besselValues[8][4] = 30.371;
    		besselValues[8][5] = 31.812;
    		
    		besselValues[9][0] = 27.494;
    		besselValues[9][1] = 29.047;
    		besselValues[9][2] = 30.569;
    		besselValues[9][3] = 32.065;
    		besselValues[9][4] = 33.537;
    		besselValues[9][5] = 34.989;
    		
    		besselValuesInitialized = true;
    	}

    	return besselValues[m][n];
    }
}
