/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.calc;

import java.lang.reflect.Method;

import org.apache.log4j.Level;

import fhm.edu.fem.calc.objects.CompilationFunctions;
import fhm.edu.fem.calc.objects.Element;
import fhm.edu.fem.calc.objects.ElementList;
import fhm.edu.fem.calc.objects.MathObject;
import fhm.edu.fem.calc.objects.Node;
import fhm.edu.fem.calc.objects.NodeList;
import fhm.edu.fem.calc.objects.Vector;
import fhm.edu.fem.calc.objects.matrix.Matrix;
import fhm.edu.fem.constants.ApplicationConstants;
import fhm.edu.fem.constants.GaussConstants;

/**
 * Class Description:
 *		This class extends from the normal FEMCalc class taking care
 *		of the whole calculation. The difference is that in this class
 *		we perform the compilation of the complete-stiffness-matrix with
 *		the Gauss-Quadrature-Integration.
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class FEMCalcGauss extends FEMCalc {
	
	/* Inner and Outer Points ("Stuetzstellen des inneren und aeusseren Integrals") */
	protected static int INNER_POINTS = 4;
	protected static int OUTER_POINTS = 4;

    /**
     * Do the compilation with the Gauss-Quadrature-Integration
     * @return true if all went well, false otherwise
     */
    protected boolean compilation()
    {
        try {
            // Grab the approach from the manager
            int femApproach = femManager.getFEMApproach();
            
            // Grab the list of all elements
            ElementList elementlist = femManager.getElementList();
            
            // The variables to use in the Gauss Quadrature Method call
            double a = 0;
            double b = 1;
            Method h1 = null;
            Method h2 = null;
            Method q = null;
            int z1 = -1;
            int z2 = -1;
    		Vector ax = GaussConstants.get(INNER_POINTS, GaussConstants.VALUE_A);
    		Vector hx = GaussConstants.get(INNER_POINTS, GaussConstants.VALUE_H);
    		Vector ay = GaussConstants.get(OUTER_POINTS, GaussConstants.VALUE_A);
    		Vector hy = GaussConstants.get(OUTER_POINTS, GaussConstants.VALUE_H);
            MathObject firstIntegralResultObject = null;
            MathObject secondIntegralResultObject = null;
            
            // Calculate how many percentage one element makes up
            double oneElementPercentage = 100.0 / elementlist.size();
            
            // Calculate how many elements make up one percent
            double onePercentElements = elementlist.size() / 100.0;
            
            // Where we store the current actual percentage
            double currentPercentage = 0.0;
            
            // Run through all elements
            for(int elementIndex = 1; elementIndex <= elementlist.size(); elementIndex++)
            {                
                // Calculate the current percentage of elements already done in the compilation
                currentPercentage = (elementIndex-1) * oneElementPercentage;

                // Make a percentage output about every 5% of completed elements
                if(elementIndex%(onePercentElements*5.0) < 1) {
                    
                    if(localLogger.isEnabledFor(Level.INFO)) {
                        localLogger.log(Level.INFO, "Compilation at " + currentPercentage + " %");
                    }
                }

                // Get the current element
                Element tmpElement =  elementlist.getElement(elementIndex);
                
                // Store the x- and y-values of the element's nodes in the static data structure in the Class CompilationFunctions
                CompilationFunctions.storeElementXYValues(tmpElement);

                NodeList nodeList =  tmpElement.getNodeList();
                int elementType = -1;
                
                // Take a look which kind of element we have here and fill the input variables for the integration accordingly
                if(nodeList.size() == 6) {
                	
                	// Triangle
            		h1 = CompilationFunctions.getMethod(CompilationFunctions.TRIANGLE_QUADRATIC_H1);
            		h2 = CompilationFunctions.getMethod(CompilationFunctions.TRIANGLE_QUADRATIC_H2);
            		q = CompilationFunctions.getMethod(CompilationFunctions.TRIANGLE_QUADRATIC_JACOBI);
            		z1 = 6;
            		z2 = 6;
                	elementType = GaussConstants.FIGURE_TRIANGLE;
                	
                } else if(nodeList.size() == 8) {
                	
                	// Quadrat
            		h1 = CompilationFunctions.getMethod(CompilationFunctions.QUADRAT_QUADRATIC_H1);
            		h2 = CompilationFunctions.getMethod(CompilationFunctions.QUADRAT_QUADRATIC_H2);
            		q = CompilationFunctions.getMethod(CompilationFunctions.QUADRAT_QUADRATIC_JACOBI);
            		z1 = 8;
            		z2 = 8;
                	elementType = GaussConstants.FIGURE_QUADRAT;
                }
                
                // Calculate the first integral
                firstIntegralResultObject = FEMGaussQuadrature.calcIntegration(a, b, h1, h1, q, z1, z1, ax, hx, ay, hy, elementType);
                
                // Calculate the second integral
                secondIntegralResultObject = FEMGaussQuadrature.calcIntegration(a, b, h2, h2, q, z2, z2, ax, hx, ay, hy, elementType);
                
                // Check if both result objects of the first and second integral match our requirements
                if(!(firstIntegralResultObject instanceof Matrix) || !(secondIntegralResultObject instanceof Matrix)) {
                	
                	// At least one of the result objects is not a Matrix => ERROR
                    if(localLogger.isEnabledFor(Level.ERROR)) {
                        localLogger.log(Level.ERROR, "One of the integral result objects is not a Matrix !");
                    }
                    
                    return false;
                }
                
                // Combine the first and second integral result objects to one resulting element stiffness matrix
                Matrix elementStiffnessMatrix = ((Matrix)(firstIntegralResultObject)).add((Matrix)(secondIntegralResultObject));
                elementStiffnessMatrix.setName("ElementStiffnessMatrix: " + elementIndex);
                            
                if(localLogger.isEnabledFor(Level.DEBUG)) {
                    localLogger.log(Level.DEBUG, "ElementStiffnessMatrix for Element " + elementIndex + ": " + elementStiffnessMatrix);
                }
                          
                // Run through all nodes in the element to get the index i of the matrix
                for(int indexA = 1; indexA <= elementStiffnessMatrix.size(); indexA++) {
                   
                    // Grab the nodeID from the first node
                    Node tmpNodeA = nodeList.getNodeInCorrectOrder(indexA);
                    int i = tmpNodeA.getId();
                    
                    // Determine the starting valueof indexB for the next lines, depending if we use a MATRIX_STANDARD or MATRIX_HULL
                    int startOfIndexB = -1;
                    
                    if(ApplicationConstants.MATRIX_USE == ApplicationConstants.MATRIX_STANDARD) {
                        
                        // We are using a standard matrix
                        startOfIndexB = 1;
                        
                    } else if(ApplicationConstants.MATRIX_USE == ApplicationConstants.MATRIX_HULL) {
                        
                        // We are using a hull matrix
                        startOfIndexB = indexA;
                    }
                    
                   // Run through all nodes in the element to get the index j of the matrix
                   for(int indexB = startOfIndexB; indexB <= elementStiffnessMatrix.size(); indexB++) {
                       
                       Node tmpNodeB = nodeList.getNodeInCorrectOrder(indexB);
                       int j = tmpNodeB.getId();
                           
                       // Grab the current value from the element stiffness matrix and add it to the value in the complete stiffness matrix
                       double elementValue = elementStiffnessMatrix.getValue(indexA, indexB);
                       double newValue = this.completeStiffnessMatrix.getValue(i, j) + elementValue;                       
                       this.completeStiffnessMatrix.setValue(i, j, newValue);
                   }             
               }
            }
            
            if(localLogger.isEnabledFor(Level.DEBUG)) {
                localLogger.log(Level.DEBUG, "Compilation complete !");
                localLogger.log(Level.DEBUG, "Matrix size: " + this.completeStiffnessMatrix.size());
                localLogger.log(Level.DEBUG, "CompleteStiffnessMatrix: " + this.completeStiffnessMatrix);
            }

            return true;
            
        } catch (Exception e) {
            
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Failure during compilation!", e);
			}
			
			return false;
        }
    }
}
