/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.calc;

import java.lang.reflect.Method;
import java.util.Arrays;

import org.apache.log4j.Level;

import fhm.edu.fem.calc.objects.BorderList;
import fhm.edu.fem.calc.objects.CompilationFunctions;
import fhm.edu.fem.calc.objects.Element;
import fhm.edu.fem.calc.objects.ElementList;
import fhm.edu.fem.calc.objects.MathObject;
import fhm.edu.fem.calc.objects.Node;
import fhm.edu.fem.calc.objects.NodeList;
import fhm.edu.fem.calc.objects.Vector;
import fhm.edu.fem.calc.objects.algorithm.Jacobi;
import fhm.edu.fem.calc.objects.algorithm.Rayleigh;
import fhm.edu.fem.calc.objects.algorithm.RayleighS;
import fhm.edu.fem.calc.objects.algorithm.SVI1;
import fhm.edu.fem.calc.objects.algorithm.SVI2;
import fhm.edu.fem.calc.objects.algorithm.SVI3;
import fhm.edu.fem.calc.objects.matrix.ENVVector;
import fhm.edu.fem.calc.objects.matrix.Matrix;
import fhm.edu.fem.calc.objects.matrix.MatrixFactory;
import fhm.edu.fem.calc.objects.matrix.MatrixHull;
import fhm.edu.fem.constants.ApplicationConstants;
import fhm.edu.fem.constants.GaussConstants;
import fhm.edu.fem.exception.CholeskyFailedException;

/**
 * Class Description:
 *		This class is being used when we have a FEM4 files for Membran
 *		calculations. Any node designated as border node in the FEMIN File
 *		is taken out of the calculations (all other get Alias IDs)
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class FEMCalcMembrane extends FEMCalcTemperature {
	
	/* The node list holding all Node which have an alias ID != 0 */
	protected NodeList aliasIDNodeList = new NodeList();
	
    /* The complete mass matrix */
    protected Matrix completeMassMatrix = null;
    
    /* The Matrix L */
    protected Matrix L = null;
    
    /* The Matrix C - on our way to the results */
    protected Matrix C = null;
	
    /**
     * Create a new FEM Calculator for Membrane Calculations
     */
    public FEMCalcMembrane() {
        
    	// Call the super constructor
        super();
    }
    
    /**
     * Call this method to start the calculation
     * @return true if all went well, false otherwise
     */
    public boolean start() {
        
        // Remember the starting time
        this.startTime = System.currentTimeMillis();
        
        // Clear the result list of the FEMManager
        femManager.clearResultList();
        
        /**
         * ALIAS ID LIST CREATION
         */
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Creating the Alias List ...");
        }
        
        // Build the alias ID list
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.createAliasList()) {
        	
        	if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error during the creation of the alias list !");
            }
            
            return false;
        }
        
        // Calc the time needed to create the alias list
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Creation of the Alias List successfully completed: " + this.tmpTime + " ms");
        }
        
        /**
         * COMPILATION OF THE COMPLETE STIFFNESS MATRIX
         */
        
        // Do the compilation for the normal Complete-Stiffness-Matrix
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.compileStiffnessMatrix()) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error during compilation of the complete-stiffness-matrix !");
            }
            
            return false;
        }
        
        // Calc the time needed for the compilation
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Compilation of the Complete-Stiffness-Matrix successfully completed: " + this.tmpTime + " ms");
            localLogger.log(Level.INFO, "Size of the Complete-Stiffness-Matrix: " + this.completeStiffnessMatrix.size() + "x" + this.completeStiffnessMatrix.size());
            localLogger.log(Level.INFO, "Required Memory of the Complete-Stiffness-Matrix: " + this.completeStiffnessMatrix.getRequiredMemory() + " bytes");
        }
        
        /**
         * COMPILATION OF THE COMPLETE MASS MATRIX
         */
        
        // Do the compilation for the MassMatrix
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.compileMassMatrix()) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error during compilation of the complete-mass-matrix !");
            }
            
            return false;
        }
        
        // Check if the user wants to shift the point of origin (in case eigenvalue = 0 fails)
        if(ApplicationConstants.MEMBRAN_SHIFT_ORIGIN > 0) {
        	
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Shifting point of Origin for Eigenvalues by: " + ApplicationConstants.MEMBRAN_SHIFT_ORIGIN);
            }
        	
        	// Shift the complete stiffness matrix
        	for(int i = 1; i <= this.completeStiffnessMatrix.size(); i++) {
        		
        		this.completeStiffnessMatrix.setValue(i, i, this.completeStiffnessMatrix.getValue(i, i) + ApplicationConstants.MEMBRAN_SHIFT_ORIGIN);
        	}
        }
        
        // Calc the time needed for the compilation
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Compilation of the Complete-Mass-Matrix successfully completed: " + this.tmpTime + " ms");
            localLogger.log(Level.INFO, "Size of the Complete-Mass-Matrix: " + this.completeMassMatrix.size() + "x" + this.completeMassMatrix.size());
            localLogger.log(Level.INFO, "Required Memory of the Complete-Mass-Matrix: " + this.completeMassMatrix.getRequiredMemory() + " bytes");
        }
        
        /**
         * ALGORITHM DECISION - Jacobi, Rayleigh, RayleighS, SVI1, SVI2, SVI3
         */
        
        // Run the Membran Algorithm (either Jacobi, Rayleigh or SVI)        
        if(ApplicationConstants.MEMBRAN_USE == ApplicationConstants.MEMBRAN_RAYLEIGH || 
           ApplicationConstants.MEMBRAN_USE == ApplicationConstants.MEMBRAN_RAYLEIGH_S &&
           ApplicationConstants.MATRIX_USE == ApplicationConstants.MATRIX_HULL)	{
        	
        	// RAYLEIGH or RAYLEIG_S - ALGORITHM
        	this.tmpTime = System.currentTimeMillis();
        	
        	if(!this.calculateRAY()){
        		
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error during calculation of RAYLEIGH !");
                    
                }    
                return false;
        	}
        	
            // Calc the time needed for the rayleigh calculation
            this.tmpTime = System.currentTimeMillis() - this.tmpTime;
            
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Calculation of Rayleigh successfully completed: " + this.tmpTime + " ms");
            }
            
        } else if(ApplicationConstants.MEMBRAN_USE == ApplicationConstants.MEMBRAN_SVI &&
                  ApplicationConstants.MATRIX_USE == ApplicationConstants.MATRIX_HULL) {
        	
        	// SVI - ALGORITHM
        	this.tmpTime = System.currentTimeMillis();
        	
        	if(!this.calculateSVI()){
        		
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error during calculation of SVI !");
                    
                }    
                return false;
        	}
        	
            // Calc the time needed for the SVI calculation
            this.tmpTime = System.currentTimeMillis() - this.tmpTime;
            
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Calculation of SVI successfully completed: " + this.tmpTime + " ms");
            }
        	
        } else {
        	
        	// JACOBI - ALGORITHM
        	
            // Calculate C
            this.tmpTime = System.currentTimeMillis();
            
            if(!this.calculateC()) {
                
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error during calculation of c !");
                }
                
                return false;
            }
            
            // Calc the time needed for the calculation of C
            this.tmpTime = System.currentTimeMillis() - this.tmpTime;
            
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Calculation of c successfully completed: " + this.tmpTime + " ms");
            }
            
            // Calculate the Jacobi
            this.tmpTime = System.currentTimeMillis();
            
            if(!this.calculateJacobi()) {
                
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error during calculation of Jacobi !");
                }
                
                return false;
            }
            
            // Calc the time needed for the calculation Jacobi
            this.tmpTime = System.currentTimeMillis() - this.tmpTime;
            
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Calculation of Jacobi successfully completed: " + this.tmpTime + " ms");
            }
        }
        
        // Calculate the total time
        this.totalTime = System.currentTimeMillis() - this.startTime;

        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Calculation successfully completed - Total time: " + this.totalTime + " ms");
        }
        
        return true;
    }

    /**
     * Create the alias List
     * @return true if all went well, false otherwise
     */
    protected boolean createAliasList() {
    	
    	// Grab the list of all nodes
    	NodeList completeNodeList = femManager.getNodeList();
    	
    	// Grab the border list
    	BorderList borderList = femManager.getBorderList();
    	
    	// Make sure the list is empty
    	aliasIDNodeList.clear();
    	
    	// List of all border nodes
    	NodeList borderNodeList = new NodeList();
    	
    	// Run through all borders
    	for(int i = 1; i <= borderList.size(); i++) {

    		borderNodeList.addAll(borderList.getBorder(i).getNodes());
    	}
    	
		// Create the alias IDs based on the current border
		NodeList nodeList = completeNodeList.createAliasIDs(1, borderNodeList);

		// Add the node list we are getting back to the aliasIDNodeList
		aliasIDNodeList.addAll(nodeList);	
    	
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Found " + aliasIDNodeList.size() + " inner Nodes");
        }
        
        // Save the alias ID list in the FEMManager for later use
        femManager.setAliasIDNodeList(this.aliasIDNodeList);
		
    	return true;
    }
    
    /**
     * Do the compilation with the Gauss-Quadrature-Integration
     * with it's size depending on the number of inner nodes we still have after creating
     * the alias IDs
     * @return true if all went well, false otherwise
     */
    protected boolean compileStiffnessMatrix() {
    	
        try {
        	this.completeStiffnessMatrix = MatrixFactory.createUserDefinedMatrix("Complete Stiffness Matrix" , this.aliasIDNodeList.size());
        	
            // Grab the approach from the manager
            int femApproach = femManager.getFEMApproach();
            
            // Grab the list of all elements
            ElementList elementlist = femManager.getElementList();
            
            // The variables to use in the Gauss Quadrature Method call
            double a = 0;
            double b = 1;
            Method h1 = null;
            Method h2 = null;
            Method q = null;
            int z1 = -1;
            int z2 = -1;
    		Vector ax = GaussConstants.get(INNER_POINTS, GaussConstants.VALUE_A);
    		Vector hx = GaussConstants.get(INNER_POINTS, GaussConstants.VALUE_H);
    		Vector ay = GaussConstants.get(OUTER_POINTS, GaussConstants.VALUE_A);
    		Vector hy = GaussConstants.get(OUTER_POINTS, GaussConstants.VALUE_H);
            MathObject firstIntegralResultObject = null;
            MathObject secondIntegralResultObject = null;
            
            // Calculate how many percentage one element makes up
            double oneElementPercentage = 100.0 / elementlist.size();
            
            // Calculate how many elements make up one percent
            double onePercentElements = elementlist.size() / 100.0;
            
            // Where we store the current actual percentage
            double currentPercentage = 0.0;
            
            // Run through all elements
            for(int elementIndex = 1; elementIndex <= elementlist.size(); elementIndex++)
            {                
                // Calculate the current percentage of elements already done in the compilation
                currentPercentage = (elementIndex-1) * oneElementPercentage;

                // Make a percentage output about every 5% of completed elements
                if(elementIndex%(onePercentElements*5.0) < 1) {
                    
                    if(localLogger.isEnabledFor(Level.INFO)) {
                        localLogger.log(Level.INFO, "Compilation - Complete-Stiffness-Matrix at " + ApplicationConstants.convertPercentageToString(currentPercentage) + " %");
                    }
                }

                // Get the current element
                Element tmpElement =  elementlist.getElement(elementIndex);
                
                // Store the x- and y-values of the element's nodes in the static data structure in the Class CompilationFunctions
                CompilationFunctions.storeElementXYValues(tmpElement);

                NodeList nodeList =  tmpElement.getNodeList();
                int elementType = -1;
                
                // Take a look which kind of element we have here and fill the input variables for the integration accordingly
                if(nodeList.size() == 6) {
                	
                	// Triangle
            		h1 = CompilationFunctions.getMethod(CompilationFunctions.TRIANGLE_QUADRATIC_H1);
            		h2 = CompilationFunctions.getMethod(CompilationFunctions.TRIANGLE_QUADRATIC_H2);
            		q = CompilationFunctions.getMethod(CompilationFunctions.TRIANGLE_QUADRATIC_JACOBI);
            		z1 = 6;
            		z2 = 6;
                	elementType = GaussConstants.FIGURE_TRIANGLE;
                	
                } else if(nodeList.size() == 8) {
                	
                	// Quadrat
            		h1 = CompilationFunctions.getMethod(CompilationFunctions.QUADRAT_QUADRATIC_H1);
            		h2 = CompilationFunctions.getMethod(CompilationFunctions.QUADRAT_QUADRATIC_H2);
            		q = CompilationFunctions.getMethod(CompilationFunctions.QUADRAT_QUADRATIC_JACOBI);
            		z1 = 8;
            		z2 = 8;
                	elementType = GaussConstants.FIGURE_QUADRAT;
                }

                // Calculate the first integral
                firstIntegralResultObject = FEMGaussQuadrature.calcIntegration(a, b, h1, h1, q, z1, z1, ax, hx, ay, hy, elementType);
                
                // Calculate the second integral
                secondIntegralResultObject = FEMGaussQuadrature.calcIntegration(a, b, h2, h2, q, z2, z2, ax, hx, ay, hy, elementType);
                
                // Check if both result objects of the first and second integral match our requirements
                if(!(firstIntegralResultObject instanceof Matrix) || !(secondIntegralResultObject instanceof Matrix)) {
                	
                	// At least one of the result objects is not a Matrix => ERROR
                    if(localLogger.isEnabledFor(Level.ERROR)) {
                        localLogger.log(Level.ERROR, "One of the integral result objects is not a Matrix !");
                    }
                    
                    return false;
                }
                
                // Combine the first and second integral result objects to one resulting element stiffness matrix
                Matrix elementStiffnessMatrix = ((Matrix)(firstIntegralResultObject)).add((Matrix)(secondIntegralResultObject));
                elementStiffnessMatrix.setName("ElementStiffnessMatrix: " + elementIndex);
                
                if(localLogger.isEnabledFor(Level.DEBUG)) {
                    localLogger.log(Level.DEBUG, "ElementStiffnessMatrix for Element " + elementIndex + ": " + elementStiffnessMatrix);
                }
                          
                // Run through all nodes in the element to get the index i of the matrix
                for(int indexA = 1; indexA <= elementStiffnessMatrix.size(); indexA++) {
                   
                    // Grab the nodeID from the first node
                    Node tmpNodeA = nodeList.getNodeInCorrectOrder(indexA);
                    int i = tmpNodeA.getAliasId();
                    
                    if(i > 0) {
                    
                        // Determine the starting valueof indexB for the next lines, depending if we use a MATRIX_STANDARD or MATRIX_HULL
                        int startOfIndexB = -1;
                        
                        if(ApplicationConstants.MATRIX_USE == ApplicationConstants.MATRIX_STANDARD) {
                            
                            // We are using a standard matrix
                            startOfIndexB = 1;
                            
                        } else if(ApplicationConstants.MATRIX_USE == ApplicationConstants.MATRIX_HULL) {
                            
                            // We are using a hull matrix
                            startOfIndexB = indexA;
                        }
                        
                       // Run through all nodes in the element to get the index j of the matrix
                       for(int indexB = startOfIndexB; indexB <= elementStiffnessMatrix.size(); indexB++) {
                           
                           Node tmpNodeB = nodeList.getNodeInCorrectOrder(indexB);
                           int j = tmpNodeB.getAliasId();
                           
                           if(j > 0) {
                           	
                            // Grab the current value from the element stiffness matrix and add it to the value in the complete stiffness matrix
                            double elementValue = elementStiffnessMatrix.getValue(indexA, indexB);
                            double newValue = this.completeStiffnessMatrix.getValue(i, j) + elementValue;                       
                            this.completeStiffnessMatrix.setValue(i, j, newValue);
                           }
                       }
                    }             
               }
            }
            
            if(localLogger.isEnabledFor(Level.DEBUG)) {
                localLogger.log(Level.DEBUG, "Compilation complete !");
                localLogger.log(Level.DEBUG, "Matrix size: " + this.completeStiffnessMatrix.size());
                localLogger.log(Level.DEBUG, "CompleteStiffnessMatrix: " + this.completeStiffnessMatrix);
            }
            
            // Store the stiffness-matrix in the FEMManager for later use
            femManager.setCompleteStiffnessMatrix(this.completeStiffnessMatrix);

            return true;
            
        } catch (Exception e) {
            
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Failure during compilation!", e);
			}
			
			return false;
        }
    }
    
    /**
     * Compile the Complete-Mass-Matrix
     * @return true if all went well, false otherwise
     */
    protected boolean compileMassMatrix() {
    	
    	try {
    		// Create the complete mass matrix based on the length of the alias node ID list
            this.completeMassMatrix = MatrixFactory.createUserDefinedMatrix("Complete Mass Matrix" , aliasIDNodeList.size()); 
            
            // Grab the approach from the manager
            int femApproach = femManager.getFEMApproach();
            
            // Grab the list of all elements
            ElementList elementlist = femManager.getElementList();
            
            // The variables to use in the Gauss Quadrature Method call
            double a = 0;
            double b = 1;
            Method h1 = null;
            Method q = null;
            int z1 = -1;
            int z2 = -1;
    		Vector ax = GaussConstants.get(INNER_POINTS, GaussConstants.VALUE_A);
    		Vector hx = GaussConstants.get(INNER_POINTS, GaussConstants.VALUE_H);
    		Vector ay = GaussConstants.get(OUTER_POINTS, GaussConstants.VALUE_A);
    		Vector hy = GaussConstants.get(OUTER_POINTS, GaussConstants.VALUE_H);
            MathObject firstIntegralResultObject = null;
            
            // Calculate how many percentage one element makes up
            double oneElementPercentage = 100.0 / elementlist.size();
            
            // Calculate how many elements make up one percent
            double onePercentElements = elementlist.size() / 100.0;
            
            // Where we store the current actual percentage
            double currentPercentage = 0.0;
            
            // Run through all elements
            for(int elementIndex = 1; elementIndex <= elementlist.size(); elementIndex++)
            {                
                // Calculate the current percentage of elements already done in the compilation
                currentPercentage = (elementIndex-1) * oneElementPercentage;

                // Make a percentage output about every 5% of completed elements
                if(elementIndex%(onePercentElements*5.0) < 1) {
                    
                    if(localLogger.isEnabledFor(Level.INFO)) {
                        localLogger.log(Level.INFO, "Compilation - Complete-Mass-Matrix at " + ApplicationConstants.convertPercentageToString(currentPercentage) + " %");
                    }
                }

                // Get the current element
                Element tmpElement =  elementlist.getElement(elementIndex);
                
                // Store the x- and y-values of the element's nodes in the static data structure in the Class CompilationFunctions
                CompilationFunctions.storeElementXYValues(tmpElement);

                NodeList nodeList =  tmpElement.getNodeList();
                int elementType = -1;
                
                // Take a look which kind of element we have here and fill the input variables for the integration accordingly
                if(nodeList.size() == 6) {
                	
                	// Triangle
            		h1 = CompilationFunctions.getMethod(CompilationFunctions.TRIANGLE_QUADRATIC_N);
            		q = CompilationFunctions.getMethod(CompilationFunctions.TRIANGLE_QUADRATIC_ONETHROUGHJAC);
            		z1 = 6;
            		z2 = 6;
                	elementType = GaussConstants.FIGURE_TRIANGLE;
                	
                } else if(nodeList.size() == 8) {
                	
                	// Quadrat
            		h1 = CompilationFunctions.getMethod(CompilationFunctions.QUADRAT_QUADRATIC_N);
            		q = CompilationFunctions.getMethod(CompilationFunctions.QUADRAT_QUADRATIC_ONETHROUGHJAC);
            		z1 = 8;
            		z2 = 8;
                	elementType = GaussConstants.FIGURE_QUADRAT;
                }
                
                // Calculate the first integral
                firstIntegralResultObject = FEMGaussQuadrature.calcIntegration(a, b, h1, h1, q, z1, z1, ax, hx, ay, hy, elementType);
                
                // Check if both result objects of the first and second integral match our requirements
                if(!(firstIntegralResultObject instanceof Matrix)) {
                	
                	// At least one of the result objects is not a Matrix => ERROR
                    if(localLogger.isEnabledFor(Level.ERROR)) {
                        localLogger.log(Level.ERROR, "One of the integral result objects is not a Matrix !");
                    }
                    
                    return false;
                }
                
                // Combine the first and second integral result objects to one resulting element stiffness matrix
                Matrix elementMassMatrix = (Matrix)(firstIntegralResultObject);

                // Run through all nodes in the element to get the index i of the matrix
                for(int indexA = 1; indexA <= elementMassMatrix.size(); indexA++) {
                   
                    // Grab the nodeID from the first node
                    Node tmpNodeA = nodeList.getNodeInCorrectOrder(indexA);
                    int i = tmpNodeA.getAliasId();
                    
                    if(i > 0) {
                        
                        // Determine the starting valueof indexB for the next lines, depending if we use a MATRIX_STANDARD or MATRIX_HULL
                        int startOfIndexB = -1;
                        
                        if(ApplicationConstants.MATRIX_USE == ApplicationConstants.MATRIX_STANDARD) {
                            
                            // We are using a standard matrix
                            startOfIndexB = 1;
                            
                        } else if(ApplicationConstants.MATRIX_USE == ApplicationConstants.MATRIX_HULL) {
                            
                            // We are using a hull matrix
                            startOfIndexB = indexA;
                        }
                        
                       // Run through all nodes in the element to get the index j of the matrix
                       for(int indexB = startOfIndexB; indexB <= elementMassMatrix.size(); indexB++) {
                           
                           Node tmpNodeB = nodeList.getNodeInCorrectOrder(indexB);
                           int j = tmpNodeB.getAliasId();
                               
                           if(j > 0) {
                             
                               // Grab the current value from the element stiffness matrix and add it to the value in the complete stiffness matrix
                               double elementValue = elementMassMatrix.getValue(indexA, indexB);
                               double newValue = this.completeMassMatrix.getValue(i, j) + elementValue;    
                               this.completeMassMatrix.setValue(i, j, newValue);
                           }
                       }   
                    }
                }            
            }
            
            if(localLogger.isEnabledFor(Level.DEBUG)) {
                localLogger.log(Level.DEBUG, "Compilation complete !");
                localLogger.log(Level.DEBUG, "Matrix size: " + this.completeMassMatrix.size());
                localLogger.log(Level.DEBUG, "CompleteMassMatrix: " + this.completeMassMatrix);
            }

            // Store the stiffness-matrix in the FEMManager for later use
            femManager.setCompleteMassMatrix(this.completeMassMatrix);
           
            return true;
            
        } catch (Exception e) {
            
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Failure during compilation!", e);
			}
			
			return false;
        }
    }
    
    /**
     * Calculate the RAYLEIGH ALGORITHM
     * @return true if all went well, false otherwise
     */
    protected boolean calculateRAY() {
    	    	        
    	// Where we store the eigenvalues returned by the algorithm
        Vector eigenvalueVector = null;
        
        // Where we store the eigenvectors returned by the algorithm
        double[][] eigenvectorArray;
        
        // Grab all necessary data from the complete-stiffness- and complete-mass-matrix
    	ENVVector envMass = (((MatrixHull)completeMassMatrix).getENV()).getData();
    	ENVVector envStiff = (((MatrixHull)completeStiffnessMatrix).getENV()).getData();
    	int[] diagPointer = ((MatrixHull)completeMassMatrix).getDiagPointer();
    	
    	// Transform the ENVVector objects into arrays. The arrays MUST start at index 1
    	double[] massMatrixData = new double[envMass.size()+1];
    	double[] stiffnessMatrixData = new double[envStiff.size()+1];

    	for(int i = 1; i <= envMass.size(); i++) {
    		
    		massMatrixData[i] = ((Double)(envMass.get(i-1))).doubleValue();
    		stiffnessMatrixData[i] = ((Double)(envStiff.get(i-1))).doubleValue();
    	}
        
        // Check if we need to use the RAYLEIGH or RAYLEIGH_S
        if(ApplicationConstants.MEMBRAN_USE == ApplicationConstants.MEMBRAN_RAYLEIGH) {
        	
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Starting to calculate the normal Rayleigh Algorithm");
            }
            
            // Create the Rayleigh object
        	Rayleigh ray = new Rayleigh();
        	
        	// Start the calculation
        	if(!ray.compute(this.completeMassMatrix.size(), stiffnessMatrixData, massMatrixData, diagPointer,
        					ApplicationConstants.MEMBRAN_NUMBER_OF_EIGENVALUES, 
        					ApplicationConstants.MEMBRAN_ALGORITHM_EXACTNESS, 
        					ApplicationConstants.MEMBRAN_MACHINE_PRECISION)) {
        		
                if(localLogger.isEnabledFor(Level.INFO)) {
                    localLogger.log(Level.INFO, "Error while calculating the Rayleigh Algorithm !");
                }
                
                return false;
        	}
        	
        	// Grab the results from the object
        	eigenvectorArray = ray.getEigenvectorArray();
        	eigenvalueVector = ray.getEigenvalueVector();
        	
        } else if(ApplicationConstants.MEMBRAN_USE == ApplicationConstants.MEMBRAN_RAYLEIGH_S) {
        	
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Starting to calculate the Rayleigh-S Algorithm");
            }
        	
            // Create the Rayleigh-S Object
        	RayleighS rayS = new RayleighS();
        	
        	// Start the calculation
        	if(!rayS.compute(this.completeMassMatrix.size(), stiffnessMatrixData, massMatrixData, diagPointer,
        					 ApplicationConstants.MEMBRAN_NUMBER_OF_EIGENVALUES, 
        					 ApplicationConstants.MEMBRAN_ALGORITHM_EXACTNESS, 
        					 ApplicationConstants.MEMBRAN_MACHINE_PRECISION,
        					 ApplicationConstants.MEMBRAN_ALGORITHM_MAX_ITERATIONS)) {
        		
                if(localLogger.isEnabledFor(Level.INFO)) {
                    localLogger.log(Level.INFO, "Error while calculating the Rayleigh-S Algorithm !");
                }
                
                return false;
        	}
        	
        	// Grab the results from the object
        	eigenvectorArray = rayS.getEigenvectorArray();
        	eigenvalueVector = rayS.getEigenvalueVector();
            
        } else {
        	
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Unknown Rayleigh Algorithm to use, aborting !");
            }
            
            return false;
        }
        
    	// Store the eigenvalue vector in the FEMManager
    	femManager.setEigenvalueVector(eigenvalueVector);
    	
        // Calculate u for every Eigenvalue (these are stored on the diagonal of the eigenvalueMatrix)
    	for(int eigenvalueIndex = 1; eigenvalueIndex <= ApplicationConstants.MEMBRAN_NUMBER_OF_EIGENVALUES; eigenvalueIndex++) {
    		
    		// Grab the eigenvalue from the matrix
    		double eigenvalue = eigenvalueVector.getValue(eigenvalueIndex);
    		
            // Check if the user wanted to shift the point of origin for the eigenvalues (in case eigenvalue = 0 fails)
            if(ApplicationConstants.MEMBRAN_SHIFT_ORIGIN > 0) {
            	
                if(localLogger.isEnabledFor(Level.INFO)) {
                    localLogger.log(Level.INFO, "Shifting point of Origin back by: " + ApplicationConstants.MEMBRAN_SHIFT_ORIGIN);
                }
            	
            	// Shift the eigenvalue back
            	eigenvalue = eigenvalue - ApplicationConstants.MEMBRAN_SHIFT_ORIGIN;
            }
    		
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Processing Eigenvalue #" + eigenvalueIndex + " : " + eigenvalue);
            }
            
            // Build u(eigenvector) for the current eigenvalue
            Vector u = new Vector("u", eigenvectorArray.length);
            
            for(int eigenvectorIndex = 1; eigenvectorIndex <= u.size(); eigenvectorIndex++) {
            	
            	u.setValue(eigenvectorIndex, eigenvectorArray[eigenvectorIndex-1][eigenvalueIndex-1]);
            }
            
            // Transform the vector in a full result vector and store it in the FEMManager
            this.postCalculationTransformations(u);
    	}
    	
    	return true;
    }
    
    /**
     * Calculate the SVI ALGORITHM
     * @return true if all went well, false otherwise
     */
    protected boolean calculateSVI() {
        
    	// Where we store the eigenvalues returned by the algorithm
        Vector eigenvalueVector = null;
        
        // Where we store the eigenvectors returned by the algorithm
        double[][] eigenvectorArray;
        
        // Grab all necessary data from the complete-stiffness- and complete-mass-matrix
    	ENVVector envMass = (((MatrixHull)completeMassMatrix).getENV()).getData();
    	ENVVector envStiff = (((MatrixHull)completeStiffnessMatrix).getENV()).getData();
    	int[] diagPointer = ((MatrixHull)completeMassMatrix).getDiagPointer();
    	
    	// Transform the ENVVector objects into arrays. The arrays MUST start at index 1
    	double[] massMatrixData = new double[envMass.size()+1];
    	double[] stiffnessMatrixData = new double[envStiff.size()+1];

    	for(int i = 1; i <= envMass.size(); i++) {
    		
    		massMatrixData[i] = ((Double)(envMass.get(i-1))).doubleValue();
    		stiffnessMatrixData[i] = ((Double)(envStiff.get(i-1))).doubleValue();
    	}
    	
    	// Call the proper SVI 1, 2 or 3
    	if(ApplicationConstants.MEMBRAN_SVI_USE == ApplicationConstants.MEMBRAN_SVI_1) {
    		
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Starting to calculate the SVI-1 Algorithm");
            }
    		
            // Create the SVI1 Object
        	SVI1 svi1 = new SVI1();
    		
        	// Start the calculation
        	if(!svi1.compute(this.completeMassMatrix.size(), stiffnessMatrixData, massMatrixData, diagPointer,
							 ApplicationConstants.MEMBRAN_NUMBER_OF_EIGENVALUES,
							 ApplicationConstants.MEMBRAN_NUMBER_OF_EIGENVALUES,
							 ApplicationConstants.MEMBRAN_ALGORITHM_EXACTNESS,
							 ApplicationConstants.MEMBRAN_ALGORITHM_MAX_ITERATIONS)) {
        		
                if(localLogger.isEnabledFor(Level.INFO)) {
                    localLogger.log(Level.INFO, "Error while calculating the SVI-1 Algorithm !");
                }
                
                return false;
        	}
	
        	// Grab the results from the object
        	eigenvectorArray = svi1.getEigenvectorArray();
        	eigenvalueVector = svi1.getEigenvalueVector();
    		
    	} else if(ApplicationConstants.MEMBRAN_SVI_USE == ApplicationConstants.MEMBRAN_SVI_2) {
    		
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Starting to calculate the SVI-2 Algorithm");
            }
    		
            // Create the SVI2 Object
        	SVI2 svi2 = new SVI2();
    		
        	// Start the calculaton
        	if(!svi2.compute(this.completeMassMatrix.size(), ApplicationConstants.MEMBRAN_NUMBER_OF_EIGENVALUES,
        					 diagPointer, stiffnessMatrixData, massMatrixData, 
        					 ApplicationConstants.MEMBRAN_ALGORITHM_EXACTNESS, 
        					 ApplicationConstants.MEMBRAN_ALGORITHM_MAX_ITERATIONS)) {
        		
                if(localLogger.isEnabledFor(Level.INFO)) {
                    localLogger.log(Level.INFO, "Error while calculating the SVI-2 Algorithm !");
                }
                
                return false;
        	}
	
        	// Grab the results from the object
        	eigenvectorArray = svi2.getEigenvectorArray();
        	eigenvalueVector = svi2.getEigenvalueVector();
    		
    	} else if(ApplicationConstants.MEMBRAN_SVI_USE == ApplicationConstants.MEMBRAN_SVI_3) {
    		
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Starting to calculate the SVI-3 Algorithm");
            }
    		
            // Create the SVI3 Object
        	SVI3 svi3 = new SVI3();
    		
        	// Start the calculation
        	if(!svi3.compute(this.completeMassMatrix.size(), stiffnessMatrixData, massMatrixData, diagPointer,
							 ApplicationConstants.MEMBRAN_NUMBER_OF_EIGENVALUES,
							 ApplicationConstants.MEMBRAN_NUMBER_OF_EIGENVALUES, 
							 ApplicationConstants.MEMBRAN_ALGORITHM_MAX_ITERATIONS, 
							 ApplicationConstants.MEMBRAN_ALGORITHM_EXACTNESS)) {
        		
                if(localLogger.isEnabledFor(Level.INFO)) {
                    localLogger.log(Level.INFO, "Error while calculating the SVI-3 Algorithm !");
                }
                
                return false;
        	}
	
        	// Grab the results from the object
        	eigenvectorArray = svi3.getEigenvectorArray();
        	eigenvalueVector = svi3.getEigenvalueVector();
    		
    	} else {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Unknown SVI Algorithm to use, aborting !");
            }
            
            return false;
    	}
    	
    	// Store the eigenvalue vector in the FEMManager
    	femManager.setEigenvalueVector(eigenvalueVector);

        // Calculate u for every Eigenvalue (these are stored on the diagonal of the eigenvalueMatrix)
    	for(int eigenvalueIndex = 1; eigenvalueIndex <= ApplicationConstants.MEMBRAN_NUMBER_OF_EIGENVALUES; eigenvalueIndex++) {
    		
    		// Grab the eigenvalue from the matrix
    		double eigenvalue = eigenvalueVector.getValue(eigenvalueIndex);
    		
            // Check if the user wanted to shift the point of origin for the eigenvalues (in case eigenvalue = 0 fails)
            if(ApplicationConstants.MEMBRAN_SHIFT_ORIGIN > 0) {
            	
                if(localLogger.isEnabledFor(Level.INFO)) {
                    localLogger.log(Level.INFO, "Shifting point of Origin back by: " + ApplicationConstants.MEMBRAN_SHIFT_ORIGIN);
                }
            	
            	// Shift the eigenvalue back
            	eigenvalue = eigenvalue - ApplicationConstants.MEMBRAN_SHIFT_ORIGIN;
            }
    		
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Processing Eigenvalue #" + eigenvalueIndex + " : " + eigenvalue);
            }
            
            // Build u(eigenvector) for the current eigenvalue
            Vector u = new Vector("u", eigenvectorArray.length);
            
            for(int eigenvectorIndex = 1; eigenvectorIndex <= u.size(); eigenvectorIndex++) {
            	
            	u.setValue(eigenvectorIndex, eigenvectorArray[eigenvectorIndex-1][eigenvalueIndex-1]);
            } 
            
            // Transform the vector in a full result vector and store it in the FEMManager
            this.postCalculationTransformations(u);
    	}
    	
    	return true;
    }
    
    
    /**
     * PART OF THE JACBO ALGORITHM - Calculate C
     * @return true if all went well, false otherwise
     */
    protected boolean calculateC() {
    	
    	// A*x - eigenwert*B*x = 0
    	// K*u - eigenwert*M*u = 0
    	
        /**
         * Calculate the CHOLESKY
         * Calculate B = L*L^T (B = Complete-Mass-Matrix)
         * */
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Jacobi - Starting to calculate the CHOLESKY ...");
        }
        
        L = null;
        
        try {
            // Compute the CHOLESKY
            L = this.completeMassMatrix.cholesky();

            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Computed CHOLESKY for Matrix: " + this.completeMassMatrix.getName());
            }
            
            if(localLogger.isEnabledFor(Level.DEBUG)) {
                localLogger.log(Level.DEBUG, "CHOLESKY: " + L);
            }
            
        } catch (CholeskyFailedException e) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error while computing the CHOLESKY for Matrix: " + this.completeMassMatrix.getName());
            }
            
            return false;
        }
        
        /**
         * Calculate H = A*(L^T)^-1
         * C = L^-1 * A * (L^T)^-1 
         * via Formula 5.11
         */
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Jacobi - Starting to calculate Matrix H");
        }
        
        Matrix H = MatrixFactory.createUserDefinedMatrix("H", L.size());
        
        // Calculate how many percentage one line makes up
        double oneLinePercentage = 100.0 / H.size();
        
        // Calculate how many lines make up one percent
        double onePercentLines = H.size() / 100.0;
        
        // Where we store the current actual percentage
        double currentPercentage = 0.0;
        
        // Calculate : Formula 5.11
        for(int k = 1; k <= H.size(); k++) {
        	
            // Calculate the current percentage of line already done in the CHOLESKY
            currentPercentage = (k-1) * oneLinePercentage;

            // Make a percentage output about every 2% of completed lines
            if(k%(onePercentLines*2.0) < 1) {
                
                if(localLogger.isEnabledFor(Level.INFO)) {
                    localLogger.log(Level.INFO, "Calculation - Matrix H at " + ApplicationConstants.convertPercentageToString(currentPercentage) + " %");
                }
            }
        	
        	for(int i = k; i <= H.size(); i++) {
        		
        		double a = this.completeStiffnessMatrix.getValue(i, k);
        		
        		// Build the sum from j=1 till (k-1) for h(i,j)*l(k,j);
        		double sum = 0;
        		
        		for(int j = 1; j <= (k-1); j++) {
        			
        			sum += (H.getValue(i, j)*L.getValue(k, j));
        		}
        		
        		// Calculate (a-sum)/l(k,k)
        		double result = (a-sum)/L.getValue(k, k);
        		H.setValue(i, k, result);
        		H.setValue(k, i, result);
        	}
        }
        
        /**
         * Calculate C = L^-1 * H
         * via Formula 5.13
         */
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Jacobi - Starting to calculate Matrix C");
        }
        
        C = MatrixFactory.createUserDefinedMatrix("C", L.size());
        
        // Calculate how many percentage one line makes up
        oneLinePercentage = 100.0 / H.size();
        
        // Calculate how many lines make up one percent
        onePercentLines = H.size() / 100.0;
        
        // Where we store the current actual percentage
        currentPercentage = 0.0;
        
        // Calculate : Formula 5.13
        for(int i = 1; i <= H.size(); i++) {
        	
            // Calculate the current percentage of line already done in the CHOLESKY
            currentPercentage = (i-1) * oneLinePercentage;

            // Make a percentage output about every 2% of completed lines
            if(i%(onePercentLines*2.0) < 1) {
                
                if(localLogger.isEnabledFor(Level.INFO)) {
                    localLogger.log(Level.INFO, "Calculation - Matrix C at " + ApplicationConstants.convertPercentageToString(currentPercentage) + " %");
                }
            }
        	
        	for(int k = 1; k <= i; k++) {
        		
        		double h = H.getValue(i, k);
        		
        		// Build the sum from j=1 till (k-1) for l(i,j)*c(k,j);
        		double sum1 = 0;
        		
        		for(int j = 1; j <= (k-1); j++) {
        			
        			sum1 += (L.getValue(i, j)*C.getValue(k, j));
        		}
        		
        		// Build the sum from j=k till (i-1) for l(i,j)*c(j,k);
        		double sum2 = 0;
        		
        		for(int j = k; j <= (i-1); j++) {
        			
        			sum2 += (L.getValue(i, j)*C.getValue(j, k));
        		}
        		
        		// Calculate (h-sum1-sum2)/l(i,i)
        		double result = (h-sum1-sum2)/L.getValue(i, i);
        		C.setValue(i, k, result);
        		C.setValue(k, i, result);
        	}
        }
    	
    	return true;
    }
    
    /**
     * Calculate the Jacobi, you need to calculate the Matrix C first
     * @return true if all went well, false otherwise
     */
    protected boolean calculateJacobi() {   	
    	
    	// Result Matrix where the eigenvectors are stored
    	double[][] eigenvectorArray = null;
    	Vector eigenvalueVector = null;
    	
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Starting to calculate the Jacobi Algorithm");
        }
    	
    	// Create the Jacobi Object
    	Jacobi jacobi = new Jacobi();

    	// Start the calculation
		if(!jacobi.compute(C.size(), ApplicationConstants.MEMBRAN_MACHINE_PRECISION, C.asArray(), ApplicationConstants.MEMBRAN_ORDER)) {
			
	        if(localLogger.isEnabledFor(Level.INFO)) {
	            localLogger.log(Level.INFO, "Error while calculating the Jacobi Algorithm !");
	        }
	        
	        return false;
		}
		
    	// Grab the results from the object
    	eigenvectorArray = jacobi.getEigenvectorArray();
    	eigenvalueVector = jacobi.getEigenvalueVector();
    	
    	// Store a shorter eigenvalue vector in the FEMManager (because we calculated all eigenvalues)
    	Vector shortEigenvalueVector = new Vector("Eigenvalues", ApplicationConstants.MEMBRAN_NUMBER_OF_EIGENVALUES);
    	
    	for(int i = 1; i <= ApplicationConstants.MEMBRAN_NUMBER_OF_EIGENVALUES; i++) {
    		
    		shortEigenvalueVector.setValue(i, eigenvalueVector.getValue(i));
    	}
    	
    	femManager.setEigenvalueVector(shortEigenvalueVector);
    	    	
    	// Calculate u for every Eigenvalue (these are stored on the diagonal of the eigenvalueMatrix)
    	for(int eigenvalueIndex = 1; eigenvalueIndex <= ApplicationConstants.MEMBRAN_NUMBER_OF_EIGENVALUES; eigenvalueIndex++) {
    		
    		// Grab the eigenvalue from the matrix
    		double eigenvalue = eigenvalueVector.getValue(eigenvalueIndex);
    		
            // Check if the user wanted to shift the point of origin for the eigenvalues (in case eigenvalue = 0 fails)
            if(ApplicationConstants.MEMBRAN_SHIFT_ORIGIN > 0) {
            	
                if(localLogger.isEnabledFor(Level.INFO)) {
                    localLogger.log(Level.INFO, "Shifting point of Origin back by: " + ApplicationConstants.MEMBRAN_SHIFT_ORIGIN);
                }
            	
            	// Shift the eigenvalue back
            	eigenvalue = eigenvalue - ApplicationConstants.MEMBRAN_SHIFT_ORIGIN;
            }
    		
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Processing Eigenvalue #" + eigenvalueIndex + " : " + eigenvalue);
            }
            
            // Build y(eigenvector) for the current eigenvalue
            Vector y = new Vector("y", eigenvectorArray.length);
            
            for(int eigenvectorIndex = 1; eigenvectorIndex <= y.size(); eigenvectorIndex++) {
            	
            	y.setValue(eigenvectorIndex, eigenvectorArray[eigenvectorIndex-1][eigenvalueIndex-1]);
            }
            
            // Solve "L(transponated) * u = y" by calculating Vector u
            Vector u = new Vector("u", this.completeMassMatrix.size());
            
            // Run through all element positions in the vector u
            for(int i = this.completeMassMatrix.size(); i >= 1; i--) {
                
                double sum = 0;
                
                // Calculate the sum l[k][i] * u[k] from k=i+1 to size of the matrix
                for(int k = (i+1); k <= this.completeMassMatrix.size(); k++) {
                
                    sum = sum + (L.getValue(k, i) * u.getValue(k));
                }
                
                double newValue = (y.getValue(i) - sum) / L.getValue(i, i);
                u.setValue(i, newValue);
            }
         
            // Transform the vector in a full result vector and store it in the FEMManager
            this.postCalculationTransformations(u);
    	}
    	
    	return true;
    }
    
    /**
     * This is a helper method combining all post calculation transformations
     * needed by a result vector. Means it is converted from it's smaller size into
     * a full result vector we can store in the FEMOUT file (storing the results in the
     * FEMManager is also handeled by this method).
     * @param u, the u vector to handle
     */
    protected void postCalculationTransformations(Vector u) {     
    	
        // Also remember the given small result vector before adding the zeros again
        femManager.addToResultListSmall(u);
    	
        // Add all zero values again which we killed due to conditions being zero
        Vector result = new Vector("Full result Vector", femManager.getNodeList().size());
        NodeList fullNodeList = femManager.getNodeList();
        
        for(int i = 1; i <= femManager.getNodeList().size(); i++) {
        	
        	// Check if the node ID has an alias
        	Node currentNode = fullNodeList.getNodeInCorrectOrder(i);
        	
        	if(currentNode.getAliasId() == 0) {
        	
        		result.setValue(currentNode.getId(), 0.0);
        		
        	} else {
        		
        		double value = u.getValue(currentNode.getAliasId());
        		result.setValue(currentNode.getId(), value);
        	}
        }
        
        // Check if the user wants to norm the eigenvectors
        if(ApplicationConstants.MEMBRAN_NORM_EIGENVECTORS) {
        
        	// Norm the current result vector to the given value
        	result = result.normVectorTo(ApplicationConstants.MEMBRAN_NORM_EIGENVECTORS_VALUE);
        }
        
        // Calculate the difference between the normed result vector and the exact calculated results (for PKW)
        // this.calcPKWError(result);

        // Calculation for Vector u has been completed, save the results in the FEMManager
        femManager.addToResultList(result);
    }
    
    /**
     * Part 2-5 : Helper Method - Error calculation
     * ########################
     * Used to calculate the error between the calculated results in vector u and the exact
     * results calculated by sin(i*(pi/a)*x) * sin(j*(pi/b)*y)
     * @param u, the result vector u
     */
    protected void calcEigenvectorError(Vector u)	{
    	
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Part 2-5 : Helper Method - Starting Error Calculation between Algorithm u-Vector and real exact u-Values ...");
        }
    	
    	// Grab all nodes from the fem manager
    	NodeList nodeList = femManager.getNodeList();
    	
    	// Constants: Set to lambda 1,1 because we know we take a look at the first eigenvalue
    	int i = 1;
    	int j  = 1;
    	
    	// The size of the membran
    	int membranSideA = 5;
    	int membranSideB = 7;
    	
    	// Vector to hold the exact results
    	Vector resultVec = new Vector("Results", u.size());
    	
    	// Array where we store the error (absolute difference between the exact values and the calculated values
    	double[] error = new double[u.size()];
    	
    	// Run through the whole vector u
    	for(int t = 1; t <= u.size(); t++)	{

    		// Grab the actual node for the current position in U because we need the X- and Y-Values of the point
    		Node node = nodeList.getNodeInCorrectOrder(t);
    		
    		// Calculate the exact results based on the X- and Y-Values and the i/j
    		double result = this.calcExactEigenvectorValues(membranSideA, membranSideB, node.getX(), node.getY(), i, j);
    		resultVec.setValue(t, result);
    	}
    	
    	// Make sure the exact result vector is also normed to the same value we used to norm the u-Vector
    	resultVec = resultVec.normVectorTo(ApplicationConstants.MEMBRAN_NORM_EIGENVECTORS_VALUE);
    	
    	// Run through the exact result vector and calculate the absolute error compared to the original u-Vector
    	for(int t = 1; t <= resultVec.size(); t++) {
    		
    		// Grab the actual value our algorithm calculated
    		double calcU = u.getValue(t);
    		
    		// Calculate the absolute error
    		error[t-1] = Math.abs(calcU - resultVec.getValue(t));
    	}
    	
    	// Sort the list ascending
    	Arrays.sort(error);
    	
    	// Display the 10 biggest differences (they are stored at the end of the sorted array
    	for(int t = (error.length-1); t >= (error.length-11); t--) {
    		
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Error at u-Position #" + t + " = " + error[t]);
            }
    	}
    }
    
    /**
     * Calculate the exact eigenvector values via formula : sin(i*(pi/a)*x) * sin(j*(pi/b)*y)
     * @param a, length of membran side a
     * @param b, length of membran side b
     * @param x, the x value of the point we want to calculate the eigenvector value for
     * @param y, the y value of the point we want to calculate the eigenvector value for
     * @param i, the i-value we use
     * @param j, the j-value we use
     * @return double, the result
     */
    protected double calcExactEigenvectorValues(double a, double b, double  x, double  y, int i, int j) {
    	
    	return Math.sin( i*(Math.PI/a)*x ) * Math.sin( j*(Math.PI/b)*y );
    }
}
