/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.calc;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.List;

import org.apache.log4j.Level;

import fhm.edu.fem.calc.objects.Border;
import fhm.edu.fem.calc.objects.Calculation;
import fhm.edu.fem.calc.objects.CompilationFunctions;
import fhm.edu.fem.calc.objects.Condition;
import fhm.edu.fem.calc.objects.Element;
import fhm.edu.fem.calc.objects.ElementList;
import fhm.edu.fem.calc.objects.Force;
import fhm.edu.fem.calc.objects.ForceList;
import fhm.edu.fem.calc.objects.MathObject;
import fhm.edu.fem.calc.objects.Node;
import fhm.edu.fem.calc.objects.NodeList;
import fhm.edu.fem.calc.objects.Vector;
import fhm.edu.fem.calc.objects.matrix.Matrix;
import fhm.edu.fem.calc.objects.matrix.MatrixFactory;
import fhm.edu.fem.constants.ApplicationConstants;
import fhm.edu.fem.constants.GaussConstants;
import fhm.edu.fem.constants.SMatrixConstants;
import fhm.edu.fem.exception.CholeskyFailedException;
import fhm.edu.fem.exception.FunctionCalculationException;

/**
 * Class Description:
 *		This class is being used when we have FEM6 files for Spanner
 *		calculations.
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class FEMCalcSpanner extends FEMCalcCircleMembrane {
	
	/* The right-hand-side Vector r */
	protected Vector vectorR = null;
	
	/* The element index where we found the max "Betrag" of the "Spannungen" Sigma_e */
	protected int maxSigmaElementIndex = -1;

    /**
     * Create a new FEM Calculator for Spanner Calculations
     */
    public FEMCalcSpanner() {
        
    	// Call the super constructor
        super();
    }
    
    /**
     * Call this method to start the calculation
     * @return true if all went well, false otherwise
     */
    public boolean start() {
    	
        // Remember the starting time
        this.startTime = System.currentTimeMillis();
        
        // Clear the result list of the FEMManager
        femManager.clearResultList();
                
        /**
         * COMPILATION OF THE COMPLETE STIFFNESS MATRIX
         */
        
        // Do the compilation for the normal Complete-Stiffness-Matrix
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.compileStiffnessMatrix()) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error during compilation of the complete-stiffness-matrix !");
            }
            
            return false;
        }
        
        // Calc the time needed for the compilation
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Compilation of the Complete-Stiffness-Matrix successfully completed: " + this.tmpTime + " ms");
            localLogger.log(Level.INFO, "Size of the Complete-Stiffness-Matrix: " + this.completeStiffnessMatrix.size() + "x" + this.completeStiffnessMatrix.size());
            localLogger.log(Level.INFO, "Required Memory of the Complete-Stiffness-Matrix: " + this.completeStiffnessMatrix.getRequiredMemory() + " bytes");
        }
        
        /**
         * BUILD VECTOR R
         */
        
        // Build the Vector r
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.buildVectorR()) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error while building Vector r !");
            }
            
            return false;
        }
        
        // Calc the time needed for the build
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Building Vector r successfully completed: " + this.tmpTime + " ms");
            localLogger.log(Level.INFO, "Size of the Vector r: " + this.vectorR.size());
        }
        
        /**
         * FINISH THE CALCULATION
         */
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Doing the calculations ...");
        }
        
        // Run through all calculations        
        List<Calculation> calculationList = femManager.getCalculationList();
        
        for(int calcIndex = 0; calcIndex < calculationList.size(); calcIndex++) {
            
            // Remember the time at the start of this calculation
            this.tmpTime = System.currentTimeMillis();
            
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Starting calculation: " + calcIndex);
            }
            
            // Grab calculation
            Calculation calc = calculationList.get(calcIndex);
            
            // Do the calculation
            if(!this.calculate(calc)) {
                
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error during calculation: " + calcIndex);
                }
                
                return false;
            }
            
            // Calculate the time needed to complete the current calculation
            this.tmpTime = System.currentTimeMillis() - this.tmpTime;
            
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Calculation " + calcIndex + " successfully completed: " + this.tmpTime + " ms");
            }
        }
        
        /**
         * CALCULATE SIGMA
         */
        
        // Calculate sigma
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.calcSigma()) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error while calculating Sigma !");
            }
            
            return false;
        }
        
        // Calc the time needed for sigma
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Calculating Sigma successfully completed: " + this.tmpTime + " ms");
        }
        
        /**
         * MODIFY THE POSITION OF THE NODES DEPENDING ON THE RESULTING U AND V
         */
        
        // Modify the node positions
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.modifyNodePositions()) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error while modifying the node positions !");
            }
            
            return false;
        }
        
        // Calc the time needed for the modification
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Modifying Node Coordinates successfully completed: " + this.tmpTime + " ms");
        }
        
        // Calculate the total time
        this.totalTime = System.currentTimeMillis() - this.startTime;

        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Full Calculation successfully completed - Total time: " + this.totalTime + " ms");
        }
        
        return true;
    }
    
    /**
     * Call this method to restart the calculation when the forces have been changed in the animation. The new forces
     * must have been changed in the FEMManager for the changes to take effect in the results of the calculation.
     * 
     * PLEASE NOTE: 
     * YOU NEED TO START THE CALCULATION THE NORMAL WAY VIA start() AT LEAST ONCE FIRST
     * BECAUSE IN restartAfterForceChange() THE COMPLETE-STIFFNESS-MATRIX IS NOT RECALCULATED BECAUSE
     * IT DOES NOT CHANGE WHEN THE FORCES CHANGE ! IF YOU DID NOT CALL start() FIRST THE COMPLETE-STIFFNESS-MATRIX
     * IS RECALCULATED in restartAfterForceChange() AUTOMATICALLY !
     * 
     * @return true if all went well, false otherwise
     */
    public boolean restartAfterForceChange() {
    	
    	// Check if we have a complete stiffness matrix
    	if(this.completeStiffnessMatrix == null) {
    		
    		// No complete stiffness matrix, use the normal way of calculation    		
    		return this.start();
    	}
    	
    	//
    	// We already have a complete stiffness matrix, so only recalculate the Vector R and finish the calculation again
    	//
    	
        // Remember the starting time
        this.startTime = System.currentTimeMillis();
        
        // Clear the result list of the FEMManager
        femManager.clearResultList();
                        
        /**
         * BUILD VECTOR R
         */
        
        // Build the Vector r
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.buildVectorR()) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error while building Vector r !");
            }
            
            return false;
        }
        
        // Calc the time needed for the build
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Building Vector r successfully completed: " + this.tmpTime + " ms");
            localLogger.log(Level.INFO, "Size of the Vector r: " + this.vectorR.size());
        }
        
        /**
         * FINISH THE CALCULATION
         */
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Doing the calculations ...");
        }
        
        // Run through all calculations        
        List<Calculation> calculationList = femManager.getCalculationList();
        
        for(int calcIndex = 0; calcIndex < calculationList.size(); calcIndex++) {
            
            // Remember the time at the start of this calculation
            this.tmpTime = System.currentTimeMillis();
            
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Starting calculation: " + calcIndex);
            }
            
            // Grab calculation
            Calculation calc = calculationList.get(calcIndex);
            
            // Do the calculation
            if(!this.calculate(calc)) {
                
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error during calculation: " + calcIndex);
                }
                
                return false;
            }
            
            // Calculate the time needed to complete the current calculation
            this.tmpTime = System.currentTimeMillis() - this.tmpTime;
            
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Calculation " + calcIndex + " successfully completed: " + this.tmpTime + " ms");
            }
        }
        
        /**
         * CALCULATE SIGMA
         */
        
        // Calculate sigma
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.calcSigma()) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error while calculating Sigma !");
            }
            
            return false;
        }
        
        // Calc the time needed for sigma
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Calculating Sigma successfully completed: " + this.tmpTime + " ms");
        }
        
        /**
         * MODIFY THE POSITION OF THE NODES DEPENDING ON THE RESULTING U AND V
         */
        
        // Modify the node positions
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.modifyNodePositions()) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error while modifying the node positions !");
            }
            
            return false;
        }
        
        // Calc the time needed for the modification
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Modifying Node Coordinates successfully completed: " + this.tmpTime + " ms");
        }
        
        // Calculate the total time
        this.totalTime = System.currentTimeMillis() - this.startTime;

        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Re-calculation successfully completed - Total time: " + this.totalTime + " ms");
        }
        
        return true;
    }
    
    /**
     * Do the compilation with the Gauss-Quadrature-Integration
     * with it's size depending on the number of inner nodes we still have after creating
     * the alias IDs
     * @return true if all went well, false otherwise
     */
    protected boolean compileStiffnessMatrix() {
    	
        try {
        	int matrixSize = (2*femManager.getNodeList().size());
        	this.completeStiffnessMatrix = MatrixFactory.createUserDefinedMatrix("Complete Stiffness Matrix" , matrixSize);
        	
            // Grab the approach from the manager
            int femApproach = femManager.getFEMApproach();
            
            // Grab the list of all elements
            ElementList elementlist = femManager.getElementList();
            
            // Calculate how many percentage one element makes up
            double oneElementPercentage = 100.0 / elementlist.size();
            
            // Calculate how many elements make up one percent
            double onePercentElements = elementlist.size() / 100.0;
            
            // Where we store the current actual percentage
            double currentPercentage = 0.0;
            
            // Run through all elements
            for(int elementIndex = 1; elementIndex <= elementlist.size(); elementIndex++)
            {                
                // Calculate the current percentage of elements already done in the compilation
                currentPercentage = (elementIndex-1) * oneElementPercentage;

                // Make a percentage output about every 5% of completed elements
                if(elementIndex%(onePercentElements*5.0) < 1) {
                    
                    if(localLogger.isEnabledFor(Level.INFO)) {
                        localLogger.log(Level.INFO, "Compilation - Complete-Stiffness-Matrix at " + ApplicationConstants.convertPercentageToString(currentPercentage) + " %");
                    }
                }

                // Get the current element
                Element tmpElement =  elementlist.getElement(elementIndex);
                NodeList nodeList = tmpElement.getNodeList();
                
                // Check if we should calculate the element-stiffness-matrix via the S-Matrices or the Double Integrale
                Matrix elementStiffnessMatrix = null;
                
                if(ApplicationConstants.SPANNER_STIFFNESSMATRIX_USE == ApplicationConstants.SPANNER_STIFFNESSMATRIX_SMATRICES) {
                	
                	// S-Matrix Approach
                	elementStiffnessMatrix = this.calculateElementStiffnessMatrixViaSMatrix(tmpElement);
                	
                } else if(ApplicationConstants.SPANNER_STIFFNESSMATRIX_USE == ApplicationConstants.SPANNER_STIFFNESSMATRIX_INTEGRAL) {
                	
                	// Double Integral Approach
                	elementStiffnessMatrix = this.calculateElementStiffnessMatrixViaIntegral(tmpElement);
                	
                } else {
                	
                    if(localLogger.isEnabledFor(Level.INFO)) {
                        localLogger.log(Level.INFO, "Unknown way of compiling the Stiffnessmatrices required: " + ApplicationConstants.SPANNER_STIFFNESSMATRIX_USE);
                    }
                    
                    return false;
                }
                
                // Apply the factor to the element stiffness matrix: 0.5*((E*h)/(1-v^2))
                double e = femManager.getSpannerMaterialConstantE();
                double h = femManager.getSpannerMaterialConstantH();
                double v = femManager.getSpannerMaterialConstantV();
                
                double elementStiffnessFactor = ((e*h)/(1-(v*v)));
                elementStiffnessMatrix = elementStiffnessMatrix.mult(elementStiffnessFactor);

                // Run through all nodes in the element to get the index i of the matrix
                for(int indexA = 1; indexA <= elementStiffnessMatrix.size(); indexA++) {
                   
                    // Grab the nodeID from the node
                	int tmpNodeAIndex = indexA%(tmpElement.getNodeList().size()+1);
                	
                	if(indexA > (elementStiffnessMatrix.size()/2)) {
                		tmpNodeAIndex = tmpNodeAIndex+1;
                	}

                    Node tmpNodeA = nodeList.getNodeInCorrectOrder(tmpNodeAIndex);
                    
                    // Determine the starting valueof indexB for the next lines, depending if we use a MATRIX_STANDARD or MATRIX_HULL
                    int startOfIndexB = -1;
                    
                    if(ApplicationConstants.MATRIX_USE == ApplicationConstants.MATRIX_STANDARD) {
                        
                        // We are using a standard matrix
                        startOfIndexB = 1;
                        
                    } else if(ApplicationConstants.MATRIX_USE == ApplicationConstants.MATRIX_HULL) {
                        
                        // We are using a hull matrix
                        startOfIndexB = indexA;
                    }
                    
                   // Run through all nodes in the element to get the index j of the matrix
                   for(int indexB = startOfIndexB; indexB <= elementStiffnessMatrix.size(); indexB++) {
                       
                   	   // Grab the nodeID from the node
                   	   int tmpNodeBIndex = indexB%(tmpElement.getNodeList().size() + 1);

	                   	if(indexB > (elementStiffnessMatrix.size()/2)) {
	                		tmpNodeBIndex = tmpNodeBIndex+1;
	                	}

                       Node tmpNodeB = nodeList.getNodeInCorrectOrder(tmpNodeBIndex);
                       	
                        // Grab the current value from the element stiffness matrix and add it to the value in the complete stiffness matrix
                       double elementValue = elementStiffnessMatrix.getValue(indexA, indexB);

                        // Determine where to put the value
                        int rowInCompleteMatrix = -1;
                        int colInCompleteMatrix = -1;
                        
                        // ROW-Node - Is the value for U or V
                        if(indexA > (elementStiffnessMatrix.size()/2)) {
                        	
                        	// We got V
                        	rowInCompleteMatrix = tmpNodeA.getId() * 2;

                        } else {

                        	// We got U
                        	rowInCompleteMatrix = (tmpNodeA.getId() * 2) - 1;
                        }
                        
                        // COLUMN-Node - Is the value for U or V
                        if(indexB > (elementStiffnessMatrix.size()/2)) {
                        	
                        	// We got V
                        	colInCompleteMatrix = tmpNodeB.getId() * 2;
                        
                        } else {
                        	
                        	// We got U
                        	colInCompleteMatrix = (tmpNodeB.getId() * 2) - 1;
                        }
                        
                        double newValue = this.completeStiffnessMatrix.getValue(rowInCompleteMatrix, colInCompleteMatrix) + elementValue;                       
                        this.completeStiffnessMatrix.setValue(rowInCompleteMatrix, colInCompleteMatrix, newValue);
                   }             
               }
            }
            
            if(localLogger.isEnabledFor(Level.DEBUG)) {
                localLogger.log(Level.DEBUG, "Compilation complete !");
                localLogger.log(Level.DEBUG, "Matrix size: " + this.completeStiffnessMatrix.size());
                localLogger.log(Level.DEBUG, "CompleteStiffnessMatrix: " + this.completeStiffnessMatrix);
            }
            
            // Store the stiffness-matrix in the FEMManager for later use
            femManager.setCompleteStiffnessMatrix(this.completeStiffnessMatrix);

            return true;
            
        } catch (Exception e) {
            
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Failure during compilation!", e);
			}
			
			return false;
        }
    }
    
    /**
     * Create the Element-Stiffness-Matrix for the given Element via the Integral
     * @param element, the element to calculate the stiffness matrix for
     * @return the ElementStiffnessMatrix
     */
    protected Matrix calculateElementStiffnessMatrixViaIntegral(Element element) {
    	
        NodeList nodeList =  element.getNodeList();
        Matrix elementStiffnessMatrix = MatrixFactory.createUserDefinedMatrix("ElementstiffnessMatrix: " + element.getId(), 2*nodeList.size());
    	
    	try {
    	
	        // The variables to use in the Gauss Quadrature Method call
	        double a = 0;
	        double b = 1;
	        Method h1_1 = null;
	        Method h2_1 = null;
	        Method q_1 = null;
	        Method h1_2 = null;
	        Method h2_2 = null;
	        Method q_2 = null;
	        int z1 = -1;
	        int z2 = -1;
			Vector ax = GaussConstants.get(INNER_POINTS, GaussConstants.VALUE_A);
			Vector hx = GaussConstants.get(INNER_POINTS, GaussConstants.VALUE_H);
			Vector ay = GaussConstants.get(OUTER_POINTS, GaussConstants.VALUE_A);
			Vector hy = GaussConstants.get(OUTER_POINTS, GaussConstants.VALUE_H);
	        MathObject firstIntegralResultObject = null;
	        MathObject secondIntegralResultObject = null;
	        int elementType = -1;
	        
	        // The resulting Matrices
	        Matrix S11_ROOF = null;
	        Matrix S22_ROOF = null;
	        Matrix S12_ROOF = null;
	        Matrix S21_ROOF = null;   
	    	
	        // Store the x- and y-values of the element's nodes in the static data structure in the Class CompilationFunctions
	        CompilationFunctions.storeElementXYValues(element);
	        
	        // Take a look which kind of element we have here and fill the input variables for the integration accordingly
	        if(nodeList.size() == 6) {
	        	
	            //
	            // TRIANGLE
	            //
	        	
	        	/**
	        	 * MATRIX - S11 - ROOF
	        	 */
	        	
	        	h1_1 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_TRIANGLE_QUADRATIC_H1);
	    		h2_1 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_TRIANGLE_QUADRATIC_H1);
	    		q_1 = CompilationFunctions.getMethod(CompilationFunctions.ONE);
	    		h1_2 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_TRIANGLE_QUADRATIC_H2);
	    		h2_2 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_TRIANGLE_QUADRATIC_H2);
	    		q_2 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_S11_INTEGRAL_TWO_Q);
	    		z1 = 6;
	    		z2 = 6;
	        	elementType = GaussConstants.FIGURE_TRIANGLE;
	        	
	            // Calculate the first integral
	            firstIntegralResultObject = FEMGaussQuadrature.calcIntegration(a, b, h1_1, h2_1, q_1, z1, z2, ax, hx, ay, hy, elementType);
	            
	            // Calculate the second integral
	            secondIntegralResultObject = FEMGaussQuadrature.calcIntegration(a, b, h1_2, h2_2, q_2, z1, z2, ax, hx, ay, hy, elementType);
	            
	            // Check if both result objects of the first and second integral match our requirements
	            if(!(firstIntegralResultObject instanceof Matrix) || !(secondIntegralResultObject instanceof Matrix)) {
	            	
	            	// At least one of the result objects is not a Matrix => ERROR
	                if(localLogger.isEnabledFor(Level.ERROR)) {
	                    localLogger.log(Level.ERROR, "One of the integral result objects is not a Matrix !");
	                }
	            }
	            
	            // Build the full result matrix
	            S11_ROOF = (((Matrix)firstIntegralResultObject)).add(((Matrix)secondIntegralResultObject));
	        	
	            /**
	             * MATRIX - S22 - ROOF
	             */                    
	            
	        	h1_1 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_TRIANGLE_QUADRATIC_H1);
	    		h2_1 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_TRIANGLE_QUADRATIC_H1);
	    		q_1 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_S22_INTEGRAL_ONE_Q);
	    		h1_2 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_TRIANGLE_QUADRATIC_H2);
	    		h2_2 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_TRIANGLE_QUADRATIC_H2);
	    		q_2 = CompilationFunctions.getMethod(CompilationFunctions.ONE);
	    		z1 = 6;
	    		z2 = 6;
	        	elementType = GaussConstants.FIGURE_TRIANGLE;
	        	
	            // Calculate the first integral
	            firstIntegralResultObject = FEMGaussQuadrature.calcIntegration(a, b, h1_1, h2_1, q_1, z1, z2, ax, hx, ay, hy, elementType);
	            
	            // Calculate the second integral
	            secondIntegralResultObject = FEMGaussQuadrature.calcIntegration(a, b, h1_2, h2_2, q_2, z1, z2, ax, hx, ay, hy, elementType);
	            
	            // Check if both result objects of the first and second integral match our requirements
	            if(!(firstIntegralResultObject instanceof Matrix) || !(secondIntegralResultObject instanceof Matrix)) {
	            	
	            	// At least one of the result objects is not a Matrix => ERROR
	                if(localLogger.isEnabledFor(Level.ERROR)) {
	                    localLogger.log(Level.ERROR, "One of the integral result objects is not a Matrix !");
	                }
	            }
	            
	            // Build the full result matrix
	            S22_ROOF = (((Matrix)firstIntegralResultObject)).add(((Matrix)secondIntegralResultObject));
	        	
	        	/**
	        	 * MATRIX - S12 - ROOF
	        	 */
	            
	        	h1_1 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_TRIANGLE_QUADRATIC_H1);
	    		h2_1 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_TRIANGLE_QUADRATIC_H2);
	    		q_1 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_S12_INTEGRAL_ONE_Q);
	    		h1_2 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_TRIANGLE_QUADRATIC_H2);
	    		h2_2 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_TRIANGLE_QUADRATIC_H1);
	    		q_2 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_S12_INTEGRAL_TWO_Q);
	    		z1 = 6;
	    		z2 = 6;
	        	elementType = GaussConstants.FIGURE_TRIANGLE;
	        	
	            // Calculate the first integral
	            firstIntegralResultObject = FEMGaussQuadrature.calcIntegration(a, b, h1_1, h2_1, q_1, z1, z2, ax, hx, ay, hy, elementType);
	            
	            // Calculate the second integral
	            secondIntegralResultObject = FEMGaussQuadrature.calcIntegration(a, b, h1_2, h2_2, q_2, z1, z2, ax, hx, ay, hy, elementType);
	            
	            // Check if both result objects of the first and second integral match our requirements
	            if(!(firstIntegralResultObject instanceof Matrix) || !(secondIntegralResultObject instanceof Matrix)) {
	            	
	            	// At least one of the result objects is not a Matrix => ERROR
	                if(localLogger.isEnabledFor(Level.ERROR)) {
	                    localLogger.log(Level.ERROR, "One of the integral result objects is not a Matrix !");
	                }
	            }
	            
	            // Build the full result matrix
	            S12_ROOF = (((Matrix)firstIntegralResultObject)).add(((Matrix)secondIntegralResultObject));
	        	
	            /**
	             * MATRIX - S21 - ROOF
	             */
	
	            // S21 = S12^T
	            S21_ROOF = S12_ROOF.transponate();
	            
	        } else if(nodeList.size() == 8) {
	        	
	            //
	            // QUADRAT
	            //
	        	
	        	/**
	        	 * MATRIX - S11 - ROOF
	        	 */
	        	
	        	h1_1 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_QUADRAT_QUADRATIC_H1);
	    		h2_1 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_QUADRAT_QUADRATIC_H1);
	    		q_1 = CompilationFunctions.getMethod(CompilationFunctions.ONE);
	    		h1_2 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_QUADRAT_QUADRATIC_H2);
	    		h2_2 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_QUADRAT_QUADRATIC_H2);
	    		q_2 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_S11_INTEGRAL_TWO_Q);
	    		z1 = 8;
	    		z2 = 8;
	        	elementType = GaussConstants.FIGURE_QUADRAT;
	        	
	            // Calculate the first integral
	            firstIntegralResultObject = FEMGaussQuadrature.calcIntegration(a, b, h1_1, h2_1, q_1, z1, z2, ax, hx, ay, hy, elementType);
	            
	            // Calculate the second integral
	            secondIntegralResultObject = FEMGaussQuadrature.calcIntegration(a, b, h1_2, h2_2, q_2, z1, z2, ax, hx, ay, hy, elementType);
	            
	            // Check if both result objects of the first and second integral match our requirements
	            if(!(firstIntegralResultObject instanceof Matrix) || !(secondIntegralResultObject instanceof Matrix)) {
	            	
	            	// At least one of the result objects is not a Matrix => ERROR
	                if(localLogger.isEnabledFor(Level.ERROR)) {
	                    localLogger.log(Level.ERROR, "One of the integral result objects is not a Matrix !");
	                }
	            }
	            
	            // Build the full result matrix
	            S11_ROOF = (((Matrix)firstIntegralResultObject)).add(((Matrix)secondIntegralResultObject));
	        	
	            /**
	             * MATRIX - S22 - ROOF
	             */                    
	            
	        	h1_1 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_QUADRAT_QUADRATIC_H1);
	    		h2_1 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_QUADRAT_QUADRATIC_H1);
	    		q_1 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_S22_INTEGRAL_ONE_Q);
	    		h1_2 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_QUADRAT_QUADRATIC_H2);
	    		h2_2 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_QUADRAT_QUADRATIC_H2);
	    		q_2 = CompilationFunctions.getMethod(CompilationFunctions.ONE);
	    		z1 = 8;
	    		z2 = 8;
	        	elementType = GaussConstants.FIGURE_QUADRAT;
	        	
	            // Calculate the first integral
	            firstIntegralResultObject = FEMGaussQuadrature.calcIntegration(a, b, h1_1, h2_1, q_1, z1, z2, ax, hx, ay, hy, elementType);
	            
	            // Calculate the second integral
	            secondIntegralResultObject = FEMGaussQuadrature.calcIntegration(a, b, h1_2, h2_2, q_2, z1, z2, ax, hx, ay, hy, elementType);
	            
	            // Check if both result objects of the first and second integral match our requirements
	            if(!(firstIntegralResultObject instanceof Matrix) || !(secondIntegralResultObject instanceof Matrix)) {
	            	
	            	// At least one of the result objects is not a Matrix => ERROR
	                if(localLogger.isEnabledFor(Level.ERROR)) {
	                    localLogger.log(Level.ERROR, "One of the integral result objects is not a Matrix !");
	                }
	            }
	            
	            // Build the full result matrix
	            S22_ROOF = (((Matrix)firstIntegralResultObject)).add(((Matrix)secondIntegralResultObject));
	        	
	        	/**
	        	 * MATRIX - S12 - ROOF
	        	 */
	            
	        	h1_1 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_QUADRAT_QUADRATIC_H1);
	    		h2_1 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_QUADRAT_QUADRATIC_H2);
	    		q_1 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_S12_INTEGRAL_ONE_Q);
	    		h1_2 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_QUADRAT_QUADRATIC_H2);
	    		h2_2 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_QUADRAT_QUADRATIC_H1);
	    		q_2 = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_S12_INTEGRAL_TWO_Q);
	    		z1 = 8;
	    		z2 = 8;
	        	elementType = GaussConstants.FIGURE_QUADRAT;
	        	
	            // Calculate the first integral
	            firstIntegralResultObject = FEMGaussQuadrature.calcIntegration(a, b, h1_1, h2_1, q_1, z1, z2, ax, hx, ay, hy, elementType);
	            
	            // Calculate the second integral
	            secondIntegralResultObject = FEMGaussQuadrature.calcIntegration(a, b, h1_2, h2_2, q_2, z1, z2, ax, hx, ay, hy, elementType);
	            
	            // Check if both result objects of the first and second integral match our requirements
	            if(!(firstIntegralResultObject instanceof Matrix) || !(secondIntegralResultObject instanceof Matrix)) {
	            	
	            	// At least one of the result objects is not a Matrix => ERROR
	                if(localLogger.isEnabledFor(Level.ERROR)) {
	                    localLogger.log(Level.ERROR, "One of the integral result objects is not a Matrix !");
	                }
	            }
	            
	            // Build the full result matrix
	            S12_ROOF = (((Matrix)firstIntegralResultObject)).add(((Matrix)secondIntegralResultObject));
	        	
	            /**
	             * MATRIX - S21 - ROOF
	             */
	
	            // S21 = S12^T
	            S21_ROOF = S12_ROOF.transponate();
	        }
	
	        // Combine the first and second integral result objects to one resulting element stiffness matrix
	        
	        // S11_ROOF into ElementStiffnessMatrix
	        for(int i = 1; i <= S11_ROOF.size(); i++) {
	        	for(int j = 1; j <= S11_ROOF.size(); j++) {
	        		
	        		elementStiffnessMatrix.setValue(i, j, S11_ROOF.getValue(i, j));
	        	}
	        }
	        
	        // S12_ROOF into ElementStiffnessMatrix
	        for(int i = 1; i <= S12_ROOF.size(); i++) {
	        	for(int j = 1; j <= S12_ROOF.size(); j++) {
	        		
	        		elementStiffnessMatrix.setValue(i, (j+S11_ROOF.size()), S12_ROOF.getValue(i, j));
	        	}
	        }
	        
	        // S21_ROOF into ElementStiffnessMatrix
	        for(int i = 1; i <= S21_ROOF.size(); i++) {
	        	for(int j = 1; j <= S21_ROOF.size(); j++) {
	        		
	        		elementStiffnessMatrix.setValue((i+S11_ROOF.size()), j, S21_ROOF.getValue(i, j));
	        	}
	        }
	        
	        // S22_ROOF into ElementStiffnessMatrix
	        for(int i = 1; i <= S22_ROOF.size(); i++) {
	        	for(int j = 1; j <= S22_ROOF.size(); j++) {
	        		
	        		elementStiffnessMatrix.setValue((i+S12_ROOF.size()), (j+S21_ROOF.size()), S22_ROOF.getValue(i, j));
	        	}
	        }
	
	        if(localLogger.isEnabledFor(Level.DEBUG)) {
	            localLogger.log(Level.DEBUG, "S11 for Element " + element.getId() + ": " + S11_ROOF);
	        }
	        
	        if(localLogger.isEnabledFor(Level.DEBUG)) {
	            localLogger.log(Level.DEBUG, "S12 for Element " + element.getId() + ": " + S12_ROOF);
	        }
	        
	        if(localLogger.isEnabledFor(Level.DEBUG)) {
	            localLogger.log(Level.DEBUG, "S21 for Element " + element.getId() + ": " + S21_ROOF);
	        }
	        
	        if(localLogger.isEnabledFor(Level.DEBUG)) {
	            localLogger.log(Level.DEBUG, "S22 for Element " + element.getId() + ": " + S22_ROOF);
	        }
	        
	        if(localLogger.isEnabledFor(Level.DEBUG)) {
	            localLogger.log(Level.DEBUG, "ElementStiffnessMatrix for Element " + element.getId() + ": " + elementStiffnessMatrix);
	        }
	        
        } catch (Exception e) {
            
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Failure during compilation!", e);
			}
        }
        
        return elementStiffnessMatrix;
    }
    
    /**
     * Create the Element-Stiffness-Matrix for the given Element via the S-Matrices
     * @param element, the element to calculate the stiffness matrix for
     * @return the ElementStiffnessMatrix
     */
    protected Matrix calculateElementStiffnessMatrixViaSMatrix(Element element) {
    	
        NodeList nodeList =  element.getNodeList();
        Matrix elementStiffnessMatrix = MatrixFactory.createUserDefinedMatrix("ElementstiffnessMatrix: " + element.getId(), 2*nodeList.size());
    	
        try {
        	
            int elementType = -1;
            
            // Get the correct nodes from the element
            Node node1 = nodeList.getNodeInCorrectOrder(1);
            Node node2 = nodeList.getNodeInCorrectOrder(2);
            Node nodeE = null;
            
            // Check which node we need to grab in order to get nodeE
	        if(nodeList.size() == 3) {
	              
	            // Triangle with APPROACH_LINEAR
	            elementType = SMatrixConstants.FIGURE_TRIANGLE;
	            nodeE = nodeList.getNodeInCorrectOrder(3);
	              
	        } else if(nodeList.size() == 4) {
	              
	            // Parallelogram with APPROACH_LINEAR
	            elementType = SMatrixConstants.FIGURE_QUADRAT;
	            nodeE = nodeList.getNodeInCorrectOrder(4);
	              
	        } else if(nodeList.size() == 6) {
	              
	            // Triangle with APPROACH_QUADRATIC
	            elementType = SMatrixConstants.FIGURE_TRIANGLE;
	            nodeE = nodeList.getNodeInCorrectOrder(3);
	              
	        } else if(nodeList.size() == 8) {
	              
	            // Parallelogram with APPROACH_QUADRATIC
	            elementType = SMatrixConstants.FIGURE_QUADRAT;
	            nodeE = nodeList.getNodeInCorrectOrder(4);
	            
	        } else {

                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Node1 for Element " + element.getId() + " = " + node1+ " Number of Nodes in this Element = " + nodeList.size());
                }
	        }
	        
            // DEBUG Stuff
            if(localLogger.isEnabledFor(Level.DEBUG)) {
                localLogger.log(Level.DEBUG, "Node1 for Element " + element.getId() + " = " + node1);
                localLogger.log(Level.DEBUG, "Node2 for Element " + element.getId() + " = " + node2);
                localLogger.log(Level.DEBUG, "NodeE for Element " + element.getId() + " = " + nodeE);
            }
            
            // DEBUG Stuff
            if(localLogger.isEnabledFor(Level.DEBUG)) {
                localLogger.log(Level.DEBUG, "det(J) Calculation = " + (node2.getX() - node1.getX()) + " * " + (nodeE.getY() - node1.getY()) + " - " + (nodeE.getX() - node1.getX()) + " * " + (node2.getY() - node1.getY()));
            }
            
            // Calculate det(J)
            double detJ = (node2.getX() - node1.getX()) * 
            						  (nodeE.getY() - node1.getY()) - 
            						  (nodeE.getX() - node1.getX()) * 
            						  (node2.getY() - node1.getY());
            
            if(localLogger.isEnabledFor(Level.DEBUG)) {
                localLogger.log(Level.DEBUG, "det(J) for Element " + element.getId() + " = " + detJ);
            }
            
            // Check if det(J) is smaller than 0
            if(detJ < 0) {
                
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "det(J) for Element " + element.getId() + " is negative: " + detJ);
                }
            }
            
            double mu = 0.5*(1.0-femManager.getSpannerMaterialConstantV());
            double x1 = node1.getX();
            double y1 = node1.getY();
            double x2 = node2.getX();
            double y2 = node2.getY();
            double x3 = nodeE.getX();
            double y3 = nodeE.getY();
            
            // Calculate a1, b1, c1
            double a1 = (mu * Math.pow(x3-x1, 2) + Math.pow(y3 - y1, 2)) / detJ;
            double b1 = -1.0*(mu * (x2 - x1)*(x3 - x1) + (y2 - y1)*(y3 - y1)) / detJ;
            double c1 = (mu * Math.pow(x2-x1, 2) + Math.pow(y2 - y1, 2)) / detJ;
            
            // Calculate a2, b2, c2
            double a2 = (Math.pow(x3-x1, 2) + mu*Math.pow(y3 - y1, 2)) / detJ;
            double b2 = -1.0*((x2 - x1)*(x3 - x1) + mu*(y2 - y1)*(y3 - y1)) / detJ;
            double c2 = (Math.pow(x2-x1, 2) + mu*Math.pow(y2 - y1, 2)) / detJ;
            
            // Calculate a3, b3, c3, d3
            double v = femManager.getSpannerMaterialConstantV();
            double a3 = (-0.5 * (1.0 + v) * (x3 - x1) * (y3 - y1)) / detJ;
            double b3 = (v*(x2-x1)*(y3-y1) + mu*(x3-x1)*(y2-y1)) / detJ;
            double c3 = (v*(x3-x1)*(y2-y1) + mu*(x2-x1)*(y3-y1)) / detJ;
            double d3 = (-0.5 * (1.0 + v) * (x2 - x1) * (y2 - y1)) / detJ;
            
	        // The resulting Matrices
	        Matrix S11_ROOF = null;
	        Matrix S22_ROOF = null;
	        Matrix S12_ROOF = null;
	        Matrix S21_ROOF = null;
	        
	        // Get the needed S-Matrices depending on what kind of element we have here
	        Matrix S1 = null;
	        Matrix S2Star = null;
	        Matrix S2StarTransponated = null;
	        Matrix S3 = null;
	        
	        if(nodeList.size() == 6) {
	        	
	        	S1 = SMatrixConstants.getSMatrix(SMatrixConstants.APPROACH_QUADRATIC, SMatrixConstants.FIGURE_TRIANGLE, 1);
	        	S2Star = SMatrixConstants.getQuadraticTriangleSpannerS2Star();
	        	S3 = SMatrixConstants.getSMatrix(SMatrixConstants.APPROACH_QUADRATIC, SMatrixConstants.FIGURE_TRIANGLE, 3);
	        	
	        } else if(nodeList.size() == 8) {
	        	
	        	S1 = SMatrixConstants.getSMatrix(SMatrixConstants.APPROACH_QUADRATIC, SMatrixConstants.FIGURE_QUADRAT, 1);
	        	S2Star = SMatrixConstants.getQuadraticParallelogramSpannerS2Star();
	        	S3 = SMatrixConstants.getSMatrix(SMatrixConstants.APPROACH_QUADRATIC, SMatrixConstants.FIGURE_QUADRAT, 3);
	        	
	        } else {
	        	
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Unknown Element Type. Nodelist size: " + nodeList.size());
                }
	        }
	        
	        S2StarTransponated = S2Star.transponate();
	        
	        // Calculate S11_ROOF
	        Matrix S11_ROOF_PART_1 = S1.mult(a1);
	        Matrix S11_ROOF_PART_2 = S2Star.add(S2StarTransponated);
	        S11_ROOF_PART_2 = S11_ROOF_PART_2.mult(b1);
	        Matrix S11_ROOF_PART_3 = S3.mult(c1);
	        
	        S11_ROOF = (S11_ROOF_PART_1.add(S11_ROOF_PART_2)).add(S11_ROOF_PART_3);
	        
	        // Calculate S22_ROOF
	        Matrix S22_ROOF_PART_1 = S1.mult(a2);
	        Matrix S22_ROOF_PART_2 = S2Star.add(S2StarTransponated);
	        S22_ROOF_PART_2 = S22_ROOF_PART_2.mult(b2);
	        Matrix S22_ROOF_PART_3 = S3.mult(c2);
	        
	        S22_ROOF = (S22_ROOF_PART_1.add(S22_ROOF_PART_2)).add(S22_ROOF_PART_3);
	        
	        // Calculate S12_ROOF
	        Matrix S12_ROOF_PART_1 = S1.mult(a3);
	        Matrix S12_ROOF_PART_2 = S2Star.mult(b3);
	        Matrix S12_ROOF_PART_3 = S2StarTransponated.mult(c3);
	        Matrix S12_ROOF_PART_4 = S3.mult(d3);
	        
	        S12_ROOF = ((S12_ROOF_PART_1.add(S12_ROOF_PART_2)).add(S12_ROOF_PART_3)).add(S12_ROOF_PART_4);
	        
	        // Calculate S21_ROOF
	        S21_ROOF = S12_ROOF.transponate();
	        
	        // S11_ROOF into ElementStiffnessMatrix
	        for(int i = 1; i <= S11_ROOF.size(); i++) {
	        	for(int j = 1; j <= S11_ROOF.size(); j++) {
	        		
	        		elementStiffnessMatrix.setValue(i, j, S11_ROOF.getValue(i, j));
	        	}
	        }
	        
	        // S12_ROOF into ElementStiffnessMatrix
	        for(int i = 1; i <= S12_ROOF.size(); i++) {
	        	for(int j = 1; j <= S12_ROOF.size(); j++) {
	        		
	        		elementStiffnessMatrix.setValue(i, (j+S11_ROOF.size()), S12_ROOF.getValue(i, j));
	        	}
	        }
	        
	        // S21_ROOF into ElementStiffnessMatrix
	        for(int i = 1; i <= S21_ROOF.size(); i++) {
	        	for(int j = 1; j <= S21_ROOF.size(); j++) {
	        		
	        		elementStiffnessMatrix.setValue((i+S11_ROOF.size()), j, S21_ROOF.getValue(i, j));
	        	}
	        }
	        
	        // S22_ROOF into ElementStiffnessMatrix
	        for(int i = 1; i <= S22_ROOF.size(); i++) {
	        	for(int j = 1; j <= S22_ROOF.size(); j++) {
	        		
	        		elementStiffnessMatrix.setValue((i+S12_ROOF.size()), (j+S21_ROOF.size()), S22_ROOF.getValue(i, j));
	        	}
	        }
	
	        if(localLogger.isEnabledFor(Level.DEBUG)) {
	            localLogger.log(Level.DEBUG, "S11 for Element " + element.getId() + ": " + S11_ROOF);
	        }
	        
	        if(localLogger.isEnabledFor(Level.DEBUG)) {
	            localLogger.log(Level.DEBUG, "S12 for Element " + element.getId() + ": " + S12_ROOF);
	        }
	        
	        if(localLogger.isEnabledFor(Level.DEBUG)) {
	            localLogger.log(Level.DEBUG, "S21 for Element " + element.getId() + ": " + S21_ROOF);
	        }
	        
	        if(localLogger.isEnabledFor(Level.DEBUG)) {
	            localLogger.log(Level.DEBUG, "S22 for Element " + element.getId() + ": " + S22_ROOF);
	        }
	        
	        if(localLogger.isEnabledFor(Level.DEBUG)) {
	            localLogger.log(Level.DEBUG, "ElementStiffnessMatrix for Element " + element.getId() + ": " + elementStiffnessMatrix);
	        }
        	
        } catch (Exception e) {
            
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Failure during compilation!", e);
			}
        }
        
    	return elementStiffnessMatrix;
    }
    
    /**
     * Build the Vector R
     * @return true if all went well, false otherwise
     */
    protected boolean buildVectorR() {
    	
    	// Create the Vector R
    	this.vectorR = new Vector("Vector R", (2*femManager.getNodeList().size()));
    	
    	// Grab the forces
    	ForceList forceList = femManager.getSpannerForceList();
    	
    	// Run through all forces
    	for(int i = 1; i <= forceList.size(); i++) {
    		
    		Force currentForce = forceList.getForce(i);
    		
    		int forceNodeId = currentForce.getNode().getId();
    		double forceX = currentForce.getX();
    		double forceY = currentForce.getY();
    		
    		int vectorIndex = forceNodeId*2;
    		
    		this.vectorR.setValue(vectorIndex-1, forceX);
    		this.vectorR.setValue(vectorIndex, forceY);
    	}
        
    	return true;
    }
    
    /**
     * Do the calculation
     * @param calc, the calculation object
     * @return true if all went well, false otherwise
     */
    protected boolean calculate(Calculation calc) {
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Starting to clone the complete-stiffness-Matrix ...");
        }

        // Clone the complete stiffness matrix for us to change
        Matrix stiffnessMatrixClone = (Matrix)(this.completeStiffnessMatrix.clone());

        /**
         * IMPLEMENT THE BORDER CONDITIONS
         **/
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Starting to implement the border conditions ...");
        }
        
        // Grab the conditions from the actual calculation
        List<Condition> conditionList = calc.getConditions();
        
        // Run through all conditions
        for(int i = 0; i < conditionList.size(); i++)
        {
            // Grab the current condition
            Condition condition =  conditionList.get(i);
            
            // Retrieve the value of the condition
            double conditionValue = condition.getValue();
            
            // Get the border affected by this condition and grab all nodes from this border
            Border conditionBorder = condition.getBorder();
            NodeList borderNodeList = conditionBorder.getNodes();

            // Run through all nodes in the border
            for(int j = 1; j <= borderNodeList.size(); j++){
                
                // Grab the current node from the border node list
                Node tmpNode = borderNodeList.getNodeInCorrectOrder(j);

                // Retrieve the ID number of the current node
                int nodeId = tmpNode.getId();
                
                // Adjust the index
                nodeId = nodeId * 2;
                
                // Run through the vector r
                for(int k = 1; k <= this.completeStiffnessMatrix.size(); k++){
                    
                    // Calculate the new value of r[k] made up of r[k] - ( K[k][nodeID] * conditionValue )
                   double newValue = vectorR.getValue(k) - (stiffnessMatrixClone.getValue(k, nodeId) * conditionValue);    

                    // Set the new value for b
                   vectorR.setValue(k, newValue);     
                }
                
                // Set the condition value to the proper line in the vector b
                vectorR.setValue(nodeId-1, conditionValue);
                vectorR.setValue(nodeId, conditionValue);
               
                // Fill the column and the line identified by the node ID with zeros
                stiffnessMatrixClone.setLineAndColumn(nodeId-1, 0);
                stiffnessMatrixClone.setLineAndColumn(nodeId, 0);

                // Set the value "1" at the main diagonal element identified by K[nodeID][nodeID]
                stiffnessMatrixClone.setValue(nodeId-1, nodeId-1, 1);
                stiffnessMatrixClone.setValue(nodeId, nodeId, 1);
           }
        }

        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Implementing border conditions successfully completed");
            localLogger.log(Level.INFO, "New required Memory of the MODIFIED Complete-Stiffness-Matrix: " + stiffnessMatrixClone.getRequiredMemory() + " bytes");
        }

        /**
         * Calculate the CHOLESKY
         * */
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Starting to calculate the CHOLESKY ...");
        }
        
        Matrix L = null;
        
        try {
            // Compute the CHOLESKY
            L = stiffnessMatrixClone.cholesky();

            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Computed CHOLESKY for Matrix: " + stiffnessMatrixClone.getName());
            }
            
            if(localLogger.isEnabledFor(Level.DEBUG)) {
                localLogger.log(Level.DEBUG, "CHOLESKY: " + L);
            }
            
        } catch (CholeskyFailedException e) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error while computing the CHOLESKY for Matrix: " + stiffnessMatrixClone.getName());
            }
            
            return false;
        }

        /**
         * Calculate the results by solving K*u = -r
         * */
        
//    	// We need -1.0*r
//    	this.vectorR = this.vectorR.mult(-1.0);
        
        // Calculate Vector y from L*y=r
        Vector y = new Vector("y", this.completeStiffnessMatrix.size());
        
        // Run through all element positions in the vector y
        for(int i = 1; i <= this.completeStiffnessMatrix.size(); i++) {
            
            double sum = 0;
            
            // Calculate the sum l[i][k]*y[k] from k=1 to [i-1]
            for(int k = 1; k <= (i-1); k++) {
            
                sum = sum + (L.getValue(i, k) * y.getValue(k));
            }
            
            double newValue = (vectorR.getValue(i) - sum) / L.getValue(i, i);
            y.setValue(i, newValue);
        }
        
        // Solve "L(transponated) * u = y" by calculating Vector u
        Vector u = new Vector("u", this.completeStiffnessMatrix.size());
        
        // Run through all element positions in the vector u
        for(int i = this.completeStiffnessMatrix.size(); i >= 1; i--) {
            
            double sum = 0;
            
            // Calculate the sum l[k][i] * u[k] from k=i+1 to size of the matrix
            for(int k = (i+1); k <= this.completeStiffnessMatrix.size(); k++) {
            
                sum = sum + (L.getValue(k, i) * u.getValue(k));
            }
            
            double newValue = (y.getValue(i) - sum) / L.getValue(i, i);
            u.setValue(i, newValue);
        }    

        // Calculation for Vector u has been completed, save the results in the FEMManager
        femManager.addToResultList(u);
        
        if(localLogger.isEnabledFor(Level.DEBUG)) {
            localLogger.log(Level.DEBUG, "Vector u calculation successfull: " + u);
        }

        return true;
    }
    
    /**
     * Modify the positions of the nodes depending on the result vectors
     * @return true if all went well, false otherwise
     */
    protected boolean modifyNodePositions() {
    	
    	// Grab the full node list
    	NodeList fullNodeList = femManager.getNodeList();
    	
    	// Grab the result vector
    	Vector resultVector = femManager.getResultList().get(0);
		
    	// Run through all nodes
    	for(int i = 1; i <= fullNodeList.size(); i++) {
    		
    		Node currentNode = fullNodeList.getNodeInCorrectOrder(i);
    		
    		int vectorIndex = currentNode.getId()*2;
    		
    		double modXValue = resultVector.getValue(vectorIndex-1);
    		double modYValue = resultVector.getValue(vectorIndex);
    		
    		currentNode.setX(currentNode.getX()+(modXValue*ApplicationConstants.SPANNER_ENHANCE_MOVEMENT_FACTOR));
    		currentNode.setY(currentNode.getY()+(modYValue*ApplicationConstants.SPANNER_ENHANCE_MOVEMENT_FACTOR));
    	}
    	
    	// Store the full node
    	femManager.setNodeList(fullNodeList);
    	
    	/**
    	 * RESULT COMPARISON WITH THE BASIC FIGURE AND THE SCHWARZ RESULTS (Fig.6.19)
    	 */
    	
//        if(localLogger.isEnabledFor(Level.INFO)) {
//            localLogger.log(Level.INFO, "Schwarz Book Points: 'Grundeinteilung'");
//            localLogger.log(Level.INFO, "Point A = Node 1 (u/v): " + resultVector.getValue((1*2)-1) + " / " + resultVector.getValue(1*2));
//            localLogger.log(Level.INFO, "Point B = Node 26 (u/v): " + resultVector.getValue((26*2)-1) + " / " + resultVector.getValue(26*2));
//            localLogger.log(Level.INFO, "Point C = Node 45 (u/v): " + resultVector.getValue((45*2)-1) + " / " + resultVector.getValue(45*2));
//            localLogger.log(Level.INFO, "Point D = Node 87 (u/v): " + resultVector.getValue((87*2)-1) + " / " + resultVector.getValue(87*2));
//            localLogger.log(Level.INFO, "Point E = Node 96 (u/v): " + resultVector.getValue((96*2)-1) + " / " + resultVector.getValue(96*2));
//        }
        
        // APPORACH S-MATRIX
        
//        if(localLogger.isEnabledFor(Level.INFO)) {
//            localLogger.log(Level.INFO, "Approach - S-MATRIX - Differenz to Schwarz Points case III, Tab.6.19: 'Grundeinteilung'");
//            localLogger.log(Level.INFO, "Diff Point A (u/v): " + Math.abs((0.00002)-resultVector.getValue((1*2)-1)) + " / " + Math.abs((0.00171)-resultVector.getValue(1*2)));
//            localLogger.log(Level.INFO, "Diff Point B (u/v): " + Math.abs((-0.00030)-resultVector.getValue((26*2)-1)) + " / " + Math.abs((-0.00082)-resultVector.getValue(26*2)));
//            localLogger.log(Level.INFO, "Diff Point C (u/v): " + Math.abs((0.00027)-resultVector.getValue((45*2)-1)) + " / " + Math.abs((-0.00302)-resultVector.getValue(45*2)));
//            localLogger.log(Level.INFO, "Diff Point D (u/v): " + Math.abs((-0.00272)-resultVector.getValue((87*2)-1)) + " / " + Math.abs((-0.02297)-resultVector.getValue(87*2)));
//            localLogger.log(Level.INFO, "Diff Point E (u/v): " + Math.abs((-0.00781)-resultVector.getValue((96*2)-1)) + " / " + Math.abs((-0.04294)-resultVector.getValue(96*2)));
//        }
        
        // APPROACH INTEGRAL
        
//        if(localLogger.isEnabledFor(Level.INFO)) {
//            localLogger.log(Level.INFO, "Approach - INTEGRAL - Differenz to Schwarz Points case V, Tab.6.19: 'Grundeinteilung'");
//            localLogger.log(Level.INFO, "Diff Point A (u/v): " + Math.abs((0.00003)-resultVector.getValue((1*2)-1)) + " / " + Math.abs((0.00203)-resultVector.getValue(1*2)));
//            localLogger.log(Level.INFO, "Diff Point B (u/v): " + Math.abs((-0.00025)-resultVector.getValue((26*2)-1)) + " / " + Math.abs((-0.00098)-resultVector.getValue(26*2)));
//            localLogger.log(Level.INFO, "Diff Point C (u/v): " + Math.abs((0.00036)-resultVector.getValue((45*2)-1)) + " / " + Math.abs((-0.00341)-resultVector.getValue(45*2)));
//            localLogger.log(Level.INFO, "Diff Point D (u/v): " + Math.abs((-0.00288)-resultVector.getValue((87*2)-1)) + " / " + Math.abs((-0.02432)-resultVector.getValue(87*2)));
//            localLogger.log(Level.INFO, "Diff Point E (u/v): " + Math.abs((-0.00816)-resultVector.getValue((96*2)-1)) + " / " + Math.abs((-0.04497)-resultVector.getValue(96*2)));
//        }
    	
    	/**
    	 * RESULT COMPARISON WITH THE FINE FIGURE AND THE SCHWARZ RESULTS (Fig.6.20)
    	 */
        
//        if(localLogger.isEnabledFor(Level.INFO)) {
//            localLogger.log(Level.INFO, "Schwarz Book Points: 'Verfeinerung'");
//            localLogger.log(Level.INFO, "Point A = Node 1 (u/v): " + resultVector.getValue((1*2)-1) + " / " + resultVector.getValue(1*2));
//            localLogger.log(Level.INFO, "Point B = Node 192 (u/v): " + resultVector.getValue((192*2)-1) + " / " + resultVector.getValue(192*2));
//            localLogger.log(Level.INFO, "Point C = Node 157 (u/v): " + resultVector.getValue((157*2)-1) + " / " + resultVector.getValue(157*2));
//            localLogger.log(Level.INFO, "Point D = Node 50 (u/v): " + resultVector.getValue((50*2)-1) + " / " + resultVector.getValue(50*2));
//            localLogger.log(Level.INFO, "Point E = Node 58 (u/v): " + resultVector.getValue((58*2)-1) + " / " + resultVector.getValue(58*2));
//        }
    	
        // APPROACH S-MATRIX

//      if(localLogger.isEnabledFor(Level.INFO)) {
//          localLogger.log(Level.INFO, "Approach - S-MATRIX - Differenz to Schwarz Points case IV, Tab.6.19: 'Verfeinerung'");
//          localLogger.log(Level.INFO, "Diff Point A (u/v): " + Math.abs((0.00005)-resultVector.getValue((1*2)-1)) + " / " + Math.abs((0.00205)-resultVector.getValue(1*2)));
//          localLogger.log(Level.INFO, "Diff Point B (u/v): " + Math.abs((-0.00027)-resultVector.getValue((192*2)-1)) + " / " + Math.abs((-0.00099)-resultVector.getValue(192*2)));
//          localLogger.log(Level.INFO, "Diff Point C (u/v): " + Math.abs((0.00036)-resultVector.getValue((157*2)-1)) + " / " + Math.abs((-0.00352)-resultVector.getValue(157*2)));
//          localLogger.log(Level.INFO, "Diff Point D (u/v): " + Math.abs((-0.00295)-resultVector.getValue((50*2)-1)) + " / " + Math.abs((-0.02468)-resultVector.getValue(50*2)));
//          localLogger.log(Level.INFO, "Diff Point E (u/v): " + Math.abs((-0.00829)-resultVector.getValue((58*2)-1)) + " / " + Math.abs((-0.04546)-resultVector.getValue(58*2)));
//      }
        
        // APPROACH INTEGRAL

//        if(localLogger.isEnabledFor(Level.INFO)) {
//            localLogger.log(Level.INFO, "Approach - INTEGRAL - Differenz to Schwarz Points case VI, Tab.6.19: 'Verfeinerung'");
//            localLogger.log(Level.INFO, "Diff Point A (u/v): " + Math.abs((0.00005)-resultVector.getValue((1*2)-1)) + " / " + Math.abs((0.00214)-resultVector.getValue(1*2)));
//            localLogger.log(Level.INFO, "Diff Point B (u/v): " + Math.abs((-0.00026)-resultVector.getValue((192*2)-1)) + " / " + Math.abs((-0.00105)-resultVector.getValue(192*2)));
//            localLogger.log(Level.INFO, "Diff Point C (u/v): " + Math.abs((0.00036)-resultVector.getValue((157*2)-1)) + " / " + Math.abs((-0.00365)-resultVector.getValue(157*2)));
//            localLogger.log(Level.INFO, "Diff Point D (u/v): " + Math.abs((-0.00301)-resultVector.getValue((50*2)-1)) + " / " + Math.abs((-0.02517)-resultVector.getValue(50*2)));
//            localLogger.log(Level.INFO, "Diff Point E (u/v): " + Math.abs((-0.00842)-resultVector.getValue((58*2)-1)) + " / " + Math.abs((-0.04620)-resultVector.getValue(58*2)));
//        }
    	
    	return true;
    }
    
    /**
     * Calculate Sigma
     * @return true if all went well, false otherwise
     */
    protected boolean calcSigma() {
    	
    	// The element list
    	ElementList elementList = femManager.getElementList();
    	
    	// The only full result vector containing u and v
    	Vector result = femManager.getResultList().get(0);    
    	
    	// The max "Spannungsbetrag"
    	double maxSpannungsBetrag = -1;
    	
    	// Build Matrix D
    	Matrix D = MatrixFactory.createUserDefinedMatrix("Matrix D", 3);
    	
    	double matConstantV = femManager.getSpannerMaterialConstantV();
    	double matConstantE = femManager.getSpannerMaterialConstantE();
    	
    	double factor = matConstantE/(1- Math.pow(matConstantV,2));
    	
    	D.storeLine(1, new double[] {1, matConstantV, 0});
    	D.storeLine(2, new double[] {matConstantV, 1, 0});
    	D.storeLine(3, new double[] {0, 0, ((1-matConstantV)/2)});
    	
    	// Apply the factor for D
    	D = D.mult(factor);
    	
        // Run through all elements
        for(int elementIndex = 1; elementIndex <= elementList.size(); elementIndex++) {   
        	
        	Element currentElement = elementList.getElement(elementIndex);
        	NodeList nodeList = currentElement.getNodeList();        	

        	// Build u-schlange depending on the current element
        	Vector uS = new Vector("U-Schlange",nodeList.size());
        	Vector vS = new Vector("V-Schlange",nodeList.size());
        	
        	for(int i = 1; i <= nodeList.size(); i++) {
        		
        		Node node = nodeList.getNodeInCorrectOrder(i);
        		int nodeId = node.getId();
        		
        		// Grab U and V of this node
        		double uNode = result.getValue(2*nodeId-1);
        		double vNode = result.getValue(2*nodeId);
        		
        		uS.setValue(i,uNode);
        		vS.setValue(i,vNode);
        	}
        	
        	// The basic variables we use
            Method N_XI = null;
            Method N_ETA = null;
            Method XI_X= null;
            Method XI_Y= null;
            Method ETA_X = null;
            Method ETA_Y = null;
            int z = -1;
            double[][] figurePoints = null;
            
            // Tell the static function class about the X and Y values of the current element
            CompilationFunctions.storeElementXYValues(currentElement);
            
        	// Fill the basic variables, depending if the current Element is a Triangle or a Quadrat
        	if(currentElement.getElementType() == GaussConstants.FIGURE_QUADRAT){
        		
        		try {
        			
					N_XI = CompilationFunctions.getMethod(CompilationFunctions.QUADRAT_QUADRATIC_N_XI);
					XI_X = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_QUADRAT_QUADRATIC_XI_X);
					N_ETA = CompilationFunctions.getMethod(CompilationFunctions.QUADRAT_QUADRATIC_N_ETA);
					ETA_X = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_QUADRAT_QUADRATIC_ETA_X);
					XI_Y = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_QUADRAT_QUADRATIC_XI_Y);
					ETA_Y = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_QUADRAT_QUADRATIC_ETA_Y);
					z = 8;
					
					// These are the points transfered to the local coordinate system with XI and ETA
					//double[][] quadratPoints = {{0, 0}, {1,0}, {1,1}, {0,1}, {0.5,0}, {1,0.5},{0.5,1}, {0, 0.5}};
					double[][] quadratPoints = {{0.5, 0.5}, {0.5,0.5}, {0.5,0.5}, {0.5,0.5}, {0.5,0.5}, {0.5,0.5},{0.5,0.5}, {0.5, 0.5}};
					figurePoints = quadratPoints;
					
				} catch (NoSuchMethodException e) {
					
			        if(localLogger.isEnabledFor(Level.ERROR)) {
			            localLogger.log(Level.ERROR, "Method not found !", e);
			        }
			        
			        return false;
				}
				
        	} else if(currentElement.getElementType() == GaussConstants.FIGURE_TRIANGLE) {
        		
        		try {
					N_XI = CompilationFunctions.getMethod(CompilationFunctions.TRIANGLE_QUADRATIC_N_XI);
					XI_X = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_TRIANGLE_QUADRATIC_XI_X);
					N_ETA = CompilationFunctions.getMethod(CompilationFunctions.TRIANGLE_QUADRATIC_N_ETA);
					ETA_X = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_TRIANGLE_QUADRATIC_ETA_X);
					XI_Y = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_TRIANGLE_QUADRATIC_XI_Y);
					ETA_Y = CompilationFunctions.getMethod(CompilationFunctions.SPANNER_TRIANGLE_QUADRATIC_ETA_Y);
					z = 6;
					
					//  These are the points transfered to the local coordinate system with XI and ETA
					//double[][] trianglePoints = {{0, 0}, {1,0}, {0,1}, {0.5,0}, {0.5,0.5}, {0,0.5}};
					double[][] trianglePoints = {{1.0/3.0, 1.0/3.0}, {1.0/3.0,1.0/3.0}, {1.0/3.0,1.0/3.0}, {1.0/3.0,1.0/3.0}, {1.0/3.0,1.0/3.0}, {1.0/3.0,1.0/3.0}};
					figurePoints = trianglePoints;
					
				} catch (NoSuchMethodException e) {
					
			        if(localLogger.isEnabledFor(Level.ERROR)) {
			            localLogger.log(Level.ERROR, "Method not found !", e);
			        }
			        
			        return false;
				}
        	}
        	
        	// The variables we need to calculate to create the vector epsilon
        	double uX= 0;
        	double uY = 0;
        	double vX = 0;
        	double vY = 0;
        	
        	try {
        		
				// uXI = u_e^T * N_XI
				double uXI = 0;
				for(int i = 1; i <= z; i++) {
					uXI = uXI + uS.getValue(i) * callMethodCalculation(N_XI, figurePoints[i-1][0], figurePoints[i-1][1], i);
				}
				
				// uETA = u_e^T * N_ETA
				double uETA = 0;
				for(int i = 1; i <= z; i++) {
					uETA = uETA + uS.getValue(i) * callMethodCalculation(N_ETA, figurePoints[i-1][0], figurePoints[i-1][1], i);
				}
				
				// vXI = v_e^T * N_XI
				double vXI = 0;
				for(int i = 1; i <= z; i++) {
					vXI = vXI + vS.getValue(i) * callMethodCalculation(N_XI, figurePoints[i-1][0], figurePoints[i-1][1], i);
				}
				
				// vETA = v_e^T * N_ETA
				double vETA = 0;
				for(int i = 1; i <= z; i++) {
					vETA = vETA + vS.getValue(i) * callMethodCalculation(N_ETA, figurePoints[i-1][0], figurePoints[i-1][1], i);
				}
				
				// uX = uXI * XI_x + uETA * ETA_x
				uX = 0;
				double firstPart =0;
				double secondPart = 0;
				for(int i = 1; i <= z; i++) {
					
					firstPart = firstPart + uXI * callMethodCalculation(XI_X, figurePoints[i-1][0], figurePoints[i-1][1], i);
					secondPart = secondPart + uETA * callMethodCalculation(ETA_X, figurePoints[i-1][0], figurePoints[i-1][1], i);
				}
				uX = firstPart + secondPart;
				
				// uY = uXI * XI_y + uETA * ETA_y
				uY = 0;
				firstPart =0;
				secondPart = 0;
				for(int i = 1; i <= z; i++) {
					
					firstPart = firstPart + uXI * callMethodCalculation(XI_Y, figurePoints[i-1][0], figurePoints[i-1][1], i);
					secondPart = secondPart + uETA * callMethodCalculation(ETA_Y, figurePoints[i-1][0], figurePoints[i-1][1], i);
				}
				uY = firstPart + secondPart;
				
				// vX = vXI * XI_x + vETA * ETA_x
				vX = 0;
				firstPart = 0;
				secondPart = 0;
				for(int i = 1; i <= z; i++) {
					
					firstPart = firstPart + vXI * callMethodCalculation(XI_X, figurePoints[i-1][0], figurePoints[i-1][1], i);
					secondPart = secondPart + vETA * callMethodCalculation(ETA_X, figurePoints[i-1][0], figurePoints[i-1][1], i);
				}
				vX = firstPart + secondPart;
				
				// vY = vXI * XI_y + vETA * ETA_y
				vY = 0;
				firstPart = 0;
				secondPart = 0;
				for(int i = 1; i <= z; i++) {
					
					firstPart = firstPart + vXI * callMethodCalculation(XI_Y, figurePoints[i-1][0], figurePoints[i-1][1], i);
					secondPart = secondPart + vETA * callMethodCalculation(ETA_Y, figurePoints[i-1][0], figurePoints[i-1][1], i);
				}
				vY = firstPart + secondPart;
				
			} catch (FunctionCalculationException e) {
				
		        if(localLogger.isEnabledFor(Level.ERROR)) {
		            localLogger.log(Level.ERROR, "Function Calculation Exception !");
		        }
		        
		        // Reset the ElementIndexes
		        this.maxSigmaElementIndex = -1;
			}
        	
        	// Build the Vector Epsilon
        	Vector epsilon = new Vector("Epsilon for Element #" + elementIndex, 3);
        	epsilon.setValue(1, uX);
        	epsilon.setValue(2, vY);
        	epsilon.setValue(3, uY + vX);
        	
        	// Calculate sigma
        	Vector sigma = D.mult(epsilon);
        	sigma.setName("Sigma for Element #" + elementIndex);
        	
        	// Calculate the "Betrag" of all "Spannungen" in Sigma
        	double value = Math.sqrt(Math.pow(sigma.getValue(1), 2) + Math.pow(sigma.getValue(2), 2) + Math.pow(sigma.getValue(3), 2));        	
        	
        	// Store the "Spannungsbetrag" in the Element
        	currentElement.setSpannungsBetrag(value);
        	
        	// Check if we found a new max "Betrag"
        	if(value > maxSpannungsBetrag) {
        		
        		// Remember the value and the index of the element where it occured
        		maxSpannungsBetrag = value;
        		this.maxSigmaElementIndex = elementIndex;
        	}
        }
        
    	if(localLogger.isEnabledFor(Level.INFO)) {
    	      localLogger.log(Level.INFO, "Max 'Spannungsbetrag' found in Element #" + this.maxSigmaElementIndex + " : " + maxSpannungsBetrag);
    	  }
        
        return true;
    }
    
	/**
	 * Call the given method with the given values
	 * @param a, the method to use
	 * @param x, the x-value to use
	 * @param y, the y-value to use
	 * @param position, the position if applyable
	 * @return the resulting double value
	 */
	private static double callMethodCalculation(Method a, double x, double y, int position) throws FunctionCalculationException {
	    
	    // Create the object array
	    Object[] args = new Object[3];
	    
	    args[0] = new Double(x);
	    args[1] = new Double(y);
	    args[2] = new Integer(position);
	    
	    // Make sure to catch all exceptions
	    try {
            Object returnValue = a.invoke(null, args);
            
            // Check if we received a correct type of value
            if(returnValue instanceof Double) {
                
                // Correct double value
                return ((Double)(returnValue)).doubleValue();
                
            } else if(returnValue instanceof Integer) {
                
                // Correct double value
                return ((Integer)(returnValue)).doubleValue();
                
            } else {
                
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "The return type of the function method is neither Double nor Integer : " + returnValue);
                }
                
                throw new FunctionCalculationException();
            }
            
        } catch (IllegalArgumentException e) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "The method signature did not match: (double, double, int) !");
            }
            
            throw new FunctionCalculationException();
            
        } catch (IllegalAccessException e) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "The called method is not accessible. Perhaps private?");
            }
            
            throw new FunctionCalculationException();
            
        } catch (InvocationTargetException e) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "An exception has been thrown by an underlying method !");
            }
            
            throw new FunctionCalculationException();
            
        } catch(Exception e) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "An general exception has been thrown !");
            }
            
            throw new FunctionCalculationException();
        }
	}
	
	/**
	 * @return Returns the maxSigmaElementIndex.
	 */
	public int getMaxSigmaElementIndex() {
		return maxSigmaElementIndex;
	}
}
