/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.calc;

import java.lang.reflect.Method;
import java.util.List;

import org.apache.log4j.Level;

import fhm.edu.fem.calc.objects.Border;
import fhm.edu.fem.calc.objects.CTab;
import fhm.edu.fem.calc.objects.CTabList;
import fhm.edu.fem.calc.objects.Calculation;
import fhm.edu.fem.calc.objects.CompilationFunctions;
import fhm.edu.fem.calc.objects.Condition;
import fhm.edu.fem.calc.objects.Element;
import fhm.edu.fem.calc.objects.ElementList;
import fhm.edu.fem.calc.objects.LineIntegralFunctions;
import fhm.edu.fem.calc.objects.MathObject;
import fhm.edu.fem.calc.objects.Node;
import fhm.edu.fem.calc.objects.NodeList;
import fhm.edu.fem.calc.objects.Vector;
import fhm.edu.fem.calc.objects.matrix.Matrix;
import fhm.edu.fem.constants.ApplicationConstants;
import fhm.edu.fem.constants.GaussConstants;
import fhm.edu.fem.exception.CholeskyFailedException;

/**
 * Class Description:
 *		This class extends from the FEMCalcGauss class taking care
 *		of the whole calculation. The difference is that this class 
 *		overwrites the method start() as well as the method calculation()
 *		from FEMCalc which is the base-class. It makes use of the method
 *		compilation() from FEMCalcGauss.
 *
 *		This class is being used when we have a FEM3 File with CTabs
 *		used for temperature calculations.
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class FEMCalcTemperature extends FEMCalcGauss {
    
    /* Vector B */
    protected Vector vectorB = null;
    
    /* f(x,y) constant value */
    protected double fxyValue = -20;
    
    /* Alpha constant value */
    protected double alpha = 2;
    
	/* Inner and Outer Points ("Stuetzstellen des inneren und aeusseren Integrals") */
	protected static int OUTER_POINTS_LINE_INTEGRAL = 6;
	
    /**
     * Create a new FEM Calculator for Temperature Tasks
     */
    public FEMCalcTemperature() {
        
    	// Call the super constructor
        super();
    	
        // Initialize the vector B
        this.vectorB = new Vector("Complete-Vector B", femManager.getNodeList().size());
    }
    
    /**
     * Call this method to start the calculation
     * @return true if all went well, false otherwise
     */
    public boolean start() {
        
        // Remember the starting time
        this.startTime = System.currentTimeMillis();
        
        // Clear the result list of the FEMManager
        femManager.clearResultList();
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Doing the compilation of the complete-stiffness-matrix ...");
        }
        
        // Do the compilation
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.compilation()) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error during compilation of the complete-stiffness-matrix !");
            }
            
            return false;
        }
        
        // Calc the time needed for the compilation
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Compilation of the Complete-Stiffness-Matrix successfully completed: " + this.tmpTime + " ms");
            localLogger.log(Level.INFO, "Size of the Complete-Stiffness-Matrix: " + this.completeStiffnessMatrix.size() + "x" + this.completeStiffnessMatrix.size());
            localLogger.log(Level.INFO, "Required Memory of the Complete-Stiffness-Matrix: " + this.completeStiffnessMatrix.getRequiredMemory() + " bytes");
        }
        
        this.tmpTime = System.currentTimeMillis();
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Testing if the Complete-Stiffness-Matrix is symmetric ...");
        }
        
        // Compilation done, test if the complete stiffness matrix is symmetric
        if(!this.completeStiffnessMatrix.isSymmetric()) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Complete Stiffness Matrix is not symmetric !");
            }
            
            return false;
        }
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Testing if all elements on the main diagonal in the Complete-Stiffness-Matrix are not zero ...");
        }
        
        // Check if the main diagonal of the complete stiffness matrix is not zero
        if(!this.completeStiffnessMatrix.isMainDiagonalNotZero()) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: At least one element on the main diagonal of the complete stiffness matrix is zero !");
            }
            
            return false;
        }
        
        // Calc the time needed for the tests
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Time needed for the tests: " + this.tmpTime + " ms");
        }
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Building Vector B ...");
        }
        
        // Do the compilation
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.createVectorB()) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error during creation of Vector B !");
            }
            
            return false;
        }
        
        // Calc the time needed for the compilation
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Creation of Vector B successfully completed: " + this.tmpTime + " ms");
            localLogger.log(Level.INFO, "Size of the Vector B: " + this.vectorB.size());
        }
        
        // Build the line integral
        this.tmpTime = System.currentTimeMillis();
        
        if(!this.createLineIntegral()) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error during the creation of the line integral !");
            }
            
            return false;
        }
        
        // Calc the time needed for the creation of the line integral
        this.tmpTime = System.currentTimeMillis() - this.tmpTime;
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Creation of the Line Integral successfully completed: " + this.tmpTime + " ms");
            localLogger.log(Level.INFO, "Size of the new Complete-Stiffness-Matrix: " + this.completeStiffnessMatrix.size() + "x" + this.completeStiffnessMatrix.size());
            localLogger.log(Level.INFO, "Required Memory of the new Complete-Stiffness-Matrix: " + this.completeStiffnessMatrix.getRequiredMemory() + " bytes");
        }
        
        // Do the calculations        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Doing the calculations ...");
        }
        
        // Run through all calculations        
        List<Calculation> calculationList = femManager.getCalculationList();
        
        for(int calcIndex = 0; calcIndex < calculationList.size(); calcIndex++) {
            
            // Remember the time at the start of this calculation
            this.tmpTime = System.currentTimeMillis();
            
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Starting calculation: " + calcIndex);
            }
            
            // Grab calculation
            Calculation calc = calculationList.get(calcIndex);
            
            // Do the calculation
            if(!this.calculate(calc)) {
                
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error during calculation: " + calcIndex);
                }
                
                return false;
            }
            
            // Calculate the time needed to complete the current calculation
            this.tmpTime = System.currentTimeMillis() - this.tmpTime;
            
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Calculation " + calcIndex + " successfully completed: " + this.tmpTime + " ms");
            }
        }
        
        // Calculate the total time
        this.totalTime = System.currentTimeMillis() - this.startTime;

        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "All calculations successfully completed - Total time: " + this.totalTime + " ms");
        }
        
        return true;
    }
    
    /**
     * Do the compilation of the vector B
     * @return true if all went well, false otherwise
     */
    protected boolean createVectorB() {
        
        try {
            // Grab the approach from the manager
            int femApproach = femManager.getFEMApproach();
            
            // Grab the list of all elements
            ElementList elementlist = femManager.getElementList();
            
            // The variables to use in the Gauss Quadrature Method call
            double a = 0;
            double b = 1;
            Method h1 = null;
            Method h2 = null;
            Method q = null;
            int z1 = -1;
            int z2 = -1;
    		Vector ax = GaussConstants.get(INNER_POINTS, GaussConstants.VALUE_A);
    		Vector hx = GaussConstants.get(INNER_POINTS, GaussConstants.VALUE_H);
    		Vector ay = GaussConstants.get(OUTER_POINTS, GaussConstants.VALUE_A);
    		Vector hy = GaussConstants.get(OUTER_POINTS, GaussConstants.VALUE_H);
            MathObject integralResultObject = null;
            
            // Calculate how many percentage one element makes up
            double oneElementPercentage = 100.0 / elementlist.size();
            
            // Calculate how many elements make up one percent
            double onePercentElements = elementlist.size() / 100.0;
            
            // Where we store the current actual percentage
            double currentPercentage = 0.0;
            
            // Run through all elements
            for(int elementIndex = 1; elementIndex <= elementlist.size(); elementIndex++)
            {     
                // Calculate the current percentage of elements already done in the compilation
                currentPercentage = (elementIndex-1) * oneElementPercentage;

                // Make a percentage output about every 5% of completed elements
                if(elementIndex%(onePercentElements*5.0) < 1) {
                    
                    if(localLogger.isEnabledFor(Level.INFO)) {
                        localLogger.log(Level.INFO, "Built at " + currentPercentage + " %");
                    }
                }
                
                // Get the current element
                Element tmpElement =  elementlist.getElement(elementIndex);
                
                // Store the x- and y-values of the element nodes in the global data structure for the gauss quadrature
                CompilationFunctions.storeElementXYValues(tmpElement);
                
                // Check which kind of element we have
                int nodeListSize = tmpElement.getNodeList().size();
                int elementType = -1;
                
                if(nodeListSize == 3) {
                    
                    // Linear triangle
            		h1 = CompilationFunctions.getMethod(CompilationFunctions.TRIANGLE_LINEAR_N);
            		h2 = CompilationFunctions.getMethod(CompilationFunctions.ONE);
            		q = CompilationFunctions.getMethod(CompilationFunctions.TRIANGLE_LINEAR_ONETHROUGHJAC);
            		z1 = 3;
            		z2 = 1;
            		
            		elementType = GaussConstants.FIGURE_TRIANGLE;
                    
                } else if(nodeListSize == 4) {
                    
                    // Linear quadrat
            		h1 = CompilationFunctions.getMethod(CompilationFunctions.QUADRAT_LINEAR_N);
            		h2 = CompilationFunctions.getMethod(CompilationFunctions.ONE);
            		q = CompilationFunctions.getMethod(CompilationFunctions.QUADRAT_LINEAR_ONETHROUGHJAC);
            		z1 = 4;
            		z2 = 1;
            		
            		elementType = GaussConstants.FIGURE_QUADRAT;
                    
                } else if(nodeListSize == 6) {
                    
                    // "Krummes" triangle
            		h1 = CompilationFunctions.getMethod(CompilationFunctions.TRIANGLE_QUADRATIC_N);
            		h2 = CompilationFunctions.getMethod(CompilationFunctions.ONE);
            		q = CompilationFunctions.getMethod(CompilationFunctions.TRIANGLE_QUADRATIC_ONETHROUGHJAC);
            		z1 = 6;
            		z2 = 1;
            		
            		elementType = GaussConstants.FIGURE_TRIANGLE;
                    
                } else if(nodeListSize == 8) {
                    
                    // "Krummes" quadrat
            		h1 = CompilationFunctions.getMethod(CompilationFunctions.QUADRAT_QUADRATIC_N);
            		h2 = CompilationFunctions.getMethod(CompilationFunctions.ONE);
            		q = CompilationFunctions.getMethod(CompilationFunctions.QUADRAT_QUADRATIC_ONETHROUGHJAC);
            		z1 = 8;
            		z2 = 1;
            		
            		elementType = GaussConstants.FIGURE_QUADRAT;
                    
                } else {
                    
	                if(localLogger.isEnabledFor(Level.ERROR)) {
	                    localLogger.log(Level.ERROR, "Unknown element type. Number of nodes: " + nodeListSize);
	                }
                }

                // Calculate the integral
                integralResultObject = FEMGaussQuadrature.calcIntegration(a, b, h1, h2, q, z1, z2, ax, hx, ay, hy, elementType);

                // Check if the result object matches our requirements
                if(!(integralResultObject instanceof Vector)) {
                	
                	// At least one of the result objects is not a Vector => ERROR
                    if(localLogger.isEnabledFor(Level.ERROR)) {
                        localLogger.log(Level.ERROR, "One of the integral result objects is not a Vector: " + integralResultObject.getClass());
                    }
                    
                    return false;
                }
                
                // Calculate Vector B_e * f(x,y)
                Vector bElementVector = ((Vector)(integralResultObject)).mult(this.fxyValue);
                
                // Compile the new element Vector B_e onto the Complete-Vector-B
                NodeList elementNodeList = tmpElement.getNodeList();
                
                for(int vecIndex = 1; vecIndex <= bElementVector.size(); vecIndex++) {
                    
                    Node tmpNode = elementNodeList.getNodeInCorrectOrder(vecIndex);
                    int id = tmpNode.getId();
                        
                    // Grab the current value from the element vector b and add it to the value in the gesamt vector b
                    double elementValue = bElementVector.getValue(vecIndex);
                    double newValue = this.vectorB.getValue(id) + elementValue;                       
                    this.vectorB.setValue(id, newValue);
                }  
            }
            
        } catch (Exception e) {
            
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Failure during compilation!", e);
			}
			
			return false;
        }
        
        return true;
    }
    
    /**
     * Build the line integral and compile it onto the complete stiffness matrix
     * @return true if all went well, false otherwise
     */
    protected boolean createLineIntegral() {
        
        try {
            
            // Grab the ctab list
            CTabList ctabs = femManager.getCTabList();
            
            // Run through all CTabs
            for(int ctabIndex = 1; ctabIndex <= ctabs.size(); ctabIndex++) {
                
                // The variables to use in the Gauss Quadrature Method call
                double a = 0;
                double b = 1;
                Method h1 = LineIntegralFunctions.getMethod(LineIntegralFunctions.N_SIGMA);
                Method q = LineIntegralFunctions.getMethod(LineIntegralFunctions.Q_ROOT);
                int z1 = 3;
        		Vector ay = GaussConstants.get(OUTER_POINTS_LINE_INTEGRAL, GaussConstants.VALUE_A);
        		Vector hy = GaussConstants.get(OUTER_POINTS_LINE_INTEGRAL, GaussConstants.VALUE_H);
                MathObject integralResultObject = null;
                
                // Grab the current ctab
                CTab tmpCtab = ctabs.getCTab(ctabIndex);

                // Store the x- and y-values of the ctab nodes in the global data structure for the gauss quadrature
                LineIntegralFunctions.storeCtabXYValues(tmpCtab);
                
                // Calculate the single gauss quadrature
                integralResultObject = FEMGaussQuadrature.calcSingleIntegration(a, b, h1, h1, q, z1, z1, ay, hy);
                
                // Check if the result object matches our requirements
                if(!(integralResultObject instanceof Matrix)) {
                	
                	// At least one of the result objects is not a Matrix => ERROR
                    if(localLogger.isEnabledFor(Level.ERROR)) {
                        localLogger.log(Level.ERROR, "One of the integral result objects is not a Matrix !");
                    }
                    
                    return false;
                }
                
                // Bring the alpha constant into play
                Matrix resultMatrix = ((Matrix)(integralResultObject)).mult(this.alpha);

                // Compile the resulting matrix onto the complete stiffness matrix
                for(int j = 1; j <= 3; j++) {
                    
                    // Grab the first node ID so we know where to place the value
                    int nodeAId = tmpCtab.getNode(j).getId();
                    
                    // Determine the starting valueof index K for the next lines, depending if we use a MATRIX_STANDARD or MATRIX_HULL
                    int startOfIndexK = -1;
                    
                    if(ApplicationConstants.MATRIX_USE == ApplicationConstants.MATRIX_STANDARD) {
                        
                        // We are using a standard matrix
                        startOfIndexK = 1;
                        
                    } else if(ApplicationConstants.MATRIX_USE == ApplicationConstants.MATRIX_HULL) {
                        
                        // We are using a hull matrix
                        startOfIndexK = j;
                    }
                    
                    for(int k = startOfIndexK; k <= 3; k++) {
                        
                        // Grab the second node ID so we know where to place the value
                        int nodeBId = tmpCtab.getNode(k).getId();
                        
                        // Grab the current value from the result matrix and add it to the value in the complete stiffness matrix
                        double resultValue = resultMatrix.getValue(j, k);
                        double newValue = this.completeStiffnessMatrix.getValue(nodeAId, nodeBId) + resultValue;                       
                        this.completeStiffnessMatrix.setValue(nodeAId, nodeBId, newValue);
                    }
                }
            }
            
            if(localLogger.isEnabledFor(Level.DEBUG)) {
                localLogger.log(Level.DEBUG, "Compilation complete !");
                localLogger.log(Level.DEBUG, "Matrix size: " + this.completeStiffnessMatrix.size());
                localLogger.log(Level.DEBUG, "CompleteStiffnessMatrix: " + this.completeStiffnessMatrix);
            }
            
        } catch (Exception e) {
            
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Failure during the creation of the line integral!", e);
			}
			
			return false;
        }
        
        return true;
    }
    /**
     * Do the calculation, overwritten from FEMCalc because we still need to get the results
     * of Vector B into the calculation
     * @param calc, the calculation object
     * @return true if all went well, false otherwise
     */
    protected boolean calculate(Calculation calc) {
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Starting to clone the complete-stiffness-Matrix ...");
        }

        // Clone the complete stiffness matrix for us to change
        Matrix stiffnessMatrixClone = (Matrix)(this.completeStiffnessMatrix.clone());
        
        /**
         * IMPLEMENT THE PREVIOUS RESULTS OF VECTOR B
         * */
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Starting to implement previous results of Vector B ...");
        }
        
        // We initialize the resulting vector b with the previously created Vector B
        Vector b  = (Vector)(this.vectorB.clone());
        
        // Normally we would add the results of Vector B to the results of the Line Integral, to do that while solving the calculation
        // we bring the results of Vector B onto the left side of the equation and therefore have to mulitply it with -1.0
        // K*u = -b
        b = b.mult(-1.0);

        /**
         * IMPLEMENT THE BORDER CONDITIONS
         * */
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Starting to implement the border conditions ...");
        }
        
        // Grab the conditions from the actual calculation
        List<Condition> conditionList = calc.getConditions();
        
        // Run through all conditions
        for(int i = 0; i < conditionList.size(); i++) {
        	
            // Grab the current condition
            Condition condition =  conditionList.get(i);
            
            // Retrieve the value of the condition
            double conditionValue = condition.getValue();
            
            // Get the border affected by this condition and grab all nodes from this border
            Border conditionBorder = condition.getBorder();
            NodeList borderNodeList = conditionBorder.getNodes();

            // Run through all nodes in the border
            for(int j = 1; j <= borderNodeList.size(); j++){
                
                // Grab the current node from the border node list
                Node tmpNode = borderNodeList.getNodeInCorrectOrder(j);

                // Retrieve the ID number of the current node
                int nodeId = tmpNode.getId();
                
                // Run through the vector b
                for(int k = 1; k <= this.completeStiffnessMatrix.size(); k++){
                    
                    // Calculate the new value of b[k] made up of b[k] - ( K[k][nodeID] * conditionValue )
                    double newValue = b.getValue(k) - (stiffnessMatrixClone.getValue(k, nodeId) * conditionValue);    

                    // Set the new value for b
                    b.setValue(k, newValue);     
                }

                // Set the condition value to the proper line in the vector b
                b.setValue(nodeId, conditionValue);
               
                // Fill the column and the line identified by the node ID with zeros
                stiffnessMatrixClone.setLineAndColumn(nodeId, 0);
                
                // Set the value "1" at the main diagonal element identified by K[nodeID][nodeID]
                stiffnessMatrixClone.setValue(nodeId, nodeId, 1);
           }
        }
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Implementing border conditions successfully completed");
            localLogger.log(Level.INFO, "New required Memory of the MODIFIED Complete-Stiffness-Matrix: " + stiffnessMatrixClone.getRequiredMemory() + " bytes");
        }

        /**
         * Calculate the CHOLESKY
         * */
        
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Starting to calculate the CHOLESKY ...");
        }
        
        Matrix L = null;
        
        try {
            // Compute the CHOLESKY
            L = stiffnessMatrixClone.cholesky();

            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Computed CHOLESKY for Matrix: " + stiffnessMatrixClone.getName());
            }
            
            if(localLogger.isEnabledFor(Level.DEBUG)) {
                localLogger.log(Level.DEBUG, "CHOLESKY: " + L);
            }
            
        } catch (CholeskyFailedException e) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error while computing the CHOLESKY for Matrix: " + stiffnessMatrixClone.getName());
            }
            
            return false;
        }

        /**
         * Calculate the results by solving K*u = b
         * */
        
        // Calculate Vector y from L*y=b
        Vector y = new Vector("y", this.completeStiffnessMatrix.size());
        
        // Run through all element positions in the vector y
        for(int i = 1; i <= this.completeStiffnessMatrix.size(); i++) {
            
            double sum = 0;
            
            // Calculate the sum l[i][k]*y[k] from k=1 to [i-1]
            for(int k = 1; k <= (i-1); k++) {
            
                sum = sum + (L.getValue(i, k) * y.getValue(k));
            }
            
            double newValue = (b.getValue(i) - sum) / L.getValue(i, i);
            y.setValue(i, newValue);
        }
        
        // Solve "L(transponated) * u = y" by calculating Vector u
        Vector u = new Vector("u", this.completeStiffnessMatrix.size());
        
        // Run through all element positions in the vector u
        for(int i = this.completeStiffnessMatrix.size(); i >= 1; i--) {
            
            double sum = 0;
            
            // Calculate the sum l[k][i] * u[k] from k=i+1 to size of the matrix
            for(int k = (i+1); k <= this.completeStiffnessMatrix.size(); k++) {
            
                sum = sum + (L.getValue(k, i) * u.getValue(k));
            }
            
            double newValue = (y.getValue(i) - sum) / L.getValue(i, i);
            u.setValue(i, newValue);
        }    
        
        // Calculation for Vector u has been completed, save the results in the FEMManager
        femManager.addToResultList(u);
        
        if(localLogger.isEnabledFor(Level.DEBUG)) {
            localLogger.log(Level.DEBUG, "Vector u calculation successfull: " + u);
        }
        
        // Display the results for debug purposes
        FEMCalcTemperature.displaySpecificResults(u);

        return true;
    }
    
    /**
     * This method displays the specific results for certain nodes
     * we need to compare the results with the ones from the "Schwartz" book
     * @param u, the result vector
     */
    protected static void displaySpecificResults(Vector u) {
        
//        if(localLogger.isEnabledFor(Level.INFO)) {
//            localLogger.log(Level.INFO, "Specific Results: NORMAL VERSION");
//            localLogger.log(Level.INFO, "	A = " + u.getValue(1));
//            localLogger.log(Level.INFO, "	B = " + u.getValue(11));
//            localLogger.log(Level.INFO, "	C = " + u.getValue(56));
//            localLogger.log(Level.INFO, "	D = " + u.getValue(154));
//            localLogger.log(Level.INFO, "	E = " + u.getValue(205));
//            localLogger.log(Level.INFO, "	F = " + u.getValue(193));
//            localLogger.log(Level.INFO, "	H = " + u.getValue(150));
//            localLogger.log(Level.INFO, "	I = " + u.getValue(111));
//            localLogger.log(Level.INFO, "	K = " + u.getValue(86));
//            localLogger.log(Level.INFO, "	L = " + u.getValue(110));
//            localLogger.log(Level.INFO, "	M = " + u.getValue(57));
//        }
        
//        if(localLogger.isEnabledFor(Level.INFO)) {
//            localLogger.log(Level.INFO, "Specific Results: FINE VERSION");
//            localLogger.log(Level.INFO, "	A = " + u.getValue(1));
//            localLogger.log(Level.INFO, "	B = " + u.getValue(35));
//            localLogger.log(Level.INFO, "	C = " + u.getValue(203));
//            localLogger.log(Level.INFO, "	D = " + u.getValue(559));
//            localLogger.log(Level.INFO, "	E = " + u.getValue(745));
//            localLogger.log(Level.INFO, "	F = " + u.getValue(704));
//            localLogger.log(Level.INFO, "	H = " + u.getValue(555));
//            localLogger.log(Level.INFO, "	I = " + u.getValue(409));
//            localLogger.log(Level.INFO, "	K = " + u.getValue(322));
//            localLogger.log(Level.INFO, "	L = " + u.getValue(410));
//            localLogger.log(Level.INFO, "	M = " + u.getValue(201));
//        }
    }
}