/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.calc;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import fhm.edu.fem.app.FEMStartup;
import fhm.edu.fem.calc.objects.MathObject;
import fhm.edu.fem.calc.objects.Vector;
import fhm.edu.fem.calc.objects.matrix.Matrix;
import fhm.edu.fem.calc.objects.matrix.MatrixStandard;
import fhm.edu.fem.constants.GaussConstants;
import fhm.edu.fem.exception.FunctionCalculationException;
import fhm.edu.fem.util.logging.LoggingManager;

/**
 * Class Description:
 *		Calculates the Gauss Quadrature (2.Leistungsnachweis)
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class FEMGaussQuadrature {
    
	/* Our local logger object */
	protected static Logger localLogger = LoggingManager.getLogger(FEMGaussQuadrature.class.getName());

	/** 
	 * To make sure noone is using this constructor.
	 * Else the compiler would create a public one.
	 * */
	private FEMGaussQuadrature() {}
	
	/**
	 * Calculate the Gauss-Quadrature to solve an integral
	 * @param a, lower-border of the outer-integral
	 * @param b, upper-border of the outer-integral
	 * @param c, the lower-border of the inner-integral
	 * @param d, the upper-border of the outer-integral
	 * @param h1, first part of the integral
	 * @param h2, second part of the integral
	 * @param q, third part of the integral
	 * @param z1, size of the vector in h1 (used to determine the position integer)
	 * @param z2, size of the vector in h2 (used to determine the position integer)
	 * @param ax, "stuetzstellen" vector of "a" for x
	 * @param hx "gewichtung" vector of "h" for x
	 * @param ay "stuetzstellen" vector of "a" for y
	 * @param hy "gewichtung" vector of "h" for y
	 * @param figure, may be FIGURE_TRIANGLE or FIGURE_PARALLELOGRAM
	 * @return result matrix or vector or value or whatever it calculates
	 */
	public static MathObject calcDoubleIntegration(double a, double b, double c, double d, Method h1, Method h2, Method q, 
	        								 									   			   int z1, int z2, Vector ax, Vector hx, Vector ay, Vector hy) {

	    // Create the resulting MathObject
	    MathObject resultObject = null;
	    
	    if(z1 == z2) {
	      
	        // It's a matrix
	        resultObject =  new MatrixStandard("GaussQuadrature Result", z1);
	        
	    } else if(z1 == 1 || z2 == 1) {
	        
	        // Take the bigger one
	        if(z1 > z2) {
	            
		        // It's a vector
		        resultObject = new Vector("GaussQuadrature Result", z1);
	            
	        } else {
	            
		        // It's a vector
		        resultObject = new Vector("GaussQuadrature Result", z2);
	        }
	        
	    } else {
	        
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Unknown result object requested (z1 / z2) :" + z1 + " / " + z2);
            }
            
            return null;
	    }
	    
	    /** 
	     * Start the calculation stuff 
	     * */
	    
	    // Run through all positions in the result object
	    for(int m = 1; m <= z1; m++) {
	        for(int n = 1; n <= z2; n++) {
	            
	    	    // Calculate the sum: Hy*g(((b-a)/2)*ay+((b+a)/2)) from j=1 to my
	    	    double sum = 0;
	    	    double eta = 0;
	    	    
	    	    for(int j = 1; j <= hy.size(); j++) {
	    	        
	    	        eta = ((b-a)/2) * ay.getValue(j) + ((b+a)/2);
	    	        
	    	        sum = sum + hy.getValue(j) * calcDoubleInnerIntegration(c, d, h1, h2, q, hx, ax, eta, m, n);
	    	    }
	    	    
	    	    // Calc: ((b-a)/2)*sum
	    	    double result = ((b-a)/2)*sum;
	    	    
	    	    // Place the result into the result object
	    	    if(resultObject instanceof Matrix) {

	    	        ((Matrix)resultObject).setValue(m, n, result);
	    	        
	    	    } else if(resultObject instanceof Vector) {
	    	        
	    	        // Check which value has been greater
	    	        if(z1 > z2) {
	    	            
	    		        // It's a vector with z1 running (which is 's')
	    	            ((Vector)resultObject).setValue(m, result);
	    	            
	    	        } else {
	    	            
	    		        // It's a vector with z2 running (which is 't')
	    	            ((Vector)resultObject).setValue(n, result);
	    	        }
	    	        
	    	    }
	        }
	    }	   

	    return resultObject;
	}
	
	/**
	 * Calculate the inner integration of the double integration
	 * @param mx, amount of the "stuetzstellen" for x
	 * @param hx "gewichtung" vector of "h" for x
	 * @param ax, "stuetzstellen" vector of "a" for x
	 * @param x, the current x value
	 * @param figure, the figure we do the integration for
	 * @return the resulting double value
	 */
	private static double calcDoubleInnerIntegration(double c, double d, Method h1, Method h2, Method q, Vector hx, Vector ax, double eta, int m, int n) {
	    
	    double result = 0;
	    double sum = 0.0;
	    
	    try {
		    // Calculate sum: Hx * f( (d-c)/2 * ax + (d+c)/2, eta )
		    for(int i = 1; i <= hx.size(); i++){
		    	
		    	double xi = ((d-c)/2) * ax.getValue(i) + ((d+c)/2);
		    	
		        double h1Value = callMethodCalculation(h1, xi, eta, m);
		        double h2Value = callMethodCalculation(h2, xi, eta, n);
		        double qValue = callMethodCalculation(q, xi, eta, i);
		        
		       sum += hx.getValue(i) * h1Value *  h2Value  * qValue;
		    }
		    
		    // Calc: ((d-c)/2) * sum;
		    return ((d-c)/2) * sum;
	        
	    } catch (FunctionCalculationException e) {
	        
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "FunctionCalculationException, aborting GaussQuadrature !", e);
            }
            
            // Abort the program
            FEMStartup.abort(false);
	        
	    } catch (Exception e) {
	        
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "An error occured during the GaussQuadrature !", e);
            }
	        
            // Abort the program
            FEMStartup.abort(false);
	    }
	    
	    return -1.0;
	}
	
	/**
	 * Calculate one single integral via the gauss quadrature
	 * @param a, lower-border of the outer-integral
	 * @param b, upper-border of the outer-integral
	 * @param h1, first part of the integral
	 * @param h2, second part of the integral
	 * @param q, third part of the integral
	 * @param z1, size of the vector in h1 (used to determine the position integer)
	 * @param z2, size of the vector in h2 (used to determine the position integer)
	 * @param ay "stuetzstellen" vector of "a" for y
	 * @param hy "gewichtung" vector of "h" for y
	 * @return result matrix or vector or value or whatever it calculates
	 */
	public static MathObject calcSingleIntegration(double a, double b, Method h1, Method h2, Method q, int z1, int z2, Vector ay, Vector hy) {
	    	   
	    // Create the resulting MathObject
	    MathObject resultObject = null;
	    
	    try {
		    
		    if(z1 == z2) {
		      
		        // It's a matrix
		        resultObject = new MatrixStandard("GaussQuadrature Result", z1);
		        
		    } else if(z1 == 1 || z2 == 1) {
		        
		        // Take the bigger one
		        if(z1 > z2) {
		            
			        // It's a vector
			        resultObject = new Vector("GaussQuadrature Result", z1);
		            
		        } else {
		            
			        // It's a vector
			        resultObject = new Vector("GaussQuadrature Result", z2);
		        }
		        
		    } else {
		        
	            if(localLogger.isEnabledFor(Level.ERROR)) {
	                localLogger.log(Level.ERROR, "Unknown result object requested (z1 / z2) :" + z1 + " / " + z2);
	            }
	            
	            return null;
		    }
		    
		    /** 
		     * Start the calculation stuff 
		     * */
		    	    
		    // Run through all positions in the result object
		    for(int s = 1; s <= z1; s++) {
		        for(int t = 1; t <= z2; t++) {
		            
		            // Calculate the sum: Hy * f( ((b-a)/2) * ay + ((b+a)/2) ) from j=1 to my
		            double sum = 0;
		            double xi = 0;
		            
		            for(int j = 1; j <= hy.size(); j++) {
		                
		                xi = ((b-a)/2) * ay.getValue(j) + ((b+a)/2);
		                
				        double h1Value = callMethodCalculation(h1, xi, 0, s);
				        double h2Value = callMethodCalculation(h2, xi, 0, t);
				        double qValue = callMethodCalculation(q, xi, 0, j);
				        
				        sum = sum + hy.getValue(j) * (h1Value *  h2Value  * qValue);
		            }
		    	    
		    	    // Calc: ((b-a)/2) * sum
		    	    double result = ((b-a) / 2) * sum;
		    	    
		    	    // Place the result into the result object
		    	    if(resultObject instanceof Matrix) {
		    	        
		    	        ((Matrix)resultObject).setValue(s, t, result);
		    	        
		    	    } else if(resultObject instanceof Vector) {
		    	        
		    	        // Check which value has been greater
		    	        if(z1 > z2) {
		    	            
		    		        // It's a vector with z1 running (which is 's')
		    	            ((Vector)resultObject).setValue(s, result);
		    	            
		    	        } else {
		    	            
		    		        // It's a vector with z2 running (which is 't')
		    	            ((Vector)resultObject).setValue(t, result);
		    	        }
		    	        
		    	    }
		        }
		    }	
		    
	    } catch (FunctionCalculationException e) {
	        
	        if(localLogger.isEnabledFor(Level.ERROR)) {
	            localLogger.log(Level.ERROR, "FunctionCalculationException, aborting GaussQuadrature !", e);
	        }
	        
	        // Abort the program
	        FEMStartup.abort(false);
	        
	    } catch (Exception e) {
	        
	        if(localLogger.isEnabledFor(Level.ERROR)) {
	            localLogger.log(Level.ERROR, "An error occured during the GaussQuadrature !", e);
	        }
	        
	        // Abort the program
	        FEMStartup.abort(false);
	    }
	    
	    return resultObject;
	}
	
	/**
	 * Calculate the Gauss-Quadrature to solve an integral
	 * @param a, lower-border of the outer-integral
	 * @param b, upper-border of the outer-integral
	 * @param h1, first part of the integral
	 * @param h2, second part of the integral
	 * @param q, third part of the integral
	 * @param z1, size of the vector in h1 (used to determine the position integer)
	 * @param z2, size of the vector in h2 (used to determine the position integer)
	 * @param ax, "stuetzstellen" vector of "a" for x
	 * @param hx "gewichtung" vector of "h" for x
	 * @param ay "stuetzstellen" vector of "a" for y
	 * @param hy "gewichtung" vector of "h" for y
	 * @param figure, may be FIGURE_TRIANGLE or FIGURE_PARALLELOGRAM
	 * @return result matrix or vector or value or whatever it calculates
	 */
	public static MathObject calcIntegration(double a, double b, Method h1, Method h2, Method q, 
	        								 int z1, int z2, Vector ax, Vector hx, Vector ay, Vector hy, int figure) {
	        
	    // Check if the figure exists
	    if(figure != GaussConstants.FIGURE_TRIANGLE && figure != GaussConstants.FIGURE_QUADRAT) {
	        
	        // Figure does not exist
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Unknown figure: " + figure);
            }
            
            return null;
	    }
	    
	    // Create the resulting MathObject
	    MathObject resultObject = null;
	    
	    if(z1 == z2) {
	      
	        // It's a matrix
	        resultObject = new MatrixStandard("GaussQuadrature Result", z1);
	        
	    } else if(z1 == 1 || z2 == 1) {
	        
	        // Take the bigger one
	        if(z1 > z2) {
	            
		        // It's a vector
		        resultObject = new Vector("GaussQuadrature Result", z1);
	            
	        } else {
	            
		        // It's a vector
		        resultObject = new Vector("GaussQuadrature Result", z2);
	        }
	        
	    } else {
	        
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Unknown result object requested (z1 / z2) :" + z1 + " / " + z2);
            }
            
            return null;
	    }
	    
	    /** 
	     * Start the calculation stuff 
	     * */
	    
	    // Run through all positions in the result object
	    for(int s = 1; s <= z1; s++) {
	        for(int t = 1; t <= z2; t++) {
	            
	    	    // Calculate the sum: Hy*g(0.5*ay+0.5) from j=1 to my
	    	    double sum = 0;
	    	    double eta = 0;
	    	    
	    	    for(int j = 1; j <= hy.size(); j++) {
	    	        
	    	        eta = 0.5 * ay.getValue(j) + 0.5;
	    	        
	    	        sum = sum + hy.getValue(j) * calcInnerIntegration(hx.size(), h1, h2, q, hx, ax, eta, figure, s, t);
	    	    }
	    	    
	    	    // Calc: 0.5*sum
	    	    double result = 0.5*sum;
	    	    
	    	    // Place the result into the result object
	    	    if(resultObject instanceof Matrix) {
	    	        
	    	        ((Matrix)resultObject).setValue(s, t, result);
	    	        
	    	    } else if(resultObject instanceof Vector) {
	    	        
	    	        // Check which value has been greater
	    	        if(z1 > z2) {
	    	            
	    		        // It's a vector with z1 running (which is 's')
	    	            ((Vector)resultObject).setValue(s, result);
	    	            
	    	        } else {
	    	            
	    		        // It's a vector with z2 running (which is 't')
	    	            ((Vector)resultObject).setValue(t, result);
	    	        }
	    	        
	    	    }
	        }
	    }	   
	    
	    return resultObject;
	}
	
	/**
	 * Calculate the inner integration
	 * @param mx, amount of the "stuetzstellen" for x
	 * @param hx "gewichtung" vector of "h" for x
	 * @param ax, "stuetzstellen" vector of "a" for x
	 * @param eta, the current eta value
	 * @param figure, the figure we do the integration for
	 * @return the resulting double value
	 */
	private static double calcInnerIntegration(int mx, Method h1, Method h2, Method q, Vector hx, Vector ax, double eta, int figure, int iForH1, int iForH2) {
	    
	    double result = 0;
	    double fi = 0.0;
	    double sum = 0.0;
	    
	    try {
	        
	        // Calculate d(eta) depending on the used figure
	        double dyValue = GaussConstants.evaluateDForFigure(eta, figure);
	        
		    // Calculate sum: Hx * f( d(eta)/2 * ax + d(eta)/2, eta
		    for(int i = 1; i <= mx; i++){
		        
		        double h1Value = callMethodCalculation(h1, (dyValue/2) * ax.getValue(i) + dyValue/2, eta, iForH1);
		        double h2Value = callMethodCalculation(h2, (dyValue/2) * ax.getValue(i) + dyValue/2, eta, iForH2);
		        double qValue = callMethodCalculation(q, (dyValue/2) * ax.getValue(i) + dyValue/2, eta, i);
		        
		        fi = h1Value *  h2Value  * qValue;
		        
		       sum = sum + hx.getValue(i) * fi;
		    }
		    
		    // Calc: (dyValue/2) * sum;
		    return (dyValue/2) * sum;
	        
	    } catch (FunctionCalculationException e) {
	        
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "FunctionCalculationException, aborting GaussQuadrature !", e);
            }
            
            // Abort the program
            FEMStartup.abort(false);
	        
	    } catch (Exception e) {
	        
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "An error occured during the GaussQuadrature !", e);
            }
	        
            // Abort the program
            FEMStartup.abort(false);
	    }
	    
	    return -1.0;
	}
	
	/**
	 * Call the given method with the given values
	 * @param a, the method to use
	 * @param x, the x-value to use
	 * @param y, the y-value to use
	 * @param position, the position if applyable
	 * @return the resulting double value
	 */
	private static double callMethodCalculation(Method a, double x, double y, int position) throws FunctionCalculationException {
	    
	    // Create the object array
	    Object[] args = new Object[3];
	    
	    args[0] = new Double(x);
	    args[1] = new Double(y);
	    args[2] = new Integer(position);
	    
	    // Make sure to catch all exceptions
	    try {
            Object returnValue = a.invoke(null, args);
            
            // Check if we received a correct type of value
            if(returnValue instanceof Double) {
                
                // Correct double value
                return ((Double)(returnValue)).doubleValue();
                
            } else if(returnValue instanceof Integer) {
                
                // Correct double value
                return ((Integer)(returnValue)).doubleValue();
                
            } else {
                
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "The return type of the function method is neither Double nor Integer : " + returnValue);
                }
                
                throw new FunctionCalculationException();
            }
            
        } catch (IllegalArgumentException e) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "The method signature did not match: (double, double, int) !", e);
            }
            
            throw new FunctionCalculationException();
            
        } catch (IllegalAccessException e) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "The called method is not accessible. Perhaps private?", e);
            }
            
            throw new FunctionCalculationException();
            
        } catch (InvocationTargetException e) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "An exception has been thrown by an underlying method !", e);
            }
            
            throw new FunctionCalculationException();
            
        } catch(Exception e) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "An general exception has been thrown !", e);
            }
            
            throw new FunctionCalculationException();
        }
	}
}
