/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.calc;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import fhm.edu.fem.app.FEMStartup;
import fhm.edu.fem.calc.objects.BorderList;
import fhm.edu.fem.calc.objects.CTabList;
import fhm.edu.fem.calc.objects.Calculation;
import fhm.edu.fem.calc.objects.ElementList;
import fhm.edu.fem.calc.objects.ForceList;
import fhm.edu.fem.calc.objects.Node;
import fhm.edu.fem.calc.objects.NodeList;
import fhm.edu.fem.calc.objects.Vector;
import fhm.edu.fem.calc.objects.matrix.Matrix;
import fhm.edu.fem.util.io.FEMINLoader;
import fhm.edu.fem.util.io.FEMINSaver;
import fhm.edu.fem.util.io.FEMOUTSaver;
import fhm.edu.fem.util.logging.LoggingManager;

/**
 * Class Description:
 *		This class is the super class which stores all the elements, nodes
 *		borders, etc. and also knows about the calculations and their conditions
 *		which have been loaded from the FEMIN file
 *		We only need one FEMManager while the application runs so it is a singleton
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class FEMManager {

	/* Our local logger object */
	protected static Logger localLogger = LoggingManager.getLogger(FEMManager.class.getName());
    
    /* The sole instance of the FEM manager  */
    protected static FEMManager theInstance = new FEMManager();
    
    /* The approach to use, 2 = LINEAR, 3 = QUADRATIC */
    protected int femApproach = -1;
    
    /* The list holding all nodes */
    protected NodeList nodeList = new NodeList();
    
    /* The list holding all elements */
    protected ElementList elementList = new ElementList();
    
    /* The list holding all borders */
    protected BorderList borderList = new BorderList();
    
    /* The list holding all ctabs */
    protected CTabList ctabList = new CTabList();
    
    /* The list holding all calculations */
    protected List<Calculation> calculationList = new ArrayList<Calculation>();
    
    /* The list holding all resulting u-Vectors */
    protected List<Vector> resultList = new ArrayList<Vector>();
    
    /* The list holding all result vectors (before enlarging them via the Alias IDs) */
    protected List<Vector> resultListSmall = new ArrayList<Vector>();
    
    /* The alias ID node list */
    protected NodeList aliasIDList = new NodeList();
    
    /* The calculated Stiffness-Matrix */
    protected Matrix completeStiffnessMatrix = null;
    
    /* The calculated Mass-Matrix */
    protected Matrix completeMassMatrix = null;
    
    /* The calculated eigenvalues */
    protected Vector eigenvalueVector = null;
    
    /* For Circle Membrane - The Center Coordinates */
    protected double membraneCircleCenterX = 0.0;
    protected double membraneCircleCenterY = 0.0;
    
    /* For Circle Membrane - The Radiuses */
    protected double membraneCircleRadius1 = 0.0;
    protected double membraneCircleRadius2 = 0.0;
    
    /* For Circle Membrane - The number of Sections */
    protected int membraneCircleSections = 0;
    
    /* For Circle Membrane - The Border Condition */
    protected double membraneCircleCondition = 0.0;
    
    /* For Circle Membrane - The Center Node */
    protected Node centerNode = null;
    
    /* For Spanner - The Force List */
    protected ForceList spannerForceList = new ForceList();
    
    /* For Spanner - The Material Constants */
    protected double spannerMaterialConstantE = 0.0;
    protected double spannerMaterialConstantV = 0.0;
    protected double spannerMaterialConstantH = 0.0;
    protected double spannerMaterialConstantHS = 0.0;
    
    /**
     * Private Constructor to enforce singleton
     */
    private FEMManager() {}
    
    /**
     * Retrieve the sole instance of the manager
     * @return the FEM manager
     */
    public static FEMManager getInstance() {
        return theInstance;
    }
    
    /**
     * Load all settings from the given file (normally the FEMIN file)
     * @param inFile, the FEMIN file to get all settings from
     * @return true if all went well, false otherwise
     */
    public boolean load(File inFile) {
        try {
            // Clear the whole manager
            this.clear();
            
            // Call the FEMIN loader class
            FEMINLoader loader = new FEMINLoader();
            
            if(!loader.load(inFile)) {
                
                // Failure while loading the file, stop the program
                FEMStartup.abort(false);
            }
            
            // Grab all data from the loader
            this.femApproach = loader.getFEMApproach();
            
            if(this.femApproach == 1 || this.femApproach == 2 || this.femApproach == 3 || this.femApproach == 4) {
            	
            	this.nodeList = loader.getNodes();
                this.elementList = loader.getElements();
                this.borderList = loader.getBorders();
                this.ctabList = loader.getCTabs();            
                this.calculationList = loader.getCalculations();
                
    			if(localLogger.isEnabledFor(Level.INFO)) {
    				localLogger.log(Level.INFO, "File import complete: ");
    				localLogger.log(Level.INFO, "		Approach: " + this.femApproach);
    				localLogger.log(Level.INFO, "		Nodes: " + this.nodeList.size());
    				localLogger.log(Level.INFO, "		Elements: " + this.elementList.size());
    				localLogger.log(Level.INFO, "		Borders: " + this.borderList.size());
    				localLogger.log(Level.INFO, "		CTabs: " + this.ctabList.size());
    				localLogger.log(Level.INFO, "		Calculations: " + this.calculationList.size());
    			}
            } else if(this.femApproach == 5) {
            	
            	this.membraneCircleCenterX = loader.getMembraneCircleCenterX();
            	this.membraneCircleCenterY = loader.getMembraneCircleCenterY();
            	this.membraneCircleRadius1 = loader.getMembraneCircleRadius1();
            	this.membraneCircleRadius2 = loader.getMembraneCircleRadius2();
            	this.membraneCircleSections = loader.getMembraneCircleSections();
            	this.membraneCircleCondition = loader.getMembraneCircleCondition();
            	
            	if(localLogger.isEnabledFor(Level.INFO)) {
    				localLogger.log(Level.INFO, "File import complete: ");
    				localLogger.log(Level.INFO, "		Approach: " + this.femApproach);
    				localLogger.log(Level.INFO, "		Circle Center (x/y): " + this.membraneCircleCenterX + " / " + this.membraneCircleCenterY);
    				localLogger.log(Level.INFO, "		Radius1: " + this.membraneCircleRadius1);
    				localLogger.log(Level.INFO, "		Radius2: " + this.membraneCircleRadius2);
    				localLogger.log(Level.INFO, "		Sections: " + this.membraneCircleSections);
    				localLogger.log(Level.INFO, "		Condition: " + this.membraneCircleCondition);
    			}
            	
            } else if(this.femApproach == 6) {
            	
            	this.nodeList = loader.getNodes();
                this.elementList = loader.getElements();
                this.borderList = loader.getBorders();       
                this.calculationList = loader.getCalculations();
            	
            	this.spannerForceList = loader.getSpannerForceList();
            	this.spannerMaterialConstantE = loader.getSpannerMaterialConstantE();
            	this.spannerMaterialConstantV = loader.getSpannerMaterialConstantV();
            	this.spannerMaterialConstantH = loader.getSpannerMaterialConstantH();
            	this.spannerMaterialConstantHS = loader.getSpannerMaterialConstantHS();
            	
            	if(localLogger.isEnabledFor(Level.INFO)) {
    				localLogger.log(Level.INFO, "File import complete: ");
    				localLogger.log(Level.INFO, "		Approach: " + this.femApproach);
    				localLogger.log(Level.INFO, "		Nodes: " + this.nodeList.size());
    				localLogger.log(Level.INFO, "		Elements: " + this.elementList.size());
    				localLogger.log(Level.INFO, "		Borders: " + this.borderList.size());
    				localLogger.log(Level.INFO, "		Calculations: " + this.calculationList.size());
    				localLogger.log(Level.INFO, "		Forces: " + this.spannerForceList.size());
    				localLogger.log(Level.INFO, "		Material Constant E: " + this.spannerMaterialConstantE);
    				localLogger.log(Level.INFO, "		Material Constant V: " + this.spannerMaterialConstantV);
    				localLogger.log(Level.INFO, "		Material Constant H: " + this.spannerMaterialConstantH);
    				localLogger.log(Level.INFO, "		Material Constant 'Hoechstspannung': " + this.spannerMaterialConstantHS);
    			}
            }            
            
        } catch (Exception e) {
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Loading the file failed: " + inFile.getName(), e);
			}
			
			return false;
        }
        
        return true;
    }
    
    /**
     * Save the results to the FEMOUT file
     * @param outFile, the FEMOUT file to save to
     * @return true if all went well, false otherwise
     */
    public boolean saveFEMOUT(File outFile) {
        try {
            
			if(localLogger.isEnabledFor(Level.INFO)) {
				localLogger.log(Level.INFO, "Saving results to file: " + outFile.getName());
			}
            
            // Call the FEMOUT saver class
            FEMOUTSaver saver = new FEMOUTSaver();
            saver.save(outFile);   
            
			if(localLogger.isEnabledFor(Level.INFO)) {
				localLogger.log(Level.INFO, "Saving results complete");
			}
            
        } catch (Exception e) {
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Saving the file failed: " + outFile.getName(), e);
			}
			
			return false;
        }
        
        return true;
    }
    
    /**
     * Save the results to the FEMIN file
     * @param outFile, the FEMIN file to save to
     * @param femPlotterCompatible, true if the file should be saved compatible to FEM_PLOTTER
     * @return true if all went well, false otherwise
     */
    public boolean saveFEMIN(File outFile, boolean femPlotterCompatible) {
        try {
            
			if(localLogger.isEnabledFor(Level.INFO)) {
				localLogger.log(Level.INFO, "Saving new FEMIN to file: " + outFile.getName());
			}
            
            // Call the FEMOUT saver class
            FEMINSaver saver = new FEMINSaver();
            saver.save(outFile, femPlotterCompatible);   
            
			if(localLogger.isEnabledFor(Level.INFO)) {
				localLogger.log(Level.INFO, "Saving new FEMIN complete");
			}
            
        } catch (Exception e) {
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Saving the file failed: " + outFile.getName(), e);
			}
			
			return false;
        }
        
        return true;
    }
    
    /**
     * Clear the whole manager
     */
    public void clear() {
        try {
            
            this.nodeList.clear();
            this.elementList.clear();
            this.borderList.clear();
            this.calculationList.clear();
            this.resultList.clear();
            this.spannerForceList.clear();
            
        } catch (Exception e) {
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Clearing the FEM Manager failed !", e);
			}
        }
    }
    
	/**
	 * @return Returns the borderList.
	 */
	public BorderList getBorderList() {
		return borderList;
	}
	
	/**
	 * @return Returns the calculationList.
	 */
	public List<Calculation> getCalculationList() {
		return calculationList;
	}
	
	/**
	 * @return Returns the elementList.
	 */
	public ElementList getElementList() {
		return elementList;
	}
	
	/**
	 * @return Returns the nodeList.
	 */
	public NodeList getNodeList() {
		return nodeList;
	}
	
    /**
     * @return Returns the ctab list
     */
    public CTabList getCTabList() {
        return this.ctabList;
    }
	
	/**
	 * Clear the result list
	 */
	public void clearResultList() {
	    resultList.clear();
	}
	
	/**
	 * Add a result vector to the result list
	 * @param u, the vector to add to the result list
	 */
	public void addToResultList(Vector u) {
	    
	    // Save the vector in the result list
	    resultList.add(u);
	}
	
	/**
	 * @return Returns the resultList.
	 */
	public List<Vector> getResultList() {
	    return resultList;
	}
	
	/**
	 * Clear the small result list
	 */
	public void clearResultListSmall() {
	    resultListSmall.clear();
	}
	
	/**
	 * Add a result vector to the small result list
	 * @param u, the vector to add to the small result list
	 */
	public void addToResultListSmall(Vector u) {
	    
	    // Save the vector in the small result list
	    resultListSmall.add(u);
	}
	
	/**
	 * @return Returns the small resultList.
	 */
	public List<Vector> getResultListSmall() {
	    return resultListSmall;
	}
	
    /**
     * Retrieve the approach to use
     * @return the approach, 2 = LINEAR, 3 = QUADRATIC
     */
    public int getFEMApproach() {
        return this.femApproach;
    }   
    
    /**
     * FEMApproach can be 2 = APPROACH_LINEAR or 3 = APPROACH_QUADRATIC
     * @param femApproach The femApproach to set.
     */
    public void setFemApproach(int femApproach) {
        this.femApproach = femApproach;
    }
    
    /**
     * Retrieve alias ID list
     * @return the alias ID list
     */
    public NodeList getAliasIDNodeList() {
        return this.aliasIDList;
    }   
    
    /**
     * Set the alias ID list
     * @param aliasIDList, the new alias ID list
     */
    public void setAliasIDNodeList(NodeList aliasIDList) {
        this.aliasIDList = aliasIDList;
    }
    
	/**
	 * @return Returns the completeMassMatrix.
	 */
	public Matrix getCompleteMassMatrix() {
		return completeMassMatrix;
	}
	
	/**
	 * @param completeMassMatrix The completeMassMatrix to set.
	 */
	public void setCompleteMassMatrix(Matrix completeMassMatrix) {
		this.completeMassMatrix = completeMassMatrix;
	}
	
	/**
	 * @return Returns the completeStiffnessMatrix.
	 */
	public Matrix getCompleteStiffnessMatrix() {
		return completeStiffnessMatrix;
	}
	
	/**
	 * @param completeStiffnessMatrix The completeStiffnessMatrix to set.
	 */
	public void setCompleteStiffnessMatrix(Matrix completeStiffnessMatrix) {
		this.completeStiffnessMatrix = completeStiffnessMatrix;
	}
	
	/**
	 * @return Returns the eigenvalueVector.
	 */
	public Vector getEigenvalueVector() {
		return eigenvalueVector;
	}
	
	/**
	 * @param eigenvalueVector The eigenvalueVector to set.
	 */
	public void setEigenvalueVector(Vector eigenvalueVector) {
		this.eigenvalueVector = eigenvalueVector;
	}
	
	/**
	 * @return Returns the membraneCircleCenterX.
	 */
	public double getMembraneCircleCenterX() {
		return membraneCircleCenterX;
	}
	
	/**
	 * @param membraneCircleCenterX The membraneCircleCenterX to set.
	 */
	public void setMembraneCircleCenterX(double membraneCircleCenterX) {
		this.membraneCircleCenterX = membraneCircleCenterX;
	}
	
	/**
	 * @return Returns the membraneCircleCenterY.
	 */
	public double getMembraneCircleCenterY() {
		return membraneCircleCenterY;
	}
	
	/**
	 * @param membraneCircleCenterY The membraneCircleCenterY to set.
	 */
	public void setMembraneCircleCenterY(double membraneCircleCenterY) {
		this.membraneCircleCenterY = membraneCircleCenterY;
	}
	
	/**
	 * @return Returns the membraneCircleRadius1.
	 */
	public double getMembraneCircleRadius1() {
		return membraneCircleRadius1;
	}
	
	/**
	 * @param membraneCircleRadius1 The membraneCircleRadius1 to set.
	 */
	public void setMembraneCircleRadius1(double membraneCircleRadius1) {
		this.membraneCircleRadius1 = membraneCircleRadius1;
	}
	
	/**
	 * @return Returns the membraneCircleRadius2.
	 */
	public double getMembraneCircleRadius2() {
		return membraneCircleRadius2;
	}
	
	/**
	 * @param membraneCircleRadius2 The membraneCircleRadius2 to set.
	 */
	public void setMembraneCircleRadius2(double membraneCircleRadius2) {
		this.membraneCircleRadius2 = membraneCircleRadius2;
	}
	
	/**
	 * @return Returns the membraneCircleSections.
	 */
	public int getMembraneCircleSections() {
		return membraneCircleSections;
	}
	
	/**
	 * @param membraneCircleSections The membraneCircleSections to set.
	 */
	public void setMembraneCircleSections(int membraneCircleSections) {
		this.membraneCircleSections = membraneCircleSections;
	}
	
	/**
	 * @param borderList The borderList to set.
	 */
	public void setBorderList(BorderList borderList) {
		this.borderList = borderList;
	}
	
	/**
	 * @param elementList The elementList to set.
	 */
	public void setElementList(ElementList elementList) {
		this.elementList = elementList;
	}
	
	/**
	 * @param nodeList The nodeList to set.
	 */
	public void setNodeList(NodeList nodeList) {
		this.nodeList = nodeList;
	}
	
	/**
	 * @return Returns the membraneCircleCondition.
	 */
	public double getMembraneCircleCondition() {
		return membraneCircleCondition;
	}
	
	/**
	 * @return Returns the centerNode.
	 */
	public Node getCenterNode() {
		return centerNode;
	}
	
	/**
	 * @param centerNode The centerNode to set.
	 */
	public void setCenterNode(Node centerNode) {
		this.centerNode = centerNode;
	}
	
	/**
	 * @return Returns the spannerForceList.
	 */
	public ForceList getSpannerForceList() {
		return spannerForceList;
	}
	
	/**
	 * @param spannerForceList The spannerForceList to set.
	 */
	public void setSpannerForceList(ForceList spannerForceList) {
		this.spannerForceList = spannerForceList;
	}
	
	/**
	 * @return Returns the spannerMaterialConstant.
	 */
	public double getSpannerMaterialConstantE() {
		return spannerMaterialConstantE;
	}
	
	/**
	 * @return Returns the spannerMaterialConstant.
	 */
	public double getSpannerMaterialConstantV() {
		return spannerMaterialConstantV;
	}
	
	/**
	 * @return Returns the spannerMaterialConstant.
	 */
	public double getSpannerMaterialConstantHS() {
		return spannerMaterialConstantHS;
	}
	
	/**
	 * @return Returns the spannerMaterialConstant.
	 */
	public double getSpannerMaterialConstantH() {
		return spannerMaterialConstantH;
	}
}
