/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.calc.objects;

import java.lang.reflect.Method;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import fhm.edu.fem.calc.FEMManager;
import fhm.edu.fem.exception.FunctionCalculationException;
import fhm.edu.fem.util.logging.LoggingManager;

/**
 * Class Description:
 *		Holds all the static function methods needed for the Gauss-Quadrature
 *		Compilation of the Complete Stiffness Matrix used in FEMGaussQuadrature
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class CompilationFunctions {
	
	/* Our local logger object */
	protected static Logger localLogger = LoggingManager.getLogger(CompilationFunctions.class.getName());
	
	/* The method names - For figure Triangle - Linear approach */
	public static final String TRIANGLE_LINEAR_N = "triangleLinearN";
	public static final String TRIANGLE_LINEAR_N_XI = "triangleLinearNXI";
	public static final String TRIANGLE_LINEAR_N_ETA = "triangleLinearNETA";
	public static final String TRIANGLE_LINEAR_X_XI = "triangleLinearXXI";
	public static final String TRIANGLE_LINEAR_X_ETA = "triangleLinearXETA";
	public static final String TRIANGLE_LINEAR_Y_XI = "triangleLinearYXI";
	public static final String TRIANGLE_LINEAR_Y_ETA = "triangleLinearYETA";
	public static final String TRIANGLE_LINEAR_JACOBI = "triangleLinearJAC";
	public static final String TRIANGLE_LINEAR_ONETHROUGHJAC = "triangleLinearJAConeThrough";
	
	public static final String TRIANGLE_LINEAR_H1 = "triangleLinearH1";
	public static final String TRIANGLE_LINEAR_H2 = "triangleLinearH2";
	
	/* The method names - For figure Triangle - Quadratic approach */
	public static final String TRIANGLE_QUADRATIC_N = "triangleQuadraticN";
	public static final String TRIANGLE_QUADRATIC_N_XI = "triangleQuadraticNXI";
	public static final String TRIANGLE_QUADRATIC_N_ETA = "triangleQuadraticNETA";
	public static final String TRIANGLE_QUADRATIC_X_XI = "triangleQuadraticXXI";
	public static final String TRIANGLE_QUADRATIC_X_ETA = "triangleQuadraticXETA";
	public static final String TRIANGLE_QUADRATIC_Y_XI = "triangleQuadraticYXI";
	public static final String TRIANGLE_QUADRATIC_Y_ETA = "triangleQuadraticYETA";
	public static final String TRIANGLE_QUADRATIC_JACOBI = "triangleQuadraticJAC";
	public static final String TRIANGLE_QUADRATIC_ONETHROUGHJAC = "triangleQuadraticJAConeThrough";
	
	public static final String TRIANGLE_QUADRATIC_H1 = "triangleQuadraticH1";
	public static final String TRIANGLE_QUADRATIC_H2 = "triangleQuadraticH2";
	
	/* The method names - For figure Quadrat - Linear approach */
	public static final String QUADRAT_LINEAR_N = "quadratLinearN";
	public static final String QUADRAT_LINEAR_N_XI = "quadratLinearNXI";
	public static final String QUADRAT_LINEAR_N_ETA = "quadratLinearNETA";
	public static final String QUADRAT_LINEAR_X_XI = "quadratLinearXXI";
	public static final String QUADRAT_LINEAR_X_ETA = "quadratLinearXETA";
	public static final String QUADRAT_LINEAR_Y_XI = "quadratLinearYXI";
	public static final String QUADRAT_LINEAR_Y_ETA = "quadratLinearYETA";
	public static final String QUADRAT_LINEAR_JACOBI = "quadratLinearJAC";
	public static final String QUADRAT_LINEAR_ONETHROUGHJAC = "quadratLinearJAConeThrough";
	
	public static final String QUADRAT_LINEAR_H1 = "quadratLinearH1";
	public static final String QUADRAT_LINEAR_H2 = "quadratLinearH2";
	
	/* The method names - For figure Quadrat - Quadratic approach */
	public static final String QUADRAT_QUADRATIC_N = "quadratQuadraticN";
	public static final String QUADRAT_QUADRATIC_N_XI = "quadratQuadraticNXI";
	public static final String QUADRAT_QUADRATIC_N_ETA = "quadratQuadraticNETA";
	public static final String QUADRAT_QUADRATIC_X_XI = "quadratQuadraticXXI";
	public static final String QUADRAT_QUADRATIC_X_ETA = "quadratQuadraticXETA";
	public static final String QUADRAT_QUADRATIC_Y_XI = "quadratQuadraticYXI";
	public static final String QUADRAT_QUADRATIC_Y_ETA = "quadratQuadraticYETA";
	public static final String QUADRAT_QUADRATIC_JACOBI = "quadratQuadraticJAC";
	public static final String QUADRAT_QUADRATIC_ONETHROUGHJAC = "quadratQuadraticJAConeThrough";
	
	public static final String QUADRAT_QUADRATIC_H1 = "quadratQuadraticH1";
	public static final String QUADRAT_QUADRATIC_H2 = "quadratQuadraticH2";
	
	/* The method names - For figure Circle Membrane */
	public static final String MEMBRANE_CIRCLE_SECTOR_TRIANGLE_QUADRATIC_N = "membraneCircleSectorTriangleQuadraticN";
	public static final String MEMBRANE_CIRCLE_SECTOR_TRIANGLE_QUADRATIC_NXI = "membraneCircleSectorTriangleQuadraticNXI";
	public static final String MEMBRANE_CIRCLE_SECTOR_TRIANGLE_QUADRATIC_NETA = "membraneCircleSectorTriangleQuadraticNETA";
	public static final String MEMBRANE_CIRCLE_SECTOR_TRIANGLE_QUADRATIC_NETA_THROUGH_XI = "membraneCircleSectorTriangleQuadraticNETAThroughXI";
	public static final String MEMBRANE_CIRCLE_RING_Q_1 = "membraneCircleRingQ1";
	public static final String MEMBRANE_CIRCLE_RING_Q_2 = "membraneCircleRingQ2";
	
	/* Helper Method */
	public static final String ZERO = "zero";
	public static final String ONE = "one";
	public static final String XI = "xi";
	
	/* Global data structure to remember the current x- and y-values of the element, used in the XXI, XETA, YXI, YETA methods */
	protected static double[] elementXValues;
	protected static double[] elementYValues;
	
	/* Global data structure to remember the radius */
	protected static double radius1;
	protected static double radius2;
	
	/* Global data structure to remember the angles */
	protected static double angle1;
	protected static double angle2;
	
	/*Triangle Spanner*/
	public static final String SPANNER_TRIANGLE_QUADRATIC_XI_X = "spannerTriangleQuadraticXIX";
	public static final String SPANNER_TRIANGLE_QUADRATIC_XI_Y= "spannerTriangleQuadraticXIY";
	
	public static final String SPANNER_TRIANGLE_QUADRATIC_ETA_X = "spannerTriangleQuadraticETAX";
	public static final String SPANNER_TRIANGLE_QUADRATIC_ETA_Y= "spannerTriangleQuadraticETAY";
	
	/*Quadrat Spanner*/
	public static final String SPANNER_QUADRAT_QUADRATIC_XI_X = "spannerQuadratQuadraticXIX";
	public static final String SPANNER_QUADRAT_QUADRATIC_XI_Y = "spannerQuadratQuadraticXIY";
	
	public static final String SPANNER_QUADRAT_QUADRATIC_ETA_X = "spannerQuadratQuadraticETAX";
	public static final String SPANNER_QUADRAT_QUADRATIC_ETA_Y = "spannerQuadratQuadraticETAY";
	
	/*Quadrat Spanner*/
	public static final String SPANNER_QUADRAT_QUADRATIC_H1 = "spannerQuadratQuadraticH1";
	public static final String SPANNER_QUADRAT_QUADRATIC_H2 = "spannerQuadratQuadraticH2";
	
	/*Triangle Spanner*/
	public static final String SPANNER_TRIANGLE_QUADRATIC_H1 = "spannerTriangleQuadraticH1";
	public static final String SPANNER_TRIANGLE_QUADRATIC_H2 = "spannerTriangleQuadraticH2";
	
	/* Q Parts of the Spanner Integrals */
	public static final String SPANNER_S11_INTEGRAL_TWO_Q = "spannerS11IntegralTwoQ";
	public static final String SPANNER_S22_INTEGRAL_ONE_Q = "spannerS22IntegralOneQ";
	public static final String SPANNER_S12_INTEGRAL_ONE_Q = "spannerS12IntegralOneQ";
	public static final String SPANNER_S12_INTEGRAL_TWO_Q = "spannerS12IntegralTwoQ";
    
	/**
	 * Get the proper method
	 * @param methodName, the method name taken from the constants in this class
	 * @return the method
	 */
	public static Method getMethod(String methodName) throws SecurityException, NoSuchMethodException {
	    
	    // Build the args array
	    Class args[] = new Class[3];
	    args[0] = Double.class;
	    args[1] = Double.class;
	    args[2] = Integer.class;
	    
	    return CompilationFunctions.class.getDeclaredMethod(methodName, args);
	}
	
	/**
	 * To remember the current element's x- and y-values in the static data
	 * structure here in this class - used in the XXI, XETA, YXI, YETA methods
	 * @param element, the element which values we want to store
	 */
	public static void storeElementXYValues(Element element) {
		
		// Grab the node list from the element
		NodeList nodeList = element.getNodeList();
		
		// Create the arrays, depending on the number of element nodes
		CompilationFunctions.elementXValues = new double[nodeList.size()];
		CompilationFunctions.elementYValues = new double[nodeList.size()];
		
		// Fill the new arrays
		for(int i = 1; i <= nodeList.size(); i++) {
			
			CompilationFunctions.elementXValues[i-1] = nodeList.getNodeInCorrectOrder(i).getX();
			CompilationFunctions.elementYValues[i-1] = nodeList.getNodeInCorrectOrder(i).getY();
		}
	}
	
	/**
	 * To remember the current element's angle and radius values
	 * @param element, the element which values we want to store
	 */
	public static void storeRadiusAndAngle(double radius1, double radius2, double angle1, double angle2) {
		
		CompilationFunctions.radius1 = radius1;
		CompilationFunctions.radius2 = radius2;
		CompilationFunctions.angle1 = angle1;
		CompilationFunctions.angle2 = angle2;
	}
	
	/**
	 * Helper Method
	 * @param xi
	 * @param eta
	 * @param position
	 * @return
	 */
	public static double zero(Double xi, Double eta, Integer position) {
	    
	    return 0.0;
	}
	
	/**
	 * Helper Method
	 * @param xi
	 * @param eta
	 * @param position
	 * @return
	 */
	public static double one(Double xi, Double eta, Integer position) {
	    
	    return 1.0;
	}
	
	/*
	 * 
	 * TRIANGLE
	 * LINEAR APPROACH
	 * 
	 ***************************/
	
    /**
     * Function for Triangle, linear, normal N
     * FORMULA 2.101
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double triangleLinearN(Double xi, Double eta, Integer position) throws FunctionCalculationException {

        // Determine which position in the vector the user wants to calculate
        switch(position.intValue()) {
        
        	case 1:
        	    
        	    return (1.0 - xi.doubleValue() - eta.doubleValue());
        	
        	case 2:

        	    return (xi.doubleValue());
        	
        	case 3:
        	    
        	    return (eta.doubleValue());
        	
        	default:
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error: Unknown position in function vector given: " + position);
                }
                
                throw new FunctionCalculationException("Illegal position value: " + position);
        }
    }
    
    /**
     * Function for Triangle, linear, N derived to XI
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double triangleLinearNXI(Double xi, Double eta, Integer position) throws FunctionCalculationException {

        // Determine which position in the vector the user wants to calculate
        switch(position.intValue()) {
        
        	case 1:
        	    
        	    return (-1.0);
        	
        	case 2:

        	    return (1.0);
        	
        	case 3:
        	    
        	    return (0.0);
        	
        	default:
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error: Unknown position in function vector given: " + position);
                }
                
                throw new FunctionCalculationException("Illegal position value: " + position);
        }
    }
    
    /**
     * Function for Triangle, linear, N derived to ETA
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double triangleLinearNETA(Double xi, Double eta, Integer position) throws FunctionCalculationException {

        // Determine which position in the vector the user wants to calculate
        switch(position.intValue()) {
        
        	case 1:
        	    
        	    return (-1.0);
        	
        	case 2:

        	    return (0.0);
        	
        	case 3:
        	    
        	    return (1.0);
        	
        	default:
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error: Unknown position in function vector given: " + position);
                }
                
                throw new FunctionCalculationException("Illegal position value: " + position);
        }
    }
    
    /**
     * Function for Triangle, linear, X derived to XI
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double triangleLinearXXI(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	// Check if the number if coordinates in the static arrays match our requirements
    	if(CompilationFunctions.elementXValues.length < 3 || CompilationFunctions.elementYValues.length < 3) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
            }
            
            throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
    	}
    	
    	double sum = 0.0;
    	
    	// Calculate the sum out of the x-values in the static array multiplied by the "N derived to XI" values
    	for(int i = 0; i < 3; i++) {

    		sum = sum + (CompilationFunctions.elementXValues[i] * CompilationFunctions.triangleLinearNXI(xi, eta, new Integer(i+1)));
    	}
    	
    	return sum;        
    }
    
    /**
     * Function for Triangle, linear, X derived to ETA
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double triangleLinearXETA(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	// Check if the number if coordinates in the static arrays match our requirements
    	if(CompilationFunctions.elementXValues.length < 3 || CompilationFunctions.elementYValues.length < 3) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
            }
            
            throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
    	}
    	
    	double sum = 0.0;
    	
    	// Calculate the sum out of the x-values in the static array multiplied by the "N derived to ETA" values
    	for(int i = 0; i < 3; i++) {
    		
    		sum = sum + (CompilationFunctions.elementXValues[i] * CompilationFunctions.triangleLinearNETA(xi, eta, new Integer(i+1)));
    	}
    	
    	return sum;    
    }
    
    /**
     * Function for Triangle, linear, Y derived to XI
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double triangleLinearYXI(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	// Check if the number if coordinates in the static arrays match our requirements
    	if(CompilationFunctions.elementXValues.length < 3 || CompilationFunctions.elementYValues.length < 3) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
            }
            
            throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
    	}
    	
    	double sum = 0.0;
    	
    	// Calculate the sum out of the y-values in the static array multiplied by the "N derived to XI" values
    	for(int i = 0; i < 3; i++) {
    		
    		sum = sum + (CompilationFunctions.elementYValues[i] * CompilationFunctions.triangleLinearNXI(xi, eta, new Integer(i+1)));
    	}
    	
    	return sum;    
    }
    
    /**
     * Function for Triangle, linear, Y derived to ETA
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double triangleLinearYETA(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	// Check if the number if coordinates in the static arrays match our requirements
    	if(CompilationFunctions.elementXValues.length < 3 || CompilationFunctions.elementYValues.length < 3) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
            }
            
            throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
    	}
    	
    	double sum = 0.0;
    	
    	// Calculate the sum out of the y-values in the static array multiplied by the "N derived to ETA" values
    	for(int i = 0; i < 3; i++) {
    		
    		sum = sum + (CompilationFunctions.elementYValues[i] * CompilationFunctions.triangleLinearNETA(xi, eta, new Integer(i+1)));
    	}
    	
    	return sum;  
    }
    
    /**
     * Function for Triangle, linear, JACOBI
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double triangleLinearJAC(Double xi, Double eta, Integer position) throws FunctionCalculationException {  	
    	
    	// Calculate detJ
    	double detJ = 0.0;
    	
    	// The position says "-1" all the time because it does not really matter
    	detJ = CompilationFunctions.triangleLinearXXI(xi, eta, position) * CompilationFunctions.triangleLinearYETA(xi, eta, position);
    	detJ = detJ - (CompilationFunctions.triangleLinearXETA(xi, eta, position) * CompilationFunctions.triangleLinearYXI(xi, eta, position));
    	
    	// Make sure detJ is greater than zero
    	if(detJ <= 0.0) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Value of detJ is not > 0.0: " + detJ);
            }
            
            throw new FunctionCalculationException("Value of detJ is not > 0.0: " + detJ);
    	}
    	
    	return (1.0 / detJ);  
    }
    
    /**
     * Function for Triangle, linear, 1/ JACOBI
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double triangleLinearJAConeThrough(Double xi, Double eta, Integer position) throws FunctionCalculationException { 
        
        return (1.0 / triangleLinearJAC(xi, eta, position));
    }
    
    /**
     * Function for Triangle, linear, H1 part of the Integral
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double triangleLinearH1(Double xi, Double eta, Integer position) throws FunctionCalculationException {
    	
    	double result = CompilationFunctions.triangleLinearNXI(xi, eta, position) * CompilationFunctions.triangleLinearYETA(xi, eta, position);
    	result = result - (CompilationFunctions.triangleLinearNETA(xi, eta, position) * CompilationFunctions.triangleLinearYXI(xi, eta, position));
    	
    	return result;
    }
    
    /**
     * Function for Triangle, linear, H2 part of the Integral
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double triangleLinearH2(Double xi, Double eta, Integer position) throws FunctionCalculationException {
    	
    	double result = (-1.0 * CompilationFunctions.triangleLinearNXI(xi, eta, position)) * CompilationFunctions.triangleLinearXETA(xi, eta, position);
    	result = result + (CompilationFunctions.triangleLinearNETA(xi, eta, position) * CompilationFunctions.triangleLinearXXI(xi, eta, position));
    	
    	return result;
    }

	/*
	 * 
	 * TRIANGLE
	 * QUADRATIC APPROACH
	 * 
	 ***************************/

    /**
     * Function for Triangle, quadratic, normal N
     * FORMULA 2.102
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double triangleQuadraticN(Double xi, Double eta, Integer position) throws FunctionCalculationException {

        // Determine which position in the vector the user wants to calculate
        switch(position.intValue()) {
        
        	case 1:
        	    
        	    return ((1.0 - xi.doubleValue() - eta.doubleValue()) * (1.0 - 2.0 * xi.doubleValue() - 2.0 * eta.doubleValue()));
        	
        	case 2:

        	    return (xi.doubleValue() * (2.0 * xi.doubleValue() - 1.0));
        	
        	case 3:
        	    
        	    return (eta.doubleValue() * (2.0 * eta.doubleValue() - 1.0));
        	    
        	case 4:
        	    
        	    return (4.0 * xi.doubleValue() * (1.0 - xi.doubleValue() - eta.doubleValue()));
        	    
        	case 5:
        	    
        	    return (4.0 * xi.doubleValue() * eta.doubleValue());
        	    
        	case 6:
        	    
        	    return (4.0 * eta.doubleValue() * (1.0 - xi.doubleValue() - eta.doubleValue()));
        	
        	default:
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error: Unknown position in function vector given: " + position);
                }
                
                throw new FunctionCalculationException("Illegal position value: " + position);
        }
    }
    
    /**
     * Function for Triangle, quadratic, N derived to XI
     * FORMULA 2.102
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double triangleQuadraticNXI(Double xi, Double eta, Integer position) throws FunctionCalculationException {

        // Determine which position in the vector the user wants to calculate
        switch(position.intValue()) {
        
        	case 1:
        	    
        	    return (-3.0 + 4.0 * xi.doubleValue() + 4.0 * eta.doubleValue());
        	
        	case 2:

        	    return (-1.0 + 4.0 * xi.doubleValue());
        	
        	case 3:
        	    
        	    return (0.0);
        	    
        	case 4:
        	    
        	    return (4.0 - 8.0 * xi.doubleValue() - 4.0 * eta.doubleValue());
        	    
        	case 5:
        	    
        	    return (4.0 * eta.doubleValue());
        	    
        	case 6:
        	    
        	    return (-4.0 * eta.doubleValue());
        	
        	default:
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error: Unknown position in function vector given: " + position);
                }
                
                throw new FunctionCalculationException("Illegal position value: " + position);
        }
    }
    
    /**
     * Function for Triangle, quadratic, N derived to ETA
     * FORMULA 2.102
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double triangleQuadraticNETA(Double xi, Double eta, Integer position) throws FunctionCalculationException {

        // Determine which position in the vector the user wants to calculate
        switch(position.intValue()) {
        
        	case 1:
        	    
        	    return (-3.0 + 4.0 * xi.doubleValue() + 4.0 * eta.doubleValue());
        	
        	case 2:

        	    return (0.0);
        	
        	case 3:
        	    
        	    return (-1.0 + 4.0 * eta.doubleValue());
        	    
        	case 4:
        	    
        	    return (-4.0 * xi.doubleValue());
        	    
        	case 5:
        	    
        	    return (4.0 * xi.doubleValue());
        	    
        	case 6:
        	    
        	    return (4.0 - 4.0 * xi.doubleValue() - 8.0 * eta.doubleValue());
        	
        	default:
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error: Unknown position in function vector given: " + position);
                }
                
                throw new FunctionCalculationException("Illegal position value: " + position);
        }
    }
    
    /**
     * Function for Triangle, quadratic, X derived to XI
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double triangleQuadraticXXI(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	// Check if the number if coordinates in the static arrays match our requirements
    	if(CompilationFunctions.elementXValues.length < 6 || CompilationFunctions.elementYValues.length < 6) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
            }
            
            throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
    	}
    	
    	double sum = 0.0;
    	
    	// Calculate the sum out of the x-values in the static array multiplied by the "N derived to XI" values
    	for(int i = 0; i < 6; i++) {
    		
    		sum = sum + (CompilationFunctions.elementXValues[i] * CompilationFunctions.triangleQuadraticNXI(xi, eta, new Integer(i+1)));
    	}
    	
    	return sum;        
    }
    
    /**
     * Function for Triangle, quadratic, X derived to ETA
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double triangleQuadraticXETA(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	// Check if the number if coordinates in the static arrays match our requirements
    	if(CompilationFunctions.elementXValues.length < 6 || CompilationFunctions.elementYValues.length < 6) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
            }
            
            throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
    	}
    	
    	double sum = 0.0;
    	
    	// Calculate the sum out of the x-values in the static array multiplied by the "N derived to ETA" values
    	for(int i = 0; i < 6; i++) {
    		
    		sum = sum + (CompilationFunctions.elementXValues[i] * CompilationFunctions.triangleQuadraticNETA(xi, eta, new Integer(i+1)));
    	}
    	
    	return sum;        
    }
    
    /**
     * Function for Triangle, quadratic, Y derived to XI
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double triangleQuadraticYXI(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	// Check if the number if coordinates in the static arrays match our requirements
    	if(CompilationFunctions.elementXValues.length < 6 || CompilationFunctions.elementYValues.length < 6) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
            }
            
            throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
    	}
    	
    	double sum = 0.0;
    	
    	// Calculate the sum out of the y-values in the static array multiplied by the "N derived to XI" values
    	for(int i = 0; i < 6; i++) {
    		
    		sum = sum + (CompilationFunctions.elementYValues[i] * CompilationFunctions.triangleQuadraticNXI(xi, eta, new Integer(i+1)));
    	}
    	
    	return sum;        
    }

	/**
	 * Function for Triangle, quadratic, Y derived to ETA
	 * 
	 * @param xi
	 * @param eta
	 * @param position
	 * @return
	 */
	public static double triangleQuadraticYETA(Double xi, Double eta, Integer position) throws FunctionCalculationException {
	
		// Check if the number if coordinates in the static arrays match our requirements
		if(CompilationFunctions.elementXValues.length < 6 || CompilationFunctions.elementYValues.length < 6) {
			
	        if(localLogger.isEnabledFor(Level.ERROR)) {
	            localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
	        }
	        
	        throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
		}
		
		double sum = 0.0;
		
		// Calculate the sum out of the x-values in the static array multiplied by the "N derived to ETA" values
		for(int i = 0; i < 6; i++) {
			
			sum = sum + (CompilationFunctions.elementYValues[i] * CompilationFunctions.triangleQuadraticNETA(xi, eta, new Integer(i+1)));
		}
		
		return sum;        
	}
    
    /**
     * Function for Triangle, quadratic, JACOBI
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double triangleQuadraticJAC(Double xi, Double eta, Integer position) throws FunctionCalculationException {  	
    	
    	// Calculate detJ
    	double detJ = 0.0;
    	
    	// The position says "-1" all the time because it does not really matter
    	detJ = CompilationFunctions.triangleQuadraticXXI(xi, eta, position) * CompilationFunctions.triangleQuadraticYETA(xi, eta, position);
    	detJ = detJ - (CompilationFunctions.triangleQuadraticXETA(xi, eta, position) * CompilationFunctions.triangleQuadraticYXI(xi, eta, position));
    	
    	// Make sure detJ is greater than zero
    	if(detJ <= 0.0) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Value of detJ is not > 0.0: " + detJ);
            }
            
            throw new FunctionCalculationException("Value of detJ is not > 0.0: " + detJ);
    	}
    	
    	return (1.0 / detJ);  
    }
    
    /**
     * Function for Triangle, quadratic, 1/ JACOBI
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double triangleQuadraticJAConeThrough(Double xi, Double eta, Integer position) throws FunctionCalculationException { 
        
        return (1.0 / triangleQuadraticJAC(xi, eta, position));
    }
    
    /**
     * Function for Triangle, quadratic, H1 part of the Integral
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double triangleQuadraticH1(Double xi, Double eta, Integer position) throws FunctionCalculationException {
    	
    	double result = CompilationFunctions.triangleQuadraticNXI(xi, eta, position) * CompilationFunctions.triangleQuadraticYETA(xi, eta, position);
    	result = result - (CompilationFunctions.triangleQuadraticNETA(xi, eta, position) * CompilationFunctions.triangleQuadraticYXI(xi, eta, position));
    	
    	return result;
    }
    
    /**
     * Function for Triangle, quadratic, H2 part of the Integral
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double triangleQuadraticH2(Double xi, Double eta, Integer position) throws FunctionCalculationException {
    	
    	double result = (-1.0 * CompilationFunctions.triangleQuadraticNXI(xi, eta, position)) * CompilationFunctions.triangleQuadraticXETA(xi, eta, position);
    	result = result + (CompilationFunctions.triangleQuadraticNETA(xi, eta, position) * CompilationFunctions.triangleQuadraticXXI(xi, eta, position));
    	
    	return result;
    }
    
	/*
	 * 
	 * QUADRAT
	 * LINEAR APPROACH
	 * 
	 ***************************/
    
    /**
     * Function for Quadrat, linear, normal N
     * FORMULA 2.104
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double quadratLinearN(Double xi, Double eta, Integer position) throws FunctionCalculationException {

        // Determine which position in the vector the user wants to calculate
        switch(position.intValue()) {
        
        	case 1:
        	    
        	    return ((1.0 - xi.doubleValue()) * (1.0 - eta.doubleValue()));
        	
        	case 2:

        	    return (xi.doubleValue() * (1.0 - eta.doubleValue()));
        	
        	case 3:
        	    
        	    return (xi.doubleValue() * eta.doubleValue());
        	
        	case 4:
        		
        		return ((1.0 - xi.doubleValue()) * eta.doubleValue());
        	    
        	default:
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error: Unknown position in function vector given: " + position);
                }
                
                throw new FunctionCalculationException("Illegal position value: " + position);
        }
    }
    
    /**
     * Function for Quadrat, linear, N derived to XI
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double quadratLinearNXI(Double xi, Double eta, Integer position) throws FunctionCalculationException {

        // Determine which position in the vector the user wants to calculate
        switch(position.intValue()) {
        
        	case 1:
        	    
        	    return (-1.0 + eta.doubleValue());
        	
        	case 2:

        	    return (1.0 - eta.doubleValue());
        	
        	case 3:
        	    
        	    return (eta.doubleValue());
        	
        	case 4:
        		
        		return (-1.0 * eta.doubleValue());
        	    
        	default:
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error: Unknown position in function vector given: " + position);
                }
                
                throw new FunctionCalculationException("Illegal position value: " + position);
        }
    }
    
    /**
     * Function for Quadrat, linear, N derived to ETA
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double quadratLinearNETA(Double xi, Double eta, Integer position) throws FunctionCalculationException {

        // Determine which position in the vector the user wants to calculate
        switch(position.intValue()) {
        
        	case 1:
        	    
        	    return (1.0 + xi.doubleValue());
        	
        	case 2:

        	    return (-1.0 * xi.doubleValue());
        	
        	case 3:
        	    
        	    return (xi.doubleValue());
        	
        	case 4:
        		
        		return (1.0 - xi.doubleValue());
        	    
        	default:
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error: Unknown position in function vector given: " + position);
                }
                
                throw new FunctionCalculationException("Illegal position value: " + position);
        }
    }
    
    /**
     * Function for Quadrat, linear, X derived to XI
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double quadratLinearXXI(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	// Check if the number if coordinates in the static arrays match our requirements
    	if(CompilationFunctions.elementXValues.length < 4 || CompilationFunctions.elementYValues.length < 4) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
            }
            
            throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
    	}
    	
    	double sum = 0.0;
    	
    	// Calculate the sum out of the x-values in the static array multiplied by the "N derived to XI" values
    	for(int i = 0; i < 4; i++) {
    		
    		sum = sum + (CompilationFunctions.elementXValues[i] * CompilationFunctions.quadratLinearNXI(xi, eta, new Integer(i+1)));
    	}
    	
    	return sum;        
    }
    
    /**
     * Function for Quadrat, linear, X derived to ETA
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double quadratLinearXETA(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	// Check if the number if coordinates in the static arrays match our requirements
    	if(CompilationFunctions.elementXValues.length < 4 || CompilationFunctions.elementYValues.length < 4) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
            }
            
            throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
    	}
    	
    	double sum = 0.0;
    	
    	// Calculate the sum out of the x-values in the static array multiplied by the "N derived to ETA" values
    	for(int i = 0; i < 4; i++) {
    		
    		sum = sum + (CompilationFunctions.elementXValues[i] * CompilationFunctions.quadratLinearNETA(xi, eta, new Integer(i+1)));
    	}
    	
    	return sum;    
    }
    
    /**
     * Function for Quadrat, linear, Y derived to XI
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double quadratLinearYXI(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	// Check if the number if coordinates in the static arrays match our requirements
    	if(CompilationFunctions.elementXValues.length < 4 || CompilationFunctions.elementYValues.length < 4) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
            }
            
            throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
    	}
    	
    	double sum = 0.0;
    	
    	// Calculate the sum out of the y-values in the static array multiplied by the "N derived to XI" values
    	for(int i = 0; i < 4; i++) {
    		
    		sum = sum + (CompilationFunctions.elementYValues[i] * CompilationFunctions.quadratLinearNXI(xi, eta, new Integer(i+1)));
    	}
    	
    	return sum;    
    }
    
    /**
     * Function for Quadrat, linear, Y derived to ETA
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double quadratLinearYETA(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	// Check if the number if coordinates in the static arrays match our requirements
    	if(CompilationFunctions.elementXValues.length < 4 || CompilationFunctions.elementYValues.length < 4) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
            }
            
            throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
    	}
    	
    	double sum = 0.0;
    	
    	// Calculate the sum out of the y-values in the static array multiplied by the "N derived to ETA" values
    	for(int i = 0; i < 4; i++) {
    		
    		sum = sum + (CompilationFunctions.elementYValues[i] * CompilationFunctions.quadratLinearNETA(xi, eta, new Integer(i+1)));
    	}
    	
    	return sum;  
    }
    
    /**
     * Function for Quadrat, linear, JACOBI
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double quadratLinearJAC(Double xi, Double eta, Integer position) throws FunctionCalculationException {  	
    	
    	// Calculate detJ
    	double detJ = 0.0;
    	
    	// The position says "-1" all the time because it does not really matter
    	detJ = CompilationFunctions.quadratLinearXXI(xi, eta, position) * CompilationFunctions.quadratLinearYETA(xi, eta, position);
    	detJ = detJ - (CompilationFunctions.quadratLinearXETA(xi, eta, position) * CompilationFunctions.quadratLinearYXI(xi, eta, position));
    	
    	// Make sure detJ is greater than zero
    	if(detJ <= 0.0) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Value of detJ is not > 0.0: " + detJ);
            }
            
            throw new FunctionCalculationException("Value of detJ is not > 0.0: " + detJ);
    	}
    	
    	return (1.0 / detJ);  
    }
    
    /**
     * Function for Quadrat, linear, 1/ JACOBI
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double quadratLinearJAConeThrough(Double xi, Double eta, Integer position) throws FunctionCalculationException { 
        
        return (1.0 / quadratLinearJAC(xi, eta, position));
    }
    
    /**
     * Function for Quadrat, linear, H1 part of the Integral
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double quadratLinearH1(Double xi, Double eta, Integer position) throws FunctionCalculationException {
    	
    	double result = CompilationFunctions.quadratLinearNXI(xi, eta, position) * CompilationFunctions.quadratLinearYETA(xi, eta, position);
    	result = result - (CompilationFunctions.quadratLinearNETA(xi, eta, position) * CompilationFunctions.quadratLinearYXI(xi, eta, position));
    	
    	return result;
    }
    
    /**
     * Function for Quadrat, linear, H2 part of the Integral
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double quadratLinearH2(Double xi, Double eta, Integer position) throws FunctionCalculationException {
    	
    	double result = (-1.0 * CompilationFunctions.quadratLinearNXI(xi, eta, position)) * CompilationFunctions.quadratLinearXETA(xi, eta, position);
    	result = result + (CompilationFunctions.quadratLinearNETA(xi, eta, position) * CompilationFunctions.quadratLinearXXI(xi, eta, position));
    	
    	return result;
    }
    
	/*
	 * 
	 * QUADRAT
	 * QUADRATIC APPROACH
	 * 
	 ***************************/
    
    /**
     * Function for Quadrat, quadratic, normal N
     * FORMULA 2.105
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double quadratQuadraticN(Double xi, Double eta, Integer position) throws FunctionCalculationException {

        // Determine which position in the vector the user wants to calculate
        switch(position.intValue()) {
        
        	case 1:
        	    
        	    return ((1.0 - xi.doubleValue()) * (1.0 - eta.doubleValue()) * (1.0 - 2.0 * xi.doubleValue() - 2.0 * eta.doubleValue()));
        	
        	case 2:

        	    return ((-1.0 * xi.doubleValue()) * (1.0 - eta.doubleValue()) * (1.0 - 2.0 * xi.doubleValue() + 2.0 * eta.doubleValue()));
        	
        	case 3:
        	    
        	    return ((-1.0 * xi.doubleValue() * eta.doubleValue()) * (3.0 - 2.0 * xi.doubleValue() - 2.0 * eta.doubleValue()));
        	    
        	case 4:

        	    return ((-1.0 * eta.doubleValue()) * (1.0 - xi.doubleValue()) * (1.0 + 2.0 * xi.doubleValue() - 2.0 * eta.doubleValue()));
        	
        	case 5:
        	    
        	    return ((4.0 * xi.doubleValue()) * (1.0 - xi.doubleValue()) * (1.0 - eta.doubleValue()));
        	    
        	case 6:

        	    return ((4.0 * xi.doubleValue() * eta.doubleValue()) * (1.0 - eta.doubleValue()));
        	
        	case 7:
        	    
        	    return ((4.0 * xi.doubleValue() * eta.doubleValue()) * (1.0 - xi.doubleValue()));
        	    
        	case 8:
        	    
        	    return ((4.0 * eta.doubleValue()) * (1.0 - xi.doubleValue()) * (1.0 - eta.doubleValue()));
        	
        	default:
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error: Unknown position in function vector given: " + position);
                }
                
                throw new FunctionCalculationException("Illegal position value: " + position);
        }
    }
    
    /**
     * Function for Quadrat, quadratic, N derived to XI
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double quadratQuadraticNXI(Double xi, Double eta, Integer position) throws FunctionCalculationException {

        // Determine which position in the vector the user wants to calculate
        switch(position.intValue()) {
        
        	case 1:
        	    
        		//return (-2.0*(eta-1.0)*(xi-1.0)-(2*xi+2*eta-1.0)*(eta-1.0));
        	    return (-3.0 + 5.0 * eta.doubleValue() + 4.0 * xi.doubleValue() - 2.0 * Math.pow(eta.doubleValue(), 2) - 4.0 * eta.doubleValue() * xi.doubleValue());
        	
        	case 2:

        		//return (-2.0*(eta-1.0)*xi-(2.0*xi-2.0*eta-1.0)*(eta-1.0));
        	    return (-1.0 - eta.doubleValue() + 4.0 * xi.doubleValue() + 2.0 * Math.pow(eta.doubleValue(), 2) - 4.0 * eta.doubleValue() * xi.doubleValue());
        	
        	case 3:
        	    
        		//return (2.0*xi*eta+(2*xi+2*eta-3.0)*eta);
        	    return (-3.0 * eta.doubleValue() + 2.0 * Math.pow(eta.doubleValue(), 2) + 4.0 * eta.doubleValue() * xi.doubleValue());
        	    
        	case 4:

        		//return (2*eta*(xi-1.0)+(2.0*xi-2.0*eta+1.0)*eta);
        	    return (eta.doubleValue() * (1.0 + 2.0 * xi.doubleValue() - 2.0 * eta.doubleValue()) - 2.0 * eta.doubleValue() * (1.0 - xi.doubleValue()));
        	
        	case 5:
        	    
        		//return (4*(eta-1)*(xi-1)+4*(eta-1)*xi);
        	    return (4.0 - 4.0 * eta.doubleValue() - 8.0 * xi.doubleValue() + 8.0 * eta.doubleValue() * xi.doubleValue());
        	    
        	case 6:

        		//return (-4*eta*(eta-1));
        	    return (4.0 * eta.doubleValue() - 4.0 * Math.pow(eta.doubleValue(), 2));
        	
        	case 7:
        	    
        		//return (-4*xi*eta-4*eta*(xi-1));
        	    return (4.0 * eta.doubleValue() - 8.0 * eta.doubleValue() * xi.doubleValue());
        	    
        	case 8:
        	    
        		//return (4*eta*(eta-1));
        	    return (-4.0 * eta.doubleValue() + 4.0 * Math.pow(eta.doubleValue(), 2));
        	
        	default:
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error: Unknown position in function vector given: " + position);
                }
                
                throw new FunctionCalculationException("Illegal position value: " + position);
        }
    }
    
    /**
     * Function for Quadrat, quadratic, N derived to ETA
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double quadratQuadraticNETA(Double xi, Double eta, Integer position) throws FunctionCalculationException {

        // Determine which position in the vector the user wants to calculate
        switch(position.intValue()) {
        
        	case 1:
        	    
        		//return (-2.0*(eta-1.0)*(xi-1.0)-(2*xi+2*eta-1.0)*(xi-1.0));
        	    return (-3.0 + 5.0 * xi.doubleValue() + 4.0 * eta.doubleValue() - 2.0 * Math.pow(xi.doubleValue(), 2) - 4.0 * eta.doubleValue() * xi.doubleValue());
        	
        	case 2:

        		//return (2.0*(eta-1.0)*xi-(2.0*xi-2.0*eta-1.0)*xi);
        	    return ((-1.0 * xi.doubleValue()) - 2.0 * Math.pow(xi.doubleValue(), 2) + 4.0 * eta.doubleValue() * xi.doubleValue());
        	
        	case 3:
        	    
        		//return (2.0*xi*eta+(2.0*xi+2.0*eta-3.0)*xi);
        	    return (-3.0 * xi.doubleValue() + 2.0 * Math.pow(xi.doubleValue(), 2) + 4.0 * eta.doubleValue() * xi.doubleValue());
        	    
        	case 4:

        		//return (-2.0*eta*(xi-1.0)+(2.0*xi-2.0*eta+1.0)*(xi-1.0));
        	    return (-1.0 - xi.doubleValue() + 4.0 * eta.doubleValue() + 2.0 * Math.pow(xi.doubleValue(), 2) - 4.0 * eta.doubleValue() * xi.doubleValue());
        	
        	case 5:
        	    
        		//return (4*xi*(xi-1));
        	    return (-4.0 * xi.doubleValue() + 4.0 * Math.pow(xi.doubleValue(), 2));
        	    
        	case 6:

        		//return (-4*xi*eta-4*(eta-1)*xi);
        	    return (4.0 * xi.doubleValue() - 8.0 * eta.doubleValue() * xi.doubleValue());
        	
        	case 7:
        	    
        		//return (-4*xi*(xi-1));
        	    return (4.0 * xi.doubleValue() - 4.0 * Math.pow(xi.doubleValue(), 2));
        	    
        	case 8:
        	    
        		//return (4*(eta-1)*(xi-1)+4*eta*(xi-1));
        	    return (4.0 - 4.0 * xi.doubleValue() - 8.0 * eta.doubleValue() + 8.0 * eta.doubleValue() * xi.doubleValue());
        	
        	default:
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error: Unknown position in function vector given: " + position);
                }
                
                throw new FunctionCalculationException("Illegal position value: " + position);
        }
    }
    
    /**
     * Function for Quadrat, quadratic, X derived to XI
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double quadratQuadraticXXI(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	// Check if the number if coordinates in the static arrays match our requirements
    	if(CompilationFunctions.elementXValues.length < 8 || CompilationFunctions.elementYValues.length < 8) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
            }
            
            throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
    	}
    	
    	double sum = 0.0;
    	
    	// Calculate the sum out of the x-values in the static array multiplied by the "N derived to XI" values
    	for(int i = 0; i < 8; i++) {
    		
    		sum = sum + (CompilationFunctions.elementXValues[i] * CompilationFunctions.quadratQuadraticNXI(xi, eta, new Integer(i+1)));
    	}
    	
    	return sum;        
    }
    
    /**
     * Function for Quadrat, quadratic, X derived to ETA
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double quadratQuadraticXETA(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	// Check if the number if coordinates in the static arrays match our requirements
    	if(CompilationFunctions.elementXValues.length < 8 || CompilationFunctions.elementYValues.length < 8) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
            }
            
            throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
    	}
    	
    	double sum = 0.0;
    	
    	// Calculate the sum out of the x-values in the static array multiplied by the "N derived to ETA" values
    	for(int i = 0; i < 8; i++) {
    		
    		sum = sum + (CompilationFunctions.elementXValues[i] * CompilationFunctions.quadratQuadraticNETA(xi, eta, new Integer(i+1)));
    	}
    	
    	return sum;        
    }
    
    /**
     * Function for Quadrat, quadratic, Y derived to XI
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double quadratQuadraticYXI(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	// Check if the number if coordinates in the static arrays match our requirements
    	if(CompilationFunctions.elementXValues.length < 8 || CompilationFunctions.elementYValues.length < 8) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
            }
            
            throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
    	}
    	
    	double sum = 0.0;
    	
    	// Calculate the sum out of the y-values in the static array multiplied by the "N derived to XI" values
    	for(int i = 0; i < 8; i++) {
    		
    		sum = sum + (CompilationFunctions.elementYValues[i] * CompilationFunctions.quadratQuadraticNXI(xi, eta, new Integer(i+1)));
    	}
    	
    	return sum;        
    }
    
    /**
     * Function for Triangle, quadratic, Y derived to ETA
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double quadratQuadraticYETA(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	// Check if the number if coordinates in the static arrays match our requirements
    	if(CompilationFunctions.elementXValues.length < 8 || CompilationFunctions.elementYValues.length < 8) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
            }
            
            throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
    	}
    	
    	double sum = 0.0;
    	
    	// Calculate the sum out of the x-values in the static array multiplied by the "N derived to ETA" values
    	for(int i = 0; i < 8; i++) {
    		
    		sum = sum + (CompilationFunctions.elementYValues[i] * CompilationFunctions.quadratQuadraticNETA(xi, eta, new Integer(i+1)));
    	}
    	
    	return sum;        
    }
    
    /**
     * Function for Quadrat, quadratic, JACOBI
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double quadratQuadraticJAC(Double xi, Double eta, Integer position) throws FunctionCalculationException {  	
    	
    	// Calculate detJ
    	double detJ = 0.0;
    	
    	// The position says "-1" all the time because it does not really matter
    	detJ = CompilationFunctions.quadratQuadraticXXI(xi, eta, position) * CompilationFunctions.quadratQuadraticYETA(xi, eta, position);
    	detJ = detJ - (CompilationFunctions.quadratQuadraticXETA(xi, eta, position) * CompilationFunctions.quadratQuadraticYXI(xi, eta, position));
    	
    	// Make sure detJ is greater than zero
    	if(detJ <= 0.0) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Value of detJ is not > 0.0: " + detJ);
            }
            
            throw new FunctionCalculationException("Value of detJ is not > 0.0: " + detJ);
    	}
    	
    	return (1.0 / detJ);  
    }
    
    /**
     * Function for Quadrat, quadratic, 1/ JACOBI
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double quadratQuadraticJAConeThrough(Double xi, Double eta, Integer position) throws FunctionCalculationException { 
        
        return (1.0 / quadratQuadraticJAC(xi, eta, position));
    }
    
    /**
     * Function for Quadrat, quadratic, H1 part of the Integral
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double quadratQuadraticH1(Double xi, Double eta, Integer position) throws FunctionCalculationException {
    	
    	double result = CompilationFunctions.quadratQuadraticNXI(xi, eta, position) * CompilationFunctions.quadratQuadraticYETA(xi, eta, position);
    	result = result - (CompilationFunctions.quadratQuadraticNETA(xi, eta, position) * CompilationFunctions.quadratQuadraticYXI(xi, eta, position));
    	
    	return result;
    }
    
    /**
     * Function for Quadrat, quadratic, H2 part of the Integral
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double quadratQuadraticH2(Double xi, Double eta, Integer position) throws FunctionCalculationException {
    	
    	double result = (-1.0 * CompilationFunctions.quadratQuadraticNXI(xi, eta, position)) * CompilationFunctions.quadratQuadraticXETA(xi, eta, position);
    	result = result + (CompilationFunctions.quadratQuadraticNETA(xi, eta, position) * CompilationFunctions.quadratQuadraticXXI(xi, eta, position));
    	
    	return result;
    }
    
	/*
	 * 
	 * MEMBRANE CIRCLE
	 * 
	 ***************************/
    
    /**
     * Membrane Circle Ring Q1
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double membraneCircleRingQ1(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	return (CompilationFunctions.radius1 + (CompilationFunctions.radius2 - CompilationFunctions.radius1) * xi.doubleValue());
    }
    
    /**
     * Membrane Circle Ring Q2
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double membraneCircleRingQ2(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	return (1/(CompilationFunctions.radius1 + (CompilationFunctions.radius2 - CompilationFunctions.radius1) * xi.doubleValue()));
    }
    
    /**
     * membraneCircleSectorTriangleQuadraticN
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double membraneCircleSectorTriangleQuadraticN(Double xi, Double eta, Integer position) throws FunctionCalculationException {

        // Determine which position in the vector the user wants to calculate
        switch(position.intValue()) {
        
        	case 1:
        	    
        	    return (1.0 - (3.0 * xi.doubleValue()) + (2.0 * Math.pow(xi.doubleValue(), 2)));
        	
        	case 2:

        	    return ((-1.0 * xi.doubleValue()) + (2.0 * Math.pow(xi.doubleValue(), 2)) - (xi.doubleValue() * eta.doubleValue()) - (2.0 * Math.pow(xi.doubleValue(), 2) * eta.doubleValue()) + (2.0 * xi.doubleValue() * Math.pow(eta.doubleValue(), 2)));
        	
        	case 3:
        	    
        	    return ((-3.0 * xi.doubleValue() * eta.doubleValue()) + (2.0 * Math.pow(xi.doubleValue(), 2) * eta.doubleValue()) + (2.0 * xi.doubleValue() * Math.pow(eta.doubleValue(), 2)));
        	
        	case 4:
        		
        		return ((4.0 * xi.doubleValue()) - (4.0 * Math.pow(xi.doubleValue(), 2)) - (4.0 * xi.doubleValue() * eta.doubleValue()) + (4.0 * Math.pow(xi.doubleValue(), 2) * eta.doubleValue()));
        		
        	case 5:
        		
        		return ((4.0 * xi.doubleValue() * eta.doubleValue()) - (4.0 * xi.doubleValue() * Math.pow(eta.doubleValue(), 2)));
        		
        	case 6:
        		
        		return ((4.0 *  xi.doubleValue() * eta.doubleValue()) - (4.0 * Math.pow(xi.doubleValue(), 2) * eta.doubleValue()));
        	
        	default:
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error: Unknown position in function vector given: " + position);
                }
                
                throw new FunctionCalculationException("Illegal position value: " + position);
        }
    }
    
    /**
     * membraneCircleSectorTriangleQuadraticNXI
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double membraneCircleSectorTriangleQuadraticNXI(Double xi, Double eta, Integer position) throws FunctionCalculationException {

        // Determine which position in the vector the user wants to calculate
        switch(position.intValue()) {
        
        	case 1:
        	    
        	    return (- 3.0  + (4.0 * xi.doubleValue()));
        	
        	case 2:

        	    return (-1.0 + (4.0 * xi.doubleValue()) - eta.doubleValue() - (4.0 * xi.doubleValue() * eta.doubleValue()) + (2.0 *  Math.pow(eta.doubleValue(), 2)));
        	
        	case 3:
        	    
        	    return ((-3.0 * eta.doubleValue()) + (4.0 * xi.doubleValue() * eta.doubleValue()) + (2.0  * Math.pow(eta.doubleValue(), 2)));
        	
        	case 4:
        		
        		return (4.0 - (8.0 * xi.doubleValue()) - (4.0 * eta.doubleValue()) + (8.0 * xi.doubleValue() * eta.doubleValue()));
        		
        	case 5:
        		
        		return ((4.0  * eta.doubleValue()) - (4.0 * Math.pow(eta.doubleValue(), 2)));
        		
        	case 6:
        		
        		return ((4.0 *  eta.doubleValue()) - (8.0 * xi.doubleValue() * eta.doubleValue()));
        	
        	default:
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error: Unknown position in function vector given: " + position);
                }
                
                throw new FunctionCalculationException("Illegal position value: " + position);
        }
    }
    
    /**
     * membraneCircleSectorTriangleQuadraticNETA
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double membraneCircleSectorTriangleQuadraticNETA(Double xi, Double eta, Integer position) throws FunctionCalculationException {

        // Determine which position in the vector the user wants to calculate
        switch(position.intValue()) {
        
        	case 1:
        	    
        	    return (0.0);
        	
        	case 2:

        	    return ((-1.0 * xi.doubleValue()) - (2.0 * Math.pow(xi.doubleValue(), 2)) + (4.0 * xi.doubleValue() * eta.doubleValue()));
        	
        	case 3:
        	    
        	    return ((-3.0 * xi.doubleValue()) + (2.0 * Math.pow(xi.doubleValue(), 2)) + (4.0 * xi.doubleValue() * eta.doubleValue()));
        	
        	case 4:
        		
        		return ((-4.0 * xi.doubleValue()) + (4.0 * Math.pow(xi.doubleValue(), 2)));
        		
        	case 5:
        		
        		return ((4.0 * xi.doubleValue()) - (8.0 * xi.doubleValue() * eta.doubleValue()));
        		
        	case 6:
        		
        		return ((4.0 * xi.doubleValue()) - (4.0 * Math.pow(xi.doubleValue(), 2)));
        	
        	default:
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error: Unknown position in function vector given: " + position);
                }
                
                throw new FunctionCalculationException("Illegal position value: " + position);
        }
    }
    
    /**
     * membraneCircleSectorTriangleQuadraticNETAThroughXI
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double membraneCircleSectorTriangleQuadraticNETAThroughXI(Double xi, Double eta, Integer position) throws FunctionCalculationException {

        // Determine which position in the vector the user wants to calculate
        switch(position.intValue()) {
        
        	case 1:
        	    
        	    return (0.0);
        	
        	case 2:

        	    return ((-1.0) - (2.0 * xi.doubleValue()) + (4.0 * eta.doubleValue()));
        	
        	case 3:
        	    
        	    return ((-3.0) + (2.0 * xi.doubleValue()) + (4.0 * eta.doubleValue()));
        	
        	case 4:
        		
        		return ((-4.0) + (4.0 * xi.doubleValue()));
        		
        	case 5:
        		
        		return ((4.0) - (8.0 * eta.doubleValue()));
        		
        	case 6:
        		
        		return ((4.0) - (4.0 * xi.doubleValue()));
        	
        	default:
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error: Unknown position in function vector given: " + position);
                }
                
                throw new FunctionCalculationException("Illegal position value: " + position);
        }
    }
    
    /**
     * XI
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double xi(Double xi, Double eta, Integer position) throws FunctionCalculationException {

        return xi.doubleValue();
    }
    
    /*
     * 
     * SPANNER
     * 
     ***************/
    
    /**
     * Function for Triangle, quadratic, XI derived to X
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double spannerTriangleQuadraticXIX(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	// Check if the number if coordinates in the static arrays match our requirements
    	if(CompilationFunctions.elementXValues.length < 6 || CompilationFunctions.elementYValues.length < 6) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
            }
            
            throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
    	}
    	
    	double y1 = CompilationFunctions.elementYValues[0];
    	double yE = CompilationFunctions.elementYValues[2];
    	 	  	
    	double result = (yE-y1) / CompilationFunctions.triangleQuadraticJAConeThrough(xi, eta, position);
    	
    	return result;        
    }
    
    /**
     * Function for Triangle, quadratic, XI derived to X
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double spannerTriangleQuadraticXIY(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	// Check if the number if coordinates in the static arrays match our requirements
    	if(CompilationFunctions.elementXValues.length < 6 || CompilationFunctions.elementYValues.length < 6) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
            }
            
            throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
    	}
    	
    	double x1 = CompilationFunctions.elementXValues[0];
    	double xE = CompilationFunctions.elementXValues[2];
    	 	  	
    	double result = -1.0*((xE-x1)/ CompilationFunctions.triangleQuadraticJAConeThrough(xi,eta,position));
    	
    	return result;        
    }
    
    /**
     * Function for Triangle, quadratic, XI derived to X
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double spannerTriangleQuadraticETAX(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	// Check if the number if coordinates in the static arrays match our requirements
    	if(CompilationFunctions.elementXValues.length < 6 || CompilationFunctions.elementYValues.length < 6) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
            }
            
            throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
    	}
    	
    	double y1 = CompilationFunctions.elementYValues[0];
    	double y2 = CompilationFunctions.elementYValues[1];
    	
    	double result = -1.0*((y2-y1)/ CompilationFunctions.triangleQuadraticJAConeThrough(xi,eta,position));
    	
    	return result;        
    }
    
    /**
     * Function for Triangle, quadratic, XI derived to X
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double spannerTriangleQuadraticETAY(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	// Check if the number if coordinates in the static arrays match our requirements
    	if(CompilationFunctions.elementXValues.length < 6 || CompilationFunctions.elementYValues.length < 6) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
            }
            
            throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
    	}
    	
    	double x1 = CompilationFunctions.elementXValues[0];
    	double x2 = CompilationFunctions.elementXValues[1];
    	
    	double result = (x2-x1)/ CompilationFunctions.triangleQuadraticJAConeThrough(xi,eta,position);
    	
    	return result;        
    }
    
    /**
     * Function for Triangle, quadratic, XI derived to X
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double spannerQuadratQuadraticXIX(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	// Check if the number if coordinates in the static arrays match our requirements
    	if(CompilationFunctions.elementXValues.length < 8 || CompilationFunctions.elementYValues.length < 8) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
            }
            
            throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
    	}
    	
    	double y1 = CompilationFunctions.elementYValues[0];
    	double y4 = CompilationFunctions.elementYValues[3];
    	 	  	
    	double result = (y4-y1)/ CompilationFunctions.quadratQuadraticJAConeThrough(xi,eta,position);
    	
    	return result;        
    }
    
    /**
     * Function for Triangle, quadratic, XI derived to X
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double spannerQuadratQuadraticXIY(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	// Check if the number if coordinates in the static arrays match our requirements
    	if(CompilationFunctions.elementXValues.length < 8 || CompilationFunctions.elementYValues.length < 8) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
            }
            
            throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
    	}
    	
    	double x1 = CompilationFunctions.elementXValues[0];
    	double x4 = CompilationFunctions.elementXValues[3];
    	 	  	
    	double result = -1.0*((x4-x1)/ CompilationFunctions.quadratQuadraticJAConeThrough(xi,eta,position));
    	
    	return result;        
    }
    
    /**
     * Function for Triangle, quadratic, XI derived to X
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double spannerQuadratQuadraticETAX(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	// Check if the number if coordinates in the static arrays match our requirements
    	if(CompilationFunctions.elementXValues.length < 8 || CompilationFunctions.elementYValues.length < 8) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
            }
            
            throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
    	}
    	
    	double y1 = CompilationFunctions.elementYValues[0];
    	double y2 = CompilationFunctions.elementYValues[1];
    	
    	double result = -1.0*((y2-y1)/ CompilationFunctions.quadratQuadraticJAConeThrough(xi,eta,position));
    	
    	return result;        
    }
    
    /**
     * Function for Triangle, quadratic, XI derived to X
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double spannerQuadratQuadraticETAY(Double xi, Double eta, Integer position) throws FunctionCalculationException {

    	// Check if the number if coordinates in the static arrays match our requirements
    	if(CompilationFunctions.elementXValues.length < 8 || CompilationFunctions.elementYValues.length < 8) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
            }
            
            throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
    	}
    	
    	double x1 = CompilationFunctions.elementXValues[0];
    	double x2 = CompilationFunctions.elementXValues[1];
    	
    	double result = (x2-x1)/CompilationFunctions.quadratQuadraticJAConeThrough(xi,eta,position);
    	
    	return result;        
    }
    
    /**
     * Function for Spanner Triangle, quadratic, H1 part of the Integral
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double spannerTriangleQuadraticH1(Double xi, Double eta, Integer position) throws FunctionCalculationException {
    	
    	double result = (CompilationFunctions.triangleQuadraticYETA(xi,eta,position) * CompilationFunctions.triangleQuadraticNXI(xi,eta,position)); 
    	result = result - (CompilationFunctions.triangleQuadraticYXI(xi,eta,position) * CompilationFunctions.triangleQuadraticNETA(xi,eta,position));
    	
    	// JACOBI = 1/detJ, which makes detJ = 1/JAC
    	result = result * (1.0/Math.sqrt(CompilationFunctions.triangleQuadraticJAConeThrough(xi,eta,position)));
    	
    	return result;
    }
    
    /**
     * Function for Spanner Triangle, quadratic, H2 part of the Integral
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double spannerTriangleQuadraticH2(Double xi, Double eta, Integer position) throws FunctionCalculationException {
    	
    	double result = ((-1.0*CompilationFunctions.triangleQuadraticXETA(xi,eta,position)) * CompilationFunctions.triangleQuadraticNXI(xi,eta,position)); 
    	result = result + (CompilationFunctions.triangleQuadraticXXI(xi,eta,position)*CompilationFunctions.triangleQuadraticNETA(xi,eta,position));
    	
    	// JACOBI = 1/detJ, which makes detJ = 1/JAC
    	result = result * (1.0/Math.sqrt(CompilationFunctions.triangleQuadraticJAConeThrough(xi,eta,position)));
    	
    	return result;
    }
    
    /**
     * Function for Spanner Triangle, quadratic, Q part of the second Integral for Matrix S11_ROOF
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double spannerS11IntegralTwoQ(Double xi, Double eta, Integer position) throws FunctionCalculationException {
    	
    	return (0.5*(1.0-FEMManager.getInstance().getSpannerMaterialConstantV()));
    }
    
    /**
     * Function for Spanner Triangle, quadratic, Q part of the first Integral for Matrix S22_ROOF
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double spannerS22IntegralOneQ(Double xi, Double eta, Integer position) throws FunctionCalculationException {
    	
    	return (0.5*(1.0-FEMManager.getInstance().getSpannerMaterialConstantV()));
    }
    
    /**
     * Function for Spanner Triangle, quadratic, Q part of the first Integral for Matrix S12_ROOF
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double spannerS12IntegralOneQ(Double xi, Double eta, Integer position) throws FunctionCalculationException {
    	
    	return FEMManager.getInstance().getSpannerMaterialConstantV();
    }
    
    /**
     * Function for Spanner Triangle, quadratic, Q part of the second Integral for Matrix S12_ROOF
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double spannerS12IntegralTwoQ(Double xi, Double eta, Integer position) throws FunctionCalculationException {
    	
    	return (0.5*(1.0-FEMManager.getInstance().getSpannerMaterialConstantV()));
    }
    
    /**
     * Function for Spanner Quadrat, quadratic, H1 part of the Integral
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double spannerquadratQuadraticH1(Double xi, Double eta, Integer position) throws FunctionCalculationException {
    	
    	double result = (CompilationFunctions.quadratQuadraticYETA(xi,eta,position) * CompilationFunctions.quadratQuadraticNXI(xi,eta,position)); 
    	result = result - (CompilationFunctions.quadratQuadraticYXI(xi,eta,position) * CompilationFunctions.quadratQuadraticNETA(xi,eta,position));
    	
    	// JACOBI = 1/detJ, which makes detJ = 1/JAC
    	result = result * (1.0/Math.sqrt(CompilationFunctions.quadratQuadraticJAConeThrough(xi,eta,position)));
    	
    	return result;
    }
    
    /**
     * Function for Spanner Quadrat, quadratic, H2 part of the Integral
     * 
     * @param xi
     * @param eta
     * @param position
     * @return
     */
    public static double spannerquadratQuadraticH2(Double xi, Double eta, Integer position) throws FunctionCalculationException {
    	
    	double result = ((-1.0*CompilationFunctions.quadratQuadraticXETA(xi,eta,position)) * CompilationFunctions.quadratQuadraticNXI(xi,eta,position)); 
    	result = result + (CompilationFunctions.quadratQuadraticXXI(xi,eta,position)*CompilationFunctions.quadratQuadraticNETA(xi,eta,position));
    	
    	// JACOBI = 1/detJ, which makes detJ = 1/JAC
    	result = result * (1.0/Math.sqrt(CompilationFunctions.quadratQuadraticJAConeThrough(xi,eta,position)));
    	
    	return result;
    }
}
