/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.calc.objects;

import java.util.ArrayList;
import java.util.Iterator;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import fhm.edu.fem.constants.SMatrixConstants;
import fhm.edu.fem.util.logging.LoggingManager;

/**
 * Class Description:
 *		This class acts as element list. It takes the element index from the element itself
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class ElementList {
    
	/* Our local logger object */
	protected static Logger localLogger = LoggingManager.getLogger(ElementList.class.getName());

	/* The List object where we store our elements */
	protected ArrayList<Element> elements = new ArrayList<Element>();
	
	/**
	 * Create a new element list
	 */
	public ElementList() {}
	
	/**
	 * Set a new element in the list. It will be set at the index specified in the element
	 * @param element, the element to set
	 */
	public void setElement(Element element) {
		elements.add(element);
	}
	
	/**
	 * Retrieve a element with the given index from the list
	 * @param elementIndex
	 * @return the found element
	 */
	public Element getElement(int elementIndex) {
		
		// Run through the list and find the element with the given index
		Iterator it = elements.iterator();
		Element tmpElement = null;
		
		while(it.hasNext()) {
			// Grab the next element from the list
			tmpElement = (Element)(it.next());
			
			// Check if it is the element we are searching for
			if(tmpElement != null) {
				
				if(elementIndex == tmpElement.getId()) {
					
					// We found the correct element, return it
					return tmpElement;
				}
			}
		}
		
		// We found no proper element
		return null;
	}
	
	/**
	 * Convert this list to an array of elements
	 * @return the element array
	 */
	public Object[] toArray() {
		return this.elements.toArray();
	}
	
	/**
	 * Clear the list
	 */
	public void clear() {
		this.elements.clear();
	}
	
	/**
	 * Retrieve the number of elements in the list
	 * @return the number of elements
	 */
	public int size() {
		return this.elements.size();
	}
	
	/**
	 * Clone Implementation
	 * @return a clone of this list
	 */
	public Object clone() {
		// Create the new list
		ElementList cloneList = new ElementList();
		
		// Fill the new list with the same elements as the current list
		Object[] elements = this.toArray();
		
		for(int i = 0; i < elements.length; i++) {
			cloneList.setElement((Element)(elements[i]));
		}
		
		return cloneList;
	}
	
	/**
	 * Checks if somewhere in this elementlist does a node already exist between
	 * the given nodes (in any element)
	 * @param a, the first node
	 * @param b, the second node
	 * @return null if no node has been found, or the node itself
	 */
	public Node doesNodeExistBetween(Node a, Node b) {
	    
	    // Run through all elements
	    for(int i = 1; i <= this.size(); i++) {
	        
	        // Grab the current element and it's node list
	        Element tmpElement = this.getElement(i);
	        NodeList nodeList = tmpElement.getNodeList();
	        
	        // The type of the element
	        int elementType = tmpElement.getElementType();
	        
	        // Set the node list size based on the element type
	        int nodeListSize = -1;
	        
	        // Make sure we only run through the non-pseudo nodes
	        if(elementType == SMatrixConstants.FIGURE_TRIANGLE) {
	            
	            nodeListSize = 3;
	            
	        } 
	        else if(elementType == SMatrixConstants.FIGURE_QUADRAT) {
	            
	            nodeListSize = 4;
	        } 
	        else 
	        {
				if(localLogger.isEnabledFor(Level.ERROR)) {
					localLogger.log(Level.ERROR, "Encountered unknown element type. Nodelist size: " + nodeList.size());
				}
				
				return null;
	        }
	        
            // Let's see if the nodes 'a' and 'b' are in this element somewhere
            for(int j = 1; j <= nodeListSize; j++) {
                
                // Grab the first and second node
                Node firstNode = nodeList.getNodeInCorrectOrder(j);
                Node secNode = null;
                
                if((j+1) > nodeListSize) {
                    
                    secNode = nodeList.getNodeInCorrectOrder(1); 
                }
                else
                {
                    secNode = nodeList.getNodeInCorrectOrder(j+1);  
                }
                
                // Compare both the first and second node if they match node 'a' and node 'b'
                if((a.equals(firstNode) && b.equals(secNode)) || (a.equals(secNode) && b.equals(firstNode))) {
                    
                    // We found the proper nodes, calc the index of the pseudo node
                    int pseudoNodeIndex = j + elementType;
                    
                    // Check if the index is in a proper range
                    if(pseudoNodeIndex <= nodeList.size() && pseudoNodeIndex >= (elementType)) {
                        
                        // Grab the node
                        return nodeList.getNodeInCorrectOrder(pseudoNodeIndex);
                    }
                }
            }
	    }
	    
	    // No node found
	    return null;
	}
}
