/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.calc.objects;

import java.lang.reflect.Method;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import fhm.edu.fem.exception.FunctionCalculationException;
import fhm.edu.fem.util.logging.LoggingManager;

/**
 * Class Description:
 *		Holds all the static function methods needed for the
 *		calculation of the line integral when we deal with
 *		temperature problems via FEMCalcTemperature
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class LineIntegralFunctions {

	/* Our local logger object */
	protected static Logger localLogger = LoggingManager.getLogger(LineIntegralFunctions.class.getName());
	
	/* The method names */
	public static final String N_SIGMA = "nSigma";
	public static final String Q_ROOT = "qRoot";
	
	/* Global data structure to remember the current x- and y-values of the element, used in the Q_ROOT method */
	protected static double[] ctabXValues;
	protected static double[] ctabYValues;
	
	/**
	 * Get the proper method
	 * @param methodName, the method name taken from the constants in this class
	 * @return the method
	 */
	public static Method getMethod(String methodName) throws SecurityException, NoSuchMethodException {
	    
	    // Build the args array
	    Class args[] = new Class[3];
	    args[0] = Double.class;
	    args[1] = Double.class;
	    args[2] = Integer.class;
	    
	    return LineIntegralFunctions.class.getDeclaredMethod(methodName, args);
	}
	
	/**
	 * To remember the current ctabs x- and y-values in the static data
	 * structure here in this class - used in the XXI, XETA, YXI, YETA methods
	 * @param element, the element which values we want to store
	 */
	public static void storeCtabXYValues(CTab ctab) {
		
		// Create the arrays, depending on the number of element nodes
	    LineIntegralFunctions.ctabXValues = new double[3];
	    LineIntegralFunctions.ctabYValues = new double[3];
	    
		// Grab the nodes from the ctab and store them in the double arrays
	    Node tmpNode = ctab.getLeftNode();
	    LineIntegralFunctions.ctabXValues[0] = tmpNode.getX();
	    LineIntegralFunctions.ctabYValues[0] = tmpNode.getY();

	    tmpNode = ctab.getMidNode();
	    LineIntegralFunctions.ctabXValues[1] = tmpNode.getX();
	    LineIntegralFunctions.ctabYValues[1] = tmpNode.getY();

	    tmpNode = ctab.getRightNode(); 
	    LineIntegralFunctions.ctabXValues[2] = tmpNode.getX();
	    LineIntegralFunctions.ctabYValues[2] = tmpNode.getY();
	}
	
    /**
     * Function for N derived to Sigma
     * 
     * @param sigma
     * @param empty
     * @param position
     * @return
     */
    public static double nSigma(Double sigma, Double empty, Integer position) throws FunctionCalculationException {

        // Determine which position in the vector the user wants to calculate
        switch(position.intValue()) {
        
        	case 1:
        	    
        	    return (1.0 - 3.0*sigma.doubleValue() + 2.0*Math.pow(sigma.doubleValue(), 2));
        	
        	case 2:

        	    return (4.0*sigma.doubleValue() - 4.0*Math.pow(sigma.doubleValue(), 2));
        	
        	case 3:
        	    
        	    return (-1.0 * sigma.doubleValue() + 2.0*Math.pow(sigma.doubleValue(), 2));
        	
        	default:
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error: Unknown position in function vector given: " + position);
                }
                
                throw new FunctionCalculationException("Illegal position value: " + position);
        }
    }
    
    /**
     * Function for q Root
     * 
     * @param sigma
     * @param empty
     * @param position
     * @return
     */
    public static double qRoot(Double sigma, Double empty, Integer position) throws FunctionCalculationException {

    	// Check if the number if coordinates in the static arrays match our requirements
    	if(LineIntegralFunctions.ctabXValues.length < 3 || LineIntegralFunctions.ctabYValues.length < 3) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error: Insufficient number of x- and y-coordinates stored in the static arrays !");
            }
            
            throw new FunctionCalculationException("Insufficient number of coordinates in static arrays !");
    	}
    	
    	double valueX = 0.0;
    	double valueY = 0.0;
    	
    	// Calculate (x * N')^2 + (y * N')^2
    	valueX = LineIntegralFunctions.ctabXValues[0] * (-3.0 + 4.0*sigma.doubleValue()) + LineIntegralFunctions.ctabXValues[1] * (4.0 - 8.0*sigma.doubleValue()) + LineIntegralFunctions.ctabXValues[2] * (-1.0 + 4.0*sigma.doubleValue());
    	valueY = LineIntegralFunctions.ctabYValues[0] * (-3.0 + 4.0*sigma.doubleValue()) + LineIntegralFunctions.ctabYValues[1] * (4.0 - 8.0*sigma.doubleValue()) + LineIntegralFunctions.ctabYValues[2] * (-1.0 + 4.0*sigma.doubleValue());
    	 
    	return Math.sqrt(valueX*valueX + valueY*valueY);
    }
}
