/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.calc.objects;

import org.apache.log4j.Logger;

import fhm.edu.fem.calc.FEMManager;
import fhm.edu.fem.constants.ApplicationConstants;
import fhm.edu.fem.constants.SMatrixConstants;
import fhm.edu.fem.util.logging.LoggingManager;

/**
 * Class Description:
 *		This class acts as membrane circle in case of FEM5. It takes care of calculating the point
 *		coordinates out of the polar-coordinates and makes sure that the elements and borders are
 *		also created correctly.
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class MembraneCircle {
	
	/* Our local logger object */
	protected static Logger localLogger = LoggingManager.getLogger(MembraneCircle.class.getName());
	
    /* For Circle Membrane - The Center Coordinates */
    protected double membraneCircleCenterX = 0.0;
    protected double membraneCircleCenterY = 0.0;
    
    /* For Circle Membrane - The Radiuses */
    protected double membraneCircleRadius1 = 0.0;
    protected double membraneCircleRadius2 = 0.0;
    
    /* For Circle Membrane - The number of Sections */
    protected int membraneCircleSections = 0;
    
    /* Remember if we already made the calculations */
    protected boolean alreadyCalculated = false;
    
    /* The Center Node */
    protected Node centerNode = null;

    /**
     * Construct a new membrane circle from the data in the FEMManager
     * @param femManager
     */
    public MembraneCircle(FEMManager femManager) {
        
    	this.membraneCircleCenterX = femManager.getMembraneCircleCenterX();
    	this.membraneCircleCenterY = femManager.getMembraneCircleCenterY();
    	this.membraneCircleRadius1 = femManager.getMembraneCircleRadius1();
    	this.membraneCircleRadius2 = femManager.getMembraneCircleRadius2();
    	this.membraneCircleSections = femManager.getMembraneCircleSections();
    }
    
	/**
	 * @return Returns the membraneCircleCenterX.
	 */
	public double getCenterX() {
		return membraneCircleCenterX;
	}
	
	/**
	 * @return Returns the membraneCircleCenterY.
	 */
	public double getCenterY() {
		return membraneCircleCenterY;
	}
	
	/**
	 * @return Returns the membraneCircleRadius1.
	 */
	public double getRadius1() {
		return membraneCircleRadius1;
	}
	
	/**
	 * @return Returns the membraneCircleRadius2.
	 */
	public double getRadius2() {
		return membraneCircleRadius2;
	}
	
	/**
	 * @return Returns the membraneCircleSections.
	 */
	public int getSections() {
		return membraneCircleSections;
	}
	
	/**
	 * Get the angle based on the number of sections with the specific ID
	 * @param id, starting at 1, till number of sections
	 * @return
	 */
	public double getAngle(int id) throws IndexOutOfBoundsException {
		
	    // Take a look if the index is correct
	    if(id < 1 || id > this.membraneCircleSections) {

	        // Not a correct index, throw exception
	        throw new IndexOutOfBoundsException();
	    }
		
		// Calculate the new angle
		double angle = ((2*Math.PI) / this.membraneCircleSections) * (id-1);
		
		return angle;
	}
	
	/**
	 * Calculate the nodes based on the settings
	 */
	public void calculateAll() {
		
		// Check if we already made these calculation
		if(this.alreadyCalculated) {
			
			return;
		}
	
		// The new nodelist
		NodeList nodeList = new NodeList();
		
		// The  new element list
		ElementList sectionList = new ElementList();
		
		// The new border list
		BorderList borderList = new BorderList();
		NodeList borderNodeList = new NodeList();
		
		// The node ID
		int nodeId = 1;
		
		// The element ID
		int elementId = 1;
		
		// Create the center node and store it in the node list
		Node center = new Node(nodeId, this.membraneCircleCenterX, this.membraneCircleCenterY, false);
		nodeList.setNode(center);
		nodeId++;
		this.centerNode = center;
		
		Node newNode1 = null;
		Node newNode2 = null;
		
		// Run through all sections
		for(int i = 1; i <= this.membraneCircleSections; i++) {
			
			// Grab the angle of the current section
			double angle = this.getAngle(i);
			
			if(newNode1 == null && newNode2 == null) {
			
				double node1X = this.membraneCircleRadius1 * Math.cos(angle);
				double node1Y = this.membraneCircleRadius1 * Math.sin(angle);
				
				node1X = node1X + this.membraneCircleCenterX;
				node1Y = node1Y + this.membraneCircleCenterY;
				
				newNode1 = new Node(nodeId, node1X, node1Y, false);
				nodeId++;
				
				double node2X = this.membraneCircleRadius2 * Math.cos(angle);
				double node2Y = this.membraneCircleRadius2 * Math.sin(angle);
				
				node2X = node2X + this.membraneCircleCenterX;
				node2Y = node2Y + this.membraneCircleCenterY;
				
				newNode2 = new Node(nodeId, node2X, node2Y, false);
				nodeId++;
				
				// Store the new nodes in the list
				nodeList.setNode(newNode1);
				nodeList.setNode(newNode2);
				
				i++;
				
				//	Grab the angle of the current section
				angle = this.getAngle(i);
			}
			
			double node3X = this.membraneCircleRadius1 * Math.cos(angle);
			double node3Y = this.membraneCircleRadius1 * Math.sin(angle);
			
			node3X = node3X + this.membraneCircleCenterX;
			node3Y = node3Y + this.membraneCircleCenterY;
			
			Node newNode3 = new Node(nodeId, node3X, node3Y, false);
			nodeId++;
			
			double node4X = this.membraneCircleRadius2 * Math.cos(angle);
			double node4Y = this.membraneCircleRadius2 * Math.sin(angle);
			
			node4X = node4X + this.membraneCircleCenterX;
			node4Y = node4Y + this.membraneCircleCenterY;
			
			Node newNode4 = new Node(nodeId, node4X, node4Y, false);
			nodeId++;
			
			// Store the new nodes in the list
			nodeList.setNode(newNode3);
			nodeList.setNode(newNode4);
			
			// Store the four nodes in the small section node list
			NodeList sectionNodeList = new NodeList();
			
			sectionNodeList.setNode(newNode1);
			sectionNodeList.setNode(newNode2);
			sectionNodeList.setNode(newNode4);
			sectionNodeList.setNode(newNode3);
			
			// Store the border elements
			if(!borderNodeList.contains(newNode2)) {
				borderNodeList.setNode(newNode2);
			}
			if(!borderNodeList.contains(newNode4)) {
				borderNodeList.setNode(newNode4);
			}
			
			// Create the new Section (Element)
			Element newSection = new Element(elementId, -1, sectionNodeList);
			elementId++;
			sectionList.setElement(newSection);
			
			newNode1 = newNode3;
			newNode2 = newNode4;
		}
		
		// Handle the last element by hand
		Element firstElement = sectionList.getElement(1);
		NodeList firstNodeList = firstElement.getNodeList();
		Node newNode3 = firstNodeList.getNodeInCorrectOrder(1);
		Node newNode4 = firstNodeList.getNodeInCorrectOrder(2);
		
		NodeList sectionNodeList = new NodeList();
		
		// Add the real nodes which coordinates we really know
		sectionNodeList.setNode(newNode1);
		sectionNodeList.setNode(newNode2);
		sectionNodeList.setNode(newNode4);
		sectionNodeList.setNode(newNode3);
		
		// Create the element with these nodes and add it to the element list
		Element newSection = new Element(elementId, -1, sectionNodeList);
		sectionList.setElement(newSection);
		
		// Create the border and store it
		Border mainBorder = new Border(1, borderNodeList);
		borderList = new BorderList();
		borderList.setBorder(mainBorder);
		
		// Store all calculation results in the FEMManager
		FEMManager.getInstance().setBorderList(borderList);
		FEMManager.getInstance().setNodeList(nodeList);
		FEMManager.getInstance().setElementList(sectionList);
		FEMManager.getInstance().setCenterNode(center);
		
		// Calculate all PSEUDO nodes laying between two other nodes but not having any coordinates
		calculatePseudoNodesForQuadrats();
		
		// Calculate Triangles
		calculateTriangles();
		
		if(!ApplicationConstants.MEMBRAN_CIRCLE_OVERLAPPING_ELEMENTS) {
			calculatePseudoNodesForTriangles();
		}

		this.alreadyCalculated = true;
	}
	
	/**
	 * Calculate the Pseudo Nodes which lay between two other nodes.
	 * These are not the same as in the first exercises but are there for storing the results
	 * we get. They do not have any real coordinates.
	 */
	protected void calculatePseudoNodesForQuadrats() {
		
        // Grab the necessary data
        ElementList elementList = FEMManager.getInstance().getElementList();
        NodeList nodeList = FEMManager.getInstance().getNodeList();
        
        for(int i = 1; i <= elementList.size();i++){
            
            Element curElement = elementList.getElement(i);
            
            if(curElement.getElementType() == SMatrixConstants.FIGURE_QUADRAT) {
            	
                // Grab the node list size, can only be 4 for QUADS
                int nodeListSize =  curElement.getNodeList().size();
                
                for(int j = 1; j <= nodeListSize; j++){
                    
                    Node firstNode = curElement.getNodeList().getNodeInCorrectOrder(j);
                    Node secNode = null;
                    
                    if((j+1) > nodeListSize ) {
                        secNode = curElement.getNodeList().getNodeInCorrectOrder(1); 
                    } else {
                        secNode = curElement.getNodeList().getNodeInCorrectOrder(j+1);  
                    }                    

                    // Check if we already have calculated the pseudo node in some other element
                    Node foundNode = elementList.doesNodeExistBetween(firstNode, secNode);
                    
                    // Check if no node has been found
                    if(foundNode == null) {
                    	
                        // Calculate the new x/y values (these are only used for drawing purposes)
                        double x1 = firstNode.getX();
                        double y1 = firstNode.getY();
                        
                        double x2 = secNode.getX();
                        double y2 = secNode.getY();
                        
                        double newX = ((x1 - x2)/2) + x2;
                        double newY = ((y1 - y2)/2 ) + y2;
                        
                        // Create the new node
                        Node newNode = new Node(nodeList.size()+1, newX, newY, false);

                        // Store the new node in the local element list
                        curElement.getNodeList().setNode(newNode);
                        
                        // Store the new node in the global node list
                        nodeList.setNode(newNode);
                        
                    } else {

                        // Store the new found node in the local element list
                    	if(!curElement.getNodeList().contains(foundNode)) {
                    		curElement.getNodeList().setNode(foundNode);
                    	}
                    } 
                }
            }
        }   
        
        //
        // Do the same for the borders
        //
        
        // Grab the border data
        BorderList borderList = FEMManager.getInstance().getBorderList();
        
        // Run through all borders
        for(int i = 1; i <= borderList.size(); i++) {

            Border currentBorder = borderList.getBorder(i);
            
            // Grab the number of existing nodes in the border
            nodeList = currentBorder.getNodes();
            int nodeListSize = nodeList.size();
            
            // Run through all basic nodes in the border
            for(int j = 1; j <= nodeListSize; j++) {
     
                // Grab first and second current node
                Node firstNode = nodeList.getNodeInCorrectOrder(j);
           
                // Again run through all basic nodes in the border
                for(int o = j; o <= nodeListSize; o++) {
                    
                    Node secondNode = nodeList.getNodeInCorrectOrder(o);

                    // Get the node in between
                    Node midNode = elementList.doesNodeExistBetween(firstNode, secondNode);

                    if(midNode != null) {
                        
                        // Add the found middle node to the node list of the border
                    	if(!nodeList.contains(midNode)) {
                    		nodeList.setNode(midNode); 			
                    	}
                    }
                }
            }
        }
	}
	
	/**
	 * Calculate all triangles. These also occupy the region of the quadrats
	 */
	protected void calculateTriangles() {
		
		// Run through the element list which holds only quadrats at this time
		ElementList elementList = FEMManager.getInstance().getElementList();
		
		// Store the size here to make sure we are only running through the already existing quadrats
		int elementListSize = elementList.size();
		
		for(int elementIndex = 1; elementIndex <= elementListSize; elementIndex++) {
			
			Element currentElement = elementList.getElement(elementIndex);
			
			// Grab the nodelist and clone it
			NodeList elementNodeList = currentElement.getNodeList();
			
			// Transform the node list
			NodeList transformedNodeList = transformNodeListFromQuadratToTriangle(elementNodeList);
			
			// Create the new triangle element and add it to the element list
			Element newTriangle = new Element(elementList.size()+1, SMatrixConstants.FIGURE_TRIANGLE, transformedNodeList);
			elementList.setElement(newTriangle);
		}
		
		// Store the changed element list in the FEMManager
		FEMManager.getInstance().setElementList(elementList);
	}
	
    /**
     * Adjust the Order of the node list
     * @param nodeList, the node list to adjust (it is not changed)
     * @return the new adjusted node list
     */
    protected NodeList transformNodeListFromQuadratToTriangle(NodeList nodeList) {

    	// The full node list
    	NodeList fullNodeList = FEMManager.getInstance().getNodeList();
    	
    	// The new nodelist
    	NodeList newNodeList = new NodeList();
    	
    	if(ApplicationConstants.MEMBRAN_CIRCLE_OVERLAPPING_ELEMENTS) {
    		
	    	newNodeList.setNode(this.getCenterNode());
	    	newNodeList.setNode(nodeList.getNodeInCorrectOrder(2));
	    	newNodeList.setNode(nodeList.getNodeInCorrectOrder(3));
	    	newNodeList.setNode(nodeList.getNodeInCorrectOrder(1));
	    	newNodeList.setNode(nodeList.getNodeInCorrectOrder(6));
	    	newNodeList.setNode(nodeList.getNodeInCorrectOrder(4));
	    	
    	} else {
    		
        	// P1 is the center point
    		newNodeList.setNode(this.getCenterNode());
        	
        	// P1->P2
    		Node node = nodeList.getNodeInCorrectOrder(1);
        	newNodeList.setNode(node);
        	
        	// P4->P3
        	node = nodeList.getNodeInCorrectOrder(4);
        	newNodeList.setNode(node);
    	}

    	return newNodeList;
    }
    
	/**
	 * Calculate the Pseudo Nodes which lay between two other nodes.
	 * These are not the same as in the first exercises but are there for storing the results
	 * we get. They do not have any real coordinates.
	 */
	protected void calculatePseudoNodesForTriangles() {
		
        // Grab the necessary data
        ElementList elementList = FEMManager.getInstance().getElementList();
        NodeList nodeList = FEMManager.getInstance().getNodeList();
        
        for(int i = 1; i <= elementList.size();i++){
            
            Element curElement = elementList.getElement(i);
            
            if(curElement.getElementType() == SMatrixConstants.FIGURE_TRIANGLE) {
            	
                // Grab the node list size, can only be 4 for QUADS
                int nodeListSize =  curElement.getNodeList().size();
                
                for(int j = 1; j <= nodeListSize; j++){
                    
                    Node firstNode = curElement.getNodeList().getNodeInCorrectOrder(j);
                    Node secNode = null;
                    
                    if((j+1) > nodeListSize ) {
                        secNode = curElement.getNodeList().getNodeInCorrectOrder(1); 
                    } else {
                        secNode = curElement.getNodeList().getNodeInCorrectOrder(j+1);  
                    }                    

                    // Check if we already have calculated the pseudo node in some other element
                    Node foundNode = elementList.doesNodeExistBetween(firstNode, secNode);
                    
                    // Check if no node has been found
                    if(foundNode == null) {
                    	
                        // Calculate the new x/y values (these are only used for drawing purposes)
                        double x1 = firstNode.getX();
                        double y1 = firstNode.getY();
                        
                        double x2 = secNode.getX();
                        double y2 = secNode.getY();
                        
                        double newX = ((x1 - x2)/2) + x2;
                        double newY = ((y1 - y2)/2 ) + y2;
                        
                        // Create the new node
                        Node newNode = new Node(nodeList.size()+1, newX, newY, false);

                        // Store the new node in the local element list
                        curElement.getNodeList().setNode(newNode);
                        
                        // Store the new node in the global node list
                        nodeList.setNode(newNode);
                        
                    } else {

                        // Store the new found node in the local element list
                    	if(!curElement.getNodeList().contains(foundNode)) {
                    		curElement.getNodeList().setNode(foundNode);
                    	}
                    } 
                }
            }
        }   
	}
	
	/**
	 * @return Returns the centerNode.
	 */
	public Node getCenterNode() {
		return centerNode;
	}
}
