/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.calc.objects;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * Class Description:
 *		This class acts as node list. It takes the node index from the node itself
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class NodeList {

	/* The List object where we store our nodes */
	protected ArrayList<Node> nodes = new ArrayList<Node>();
	
	/**
	 * Create a new node list
	 */
	public NodeList() {}
	
	/**
	 * Create the Alias IDs for all nodes in this list
	 * @param int, the start alias ID
	 * @param borderNodeList, the list of all border nodes
	 * @return NodeList, list containing nodes with an alias ID of != 0 only
	 */
	public NodeList createAliasIDs(int aliasIDCounter, NodeList borderNodeList) {
		
		// The list of nodes we return (aliasID Nodes only)
		NodeList aliasIDNodeList = new NodeList();
		
		// Run through all nodes
		for(int i = 1; i <= this.size(); i++) {
			
			// Get the current node
			Node currentNode = this.getNode(i);
			
			// Check if the current node is not in the border node list
			if(!(borderNodeList.contains(currentNode))) {
				
				currentNode.setAliasId(aliasIDCounter);
				aliasIDCounter++;
				
				// Also set the Node into the new aliasIDList
				aliasIDNodeList.setNode(currentNode);
			}
		}
		
		return aliasIDNodeList;
	}
	
	/**
	 * Search for a specific node
	 * @param node, the node to search
	 * @return boolean, true if it contains the list, false otherwise
	 */
	public boolean contains(Node node) {
		return this.nodes.contains(node);
	}
	
	/**
	 * Add all nodes from the given list to this list
	 * @param list, the node list
	 */
	public void addAll(NodeList list) {

		// Run through the list
		for(int i = 1; i <= list.size(); i++) {
			
			// Grab the node
			Node currentNode = list.getNodeInCorrectOrder(i);
			
			// Check if the node is not already in the list
			if(!(this.contains(currentNode))) {
				
				// Add the node to the list
				this.setNode(currentNode);
			}
		}
	}
	
	/**
	 * @return Get the ArrayList
	 */
	public ArrayList<Node> getList() {	
		return this.nodes;
	}
	
	/**
	 * Set a new node in the list. It will be set at the index specified in the node
	 * @param node, the node to set
	 */
	public void setNode(Node node) {
		nodes.add(node);
	}
	
	/**
	 * Retrieve a node with the given index from the list
	 * @param nodeIndex
	 * @return the found node or 'null' if no node has been found
	 */
	public Node getNode(int nodeIndex) {
		
		// Run through the list and find the node with the given index
		Iterator it = nodes.iterator();
		Node tmpNode = null;
		
		while(it.hasNext()) {
			// Grab the next node from the list
			tmpNode = (Node)(it.next());
			
			// Check if it is the node we are searching for
			if(tmpNode != null) {
				
				if(nodeIndex == tmpNode.getId()) {
					
					// We found the correct node, return it
					return tmpNode;
				}
			}
		}
		
		// We found no proper node
		return null;
	}
	
	/**
	 * Set the center node at the beginning of the list
	 * @param node, the center node
	 */
	public void setCenterNodeAtBeginning(Node node) {
		
		this.nodes.add(0, node);
	}
	
	/**
	 * Retrieve the nodes in the correct order
	 * @param i, the index (not the node's index)
	 * @return the node
	 */
	public Node getNodeInCorrectOrder(int i) throws IndexOutOfBoundsException {
	    
	    // Take a look if the index is correct
	    if(i < 1 || i > this.nodes.size()) {
	        
	        // Not a correct index, throw exception
	        throw new IndexOutOfBoundsException();
	    }
	    
	    // Correct the index because we start with 1 not 0
	    return (Node)(this.nodes.get(i-1));
	}
	
	/**
	 * Clear the list
	 */
	public void clear() {
		this.nodes.clear();
	}
	
	/**
	 * Retrieve the x-value of a specific node from the list
	 * @param nodeIndex
	 * @return the x-value
	 */
	public double getX(int nodeIndex) {
		
		// Find the proper node
		Node tmpNode = this.getNode(nodeIndex);
		
		// Check if we found a node
		if(tmpNode == null) {
			
			return -1;
			
		}
		
		// We found a node, return the value
		return tmpNode.getX();
	}
	
	/**
	 * Retrieve the y-value of a specific node from the list
	 * @param nodeIndex
	 * @return the y-value
	 */
	public double getY(int nodeIndex) {
		
		// Find the proper node
		Node tmpNode = this.getNode(nodeIndex);
		
		// Check if we found a node
		if(tmpNode == null) {
			
			return -1;
			
		}
		
		// We found a node, return the value
		return tmpNode.getY();
	}
	
	/**
	 * Convert this list to an array of nodes
	 * @return the node array
	 */
	public Object[] toArray() {
		return this.nodes.toArray();
	}
	
	/**
	 * Retrieve the number of nodes in the list
	 * @return the number of nodes
	 */
	public int size() {
		return this.nodes.size();
	}
	
	/**
	 * Clone Implementation
	 * @return a clone of this list
	 */
	public Object clone() {
		// Create the new list
		NodeList cloneList = new NodeList();
		
		// Fill the new list with the same nodes as the current list
		Object[] nodes = this.toArray();
		
		for(int i = 0; i < nodes.length; i++) {
			cloneList.setNode((Node)(nodes[i]));
		}
		
		return cloneList;
	}
}
