/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.calc.objects;

import fhm.edu.fem.calc.objects.matrix.Matrix;

/**
 * Class Description:
 *		This class represents a vector. Do not confuse it with the Vector
 *		class of the Java API which is a thread-safe implementation of ArrayList.
 *		The index i starts at 1 not at 0.
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class Vector implements MathObject {

    // The list where we store all our vector data
    protected double[] data = null;
    
    // The size of the vector
    protected int size;
    
    // The name of the vector
    protected String name = "";
    
    /**
     * Construct a new vector
     * @param name, the name of the vector
     */
    public Vector(String name) {
        this(name, 1);
    }
    
    /**
     * Construct a new vector with the given initiale size
     * @param name, the name of the vector
     * @param size, the number of expected elements in the vector
     */
    public Vector(String name, int size) {
        this.name = name;
        this.size = size;
        this.data = new double[size];
    }
    
    /**
     * Construct a new vector with the given initial data
     * @param name, the name of the vector
     * @param double[], the data array
     */
    public Vector(String name, double[] array) {
        this.name = name;
        this.size = array.length;
        this.data = new double[size];
        
        // Fill the data array
        for(int i = 0; i < this.size(); i++) {
        	
        	this.data[i] = array[i];
        }
    }
    
    /**
     * Retrieve a value from the vector
     * @param index, the index of the value to retrieve (first index is 1 not 0)
     * @return the retrieved value
     * @throws IndexOutOfBoundsException
     */
    public double getValue(int index) throws IndexOutOfBoundsException {
        
        // Check if we got a proper index
        if(index <= 0 || index > this.size){
            throw new IndexOutOfBoundsException();
            
        } else {
            
            return (this.data[index-1]);
        }
    }
    
    /**
     * Replace an existing value
     * @param index, the index of the value (first index is 1 not 0)
     * @param value, the new value
     * @throws IndexOutOfBoundsException
     */
    public void setValue(int index, double value) throws IndexOutOfBoundsException {
        
        // Check if we got a proper index
        if(index <= 0 || index > this.size) {
            throw new IndexOutOfBoundsException();
            
        } else {
            
            // Replace the existing value
            this.data[index-1]= value;
        }
    }
    
    /**
     * Multiply all values in the vector with the given value and return the new resulting vector
     * @param value, the value to multiply the vector with
     * @return the new resulting vector
     */
    public Vector mult(double value) {
        
        Vector result = new Vector(this.name, this.size);
        
        for(int i = 0; i < this.size; i++) {

            double multValue = this.data[i] * value;
            result.setValue(i+1, multValue);
        }
        
        return result;
    }
    
    /**
     * Multiply this vector with the given matrix and return the new resulting vector
     * @param matrix, the matrix to multiply the vector with
     * @param boolean, true if the vector is transponated, false otherwise
     * @return the new resulting vector
     */
    public Vector mult(Matrix matrix, boolean transponate) throws IndexOutOfBoundsException {
    	
    	// Check if the dimensions match
    	if(this.size() != matrix.size()) {
    		
    		 throw new IndexOutOfBoundsException();
    	}
    	
    	Vector result = new Vector(this.name, this.size);
    	
    	for(int j = 1; j <= this.size(); j++) {
    		
    		double sum = 0;
    		
    		for(int i = 1; i <= this.size(); i++) {
    		
    			if(transponate) {
    				sum += this.getValue(i) * matrix.getValue(j, i);
    			} else {
    				sum += this.getValue(j) * matrix.getValue(j, i);
    			}
    			
    		}
    		
    		result.setValue(j, sum);
    	}
    	
    	return result;
    }
    
    /**
     * Multiply this vector with the given vector and return the new resulting vector
     * @param vector, the vector to multiply the vector with
     * @return the result number
     */
    public double mult(Vector vector) throws IndexOutOfBoundsException {
    	
    	// Check if the dimensions match
    	if(this.size() != vector.size()) {
    		
    		 throw new IndexOutOfBoundsException();
    	}
    	
    	double result = 0;
    	
    	for(int i = 1; i <= this.size(); i++) {
    		result += this.getValue(i) * vector.getValue(i);
    	}
    	
    	return result;
    }
    
    /**
     * Retrieve the current number of elements in the vector
     * @return
     */
    public int size() {
        return this.size;
    }
    
    /**
     * Retrieve the name of this matrix
     * @return the name
     */
    public String getName() {
        return this.name;
    }
    
    /**
     * Set a new name for this matrix
     * @param name
     */
    public void setName(String name) {
        this.name = name;
    }
    
    /**
     * toString Implementation
     * @return the string
     */
    public String toString() {
        
        String resultString = "[Vector - \nName: " + this.name + "\nSize: " + this.size + "\nData: \n";
        
        // Check if we have data in the vector
        if(this.size== 0) {
            
            resultString = resultString + "EMPTY";
            
        } else {
            
            // Run through the vector and display it's data together with the index
            for(int i = 0; i < this.size; i++) {
                
                resultString = resultString + "" + this.data[i] + '\n';
            }
        }
        
        return resultString + "]";
    }
    
    /**
     * toString Implementation
     * @param int, the number of values to display
     * @return the string
     */
    public String toString(int numberOfValues) {
        
        String resultString = "[Vector - \nName: " + this.name + "\nSize: " + this.size + "\nData: \n";
        
        // Check if we have data in the vector
        if(this.size == 0) {
            
            resultString = resultString + "EMPTY";
            
        } else {
            
            // Run through the vector and display it's data together with the index
            for(int i = 0; i < numberOfValues; i++) {
                
                resultString = resultString + "" + this.data[i] + '\n';
            }
        }
        
        return resultString + "... ]";
    }
    
    /**
     * Norm the vector to the given number (meaning: the max value is equal to the
     * given number, all other smaller values are adjusted based on the max value)
     * @param normValue, the max value to norm to
     * @return Vector, the new adjusted vector (the original one is not changed)
     */
    public Vector normVectorTo(double normValue) {
    	
    	// Search the max value in the vector
    	double maxValue = Double.MIN_VALUE;
    	for(int i = 1; i <= this.size(); i++) {
    		
    		double currentValue = this.getValue(i);
    		
    		if(maxValue < currentValue) {
    			maxValue = currentValue;
    		}
    	}
    	
    	// Let's calculate the multiplicator which we apply to the whole vector
    	double multiplicator = normValue / maxValue;
    	
    	// Apply the multiplicator
    	Vector normVec = this.mult(multiplicator);
    	
    	// Set a new name for this vector
    	normVec.setName(this.getName() + " normalized to " + normValue);

    	return normVec;
    }
    
    /**
     * Clone the vector
     * @return the cloned vector
     */
     public Object clone() {
         
         // Make a clone vector
         Vector clone = new Vector(this.name, this.size);
         
         // Run through all items in the vector
         for(int i = 1 ; i <= this.size(); i++)
         {
         	clone.setValue(i, this.getValue(i));
         }
         
         // Return the cloned vector
         return clone;
     }
}
