/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.calc.objects.algorithm;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import fhm.edu.fem.calc.objects.Vector;
import fhm.edu.fem.constants.ApplicationConstants;
import fhm.edu.fem.util.logging.LoggingManager;

/**
 * Class Description:
 *		This class offers methods to calculate Eigenvalues and Eigenvectors
 *		via the Jacobi Algorithm.
 *
 *		The algorithms used in this class are Copyright by Ewe Leonties
 *		taken from the C-Classes he wrote as diploma thesis.
 *
 *		The conversion to Java and the percentage displays have been done
 *		by the authors mentioned below.
 *
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class Jacobi {
	
	/* Our local logger object */
	protected static Logger localLogger = LoggingManager.getLogger(Jacobi.class.getName());
	
	/* The eigenvalue vector */
	protected Vector eigenvalueVector = null;
	
	/* The eigenvector array */
	protected double[][] V = null;
	
	/**
	 * Constructor
	 */
	public Jacobi() {}

	/**
	 * Calculates the Eigenvalues and Eigenvectors of the symmetric Matrix A
	 * After completing this calculation you can also grab all calculated Eigenvalues via
	 * the associated Getter-Method in this class
	 * @param n, size of Matrix A
	 * @param gen, machine precision
	 * @param A, double[][] array of Matrix A
	 * @param ord, if >0 the eigenvalues and associated eigenvectors are sorted from upwards, else downwards
	 * @return bool, true if all went well, false otherwise
	 */
	public boolean compute(int n, double gen, double[][] A, int ord) {
		
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Jacobi Algorithm - by Ewe Leonties");
        }

        // Calculate how many percentage one line makes up
        double oneLinePercentage = 100.0 / n;
        
        // Calculate how many lines make up one percent
        double onePercentLines = n / 100.0;
        
        // Where we store the current actual percentage
        double currentPercentage = 0.0;
        
        // Jacobi Step Counter
        int step = 0;
		
        /**
         * START OF ALGORITHM
         */
        
		int i, j, k, p, q, jj;
		
		double norma, sum, theta, t, s, c, h, eps;
		V = new double[n][n];
		
		norma = 0;
		for (i = 0; i < n; i++) {
			for (j = 0; j < i; j++) {
				V[i][j] = 0;
				V[j][i] = 0;
				if (norma < Math.abs(A[i][j]))
					norma = Math.abs(A[i][j]);
			}
			V[i][i] = 1;
		}

		eps = gen * n * norma;
		do {
			// Raise the step counter
			step++;
			
			for (p = 0; p < n - 1; p++) {
				
	            // Calculate the current percentage of line already done in the CHOLESKY
	            currentPercentage = (p) * oneLinePercentage;

	            // Make a percentage output about every 5% of completed lines
	            if(p%(onePercentLines*5.0) < 1) {
	                
	                if(localLogger.isEnabledFor(Level.INFO)) {
	                    localLogger.log(Level.INFO, "Jacobi - Step #" + step + " at " + ApplicationConstants.convertPercentageToString(currentPercentage) + " %");
	                }
	            }
				
				for (q = p + 1; q < n; q++) {
					if (Math.abs(A[q][p]) > eps) {
						theta = 0.5 * (A[q][q] - A[p][p]) / A[q][p];
						t = 1;
						if (Math.abs(theta) > gen) {
							t = 1 / (Math.abs(theta) + Math.sqrt(1 + theta * theta));
							if (theta < 0)
								t = -t;
						}
						c = 1 / Math.sqrt(1 + t * t);
						s = c * t;
						A[p][p] -= t * A[q][p];
						A[q][q] += t * A[q][p];
						A[q][p] = 0;
						for (i = 0; i < p; i++) {
							h = c * A[p][i] - s * A[q][i];
							A[q][i] = s * A[p][i] + c * A[q][i];
							A[p][i] = h;
						}
						for (i = p + 1; i < q; i++) {
							h = c * A[i][p] - s * A[q][i];
							A[q][i] = s * A[i][p] + c * A[q][i];
							A[i][p] = h;
						}
						for (i = q + 1; i < n; i++) {
							h = c * A[i][p] - s * A[i][q];
							A[i][q] = s * A[i][p] + c * A[i][q];
							A[i][p] = h;
						}
						for (i = 0; i < n; i++) {
							h = c * V[i][p] - s * V[i][q];
							V[i][q] = s * V[i][p] + c * V[i][q];
							V[i][p] = h;
						}
					}
				}
			}

			sum = 0;
			for (i = 1; i < n; i++) {
				for (j = 0; j < i; j++) {
					if (Math.abs(A[i][j]) <= eps)
						A[i][j] = 0;
					sum += A[i][j] * A[i][j];
				}
			}
        	// Jacobi Status
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Jacobi - Step complete - " + sum + " > " + (eps * eps / 2));
            }
		} while (sum > eps * eps / 2);

		for (i = 1; i < n; i++)
			for (j = n - 1; j >= i; j--) {
				jj = j - 1;
				if (ord > 0) {
					s = A[jj][jj];
					c = A[j][j];
				} else {
					s = A[j][j];
					c = A[jj][jj];
				}
				if (s > c) {
					h = A[jj][jj];
					A[jj][jj] = A[j][j];
					A[j][j] = h;
					for (k = 0; k < n; k++) {
						h = V[k][jj];
						V[k][jj] = V[k][j];
						V[k][j] = h;
					}
				}
			}
		
		/**
		 * END OF ALGORITHM
		 */
		
		// Save the eigenvalues into the vector
		eigenvalueVector = new Vector("Eigenvalues", A.length);
		
		for(int a = 0; a < A.length; a++) {
			eigenvalueVector.setValue(a+1, A[a][a]);
		}
		
		return true;
	}

	/**
	 * Grab the vector holding the calculated eigenvalues (you need to call compute()) first
	 * @return Vector, the vector holding all eigenvalues
	 */
	public Vector getEigenvalueVector() {
		return eigenvalueVector;
	}
	
	/**
	 * Grab the double array holding the calculated eigenvectors (you need to call compute()) first
	 * @return double[][], the array holding all eigenvectors
	 */
	public double[][] getEigenvectorArray() {
		return V;
	}
}
