/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.calc.objects.algorithm;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import fhm.edu.fem.calc.objects.Vector;
import fhm.edu.fem.util.logging.LoggingManager;

/**
 * Class Description:
 *		This class offers methods to calculate Eigenvalues and Eigenvectors
 *		via the Rayleigh Algorithm.
 *
 *		The algorithms used in this class are Copyright by Ewe Leonties
 *		taken from the C-Classes he wrote as diploma thesis.
 *
 *		The conversion to Java and the percentage displays have been done
 *		by the authors mentioned below.
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class Rayleigh {
	
	/* Our local logger object */
	protected static Logger localLogger = LoggingManager.getLogger(Rayleigh.class.getName());
	
	/* The eigenvalue vector */
	protected Vector eigenvalueVector = null;
	
	/* The eigenvector array */
	protected double[][] EV = null;
	
	/**
	 * Constructor
	 */
	public Rayleigh() {}

	/**
	 * Calculates the Eigenvalues and Eigenvectors via the preconditioned
	 * Rayleigh-Quotient-Minimization via konjugented Gradients of the general
	 * EWA A * x = lambda * B * x
	 * Only works with HULL MATRICES !
	 * After completing this calculation you can also grab all calculated Eigenvalues via
	 * the associated Getter-Method in this class
	 * @param n, size of the Matrix A and B as Hull
	 * @param A, the Matrix A in the Hull Matrix format
	 * @param B, the Matrix B in the Hull Matrix format
	 * @param D, the diagonal pointer vector (both A and B have the same)
	 * @param neig, the number of eigenvalues and eigenvectors to calculate
	 * @param tol, the calculation precision
	 * @param d, the "Spektralverschiebung"
	 * @return bool, true if all went well, false otherwise
	 */
	public boolean compute(int n, double[] A, double[] B, int[] D, int neig, double tol, double d) {
		
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Rayleigh Algorithm - by Ewe Leonties");
        }
        
        /**
         * START OF ALGORITHM
         */

		int neust = 10;
		int nzyk = 100; // Raised the number of cycles to make convergence possible (something is wrong with resetting this value after each pair)
		double[][] Y = new double[n][neig];
		double[][] BEV = new double[n][neig];
		double[] KAP = new double[n];
		double[] V = new double[n];
		double[] V1 = new double[n];
		double[] G = new double[n];
		double[] H = new double[n];
		double[] S = new double[n];
		double[] W = new double[n];
		double[] W1 = new double[n];
		double[] X = new double[n];
		double[] C = new double[D[n] + 1];
		double aux, gtg, qa, za, f, q, z;
		double alf, bet, gam, sig, tau, rho, phi;
		double a1, a2, a3, dk, eps, del, fakd;
		int i, j, k, l, flg, ini;
		int izykl = 0;

		EV = new double[n][neig];
		double[] LAM = new double[neig];

		for (i = 1; i <= D[n]; i++) {
			C[i] = A[i];
		}

		if (cholesky(n, C, D) == 0) {
			
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error while calculating the Cholesky !");
            }
			
			return false; // fehler(200,0);
		}
		ini = 1;
		fakd = 10;
		for (l = 0; l < neig; l++) {
			
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "Rayleigh - Calculating Eigenvalue #" + l + " of " + neig);
            }
			
			do // Initialisierung der
			// Iteration
			{ // fuer den l-ten EW
				if (ini != 0) {
					for (i = 0; i < n; i++)
						X[i] = randomx(i + l);
					izykl = 0;
					ini = 0;
				}
				za = 1; // Neustart
				abpz(n, A, B, D, X, V, V1);
				for (j = 0; j < l; j++) {
					phi = 0;
					for (i = 0; i < n; i++)
						phi += BEV[i][j] * X[i];
					phi = phi * (d - LAM[j]);
					for (i = 0; i < n; i++)
						V[i] += phi * BEV[i][j];
				}
				alf = rho = 0;
				for (i = 0; i < n; i++) {
					alf += X[i] * V[i];
					rho += X[i] * V1[i];
					S[i] = 0;
				}
				q = alf / rho;

				k = 0; // Iterationsschritte
				do // fuer den l-ten EW
				{
					flg = 0;
					gtg = 0;
					f = 2 / rho;
					for (i = 0; i < n; i++) {
						G[i] = (V[i] - q * V1[i]) * f;
						gtg += G[i] * G[i];
					}
					vrenv(n, C, D, G, H);

					for (j = 0; j < l; j++) // Nachfuehrung der
					{ // Vorkonditionierung
						phi = 0;
						for (i = 0; i < n; i++)
							phi += H[i] * BEV[i][j];
						phi = phi * (d - LAM[j]) / (1 + (d - LAM[j]) * KAP[j]);
						for (i = 0; i < n; i++)
							H[i] -= phi * Y[i][j];
					}

					z = 0; // Berechnung
					// des
					for (i = 0; i < n; i++)
						z += H[i] * G[i]; // iterierten
					// Vektors
					eps = z / za; // Konvergenztest
					for (i = 0; i < n; i++)
						S[i] = -H[i] + eps * S[i];
					abpz(n, A, B, D, S, W, W1);

					for (j = 0; j < l; j++) {
						phi = 0;
						for (i = 0; i < n; i++)
							phi += BEV[i][j] * S[i];
						phi = phi * (d - LAM[j]);
						for (i = 0; i < n; i++)
							W[i] += phi * BEV[i][j];
					}
					bet = gam = sig = tau = 0;
					for (i = 0; i < n; i++) {
						bet += X[i] * W[i];
						gam += S[i] * W[i];
						sig += X[i] * W1[i];
						tau += S[i] * W1[i];
					}
					a1 = alf * tau - gam * rho;
					a2 = gam * sig - bet * tau;
					a3 = bet * rho - alf * sig;
					del = a1 * a1 - 4 * a2 * a3;
					if (a1 >= 0)
						dk = (a1 + Math.sqrt(del)) / (2 * a2);
					else
						dk = (2 * a3) / (a1 - Math.sqrt(del));

					alf = rho = 0;
					for (i = 0; i < n; i++) {
						X[i] += dk * S[i];
						V[i] += dk * W[i];
						V1[i] += dk * W1[i];
						alf += X[i] * V[i];
						rho += X[i] * V1[i];
					}
					qa = q;
					za = z;
					q = alf / rho;
					z = Math.abs(1 - qa / q);
					//System.out.println("k:" + k + " RQ" + l + ":" + q + " eps: " + z);
					if ((q > qa) || (z < tol))
						flg = 1;
					k++;
				} while ((k < neust) && (flg == 0));

				izykl++;

				if (izykl > nzyk) // return(500);
				{
					
		            if(localLogger.isEnabledFor(Level.ERROR)) {
		                localLogger.log(Level.ERROR, "Error in Jacobi - No Convergence at Eigenvalue #" + l + " !");
		            }
					return false;
				}
				if (k > neust) {
					//System.out.println("Neustart");
					flg = 0;
				}

				if ((l > 0) && (q > (0.98 * d))) {
					d = d + d;
					//System.out.println("neue Spv: " + d);
					ini = 1;
					flg = 0;
				}
				
	            if(localLogger.isEnabledFor(Level.INFO)) {
	                localLogger.log(Level.INFO, "Rayleigh - Eigenvalue #" + l + " - Step complete");
	            }
				
			} while ((flg == 0) && (izykl < nzyk));

			LAM[l] = q;
			if (d < (fakd * LAM[l])) {
				d = fakd * LAM[l];
				fakd = Math.sqrt(fakd) + 0.3;
				ini = 1;
				//System.out.println("neue Spv: " + d + " und neuer Fakd: " + fakd);
			}

			z = Math.sqrt(rho); // Normierung des EV
			for (i = 0; i < n; i++) // Berechnung von Y
			{
				EV[i][l] = X[i] / z;
				BEV[i][l] = V1[i] / z;
				G[i] = BEV[i][l];
			}
			if (l < neig - 1) {
				vrenv(n, C, D, G, H);
				phi = 0;
				for (i = 0; i < n; i++) {
					Y[i][l] = H[i];
					phi += H[i] * BEV[i][l];
				}
				KAP[l] = phi;
			}
		}    	
		
		/**
		 * END OF ALGORITHM
		 */
		   
		// Save the eigenvalues into the vector
		eigenvalueVector = new Vector("Eigenvalues", LAM.length);
		
		for(int a = 0; a < LAM.length; a++) {
			eigenvalueVector.setValue(a+1, LAM[a]);
		}
	   
	   
	   return true;
	}

	
	/**
	 * Grab the vector holding the calculated eigenvalues (you need to call
	 * compute()) first
	 * 
	 * @return Vector, the vector holding all eigenvalues
	 */
	public Vector getEigenvalueVector() {
		return eigenvalueVector;
	}
	
	/**
	 * Grab the double array holding the calculated eigenvectors (you need to
	 * call compute()) first
	 * 
	 * @return double[][], the array holding all eigenvectors
	 */
	public double[][] getEigenvectorArray() {
		return EV;
	}
	
	/**
	 * Calculate the Cholesky (by Ewe Leonties)
	 * A is changed in the progress containing the result
	 * @param n, size of Matrix A
	 * @param A, Matrix A in HullMatrix style
	 * @param D, diagonal pointer vector
	 * @return int, 0 = ERROR, 1 = SUCCESS
	 */
	protected int cholesky(int n, double[] A, int[] D)
	{
		double s;
		int i,j,m,k,io,jo,fi,fj;

		for(i=1;i<=n;i++)
		{
			io=D[i]-i;
			fi=D[i-1]+1-io;
			for(j=fi;j<=i;j++)
			{
					jo=D[j]-j;
					fj=D[j-1]-jo+1;
					s=A[io+j];
					if(fi>fj) m=fi;
					else m=fj;
		
					for(k=m;k<j;k++) s-=A[io+k]*A[jo+k];
						if(j<i) A[io+j]=s/A[D[j]];
						else
						{
							if(s<0) return(0);
							A[D[i]]=Math.sqrt(s);
						}
			}
		}
		return(1);
	}
	
	/**
	 * Generated pseudo random numbers (0 <= x <= 1) (by Ewe Leonties)
	 * @param ini, initialisation value
	 * @return double, a random number
	 */
	protected double randomx(int ini)
	{
		int i,h;
		h=11;
		for(i=0;i<=ini;i++) h=(h*25173+13849)%65563;
		return((double)h/65563);
	}
	
	/**
	 * Calculates the products A*p = z2 and B*p = z1 (by Ewe Leonties)
	 * @param n, size of matrix A and B
	 * @param A, Matrix A in HullMatrix style
	 * @param B, Matrix B in HullMatrix style
	 * @param D, the diagonal pointer vector
	 * @param P, Constantsvector
	 * @param Z2, RETURN VALUE
	 * @param Z1, RETURN VALUE
	 */
	protected void abpz(int n, double[] A, double[] B, int[] D, double[] P, double[] Z2, double[] Z1)
	{
	   int i,j,jlo,jup,ih;
	
	   for(i=0;i<n;i++) Z2[i]=Z1[i]=0;
	   for(i=0;i<n;i++)
	      {
	      jlo=D[i]+1;
	      jup=D[i+1];
	      Z2[i]+=A[jup]*P[i];
	      Z1[i]+=B[jup]*P[i];
	      for(j=jlo;j<jup;j++)
	         {
	         ih=j-jup+i;
	         Z2[ih]+=A[j]*P[i];
	         Z1[ih]+=B[j]*P[i];
	         Z2[i]+=A[j]*P[ih];
	         Z1[i]+=B[j]*P[ih];
	         }
	      }
	}
	
	/**
	 * Calculate L*LT*x=b via "Vorwaerts- und Rueckwaertseinsetzen" (by Ewe Leonties)
	 * @param n, size of Matrix L
	 * @param L, Matrix L in HullMatrix style
	 * @param D, the diagonal pointer vector
	 * @param B, Constantsvector
	 * @param X, RETURN VALUE
	 */
	protected void vrenv(int n,double[] L,int[] D,double[] B,double[] X)
	{
		int i,j,io,fi;
		double s;
	
		for(i=0;i<n;i++)
		{
			s=B[i];
			io=D[i+1]-i;
			fi=D[i]+1-io;
			for(j=fi;j<i;j++) s-=L[io+j]*X[j];
			X[i]=s/L[D[i+1]];
		}
		for(i=n-1;i>=0;i--)
		{
			X[i]=X[i]/L[D[i+1]];
			io=D[i+1]-i;
			fi=D[i]+1-io;
			for(j=fi;j<i;j++) X[j]-=L[io+j]*X[i];
		}
	}
}
