/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.calc.objects.algorithm;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import fhm.edu.fem.calc.objects.Vector;
import fhm.edu.fem.util.logging.LoggingManager;

/**
 * Class Description:
 *		This class offers methods to calculate Eigenvalues and Eigenvectors
 *		via the Rayleigh-S Algorithm.
 *
 *		The algorithms used in this class are Copyright by Ewe Leonties
 *		taken from the C-Classes he wrote as diploma thesis.
 *
 *		The conversion to Java and the percentage displays have been done
 *		by the authors mentioned below.
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class RayleighS {

	/* Our local logger object */
	protected static Logger localLogger = LoggingManager.getLogger(RayleighS.class.getName());

	/* The eigenvalue vector */
	protected Vector eigenvalueVector = null;
	
	/* The eigenvector array */
	protected double[][] X = null;

	/**
	 * Constructor
	 */
	public RayleighS() {}
	
	/**
	 * Preconditioned simultane Rayleigh-Quotient-Minimization via the
	 * konjugated gradients of the general EWA A*x=lambda*B*x
	 * @param n, the size of the matrix A and B
	 * @param A, matrix B in HullMatrix style
	 * @param B, matrix B in HullMatrix style
	 * @param D, the diagonal pointer vector (same for both A and B)
	 * @param spv, the "Spektralverschiebung"
	 * @param tol, the calculation precision
	 * @param maxit, the maximum number of steps
	 * @return bool, true if all went well, false otherwise
	 */
	public boolean compute(int n, double[] A, double[] B, int[] D, int p, double spv, double tol, int maxit) {

        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "Rayleigh-S Algorithm - by Ewe Leonties");
        }
        
        /**
         * START OF ALGORITHM
         */
		
		double gen = 5e-16;
		double neu = 1e1; // Neustartgrenze
		double tolj = 5e-15; // Rechengenauigkeit jacobi
		int zk, zl, ze, mon;
		int i, j, k, lh, ih, jlo, jup;
		double mra, alp, epi, bet, sig, gam, tau, rho, del, h, z, s;
		double a1, a2, a3;

		int[] P = new int[p];
		double[] M = new double[D[n] + 1];
		double[] G = new double[n];
		double[] H = new double[n];
		double[] WA = new double[n];
		double[] WB = new double[n];
		double[] NY = new double[p];
		double[][] U = new double[p][p];
		double[][] AS = new double[p][p];
		double[][] VA = new double[n][p];
		double[][] VB = new double[n][p];
		double[][] S = new double[n][p];

		X = new double[n][p];
		double[] Q = new double[p];

		ze = 0;
		if (spv != 0) {
			for (i = 1; i <= D[n]; i++) {
				A[i] += spv * B[i];
			}
		}
		for (i = 1; i <= D[n]; i++) {
			M[i] = A[i];
		}
		if (cholesky(n, M, D) == 0) {
			
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error while calculating the Cholesky !");
            }
			
			return false; // fehler(200,0);
		}
		initx(p, n, X);
		abxvavb(ze, p, n, A, B, D, X, VA, VB);
		borthonormx(p, n, X, VA, VB);
		sqny(ze, p, n, VA, X, S, NY, Q);

		for (i = 0; i < p; i++) {
			P[i] = 1;
		}
		zk = 0;
		mra = 1;
		while ((zk < maxit) && (mra > tol)) // Neustart
		{
			if (zk > 0) {
				aschlange(ze, p, n, A, D, X, S, AS);
				jacobi(p - ze, gen, AS, U, 1);
				ritzmat(ze, p, n, X, U);
				abxvavb(ze, p, n, A, B, D, X, VA, VB);
				sqny(ze, p, n, VA, X, S, NY, Q);
			}
			zl = 0;
			mon = 1;
			while ((zl < neu) && (mon == 1) && (mra > tol) && (zk < maxit)) {
				zl++;
				zk++;
				mra = 0;
				for (k = 0; k < p; k++) {
					if (P[k] != 0) {
						alp = Q[k];
						for (i = 0; i < n; i++) {
							G[i] = 2 * (VA[i][k] - alp * VB[i][k]);
						}
						vrenv(n, M, D, G, H); // Vorkonditionierung
						bog(k, n, H, X, VB);
						z = 0;
						for (i = 0; i < n; i++) {
							z += G[i] * H[i];
						}
						epi = z / NY[k];
						NY[k] = z;
						for (i = 0; i < n; i++) {
							S[i][k] = epi * S[i][k] - H[i];
						}

						for (i = 0; i < n; i++) {
							WA[i] = WB[i] = 0;
						}
						for (i = 0; i < n; i++) {
							jlo = D[i] + 1;
							jup = D[i + 1];
							WA[i] += A[jup] * S[i][k];
							WB[i] += B[jup] * S[i][k];
							for (j = jlo; j < jup; j++) {
								ih = j - jup + i;
								WA[ih] += A[j] * S[i][k];
								WB[ih] += B[j] * S[i][k];
								WA[i] += A[j] * S[ih][k];
								WB[i] += B[j] * S[ih][k];
							}
						}

						rho = 1;
						bet = gam = sig = tau = 0;
						for (i = 0; i < n; i++) {
							bet += X[i][k] * WA[i];
							gam += S[i][k] * WA[i];
							sig += X[i][k] * WB[i];
							tau += S[i][k] * WB[i];
						}

						a1 = alp * tau - gam * rho; // Berechnung von delta
						a2 = gam * sig - bet * tau;
						a3 = bet * rho - alp * sig;
						del = a1 * a1 - 4 * a2 * a3;
						if (a1 >= 0) {
							del = (a1 + Math.sqrt(del)) / (2 * a2);
						} else {
							del = (2 * a3) / (a1 - Math.sqrt(del));
						}
						for (i = 0; i < n; i++) // Iterationsformel
						{
							X[i][k] += del * S[i][k];
							VA[i][k] += del * WA[i];
							VB[i][k] += del * WB[i];
						}

						for (j = 0; j < k; j++) // B-Orthonormierung von x
						{ // Nachfuehrung von va und
							// vb
							s = 0;
							for (i = 0; i < n; i++) {
								s += VB[i][k] * X[i][j];
							}
							for (i = 0; i < n; i++) {
								X[i][k] -= s * X[i][j];
								VA[i][k] -= s * VA[i][j];
								VB[i][k] -= s * VB[i][j];
							}
						}
						s = 0;
						for (i = 0; i < n; i++) {
							s += X[i][k] * VB[i][k];
						}
						s = Math.sqrt(s);
						for (i = 0; i < n; i++) {
							X[i][k] /= s;
							VB[i][k] /= s;
							VA[i][k] /= s;
						}

						Q[k] = 0; // neuer RQ
						for (i = 0; i < n; i++) {
							Q[k] += X[i][k] * VA[i][k];
						}

						if ((k > 0) && (P[k - 1] != 0)) {
							if (Q[k - 1] > Q[k]) {
								mon = 0;
							}
						}
						h = Math.abs((Q[k] - alp) / Q[k]);
						if (h > mra) {
							mra = h;
						}
						if (h < tol) {
							P[k] = 0;
							ze = k + 1;
							
					        if(localLogger.isEnabledFor(Level.INFO)) {
					            localLogger.log(Level.INFO, "Rayleigh-S - Completed Eigenvalue-Pair #" + k);
					        }
						}
					}
				}
			}
		}
		maxit = zk;

		if (spv != 0) {
			for (i = 0; i < p; i++) {
				Q[i] -= spv; // Spektralverschiebung
			}
		}
		
        /**
         * END OF ALGORITHM
         */
		
		// Save the eigenvalues into the vector
		eigenvalueVector = new Vector("Eigenvalues", Q.length);
		
		for(int a = 0; a < Q.length; a++) {
			eigenvalueVector.setValue(a+1, Q[a]);
		}

		return true;
	}

	/**
	 * Grab the vector holding the calculated eigenvalues (you need to call
	 * compute()) first
	 * 
	 * @return Vector, the vector holding all eigenvalues
	 */
	public Vector getEigenvalueVector() {
		return eigenvalueVector;
	}
	
	/**
	 * Grab the double array holding the calculated eigenvectors (you need to
	 * call compute()) first
	 * 
	 * @return double[][], the array holding all eigenvectors
	 */
	public double[][] getEigenvectorArray() {
		return X;
	}

	/**
	 * Calculate the Cholesky (by Ewe Leonties)
	 * A is changed in the progress containing the result
	 * @param n, size of Matrix A
	 * @param A, Matrix A in HullMatrix style
	 * @param D, diagonal pointer vector
	 * @return int, 0 = ERROR, 1 = SUCCESS
	 */
	protected int cholesky(int n, double[] A, int[] D)
	{
		double s;
		int i, j, m, k, io, jo, fi, fj;

		for (i = 1; i <= n; i++) {
			io = D[i] - i;
			fi = D[i - 1] + 1 - io;
			for (j = fi; j <= i; j++) {
				jo = D[j] - j;
				fj = D[j - 1] - jo + 1;
				s = A[io + j];
				if (fi > fj)
					m = fi;
				else
					m = fj;

				for (k = m; k < j; k++)
					s -= A[io + k] * A[jo + k];
				if (j < i)
					A[io + j] = s / A[D[j]];
				else {
					if (s < 0)
						return (0);
					A[D[i]] = Math.sqrt(s);
				}
			}
		}
		return (1);
	}

	/**
	 * Init the Matrix X with random numbers 0 <= xij <= 1
	 * @param p, the dimension of X
	 * @param n, the dimension of X
	 * @param X, RETURN VALUE, linear independent start vectors
	 */
	protected void initx(int p, int n, double[][] X)
	{
		int i, j, h;

		h = 11;
		for (i = 0; i < p; i++)
			for (j = 0; j < n; j++) {
				h = (h * 25173 + 13849) % 65563;
				X[j][i] = (double) h / 65563;
			}
	}

	/**
	 * Builds the matrix product A*X=VA and B*X=VB
	 * @param pof, offset in matrix X
	 * @param p, columns in matrix X
	 * @param n, size of matrix A and B
	 * @param A, matrix A in HullMatrix style
	 * @param B, matrix B in HullMatrix style
	 * @param D, the diagonal pointer vector
	 * @param X, iteration matrix
	 * @param VA, RETURN VALUE
	 * @param VB, RETURN VALUE
	 */
	protected void abxvavb(int pof, int p, int n, double[] A, double[] B, int[] D, double[][] X, double[][] VA, double[][] VB)
	{
		int i, j, k, l, jh, klo, kup;

		for (i = pof; i < p; i++) {
			for (j = 0; j < n; j++)
				VA[j][i] = VB[j][i] = 0;
			for (j = 0; j < n; j++) {
				klo = D[j] + 1;
				kup = D[j + 1];
				VA[j][i] += A[kup] * X[j][i];
				VB[j][i] += B[kup] * X[j][i];
				for (k = klo; k < kup; k++) {
					jh = k - kup + j;
					VA[jh][i] += A[k] * X[j][i];
					VB[jh][i] += B[k] * X[j][i];
					VA[j][i] += A[k] * X[jh][i];
					VB[j][i] += B[k] * X[jh][i];
				}
			}
		}
	}

	/**
	 * b-orthonomizes the matrix X and leads the matrices VA and VB to
	 * @param p, dimension of matrices X, VA and VB
	 * @param n, dimension of matrices X, VA and VB
	 * @param X, RETURN VALUE
	 * @param VA, RETURN VALUE
	 * @param VB, RETURN VALUE
	 */
	protected void borthonormx(int p, int n, double[][] X, double[][] VA, double[][] VB)
	{
		int i, j, k;
		double s;

		for (k = 0; k < p; k++) {
			for (j = 0; j < k; j++) {
				s = 0;
				for (i = 0; i < n; i++)
					s += VB[i][k] * X[i][j];
				for (i = 0; i < n; i++) {
					X[i][k] -= s * X[i][j];
					VA[i][k] -= s * VA[i][j];
					VB[i][k] -= s * VB[i][j];
				}
			}
			s = 0;
			for (i = 0; i < n; i++)
				s += X[i][k] * VB[i][k];
			s = Math.sqrt(s);
			for (i = 0; i < n; i++) {
				X[i][k] /= s;
				VB[i][k] /= s;
				VA[i][k] /= s;
			}
		}
	}

	/**
	 * Reserves rayleigh's quotients and helpvector
	 * Initializes matrix in downward-direction
	 * @param pof, offset in matrix X and S
	 * @param p, dimension of matrices VA, X and S
	 * @param n, dimension of matrices VA, X and S
	 * @param VA
	 * @param X
	 * @param S, RETURN VALUE
	 * @param NY, RETURN VALUE
	 * @param Q, RETURN VALUE, Quotientvector
	 */
	protected void sqny(int pof, int p, int n, double[][] VA, double[][] X, double[][] S, double[] NY, double[] Q)
	{
		int i, j;

		for (i = pof; i < p; i++) {
			for (j = 0; j < n; j++)
				S[j][i] = 0;
			Q[i] = 0;
			for (j = 0; j < n; j++)
				Q[i] += X[j][i] * VA[j][i];
			NY[i] = 1;
		}
	}

	/**
	 * Calculates the projection matrix AS = XT*A*X
	 * @param pof, offset in matrix X
	 * @param p, amount of columns in X
	 * @param n, size of matrix A
	 * @param A, matrix A in HullMatrix style
	 * @param D, diagonal pointer vector
	 * @param X, iterations vector matrix
	 * @param S, helper matrix
	 * @param AS, RETURN VALUE
	 */
	protected void aschlange(int pof, int p, int n, double[] A, int[] D, double[][] X, double[][] S, double[][] AS)
	{
		int i, j, l, ih, jh, llo, lup;

		p -= pof;
		for (i = 0; i < n; i++)
			for (j = 0; j < p; j++)
				S[i][j] = 0;

		for (i = 0; i < n; i++) // Berechnung A * X
		{
			llo = D[i] + 1;
			lup = D[i + 1];
			for (j = 0; j < p; j++) {
				jh = j + pof;
				S[i][j] += A[lup] * X[i][jh];
				for (l = llo; l < lup; l++) {
					ih = l - lup + i;
					S[ih][j] += A[l] * X[i][jh];
					S[i][j] += A[l] * X[ih][jh];
				}
			}
		}

		for (i = 0; i < p; i++)
			// Berechnung XT * S
			for (j = 0; j < p; j++) {
				AS[i][j] = 0;
				for (l = 0; l < n; l++)
					AS[i][j] += X[l][i + pof] * S[l][j];
			}
	}

	/**
	 * Calculates the Eigenvalues and Eigenvectors of the symmetric Matrix A
	 * @param n, size of Matrix A
	 * @param gen, machine precision
	 * @param A, double[][] array of Matrix A
	 * @param V, RETURN VALUE, matrix of eigenvectors
	 * @param ord, if >0 the eigenvalues and associated eigenvectors are sorted from upwards, else downwards
	 */
	protected void jacobi(int n, double gen, double[][] A, double[][] V, int ord)
	{
		int i, j, jj, p, q, k;
		double norma, sum, theta, t, s, c, h, eps;

		norma = 0;
		for (i = 0; i < n; i++) {
			for (j = 0; j < i; j++) {
				V[i][j] = 0;
				V[j][i] = 0;
				if (norma < Math.abs(A[i][j]))
					norma = Math.abs(A[i][j]);
			}
			V[i][i] = 1;
		}

		eps = gen * n * norma;
		do {
			for (p = 0; p < n - 1; p++) {
				for (q = p + 1; q < n; q++) {
					if (Math.abs(A[q][p]) > eps) {
						theta = 0.5 * (A[q][q] - A[p][p]) / A[q][p];
						t = 1;
						if (Math.abs(theta) > gen) {
							t = 1 / (Math.abs(theta) + Math.sqrt(1 + theta
									* theta));
							if (theta < 0)
								t = -t;
						}
						c = 1 / Math.sqrt(1 + t * t);
						s = c * t;
						A[p][p] -= t * A[q][p];
						A[q][q] += t * A[q][p];
						A[q][p] = 0;
						for (i = 0; i < p; i++) {
							h = c * A[p][i] - s * A[q][i];
							A[q][i] = s * A[p][i] + c * A[q][i];
							A[p][i] = h;
						}
						for (i = p + 1; i < q; i++) {
							h = c * A[i][p] - s * A[q][i];
							A[q][i] = s * A[i][p] + c * A[q][i];
							A[i][p] = h;
						}
						for (i = q + 1; i < n; i++) {
							h = c * A[i][p] - s * A[i][q];
							A[i][q] = s * A[i][p] + c * A[i][q];
							A[i][p] = h;
						}
						for (i = 0; i < n; i++) {
							h = c * V[i][p] - s * V[i][q];
							V[i][q] = s * V[i][p] + c * V[i][q];
							V[i][p] = h;
						}
					}
				}
			}

			sum = 0;
			for (i = 1; i < n; i++) {
				for (j = 0; j < i; j++) {
					if (Math.abs(A[i][j]) <= eps)
						A[i][j] = 0;
					sum += A[i][j] * A[i][j];
				}
			}
		} while (sum > eps * eps / 2);

		for (i = 1; i < n; i++)
			for (j = n - 1; j >= i; j--) {
				jj = j - 1;
				if (ord > 0) {
					s = A[jj][jj];
					c = A[j][j];
				} else {
					s = A[j][j];
					c = A[jj][jj];
				}
				if (s > c) {
					h = A[jj][jj];
					A[jj][jj] = A[j][j];
					A[j][j] = h;
					for (k = 0; k < n; k++) {
						h = V[k][jj];
						V[k][jj] = V[k][j];
						V[k][j] = h;
					}
				}
			}
	}

	/**
	 * Calculates the matrix product Y = X*U
	 * @param pof, offset in matrix x
	 * @param p, size of matrix U
	 * @param n, amount of rows in X
	 * @param X, RETURN VALUE, Y instead of X
	 * @param U, the matrix U
	 */
	protected void ritzmat(int pof, int p, int n, double[][] X, double[][] U)
	{
		int i, l, j;
		double h;

		p -= pof;
		for (i = 0; i < n; i++)
			for (j = 0; j < p; j++) {
				h = 0;
				for (l = 0; l < p; l++)
					h += X[i][l + pof] * U[l][j];
				X[i][j + pof] = h;
			}
	}

	/**
	 * Solves L * LT * x = b via "Vorwaerts- und Rueckwaertseinsetzen"
	 * @param n, size of matrix L
	 * @param L, matrix L in HullMatrix style
	 * @param D, diagonal pointer vector
	 * @param B, constant vector
	 * @param X, RETUrN VALUE
	 */
	protected void vrenv(int n, double[] L, int[] D, double[] B, double[] X)
	{
		int i, j, io, fi;
		double s;

		for (i = 0; i < n; i++) {
			s = B[i];
			io = D[i + 1] - i;
			fi = D[i] + 1 - io;
			for (j = fi; j < i; j++)
				s -= L[io + j] * X[j];
			X[i] = s / L[D[i + 1]];
		}
		for (i = n - 1; i >= 0; i--) {
			X[i] = X[i] / L[D[i + 1]];
			io = D[i + 1] - i;
			fi = D[i] + 1 - io;
			for (j = fi; j < i; j++)
				X[j] -= L[io + j] * X[i];
		}
	}

	/**
	 * b-orthogonalization of h in relation to the k-first vectors in matrix X
	 * @param k, column index
	 * @param n, dimension of vector H
	 * @param H, vector (is changed in the progress and used as RETURN VALUE)
	 * @param X, matrix of iterationvectors
	 * @param VB, Matrix of the product VB = B * X
	 */
	protected void bog(int k, int n, double[] H, double[][] X, double[][] VB)
	{
		int i, j;
		double s;

		for (j = 0; j < k; j++) {
			s = 0;
			for (i = 0; i < n; i++)
				s += H[i] * VB[i][j];
			for (i = 0; i < n; i++)
				H[i] -= s * X[i][j];
		}
	}
}
