/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.calc.objects.algorithm;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import fhm.edu.fem.calc.objects.Vector;
import fhm.edu.fem.util.logging.LoggingManager;

/**
 * Class Description:
 *		This class offers methods to calculate Eigenvalues and Eigenvectors
 *		via the SVI1 Algorithm.
 *
 *		The algorithms used in this class are Copyright by Ewe Leonties
 *		taken from the C-Classes he wrote as diploma thesis.
 *
 *		The conversion to Java and the percentage displays have been done
 *		by the authors mentioned below.
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class SVI1 {

	/* Our local logger object */
	protected static Logger localLogger = LoggingManager.getLogger(SVI1.class.getName());

	/* The eigenvalue vector */
	protected Vector eigenvalueVector = null;
	
	/* The eigenvector array */
	protected double[][] X = null;

	/**
	 * Constructor
	 */
	public SVI1() {}

	/**
	 * Simultane vector-iteration to solve A*x=lam*B*x
	 * @param n, dimension of matrix
	 * @param LA, Matrix A in HullMatrix style
	 * @param B, Matrix B in HullMatrix style
	 * @param D, diagonal pointer vector
	 * @param p1, needed amount of eigenvectors
	 * @param p2, simultane iterated eigenvectors
	 * @param tol, calculation precision
	 * @param maxit, maximum number of steps
	 * @return boolean, true if all went well false otherwise
	 */
	public boolean compute(int n, double[] LA, double[] B, int[] D, int p1, int p2, double tol, int maxit) {
		
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "SVI-1 Algorithm - by Ewe Leonties");
        }
        
        /**
         * START OF ALGORITHM
         */
		
		double gen = 5e-16;

		double[][] h;
		int i, j, q, it;
		double qn, qa, del;

		double[][] Y = new double[n][p2];
		double[][] DK = new double[p2][p2];
		double[][] GK = new double[p2][p2];
		double[][] AS = new double[p2][p2];

		X = new double[n][p1];
		double[] Q = new double[p1];

		if (cholesky(n, LA, D) == 0) {
			
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Error while calculating the Cholesky !");
            }
			
			return false; // fehler(200,0);
		}

		initx(p2, n, X);
		q = p1 - 1;
		qa = 0;
		it = 0;
		while (it < maxit) {
			bxvb(p2, n, B, D, X, Y);
			qn = 0;
			for (i = 0; i < n; i++)
				qn += Y[i][q] * X[i][q];
			del = Math.abs((qa - qn) / qn);
			qa = qn;
			if (del > tol) // Iterationsschritt
			{
	            if(localLogger.isEnabledFor(Level.INFO)) {
	                localLogger.log(Level.INFO, "SVI-1 - Starting step - (q/eps) " + 1/qn + " / " + del);
	            }
				
				mvrenv(n, p2, LA, D, Y, X);
				asch(p2, n, X, Y, AS);
				jacobi(p2, gen, AS, GK, 0);
				for (i = 0; i < p2; i++)
					AS[i][i] = 1 / Math.sqrt(AS[i][i]);
				zgkdkz(p2, n, X, GK, AS, Y);
				h = X;
				X = Y;
				Y = h;
				it++;
			} else // Ausgabe
			{
				Q[q] = 1 / qn;
				for (j = 0; j < q; j++) {
					qn = 0;
					for (i = 0; i < n; i++)
						qn += Y[i][j] * X[i][j];
					Q[j] = 1 / qn;
				}
				maxit = it; // Y zeigt auf X
				for (i = 0; i < n; i++)
					for (j = 0; j < p1; j++)
						Y[i][j] = X[i][j];
				break;
			}
		}
		
        /**
         * END OF ALGORITHM
         */

		// Save the eigenvalues into the vector
		eigenvalueVector = new Vector("Eigenvalues", Q.length);

		for (int a = 0; a < Q.length; a++) {
			eigenvalueVector.setValue(a + 1, Q[a]);
		}

		return true;
	}

	/**
	 * Grab the vector holding the calculated eigenvalues (you need to call compute()) first
	 * @return Vector, the vector holding all eigenvalues
	 */
	public Vector getEigenvalueVector() {
		return eigenvalueVector;
	}
	
	/**
	 * Grab the double array holding the calculated eigenvectors (you need to call compute()) first
	 * @return double[][], the array holding all eigenvectors
	 */
	public double[][] getEigenvectorArray() {
		return X;
	}

	/**
	 * Calculate the Cholesky (by Ewe Leonties)
	 * A is changed in the progress containing the result
	 * @param n, size of Matrix A
	 * @param A, Matrix A in HullMatrix style
	 * @param D, diagonal pointer vector
	 * @return int, 0 = ERROR, 1 = SUCCESS
	 */
	protected int cholesky(int n, double[] A, int[] D)
	{
		double s;
		int i, j, m, k, io, jo, fi, fj;

		for (i = 1; i <= n; i++) {
			io = D[i] - i;
			fi = D[i - 1] + 1 - io;
			for (j = fi; j <= i; j++) {
				jo = D[j] - j;
				fj = D[j - 1] - jo + 1;
				s = A[io + j];
				if (fi > fj)
					m = fi;
				else
					m = fj;

				for (k = m; k < j; k++)
					s -= A[io + k] * A[jo + k];
				if (j < i)
					A[io + j] = s / A[D[j]];
				else {
					if (s < 0)
						return (0);
					A[D[i]] = Math.sqrt(s);
				}
			}
		}
		return (1);
	}

	/**
	 * Init the Matrix X with random numbers 0 <= xij <= 1
	 * @param p, the dimension of X
	 * @param n, the dimension of X
	 * @param X, RETURN VALUE, linear independent start vectors
	 */
	protected void initx(int p, int n, double[][] X)
	{
		int i, j, h;

		h = 11;
		for (i = 0; i < p; i++)
			for (j = 0; j < n; j++) {
				h = (h * 25173 + 13849) % 65563;
				X[j][i] = (double) h / 65563;
			}
	}

	/**
	 * Calculates the Eigenvalues and Eigenvectors of the symmetric Matrix A
	 * @param n, size of Matrix A
	 * @param gen, machine precision
	 * @param A, double[][] array of Matrix A
	 * @param V, RETURN VALUE, matrix of eigenvectors
	 * @param ord, if >0 the eigenvalues and associated eigenvectors are sorted from upwards, else downwards
	 */
	protected void jacobi(int n, double gen, double[][] A, double[][] V, int ord)
	{
		int i, j, jj, p, q, k;
		double norma, sum, theta, t, s, c, h, eps;

		norma = 0;
		for (i = 0; i < n; i++) {
			for (j = 0; j < i; j++) {
				V[i][j] = 0;
				V[j][i] = 0;
				if (norma < Math.abs(A[i][j]))
					norma = Math.abs(A[i][j]);
			}
			V[i][i] = 1;
		}

		eps = gen * n * norma;
		do {
			for (p = 0; p < n - 1; p++) {
				for (q = p + 1; q < n; q++) {
					if (Math.abs(A[q][p]) > eps) {
						theta = 0.5 * (A[q][q] - A[p][p]) / A[q][p];
						t = 1;
						if (Math.abs(theta) > gen) {
							t = 1 / (Math.abs(theta) + Math.sqrt(1 + theta
									* theta));
							if (theta < 0)
								t = -t;
						}
						c = 1 / Math.sqrt(1 + t * t);
						s = c * t;
						A[p][p] -= t * A[q][p];
						A[q][q] += t * A[q][p];
						A[q][p] = 0;
						for (i = 0; i < p; i++) {
							h = c * A[p][i] - s * A[q][i];
							A[q][i] = s * A[p][i] + c * A[q][i];
							A[p][i] = h;
						}
						for (i = p + 1; i < q; i++) {
							h = c * A[i][p] - s * A[q][i];
							A[q][i] = s * A[i][p] + c * A[q][i];
							A[i][p] = h;
						}
						for (i = q + 1; i < n; i++) {
							h = c * A[i][p] - s * A[i][q];
							A[i][q] = s * A[i][p] + c * A[i][q];
							A[i][p] = h;
						}
						for (i = 0; i < n; i++) {
							h = c * V[i][p] - s * V[i][q];
							V[i][q] = s * V[i][p] + c * V[i][q];
							V[i][p] = h;
						}
					}
				}
			}

			sum = 0;
			for (i = 1; i < n; i++) {
				for (j = 0; j < i; j++) {
					if (Math.abs(A[i][j]) <= eps)
						A[i][j] = 0;
					sum += A[i][j] * A[i][j];
				}
			}
		} while (sum > eps * eps / 2);

		for (i = 1; i < n; i++)
			for (j = n - 1; j >= i; j--) {
				jj = j - 1;
				if (ord > 0) {
					s = A[jj][jj];
					c = A[j][j];
				} else {
					s = A[j][j];
					c = A[jj][jj];
				}
				if (s > c) {
					h = A[jj][jj];
					A[jj][jj] = A[j][j];
					A[j][j] = h;
					for (k = 0; k < n; k++) {
						h = V[k][jj];
						V[k][jj] = V[k][j];
						V[k][j] = h;
					}
				}
			}
	}

	/**
	 * Multiplies the matrices ZT * Y = AS
	 * @param p, dimension of Y
	 * @param n, dimension of Y
	 * @param Z, the matrix Z
	 * @param Y, the matrix Y
	 * @param AS, RETURN VALUE
	 */
	protected void asch(int p, int n, double[][] Z, double[][] Y, double[][] AS)
	{
		int i, j, l;

		for (i = 0; i < p; i++)
			for (j = 0; j < p; j++) {
				AS[i][j] = 0;
				for (l = 0; l < n; l++)
					AS[i][j] += Z[l][i] * Y[l][j];
			}
	}

	/**
	 * Multiplies the matrices Z * G * D = X
	 * @param p, dimension of Z
	 * @param n, dimension of Z
	 * @param Z, nxp
	 * @param G, pxp
	 * @param D, diagonalmatrix pxp
	 * @param H, RETURN VALUE
	 */
	protected void zgkdkz(int p, int n, double[][] Z, double[][] G, double[][] D, double[][] H)
	{
		int i, l, j;
		double h;

		for (i = 0; i < n; i++)
			for (j = 0; j < p; j++) {
				h = 0;
				for (l = 0; l < p; l++)
					h += Z[i][l] * G[l][j];
				H[i][j] = h * D[j][j];
			}
	}

	/**
	 * Multiplies the matrices X * B = VB
	 * @param p, dimension of X
	 * @param n, dimension of X
	 * @param B, matrix B in HullMatrix style
	 * @param D, diagonal pointer vector
	 * @param X, Iteration-Vector-Matrix
	 * @param VB, RETURN VALUE, Product-Matrix
	 */
	protected void bxvb(int p, int n, double[] B, int[] D, double[][] X, double[][] VB)
	{
		int i, j, k, l, jh, klo, kup;

		for (i = 0; i < p; i++) {
			for (j = 0; j < n; j++)
				VB[j][i] = 0;
			for (j = 0; j < n; j++) {
				klo = D[j] + 1;
				kup = D[j + 1];
				VB[j][i] += B[kup] * X[j][i];
				for (k = klo; k < kup; k++) {
					jh = k - kup + j;
					VB[jh][i] += B[k] * X[j][i];
					VB[j][i] += B[k] * X[jh][i];
				}
			}
		}
	}

	/**
	 * Solves the matrix calculation L * LT * X = B via "Vorwaerts- und Rueckwaertseinsetzen"
	 * @param n, dimension of X
	 * @param p, dimension of X
	 * @param L, left-triangle-matrix in HullMatrix style
	 * @param D, diagonal pointer vector
	 * @param B, right side matrix
	 * @param X, RETURN VALUE
	 */
	protected void mvrenv(int n, int p, double[] L, int[] D, double[][] B, double[][] X)
	{
		int i, j, k, io, fi, di;
		double s;

		for (i = 0; i < n; i++) {
			di = D[i + 1];
			io = D[i + 1] - i;
			fi = D[i] + 1 - io;
			for (k = 0; k < p; k++) {
				s = B[i][k];
				for (j = fi; j < i; j++)
					s -= L[io + j] * X[j][k];
				X[i][k] = s / L[di];
			}
		}
		for (i = n - 1; i >= 0; i--) {
			di = D[i + 1];
			io = D[i + 1] - i;
			fi = D[i] + 1 - io;
			for (k = 0; k < p; k++) {
				X[i][k] = X[i][k] / L[di];
				for (j = fi; j < i; j++)
					X[j][k] -= L[io + j] * X[i][k];
			}
		}
	}
}
