/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.calc.objects.algorithm;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import fhm.edu.fem.calc.objects.Vector;
import fhm.edu.fem.util.logging.LoggingManager;

/**
 * Class Description:
 *		This class offers methods to calculate Eigenvalues and Eigenvectors
 *		via the SVI2 Algorithm.
 *
 *		The algorithms used in this class are Copyright by Ewe Leonties
 *		taken from the C-Classes he wrote as diploma thesis.
 *
 *		The conversion to Java and the percentage displays have been done
 *		by the authors mentioned below.
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class SVI2 {

	/* Our local logger object */
	protected static Logger localLogger = LoggingManager.getLogger(SVI2.class.getName());

	/* The eigenvalue vector */
	protected Vector eigenvalueVector = null;
	
	/* The eigenvector array */
	protected double[][] X = null;

	/**
	 * Constructor
	 */
	public SVI2() {}

	/**
	 * Simultane vector-iteration to solve the general EWA A * x = lam * B * x
	 * @param n, size of matrix A and B
	 * @param p, number of wanted eigenvalues and eigenvectors
	 * @param D, diagonal pointer vector
	 * @param LA, matrix A in HullMatrix style
	 * @param LB, matrix B in hullMatrix style
	 * @param eps, calculation precision
	 * @param maxit, maximum number of iterations
	 * @return boolean, true if all went well, false otherwise
	 */
	public boolean compute(int n,int p,int[] D,double[] LA,double[] LB,double eps,int maxit) {
		
        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "SVI-2 Algorithm - by Ewe Leonties");
        }
        
        /**
         * START OF ALGORITHM
         */
		
		int i, it;
		double[][] Y = new double[n][p];
		double[][] h;
		double[] QA = new double[p];
		double del, tol;

		X = new double[n][p];
		double[] Q = new double[p];

		initx(p, n, X);

		if (cholesky(n, LA, D) == 0) {
			if (localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Error while calculating the first Cholesky !");
			}

			return false; // fehler(200,0);
		}
		if (cholesky(n, LB, D) == 0) {
			if (localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Error while calculating the second Cholesky !");
			}

			return false; // fehler(205,0);
		}

		for (i = 0; i < p; i++)
			QA[i] = 0;
		it = 0;
		do {
			mlenv(n, p, LB, D, X, Y);
			mvrenv(n, p, LA, D, Y, X);
			mltenv(n, p, LB, D, X, Y);
			h = X;
			X = Y;
			Y = h;
			ortnor(n, p, X, Q);
			del = 0;
			for (i = 0; i < p; i++) {
				tol = Math.abs((QA[i] - Q[i]) / Q[i]);
				if (tol > del)
					del = tol;
				QA[i] = Q[i];
			}
			
            if(localLogger.isEnabledFor(Level.INFO)) {
                localLogger.log(Level.INFO, "SVI-2 - Step #" + it + " complete - eps: " + del);
            }
			
			it++;
		} while ((del > eps) && (it < maxit));
		renv(n, p, D, LB, X);
		maxit = it;
		   
        /**
         * END OF ALGORITHM
         */

		// Save the eigenvalues into the vector
		eigenvalueVector = new Vector("Eigenvalues", Q.length);

		for (int a = 0; a < Q.length; a++) {
			eigenvalueVector.setValue(a + 1, Q[a]);
		}

		return true;
	}

	/**
	 * Grab the vector holding the calculated eigenvalues (you need to call compute()) first
	 * @return Vector, the vector holding all eigenvalues
	 */
	public Vector getEigenvalueVector() {
		return eigenvalueVector;
	}
	
	/**
	 * Grab the double array holding the calculated eigenvectors (you need to call compute()) first
	 * @return double[][], the array holding all eigenvectors
	 */
	public double[][] getEigenvectorArray() {
		return X;
	}

	/**
	 * Calculate the Cholesky (by Ewe Leonties)
	 * A is changed in the progress containing the result
	 * @param n, size of Matrix A
	 * @param A, Matrix A in HullMatrix style
	 * @param D, diagonal pointer vector
	 * @return int, 0 = ERROR, 1 = SUCCESS
	 */
	protected int cholesky(int n, double[] A, int[] D)
	{
		double s;
		int i, j, m, k, io, jo, fi, fj;

		for (i = 1; i <= n; i++) {
			io = D[i] - i;
			fi = D[i - 1] + 1 - io;
			for (j = fi; j <= i; j++) {
				jo = D[j] - j;
				fj = D[j - 1] - jo + 1;
				s = A[io + j];
				if (fi > fj)
					m = fi;
				else
					m = fj;

				for (k = m; k < j; k++)
					s -= A[io + k] * A[jo + k];
				if (j < i)
					A[io + j] = s / A[D[j]];
				else {
					if (s < 0)
						return (0);
					A[D[i]] = Math.sqrt(s);
				}
			}
		}
		return (1);
	}

	/**
	 * Init the Matrix X with random numbers 0 <= xij <= 1
	 * @param p, the dimension of X
	 * @param n, the dimension of X
	 * @param X, RETURN VALUE, linear independent start vectors
	 */
	protected void initx(int p, int n, double[][] X)
	{
		int i, j, h;

		h = 11;
		for (i = 0; i < p; i++)
			for (j = 0; j < n; j++) {
				h = (h * 25173 + 13849) % 65563;
				X[j][i] = (double) h / 65563;
			}
	}

	/**
	 * Multiplies the left-triangle-matrix L with matrix X
	 * @param n, dimension of matrix x
	 * @param p, dimension of matrix x
	 * @param L, left-triangle-matrix in HullMatrix style
	 * @param D, diagonal pointer vector
	 * @param X, matrix X
	 * @param Y, RETURN VALUE
	 */
	protected void mlenv(int n, int p, double[] L, int[] D, double[][] X, double[][] Y)
	{
		int i, j, k, ih, jlo, jup;

		for (k = 0; k < p; k++) {
			for (i = 0; i < n; i++) {
				Y[i][k] = 0;
				jlo = D[i] + 1;
				jup = D[i + 1];
				for (j = jlo; j <= jup; j++) {
					ih = j - jup + i;
					Y[i][k] += L[j] * X[ih][k];
				}
			}
		}
	}

	/**
	 * Solves LT * H = X
	 * @param n, dimension of matrix X
	 * @param p, dimension of matrix X
	 * @param D, diagonal pointer vector
	 * @param L, left-triangle-matrix in HullMatrix style
	 * @param X, RETURN VALUE, result matrix H replaces X
	 */
	protected void renv(int n, int p, int[] D, double[] L, double[][] X)
	{
		int i, k, j, di, io, fi;

		for (i = n - 1; i >= 0; i--) {
			di = D[i + 1];
			io = D[i + 1] - i;
			fi = D[i] + 1 - io;
			for (k = 0; k < p; k++) {
				X[i][k] = X[i][k] / L[di];
				for (j = fi; j < i; j++)
					X[j][k] -= L[io + j] * X[i][k];
			}
		}
	}

	/**
	 * Orthonormizes the Matrix Z
	 * @param n, dimension of Z
	 * @param p, dimension of Z
	 * @param Z, matrix Z with p linear independent vectors, RETURN VALUE
	 * @param Q
	 */
	protected void ortnor(int n, int p, double[][] Z, double[] Q)
	{
		int i, j, k;
		double g;

		for (i = 0; i < p; i++) {
			for (j = 0; j < i; j++) {
				g = 0;
				for (k = 0; k < n; k++)
					g += Z[k][j] * Z[k][i];
				for (k = 0; k < n; k++)
					Z[k][i] -= g * Z[k][j];
			}
			g = 0;
			for (k = 0; k < n; k++)
				g += Z[k][i] * Z[k][i];
			g = 1 / Math.sqrt(g);
			for (k = 0; k < n; k++)
				Z[k][i] *= g;
			Q[i] = g;
		}
	}

	/**
	 * Solves the matrix calculation L * LT * X = B via "Vorwaerts- und Rueckwaertseinsetzen"
	 * @param n, dimension of X
	 * @param p, dimension of X
	 * @param L, left-triangle-matrix in HullMatrix style
	 * @param D, diagonal pointer vector
	 * @param B, right side matrix
	 * @param X, RETURN VALUE
	 */
	protected void mvrenv(int n, int p, double[] L, int[] D, double[][] B, double[][] X)
	{
		int i, j, k, io, fi, di;
		double s;
	
		for (i = 0; i < n; i++) {
			di = D[i + 1];
			io = D[i + 1] - i;
			fi = D[i] + 1 - io;
			for (k = 0; k < p; k++) {
				s = B[i][k];
				for (j = fi; j < i; j++)
					s -= L[io + j] * X[j][k];
				X[i][k] = s / L[di];
			}
		}
		for (i = n - 1; i >= 0; i--) {
			di = D[i + 1];
			io = D[i + 1] - i;
			fi = D[i] + 1 - io;
			for (k = 0; k < p; k++) {
				X[i][k] = X[i][k] / L[di];
				for (j = fi; j < i; j++)
					X[j][k] -= L[io + j] * X[i][k];
			}
		}
	}

	/**
	 * Multiplies the left-triangle-matrix LT with Matrix X
	 * @param n, dimension of X
	 * @param p, dimension of X
	 * @param L, left-triangle-matrix in HullMatrix style
	 * @param D, diagonal pointer vector
	 * @param X, matrix X
	 * @param Y, RETURN VALUE
	 */
	protected void mltenv(int n, int p, double[] L, int[] D, double[][] X, double[][] Y)
	{
		int i, j, k, ih, jlo, jup;

		for (k = 0; k < p; k++)
			for (i = 0; i < n; i++) {
				Y[i][k] = 0;
				jlo = D[i] + 1;
				jup = D[i + 1];
				for (j = jlo; j <= jup; j++) {
					ih = j - jup + i;
					Y[ih][k] += L[j] * X[i][k];
				}
			}
	}
}
