/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.calc.objects.algorithm;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import fhm.edu.fem.calc.objects.Vector;
import fhm.edu.fem.constants.ApplicationConstants;
import fhm.edu.fem.util.logging.LoggingManager;

/**
 * Class Description:
 *		This class offers methods to calculate Eigenvalues and Eigenvectors
 *		via the SVI3 Algorithm.
 *
 *		The algorithms used in this class are Copyright by Ewe Leonties
 *		taken from the C-Classes he wrote as diploma thesis.
 *
 *		The conversion to Java and the percentage displays have been done
 *		by the authors mentioned below.
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class SVI3 {

	/* Our local logger object */
	protected static Logger localLogger = LoggingManager.getLogger(SVI3.class.getName());

	/* The eigenvalue vector */
	protected Vector eigenvalueVector = null;
	
	/* The eigenvector array */
	protected double[][] X = null;

	/**
	 * Constructor
	 */
	public SVI3() {}

	/**
	 * Simultamne vector-iteration with Rayleigh-Ritz-Step
	 * @param n, dimension of matrix A and B
	 * @param A, matrix A in HullMatrix style
	 * @param B, matrix B in HullMatrix style
	 * @param D, diagonal pointer vector
	 * @param p, number of wanted eigen-pairs
	 * @param pt, number of simultane iterated eigen-pairs
	 * @param nitm, maximum number of iterations
	 * @param eps, calculation precision
	 * @return boolean, true if all went well, false otherwise
	 */
	public boolean compute(int n, double[] A, double[] B, int[] D, int p, int pt, int nitm, double eps) {

        if(localLogger.isEnabledFor(Level.INFO)) {
            localLogger.log(Level.INFO, "SVI-3 Algorithm - by Ewe Leonties");
        }
        
        // Calculate how many percentage one step makes up
        double oneStepPercentage = 100.0 / nitm;
        
        // Calculate how many steps make up one percent
        double onePercentSteps = nitm / 100.0;
        
        // Where we store the current actual percentage
        double currentPercentage = 0.0;
        
        /**
         * START OF ALGORITHM
         */
		
		double gen = 5e-16;
		int nritz = 4;
		double[][] h;
		double s, ly, epsm;
		int i, j, k, l;
		int it, jt, kt, ih, jlo, jup;

		double[] H1 = new double[n];
		double[] H2 = new double[n];
		double[][] Y = new double[n][pt];
		double[][] V = new double[pt][pt];
		double[][] G = new double[pt][pt];

		X = new double[n][p];
		double[] LAM = new double[p];

		initx(pt, n, X);

		if (cholesky(n, A, D) == 0) {
			if (localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Error while calculating the first Cholesky !");
			}

			return false; // fehler(200,0);
		}
		if (cholesky(n, B, D) == 0) {
			if (localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Error while calculating the second Cholesky !");
			}

			return false; // fehler(205,0);
		}

		jt = 0;
		it = 0;
		mltenv(n, pt, B, D, X, Y);

		while (true) {
			for (l = 0; l < pt; l++) // Orthonormierung der Spalten von Y
			{ // Gram-Schmidt modifiziert
				s = 0; // Berechnung der Matrix R, Speicherung in V
				for (i = 0; i < n; i++)
					s += Y[i][l] * Y[i][l];
				s = Math.sqrt(s);
				V[l][l] = s;
				for (i = 0; i < n; i++)
					Y[i][l] = Y[i][l] / s;
				for (j = l + 1; j < pt; j++) {
					s = 0;
					for (i = 0; i < n; i++)
						s += Y[i][l] * Y[i][j];
					V[l][j] = s;
					for (i = 0; i < n; i++)
						Y[i][j] -= s * Y[i][l];
				}
			}

			if (it % nritz == 0) // nach je n-Ritz Schritten
			{ // Rayleigh-Ritz-Schritt
				for (j = 0; j < pt; j++) {
					for (k = 0; k <= j; k++) {
						s = 0;
						for (i = j; i < pt; i++)
							s += V[j][i] * V[k][i];
						G[j][k] = s;
					}
				}
				jacobi(pt, gen, G, V, 0);
				for (k = 0; k < pt; k++)
					H1[k] = 1 / Math.sqrt(G[k][k]);
				for (i = 0; i < n; i++) {
					for (k = 0; k < pt; k++) {
						s = 0;
						for (j = 0; j < pt; j++)
							s += Y[i][j] * V[j][k];
						H1[k] = s;
					}
					for (k = 0; k < pt; k++)
						Y[i][k] = H1[k];
				}

				if (it > 0) // Test auf Konvergenz
				{ // bezueglich der zwei letzten der p Vektoren
					epsm = 0;
					kt = p - 2;
					for (k = kt; k < p; k++) {
						ly = 1;
						for (j = jt; j < p; j++) {
							s = 0;
							for (i = 0; i < n; i++)
								s += Y[i][k] * X[i][j];
							if ((Math.abs(s) < (eps * 0.01)) && (k == p - 1))
								if ((j + 1) < kt)
									jt = j + 1;
								else
									jt = kt;
							ly = Math.sqrt(Math.abs((ly + s) * (ly - s)));
						}
						if (ly > epsm)
							epsm = ly;
					}
					//System.out.println("Schritt: " + it + " eps: " + epsm);
					if ((epsm < eps) || (it > nitm)) {
						ewev(n, pt, B, A, D, H1, LAM, X, Y);
						nitm = it;
						
						// Call our method to remember the eigenvalues before returning
						this.storeEigenvalues(LAM);
						
						// All went well
						return true;
					}
				}
			}

			for (k = 0; k < pt; k++) // Iterationsschritt der pt Vektoren
			{ // in Y zu Vektoren in X
				for (i = 0; i < n; i++) {
					s = 0;
					jlo = D[i] + 1;
					jup = D[i + 1];
					for (j = jlo; j <= jup; j++) {
						ih = j - jup + i;
						s += B[j] * Y[ih][k];
					}
					H1[i] = -s;
				}
				vrenv(n, A, D, H1, H2);
				for (i = 0; i < n; i++) {
					X[i][k] = 0;
					jlo = D[i] + 1;
					jup = D[i + 1];
					for (j = jlo; j <= jup; j++) {
						ih = j - jup + i;
						X[ih][k] += B[j] * H2[i];
					}
				}
			}

			h = X; // Vertauschung iterierte Vektoren Y
			X = Y; // vorhergehende Vektoren X
			Y = h;
			it++;
			
            // Calculate the current percentage of steps already done
            currentPercentage = (it) * oneStepPercentage;

            // Make a percentage output about every 5% of completed steps
            if(it%(onePercentSteps*2.0) < 1) {
                
                if(localLogger.isEnabledFor(Level.INFO)) {
                    localLogger.log(Level.INFO, "SVI-3 - Current number of Iterations at " + ApplicationConstants.convertPercentageToString(currentPercentage) + " % (may finish before reaching 100%)");
                }
            }
			
			if (it % nritz > 0) {
				//System.out.println("Schritt: " + it);
			}
		}   
		
        /**
         * END OF ALGORITHM
         */
	}
	
	/**
	 * Special method to help store the eigenvalues after they have been
	 * calculated
	 * 
	 * @param LAM,
	 *            the vector holding all calculated eigenvalues
	 */
	protected void storeEigenvalues(double[] LAM) {
		
		// Save the eigenvalues into the vector
		eigenvalueVector = new Vector("Eigenvalues", LAM.length);

		for (int a = 0; a < LAM.length; a++) {
			eigenvalueVector.setValue(a + 1, LAM[a]);
		}
	}

	/**
	 * Grab the vector holding the calculated eigenvalues (you need to call
	 * compute()) first
	 * 
	 * @return Vector, the vector holding all eigenvalues
	 */
	public Vector getEigenvalueVector() {
		return eigenvalueVector;
	}
	
	/**
	 * Grab the double array holding the calculated eigenvectors (you need to
	 * call compute()) first
	 * 
	 * @return double[][], the array holding all eigenvectors
	 */
	public double[][] getEigenvectorArray() {
		return X;
	}
	
	
	/**
	 * Calculates the Eigenvalues and Eigenvectors of the symmetric Matrix A
	 * @param n, size of Matrix A
	 * @param gen, machine precision
	 * @param A, double[][] array of Matrix A
	 * @param V, RETURN VALUE, matrix of eigenvectors
	 * @param ord, if >0 the eigenvalues and associated eigenvectors are sorted from upwards, else downwards
	 */
	protected void jacobi(int n, double gen, double[][] A, double[][] V, int ord)
	{
		int i, j, jj, p, q, k;
		double norma, sum, theta, t, s, c, h, eps;

		norma = 0;
		for (i = 0; i < n; i++) {
			for (j = 0; j < i; j++) {
				V[i][j] = 0;
				V[j][i] = 0;
				if (norma < Math.abs(A[i][j]))
					norma = Math.abs(A[i][j]);
			}
			V[i][i] = 1;
		}

		eps = gen * n * norma;
		do {
			for (p = 0; p < n - 1; p++) {
				for (q = p + 1; q < n; q++) {
					if (Math.abs(A[q][p]) > eps) {
						theta = 0.5 * (A[q][q] - A[p][p]) / A[q][p];
						t = 1;
						if (Math.abs(theta) > gen) {
							t = 1 / (Math.abs(theta) + Math.sqrt(1 + theta
									* theta));
							if (theta < 0)
								t = -t;
						}
						c = 1 / Math.sqrt(1 + t * t);
						s = c * t;
						A[p][p] -= t * A[q][p];
						A[q][q] += t * A[q][p];
						A[q][p] = 0;
						for (i = 0; i < p; i++) {
							h = c * A[p][i] - s * A[q][i];
							A[q][i] = s * A[p][i] + c * A[q][i];
							A[p][i] = h;
						}
						for (i = p + 1; i < q; i++) {
							h = c * A[i][p] - s * A[q][i];
							A[q][i] = s * A[i][p] + c * A[q][i];
							A[i][p] = h;
						}
						for (i = q + 1; i < n; i++) {
							h = c * A[i][p] - s * A[i][q];
							A[i][q] = s * A[i][p] + c * A[i][q];
							A[i][p] = h;
						}
						for (i = 0; i < n; i++) {
							h = c * V[i][p] - s * V[i][q];
							V[i][q] = s * V[i][p] + c * V[i][q];
							V[i][p] = h;
						}
					}
				}
			}

			sum = 0;
			for (i = 1; i < n; i++) {
				for (j = 0; j < i; j++) {
					if (Math.abs(A[i][j]) <= eps)
						A[i][j] = 0;
					sum += A[i][j] * A[i][j];
				}
			}
		} while (sum > eps * eps / 2);

		for (i = 1; i < n; i++)
			for (j = n - 1; j >= i; j--) {
				jj = j - 1;
				if (ord > 0) {
					s = A[jj][jj];
					c = A[j][j];
				} else {
					s = A[j][j];
					c = A[jj][jj];
				}
				if (s > c) {
					h = A[jj][jj];
					A[jj][jj] = A[j][j];
					A[j][j] = h;
					for (k = 0; k < n; k++) {
						h = V[k][jj];
						V[k][jj] = V[k][j];
						V[k][j] = h;
					}
				}
			}
	}

	/**
	 * Calculate the Cholesky (by Ewe Leonties)
	 * A is changed in the progress containing the result
	 * @param n, size of Matrix A
	 * @param A, Matrix A in HullMatrix style
	 * @param D, diagonal pointer vector
	 * @return int, 0 = ERROR, 1 = SUCCESS
	 */
	protected int cholesky(int n, double[] A, int[] D)
	{
		double s;
		int i, j, m, k, io, jo, fi, fj;

		for (i = 1; i <= n; i++) {
			io = D[i] - i;
			fi = D[i - 1] + 1 - io;
			for (j = fi; j <= i; j++) {
				jo = D[j] - j;
				fj = D[j - 1] - jo + 1;
				s = A[io + j];
				if (fi > fj)
					m = fi;
				else
					m = fj;

				for (k = m; k < j; k++)
					s -= A[io + k] * A[jo + k];
				if (j < i)
					A[io + j] = s / A[D[j]];
				else {
					if (s < 0)
						return (0);
					A[D[i]] = Math.sqrt(s);
				}
			}
		}
		return (1);
	}

	/**
	 * Init the Matrix X with random numbers 0 <= xij <= 1
	 * @param p, the dimension of X
	 * @param n, the dimension of X
	 * @param X, RETURN VALUE, linear independent start vectors
	 */
	protected void initx(int p, int n, double[][] X)
	{
		int i, j, h;

		h = 11;
		for (i = 0; i < p; i++)
			for (j = 0; j < n; j++) {
				h = (h * 25173 + 13849) % 65563;
				X[j][i] = (double) h / 65563;
			}
	}

	/**
	 * Multiplies the left-triangle-matrix L with matrix X
	 * @param n, dimension of matrix x
	 * @param p, dimension of matrix x
	 * @param L, left-triangle-matrix in HullMatrix style
	 * @param D, diagonal pointer vector
	 * @param X, matrix X
	 * @param Y, RETURN VALUE
	 */
	protected void mlenv(int n, int p, double[] L, int[] D, double[][] X, double[][] Y)
	{
		int i, j, k, ih, jlo, jup;

		for (k = 0; k < p; k++) {
			for (i = 0; i < n; i++) {
				Y[i][k] = 0;
				jlo = D[i] + 1;
				jup = D[i + 1];
				for (j = jlo; j <= jup; j++) {
					ih = j - jup + i;
					Y[i][k] += L[j] * X[ih][k];
				}
			}
		}
	}

	/**
	 * Solves LT * H = X
	 * @param n, dimension of matrix X
	 * @param p, dimension of matrix X
	 * @param D, diagonal pointer vector
	 * @param L, left-triangle-matrix in HullMatrix style
	 * @param X, RETURN VALUE, result matrix H replaces X
	 */
	protected void renv(int n, int p, int[] D, double[] L, double[][] X)
	{
		int i, k, j, di, io, fi;

		for (i = n - 1; i >= 0; i--) {
			di = D[i + 1];
			io = D[i + 1] - i;
			fi = D[i] + 1 - io;
			for (k = 0; k < p; k++) {
				X[i][k] = X[i][k] / L[di];
				for (j = fi; j < i; j++)
					X[j][k] -= L[io + j] * X[i][k];
			}
		}
	}

	/**
	 * Orthonormizes the Matrix Z
	 * @param n, dimension of Z
	 * @param p, dimension of Z
	 * @param Z, matrix Z with p linear independent vectors, RETURN VALUE
	 * @param Q
	 */
	protected void ortnor(int n, int p, double[][] Z, double[] Q)
	{
		int i, j, k;
		double g;

		for (i = 0; i < p; i++) {
			for (j = 0; j < i; j++) {
				g = 0;
				for (k = 0; k < n; k++)
					g += Z[k][j] * Z[k][i];
				for (k = 0; k < n; k++)
					Z[k][i] -= g * Z[k][j];
			}
			g = 0;
			for (k = 0; k < n; k++)
				g += Z[k][i] * Z[k][i];
			g = 1 / Math.sqrt(g);
			for (k = 0; k < n; k++)
				Z[k][i] *= g;
			Q[i] = g;
		}
	}

	/**
	 * Multiplies the left-triangle-matrix LT with Matrix X
	 * @param n, dimension of X
	 * @param p, dimension of X
	 * @param L, left-triangle-matrix in HullMatrix style
	 * @param D, diagonal pointer vector
	 * @param X, matrix X
	 * @param Y, RETURN VALUE
	 */
	protected void mltenv(int n, int p, double[] L, int[] D, double[][] X, double[][] Y)
	{
		int i, j, k, ih, jlo, jup;
	
		for (k = 0; k < p; k++)
			for (i = 0; i < n; i++) {
				Y[i][k] = 0;
				jlo = D[i] + 1;
				jup = D[i + 1];
				for (j = jlo; j <= jup; j++) {
					ih = j - jup + i;
					Y[ih][k] += L[j] * X[i][k];
				}
			}
	}
	
	/**
	 * Solves LBT * H = X via "Rueckwaertseinsetzen"
	 * Calculation of p-Eigenvectors and Eigenvalues as rayleigh's Quotients
	 * @param n, dimension of matrix X and Y
	 * @param pt, dimension of matrix X and Y
	 * @param B, left-triangle-matrix B in HullMatrix style
	 * @param A, left-triangle-matrix A in HullMatrix style
	 * @param D, diagonal pointer vector
	 * @param H1, helper-vector
	 * @param LAM, RETURN VALUE, Vector of p-Eigenvalues
	 * @param X, RETURN VALUE, Matrix of p-Eigenvectors
	 * @param Y, helper-matrix
	 */
	protected void ewev(int n, int pt, double[] B, double[] A, int[] D, double[] H1, double[] LAM, double[][] X, double[][] Y)
	{
		int i, j, k, ij, ih, jlo, jup;
		double sa, sb;

		for (k = 0; k < pt; k++) {
			for (i = 0; i < n; i++)
				X[i][k] = Y[i][k];
			for (i = n - 1; i >= 0; i--) {
				X[i][k] = X[i][k] / B[D[i + 1]];
				jlo = D[i] - D[i + 1] + i + 1;
				jup = i - 1;
				if (jlo <= jup) {
					ij = D[i];
					for (j = jlo; j <= jup; j++) {
						ij++;
						X[j][k] -= B[ij] * X[i][k];
					}
				}
			}

			for (i = 0; i < n; i++) {
				H1[i] = 0;
				jlo = D[i] + 1;
				jup = D[i + 1];
				for (j = jlo; j <= jup; j++) {
					ih = j - jup + i;
					H1[ih] += A[j] * X[i][k];
				}
			}
			sa = sb = 0;
			for (i = 0; i < n; i++) {
				sa += H1[i] * H1[i];
				sb += Y[i][k] * Y[i][k];
			}
			LAM[k] = sa / sb;
		}
	}
	
	/**
	 * Solves L * LT * x = b via "Vorwaerts- und Rueckwaertseinsetzen"
	 * @param n, dimension of L
	 * @param L, matrix L in HullMatrix style
	 * @param D, diagonal pointer vector
	 * @param B, constants-vector
	 * @param X, RETURN VALUE
	 */
	protected void vrenv(int n, double[] L, int[] D, double[] B, double[] X)
	{
		int i, j, io, fi;
		double s;

		for (i = 0; i < n; i++) {
			s = B[i];
			io = D[i + 1] - i;
			fi = D[i] + 1 - io;
			for (j = fi; j < i; j++)
				s -= L[io + j] * X[j];
			X[i] = s / L[D[i + 1]];
		}
		for (i = n - 1; i >= 0; i--) {
			X[i] = X[i] / L[D[i + 1]];
			io = D[i + 1] - i;
			fi = D[i] + 1 - io;
			for (j = fi; j < i; j++)
				X[j] -= L[io + j] * X[i];
		}
	}
}
