/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.calc.objects.matrix;

import org.apache.log4j.Logger;

import fhm.edu.fem.constants.ApplicationConstants;
import fhm.edu.fem.util.logging.LoggingManager;

/**
 * Class Description:
 *		This class represents the ENV vector used in the HullMatrix (the index starts at 1 not 0)
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class ENVHullMatrix {
    
	/* Our local logger object */
	protected static Logger localLogger = LoggingManager.getLogger(ENVHullMatrix.class.getName());

    // Where we store our data
    protected ENVVector data = new ENVVector();

    // The vector used when enlarging the ENV
    protected ENVVector enlargeVector = new ENVVector();
    
    protected ENVHullMatrix() {}
    
    /**
     * Construct a new ENV for a HullMatrix with the given number of "0.0" elements
     * @param size, the size the ENV should have at startup
     */
    public ENVHullMatrix(int size) {
        
        // Initialize the env
        for(int i = 1; i <= size; i++) {

            this.data.add(new Double(0.0));
        }
    }
    
    /**
     * Copy Constructor
     * @param env, the env object to incorporate in the new ENVHullMatrix object
     */
    public ENVHullMatrix(ENVHullMatrix env) {
        this.data = (ENVVector)(env.getData().clone());
    }
    
    /**
     * Add a new element at a specific index of the list, causing all other elements coming after that element to be
     * shifted to the right (including the element currently at the given index)
     * @param index, the index to set the value at (the index starts at 1)
     * @param value, the new value
     */
    public void add(int index, double value) {
        this.data.add(index-1, new Double(value));
    }
    
    /**
     * Replace an already existing value in the list with a new value at the specific index
     * @param index, the index of the element to replace (the index starts at 1)
     * @param value, the value to replace the other value with
     * @return the value we replaced
     */
    public double set(int index, double value) {
        return ((Double)(data.set(index-1, new Double(value)))).doubleValue();
    }
    
    /**
     * Retrieve a value from the list
     * @param index, the index of the value to retrieve (the index starts at 1)
     * @return the value
     */
    public double get(int index) {
        return ((Double)(data.get(index-1))).doubleValue();
    }
    
    /**
     * Retrieve the current size
     * @return size
     */
    public int size() {
        return data.size();
    }
    
    /**
     * Enlarge the ENV by the given difference at the given index
     * @param index, the index we enlarge the ENV at (all other items are shifted to the right)
     * @param difference, the difference to enlarge the ENV
     */
    public void enlargeAt(int index, int difference) {
        
        // Collect all enlargements in one vector to avoid too many array-copy actions within the Java VM
        this.enlargeVector.clear();
        
        // The difference we need to enlarge between the correct and the previous hull-diagonal-element-index
        for(int i = 1; i <= difference; i++) {
            
            enlargeVector.add(new Double(0.0));
        }
        
        // Add the enlarge vector to ENV
        this.data.addAll((index-1), enlargeVector);
    }
    
    /**
     * Retrieve the amount of required memory in this matrix (in byte)
     * @return the amount of required memory in byte
     */
    public long getRequiredMemory() {
        
        long bytes = 0;

        // Add the bytes needed by all Double elements to the required amount of memory
        bytes += ApplicationConstants.SIZE_DOUBLE * this.data.size();
        
        return bytes;
    }
    
    /**
     * Retrieve the data vector from the ENV
     * @return the data vector holding "Double" objects
     */
    public ENVVector getData() {
        return this.data;
    }
    
    /**
     * Clone this object
     * @return the cloned object
     */
    public Object clone() {
        
        // Use the copy constructor for cloning
        return new ENVHullMatrix(this);
    }
}