/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.calc.objects.matrix;

import fhm.edu.fem.calc.objects.MathObject;
import fhm.edu.fem.calc.objects.Vector;
import fhm.edu.fem.exception.CholeskyFailedException;

/**
 * Class Description:
 *		The interface defining a matrix class
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public interface Matrix extends MathObject {

    /**
     * Retrieve a value from the matrix
     * @param row, the row index
     * @param colum, the column index
     * @return the value
     * @throws IndexOutOfBoundsException
     */
    public double getValue(int row, int column) throws IndexOutOfBoundsException;
    
    /**
     * Set a new value in the matrix
     * @param row, the row index
     * @param colum, the column index
     * @param value, the value to set
     * @throws IndexOutOfBoundsException
     */
    public void setValue(int row, int column, double value) throws IndexOutOfBoundsException;
    
    /**
     * Store a whole line in the matrix at the given line number
     * @param lineNumber, where to store the line
     * @param line, the array holding the line's data
     */
    public void storeLine(int lineNumber,  double[] line);
    
    /**
     * Set a whole column to the given value in the matrix at the given column number
     * @param columnNumber, where to store the column
     * @param value, the sole value we want to fill the column with
     */
    public void setLineAndColumn(int index, double value);
    
    /**
     * Get the number of rows and columns in the matrix (square matrix only!)
     * @return the number of rows and columns
     */
    public int size();
    
    /* (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    public String toString();
    
    /**
     * Variable toString method
     * @param number, the number of rows and columns to display (starting at the upper left)
     * @return String
     */
    public String toString(int number);
    
    /**
     * Multiply all values in the matrix by the given value
     * @param multiplicator, the value to multiply the matrix with
     * @return the new matrix, the original matrix is not changed
     */
    public Matrix mult(double multiplicator);
    
    /**
     * Multiply this Matrix with the given vector (have to be the same size)
     * @param m, the vector
     * @return Vector, the result vector
     */
    public Vector mult(Vector m);
    
    /**
     * Multiply this Matrix with the given Matrix (have to be the same size)
     * @param m, the matrix
     * @return Matrix, the result matrix
     */
    public Matrix mult(Matrix m);
    
    /**
     * Add a given matrix to this matrix
     * @param m, the matrix to add
     * @return the new matrix, the original matrix is not changed
     */
    public Matrix add(Matrix m);
    
    /**
     * Test if the Matrix is symmetric
     * @return false if it is not symmetric, true otherwise
     */
    public boolean isSymmetric();
    
    /**
     * Test if the main diagonal of this matrix is not zero
     * @return true if it is not zero, false otherwise
     */
    public boolean isMainDiagonalNotZero();
    
    /**
     * Clone the matrix
     * @return the colned matrix
     */
     public Object clone();
     
     /**
      * Do the cholesky for this matrix
      * @return the matrix L as result of the cholesky
      */
     public Matrix cholesky() throws CholeskyFailedException;
     
     /**
      * Transponate the Matrix
      * @return the new transponated Matrix
      */
     public Matrix transponate();
     
     /**
      * Retrieve the name of this matrix
      * @return the name
      */
     public String getName();
     
     /**
      * Set a new name for this matrix
      * @param name
      */
     public void setName(String name);
     
     /**
      * Convert all data in the matrix to a quadratic double array
      * @return double[][]
      */
     public double[][] asArray();
     
     /**
      * Retrieve the amount of required memory in this matrix (in byte)
      * @return the amount of required memory in byte
      */
     public long getRequiredMemory();
}
