/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.calc.objects.matrix;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import fhm.edu.fem.calc.objects.Vector;
import fhm.edu.fem.constants.ApplicationConstants;
import fhm.edu.fem.exception.CholeskyFailedException;
import fhm.edu.fem.util.logging.LoggingManager;

/**
 * Class Description:
 *		HullMatrix extends the basic matrix. It offers a hull-oriented
 *		memory management to safe memory.
 *		ONLY WORKS ON SYMMETRIC MATRICES
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class MatrixHull implements Matrix {
    
	/* Our local logger object */
	protected static Logger localLogger = LoggingManager.getLogger(MatrixHull.class.getName());
    
    // The rows and columns in the matrix
    protected int size;
    
    // The diagonal-pointer-vector
    protected int[] diagPointer;
    
    // The array holding all the data ("ENV")
    protected ENVHullMatrix env = null;
    
    // The name of the matrix, if it has one
    protected String name = "";
    
    // Initializing these elements here to save allocation time when running
    protected int hullDiagElementIndex = 0;
    protected int prevHullDiagElementIndex = 0;
    protected int targetHullElementIndex = 0;
    protected int difference = 0;
    
    /**
     * Matrix with the given name, filled with the given values
     * @param name, name
     * @param values, quadratic
     */
    public MatrixHull(String name, double[][] values) {
    	
        this.name = name;
        this.size = values.length;

        // Create the diagonal-pointer-vector
        this.diagPointer = new int[this.size+1];
        
        // Create the array holding the data
        this.env = new ENVHullMatrix(this.size);
        
        // Initialize the diagonal-pointer-vector
        for(int i = 0; i <= this.size; i++) {

            this.diagPointer[i] = i;
        }
        
        // Store the values
        for(int i = 0; i < values.length; i++) {
        	for(int j = 0; j < values.length; j++) {
        		
        		this.setValue(i+1, j+1, values[i][j]);
        	}
        }
    }
    
    /**
     * Create a new HullMatrix with the size (x rows and x columns) (all fields are initialized with zeros)
     * @param name, the name of the matrix
     * @param size, the number of rows and columns (square matrix only!)
     */
    public MatrixHull(String name, int size) {
        
        this.name = name;
        this.size = size;

        // Create the diagonal-pointer-vector
        this.diagPointer = new int[size+1];
        
        // Create the array holding the data
        this.env = new ENVHullMatrix(size);
        
        // Initialize the diagonal-pointer-vector
        for(int i = 0; i <= size; i++) {

            this.diagPointer[i] = i;
        }
    }
    
    protected MatrixHull() {}
    
    /**
     * Copy Constructor
     * @param m, the hull matrix to copy
     */
    public MatrixHull(MatrixHull m) {
        
        // Retrieve the name and size of the matrix to copy
        this.name = m.getName();
        this.size = m.size();
        
        // Retrieve the diag pointer of the other matrix and clone it, making it our own        
        this.diagPointer = new int[m.getDiagPointer().length];
        
        for(int i = 0; i < this.diagPointer.length; i++) {
            this.diagPointer[i] = m.getDiagPointer()[i];
        }
        
        // Retrieve the ENV object from the other matrix and make a copy
        this.env = (ENVHullMatrix)(m.getENV().clone());
    }
    
    /**
     * Retrieve a value from the HullMatrix
     * @param row, the row index (starting at 1)
     * @param column, the column index (starting at 1)
     * @return the value at this position
     */
    public double getValue(int row, int column) throws IndexOutOfBoundsException {

        // Check if we got a proper index
        if(row > this.size || column > this.size || row <= 0 || column <= 0)
        {
            throw new IndexOutOfBoundsException("Row: " + row + " Column: " + column);
        }
        
        // All matrices we work with here are symmetric
        if(column > row) {
            return this.getValue(column, row);
        }
                
        // We got a proper index for both row and column and are also lumping around in the lower triangle of the matrix
        
        // Calculate the hull-element-index where the element is located (no matter if we have it or not)
        targetHullElementIndex = this.diagPointer[row] - (row - column);

        // Check if the target index is still within the range of the previous diagonal-element index the the correct one
        if(targetHullElementIndex <= this.diagPointer[row-1]) {

            // Value is not in ENV, so it must be zero
            return 0;
            
        } else {

            // Value is stored in the ENV, grab it and return it
            return env.get(targetHullElementIndex);
        }
    }
    
    /**
     * Set a value in the HullMatrix
     * @param row, the row index (starting at 1)
     * @param column, the column index (starting at 1)
     * @param value, the value to set at the specified position in the matrix
     */
    public void setValue(int row, int column, double value) throws IndexOutOfBoundsException {

        // Check if we got a proper index
        if(row > this.size || column > this.size || row <= 0 || column <= 0)
	    {
            throw new IndexOutOfBoundsException();
	    }
        
        // Re-set the variables if needed due to symmetrie
        if(column > row) {
            this.setValue(column, row, value);
            return;
        }
        
        // Grab the position of the diagonal-element from the diagonal-pointer-vector
        hullDiagElementIndex = this.diagPointer[row];

        // Also grab the position of the previous diagonal-element from the diagonal-pointer-vector
        prevHullDiagElementIndex = this.diagPointer[row-1];
        
        // Calculate the difference between the diagonal element and the element we want to set
        difference = Math.abs(row - column);

        // Calculate the hull-element-index where the element is located (no matter if we have it or not)
        targetHullElementIndex = hullDiagElementIndex - difference;

        // Check if the target index is still within the range of the previous diagonal-element index the the correct one
        if(targetHullElementIndex < prevHullDiagElementIndex) {

            // The value does not fit in the existing ENV, enlarge it
            if(value != 0) {
                
                // Recalculate the difference so we know how many elements we need to add
                difference = (prevHullDiagElementIndex - targetHullElementIndex)+1;
                
                // Enlarge the ENV
                this.env.enlargeAt((prevHullDiagElementIndex+1), difference);
                
                // Add the enlarge vector to ENV
                this.env.set((prevHullDiagElementIndex+1), value);
                
                // Adjust the index of all affected diagonal-pointer vector entries
                for(int i = row; i < this.diagPointer.length; i++) {
                    
                    this.diagPointer[i] = this.diagPointer[i] + difference;
                }
            }
            
        } else if(targetHullElementIndex == prevHullDiagElementIndex) {
            
            if(value != 0) {
                
                // Add the value
                this.env.add((prevHullDiagElementIndex+1), value);
                
                // Adjust the index of all affected diagonal-pointer vector entries
                for(int i = row; i < this.diagPointer.length; i++) {
                    
                    this.diagPointer[i] = this.diagPointer[i] + 1;
                }
            }
            
        } else {

            // The value can be stored within the existing ENV (even if the value is 0 because it might overwrite an existing value)
            this.env.set(targetHullElementIndex, value);            
        }
    }
    
    /**
     * Multiply all values in the matrix by the given value
     * @param multiplicator, the value to multiply the matrix with
     * @return the new matrix, the original matrix is not changed
     */
    public Matrix mult(double multiplicator) {

        // The resulting matrix
        MatrixHull result = new MatrixHull(this.name, this.size());
        
        // Run through all rows and columns in the matrix
        for(int i = 1 ; i <= this.size(); i++)
        {
            for(int j = 1; j <= i; j++)
            {
                result.setValue(i, j, this.getValue(i, j) * multiplicator);
            }
        }
        
        // Return the new matrix
        return result;
    }
    
    /**
     * Multiply this Matrix with the given vector (have to be the same size)
     * @param m, the vector
     * @return Vector, the result vector
     */
    public Vector mult(Vector m) {
    	
    	// Create the result vector
    	Vector result = new Vector("Result of Matrix * Vector", this.size());
    	
    	// Check the sizes
    	if(m.size() != this.size()) {
    		
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Cannot multiply Matrix with the given Vector, the sizes do not match !");
            }
            
            return result;
    	}
    	
    	for(int i = 1; i <= this.size(); i++) {
    		
    		double resultValue = 0;
    		
    		for(int j = 1; j <= this.size(); j++) {
    			
    			resultValue += m.getValue(j) * this.getValue(i, j);
    		}
    		
    		result.setValue(i, resultValue);
    	}
    	
    	return result;
    }
    
    /**
     * Do the cholesky for this matrix
     * @return the matrix L as result of the cholesky
     */
    public Matrix cholesky() throws CholeskyFailedException {
        
        // Create the new Matrix L we are going to create
        MatrixHull L = new MatrixHull("Cholesky L for Matrix " + this.name, this.size); 
                
        // Calculate how many percentage one line makes up
        double oneLinePercentage = 100.0 / this.size;
        
        // Calculate how many lines make up one percent
        double onePercentLines = this.size / 100.0;
        
        // Where we store the current actual percentage
        double currentPercentage = 0.0;
        
        // Initialize variables here to save allocation time when running
        int startIndexForK = -1;
        int startIndexForJ = -1;
        
        // Run through all rows in the matrix
        for(int k = 1; k <= this.size; k++) {

            // Calculate the current percentage of line already done in the CHOLESKY
            currentPercentage = (k-1) * oneLinePercentage;

            // Make a percentage output about every 2% of completed lines
            if(k%(onePercentLines*2.0) < 1) {
                
                if(localLogger.isEnabledFor(Level.INFO)) {
                    localLogger.log(Level.INFO, "Cholesky at " + ApplicationConstants.convertPercentageToString(currentPercentage) + " %");
                }
            }

            // Get the value from the matrix and check if the value is greater than 0 (a MUST HAVE)
            double valueA = this.getValue(k, k);
            
            if(valueA <= 0) {

                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Error while computing the CHOLESKY: Main Diagonal Element is negative at index: " + k);
                }
                
                throw new CholeskyFailedException();
            }
            
            // Calculate the sum for the square root (Sum from i=1 till (k-1) l[k][i]^2)
            double sum = 0;
            
            // PERFORMANCE: Check where the first column starts, which we can find out at this point to save calculation time
            startIndexForK = this.getFirstNonZeroIndexForRow(k);

            for(int i = startIndexForK; i <= (k-1); i++) {
                sum = sum + Math.pow(L.getValue(k, i), 2);
            }

            double valueL = Math.sqrt(valueA - sum);

            // Set the new L value at l[k][k]
            L.setValue(k, k, valueL);

            // Calculate l[j][k] = (1/l[k][k]) * (a[j][k] - sum[from i=1 till k-1](l[j][i] * l[k][i])
            for(int j = (k+1); j <= this.size; j++) {

                // Reset the sum and calculate it
                sum = 0;
                
                // PERFORMANCE: Check where the first column starts, which we can find out at this point to save calculation time
                startIndexForJ = this.getFirstNonZeroIndexForRow(j);
                
                // PERFORMANCE: Determine which one is the maximum index because we need both in a unified for-loop               
                 for(int i = Math.max(startIndexForK, startIndexForJ); i <= (k-1); i++) {
                     sum = sum + (L.getValue(j, i) * L.getValue(k, i));
                }
                
                // Calculate the new L-value
                double newValueL = (this.getValue(j, k) - sum) / valueL;
                
                // Set l[j][k]
                L.setValue(j, k, newValueL);
            }
        }

        return L;
    }
    
    /**
     * Test if the Matrix is symmetric
     * @return false if it is not symmetric, true otherwise
     */
    public boolean isSymmetric() {
        
        // This kind of matrix is ALWAYS symmetric
        return true;
    }
    
    /**
     * Transponate the Matrix
     * @return the new transponated Matrix
     */
    public Matrix transponate() {
    	
    	// Create the new Matrix
    	MatrixHull result = new MatrixHull(this);
    	
    	return result;
    }
    
    /**
     * Multiply this Matrix with the given Matrix (have to be the same size)
     * @param m, the matrix
     * @return Matrix, the result matrix
     */
    public Matrix mult(Matrix m) {
    	
    	// Create the new Matrix
    	MatrixHull result = new MatrixHull(this.name + " * " + m.getName(), this.size());
    	
    	for(int i = 1; i <= result.size(); i++) {
    		for(int j = i; j <= result.size(); j++) {
    			
    			double sum = 0;
    			
    			for(int sumIndex = 1; sumIndex <= this.size(); sumIndex++) {
    				
    				sum = sum + this.getValue(i, sumIndex) * m.getValue(sumIndex, j);
    			}
    			
    			result.setValue(i, j, sum);
    		}
    	}
    	
    	return result;
    }
    
    /**
     * Store a whole line in the matrix at the given line number
     * @param lineNumber, where to store the line
     * @param line, the array holding the line's data
     */
    public void storeLine(int lineNumber,  double[] line)
    {
        // Run through the array and store each value in the given line
        for(int i = 1; i <= lineNumber; i++){      
            
            // Set the proper value, we need to add 1 to the index i because our matrix index is starting at 1 not 0
            this.setValue(lineNumber, i, line[i-1]);
        }
    }
    
    /**
     * Set a whole column to the given value in the matrix at the given column number
     * @param columnNumber, where to store the column
     * @param value, the sole value we want to fill the column with
     */
    public void setLineAndColumn(int index, double value)
    {
        // Run through the array and store the value in the given row
        int reduceCounter = 0;
        
        for(int i = 1; i <= this.size; i++){      
            
            // Set the proper value, we need to add 1 to the index i because our matrix index is starting at 1 not 0
            this.setValue(i, index, value);
        }
    }
    
    /**
     * Add a given matrix to this matrix
     * @param m, the matrix to add
     * @return the new matrix, the original matrix is not changed
     */
    public Matrix add(Matrix m) {
        
        // Check if both matrices have the same size
        if(this.size() != m.size()) {
            
            if(localLogger.isEnabledFor(Level.ERROR)) {
                localLogger.log(Level.ERROR, "Adding one matrix to another failed ! Not the same size !");
            }
            
            return null;
        }
        
        // Create a new Matrix holding the result
        MatrixHull result = new MatrixHull(this.name, this.size());
        
        // Run through all rows and columns in the matrix
        for(int i = 1 ; i <= this.size(); i++)
        {
            for(int j = 1; j <= i; j++)
            {
                result.setValue(i, j, this.getValue(i, j) + m.getValue(i, j));
            }
        }
        
        // Return the resulting matrix
        return result;       
    }
    
    /**
     * Test if the main diagonal of this matrix is not zero
     * @return true if it is not zero, false otherwise
     */
    public boolean isMainDiagonalNotZero() {

        // Check if the main diagonal is not zero
        for(int i = 1; i <= this.size; i++) {
            
            if(this.getValue(i, i) == 0) {
                
                // The main diagonal element is zero
                return false;
            }
        }
        
        // No main diagonal element has been zero
        return true;
    }
    
    /**
     * Get the number of rows and columns in the matrix (square matrix only!)
     * @return the number of rows and columns
     */
    public int size() {
      return this.size;  
    }
    
    /**
     * Retrieve the name of this matrix
     * @return the name
     */
    public String getName() {
        return this.name;
    }
    
    /**
     * Set a new name for this matrix
     * @param name
     */
    public void setName(String name) {
        this.name = name;
    }
    
    /**
     * Convert all data in the matrix to a quadratic double array
     * @return double[][]
     */
    public double[][] asArray() {
    	
    	// Allocate the array
    	double[][] array = new double[this.size()][this.size()];
    	
    	// Fill the array with the values
    	for(int i = 1; i <= this.size(); i++) {
    		
    		for(int j = 1; j <= this.size(); j++) {
    			
    			array[i-1][j-1] = this.getValue(i,j);
    		}
    	}
    	
    	return array;    	
    }
    
    /**
     * Clone the matrix
     * @return the cloned matrix
     */
     public Object clone() {
         
         // Make a clone matrix using the copy constructor
         MatrixHull clone = new MatrixHull(this);
         
         // Return the cloned matrix
         return clone;
     }
     
     /* (non-Javadoc)
      * @see java.lang.Object#toString()
      */
     public String toString() 
     {
         // Create a new empty string
         String stringMatrix = "[MatrixHull - \nName: " + this.name + "\nSize: " + this.size + "\nData: \n";
         
         // Build a string-matrix to display
         for(int i = 1; i <= this.size(); i++)
         {
             for(int j = 1; j <= this.size(); j++)
             {
                 stringMatrix = stringMatrix + this.getValue(i, j) + " \t ";
             }
             stringMatrix = stringMatrix + '\n';
         }
         
         return stringMatrix + "]";
     } 
     
     /**
      * Variable toString method
      * @param number, the number of rows and columns to display (starting at the upper left)
      * @return String
      */
     public String toString(int number) 
     {
         // Create a new empty string
         String stringMatrix = "[MatrixHull - \nName: " + this.name + "\nSize: " + this.size + "\nData: \n";
         
         // Check if the required display size is bigger than the actual matrix size
         if(number > this.size()) {
         		// Reset the number to our current size
         		number = this.size();
         }
         
         
         // Build a string-matrix to display
         for(int i = 1; i <= number; i++)
         {
             for(int j = 1; j <= number; j++)
             {
                 stringMatrix = stringMatrix + this.getValue(i, j) + " \t ";
             }
             stringMatrix = stringMatrix + '\n';
         }
         
         return stringMatrix + "]";
     } 

     /**
      * Retrieve the amount of required memory in this matrix (in byte)
      * @return the amount of required memory in byte
      */
     public long getRequiredMemory() {
         
         long bytes = 0;
         
//         // Add the size-integer to the required amount of bytes
//         bytes += ApplicationConstants.SIZE_INT;
         
         // Add the one-dimensional diag-pointer to the required amount of bytes
         bytes += ApplicationConstants.SIZE_INT * this.size;

         // Add the required memory of the ENVHullMatrix to the required amount of bytes
         bytes += this.env.getRequiredMemory();

//         // Add the string-name to the required amount of bytes
//         bytes += ApplicationConstants.SIZE_CHAR * this.name.length();
         
         return bytes;
     }
     
     /**
      * Retrieve  the diag pointer from the outside (used by the copy constructor)
      * @return diagPointer, the diag pointer
      */
     public int[] getDiagPointer() {
         return this.diagPointer;
     }
     
     /**
      * Retrieve the ENVHullMatrix object from the outside (used by the copy constructor)
      * @return env, the env object
      */
     public ENVHullMatrix getENV() {
         return this.env;
     }
     
     /**
      * Retrieve the first index in the given row where we have a valid value (means != 0)
      * @param row, the row index (starting at 1)
      * @return the first index with a valid (means != 0) value
      */
     protected int getFirstNonZeroIndexForRow(int row) throws IndexOutOfBoundsException {
         
         // Check if we got a proper index
         if(row > this.size || row <= 0)
 	    {
             throw new IndexOutOfBoundsException();
 	    }
         
         // Calculate the index by retrieving the number of elements saved for that row, subtracting it from the row index + 1 (+1 because we start at index 1 not 0)
         return ((row - (this.diagPointer[row] - this.diagPointer[row-1]))+1);
     }
}
