/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.constants;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import fhm.edu.fem.calc.objects.matrix.Matrix;
import fhm.edu.fem.calc.objects.matrix.MatrixFactory;
import fhm.edu.fem.util.logging.LoggingManager;

/**
 * Class Description:
 *		Stores all S-Matrices
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class SMatrixConstants {
    
	/* Our local logger object */
	protected static Logger localLogger = LoggingManager.getLogger(SMatrixConstants.class.getName());
    
    // The approach constants
    public static final int APPROACH_LINEAR = 2;
    public static final int APPROACH_QUADRATIC = 3;
    
    // The figure constants
    public static final int FIGURE_TRIANGLE = 3;
    public static final int FIGURE_QUADRAT = 4;
    
    // All S-Matrices
    private static Matrix linearTriangleS1 = null;
    private static Matrix linearTriangleS2 = null;
    private static Matrix linearTriangleS3 = null;
    private static Matrix linearTriangleS4 = null;
    
    private static Matrix linearParallelogramS1 = null;
    private static Matrix linearParallelogramS2 = null;
    private static Matrix linearParallelogramS3 = null;
    private static Matrix linearParallelogramS4 = null;
    
    private static Matrix quadraticTriangleS1 = null;
    private static Matrix quadraticTriangleS2 = null;
    private static Matrix quadraticTriangleS3 = null;
    private static Matrix quadraticTriangleS4 = null;
    
    private static Matrix quadraticParallelogramS1 = null;
    private static Matrix quadraticParallelogramS2 = null;
    private static Matrix quadraticParallelogramS3 = null;
    private static Matrix quadraticParallelogramS4 = null;
    
    private static Matrix quadraticTriangleS2Star = null;
    private static Matrix quadraticParallelogramS2Star = null;
    
	/** 
	 * To make sure noone is using this constructor.
	 * Else the compiler would create a public one.
	 * */
	private SMatrixConstants() {}
	
	/**
	 * Retrieve a certain S-Matrix
	 * @param approach, the approach either LINEAR_APPROACH or QUADRATIC_APPROACH
	 * @param figure, the figure either TRIANGLE, or PARALLELOGRAM
	 * @param sIndex
	 * @return
	 */
	public static Matrix getSMatrix(int approach, int figure, int sIndex) {
	    
	    // Check which approach the user wants
	    switch(approach) {
	    
	    	case APPROACH_LINEAR:
	    	    
	    	    // We got a linear approach, check which figure the user wants
	    	    if(figure == FIGURE_TRIANGLE) {
	    	        
	    	        // Return the correct S-Matrix
	    	        return getLinearTriangle(sIndex);
	    	    }
	    	    else if(figure == FIGURE_QUADRAT) {
	    	        
	    	        // Return the correct S-Matrix
	    	        return getLinearParallelogram(sIndex);
	    	    }
	    	    else {
	    	        
	    	        // Unknown figure
	                if(localLogger.isEnabledFor(Level.ERROR)) {
	                    localLogger.log(Level.ERROR, "Requested S-Matrix for unknown figure: " + figure);
	                }
	                return null;
	    	    }
	    	    
	    	
	    	case APPROACH_QUADRATIC:
	    	    
	    	    // We got a linear approach, check which figure the user wants
	    	    if(figure == FIGURE_TRIANGLE) {
	    	        
	    	        // Return the correct S-Matrix
	    	        return getQuadraticTriangle(sIndex);
	    	    }
	    	    else if(figure == FIGURE_QUADRAT) {
	    	        
	    	        // Return the correct S-Matrix
	    	        return getQuadraticParallelogram(sIndex);
	    	    }
	    	    else {
	    	        
	    	        // Unknown figure
	                if(localLogger.isEnabledFor(Level.ERROR)) {
	                    localLogger.log(Level.ERROR, "Requested S-Matrix for unknown figure: " + figure);
	                }
	                return null;
	    	    }
	    	    	    	
	    	default:
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Requested S-Matrix for unknown approach: " + approach);
                }
                
                return null;
	    }
	}
	
	/**
	 * Get a S-Matrix for the linear approach using a triangle
	 * @param sIndex, the index of the S-Matrix (1-4)
	 * @return the S-Matrix
	 */
	private static Matrix getLinearTriangle(int sIndex) {
	    
	    // Check which S-Matrix the user needs
	    switch(sIndex) {
	    
	    	case 1:
	    	    
	    	    // Check if we have already built the matrix
	    	    if(linearTriangleS1 == null) {
	    	        
	    	        // Build the matrix
	    	        buildLinearTriangleS1();
	    	    }
	    	    
	    	    return linearTriangleS1;
	    	
	    	case 2:
	    	    
	    	    // Check if we have already built the matrix
	    	    if(linearTriangleS2 == null) {
	    	        
	    	        // Build the matrix
	    	        buildLinearTriangleS2();
	    	    }
	    	    
	    	    return linearTriangleS2;
	    	
	    	case 3:
	    	    
	    	    // Check if we have already built the matrix
	    	    if(linearTriangleS3 == null) {
	    	        
	    	        // Build the matrix
	    	        buildLinearTriangleS3();
	    	    }
	    	    
	    	    return linearTriangleS3;
	    	
	    	case 4:
	    	    // Check if we have already built the matrix
	    	    if(linearTriangleS4 == null) {
	    	        
	    	        // Build the matrix
	    	        buildLinearTriangleS4();
	    	    }
	    	    
	    	    return linearTriangleS4;
	    	
	    	default:
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Requested S-Matrix for unknown S-Index: " + sIndex);
                }
                return null;
	    }
	}
	
	/**
	 * Get a S-Matrix for the linear approach using a parallelogram
	 * @param sIndex, the index of the S-Matrix (1-4)
	 * @return the S-Matrix
	 */
	private static Matrix getLinearParallelogram(int sIndex) {
	   
	    switch(sIndex) {
	    
	    	case 1:
	    	    
	    	    // Check if we have already built the matrix
	    	    if(linearParallelogramS1 == null) {
	    	        
	    	        // Build the matrix
	    	        buildLinearParallelogramS1();
	    	    }
	    	    
	    	    return linearParallelogramS1;   	    
	    	    
	    	case 2:
	    	    // Check if we have already built the matrix
	    	    if(linearParallelogramS2 == null) {
	    	        
	    	        // Build the matrix
	    	        buildLinearParallelogramS2();
	    	    }
	    	    
	    	    return linearParallelogramS2;
	    	    
	    	case 3:
	    	    
	    	    // Check if we have already built the matrix
	    	    if(linearParallelogramS3 == null) {
	    	        
	    	        // Build the matrix
	    	        buildLinearParallelogramS3();
	    	    }
	    	    
	    	    return linearParallelogramS3;
	    	    
	    	case 4:
	    	    	    
	    	    // Check if we have already built the matrix
	    	    if(linearParallelogramS4== null) {
	    	        
	    	        // Build the matrix
	    	        buildLinearParallelogramS4();
	    	    }
	    	    
	    	    return linearParallelogramS4;
	    	    
	    	default:
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Requested S-Matrix for unknown S-Index: " + sIndex);
                }
                return null;
	    	        
	    }
	}
	
	/**
	 * Get a S-Matrix for the quadratic approach using a triangle
	 * @param sIndex, the index of the S-Matrix (1-4)
	 * @return the S-Matrix
	 */
	private static Matrix getQuadraticTriangle(int sIndex) {

	    switch(sIndex) {
	    
	    	case 1:
	    	    // Check if we have already built the matrix
	    	    if(quadraticTriangleS1 == null) {
	    	        
	    	        // Build the matrix
	    	        buildQuadraticTriangleS1();
	    	    }
	    	    
	    	    return quadraticTriangleS1;  
	    	    
	    	case 2:
	    	    // Check if we have already built the matrix
	    	    if(quadraticTriangleS2 == null) {
	    	        
	    	        // Build the matrix
	    	        buildQuadraticTriangleS2();
	    	    }
	    	    
	    	    return quadraticTriangleS2; 
	    	    
	    	case 3:
	    	    // Check if we have already built the matrix
	    	    if(quadraticTriangleS3 == null) {
	    	        
	    	        // Build the matrix
	    	        buildQuadraticTriangleS3();
	    	    }
	    	    
	    	    return quadraticTriangleS3; 
	    	    
	    	case 4:
	    	    // Check if we have already built the matrix
	    	    if(quadraticTriangleS4 == null) {
	    	        
	    	        // Build the matrix
	    	        buildQuadraticTriangleS4();
	    	    }
	    	    
	    	    return quadraticTriangleS4;
	    	   default:
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Requested S-Matrix for unknown S-Index: " + sIndex);
                }
                return null;
	    	    
	    }
	}
	
	/**
	 * Get a S-Matrix for the quadratic approach using a parallelogram
	 * @param sIndex, the index of the S-Matrix (1-4)
	 * @return the S-Matrix
	 */
	private static Matrix getQuadraticParallelogram(int sIndex) {
	    
	    switch(sIndex) {
	    
	    	case 1:
	    	    // Check if we have already built the matrix
	    	    if(quadraticParallelogramS1 == null) {
	    	        
	    	        // Build the matrix
	    	        buildQuadraticParallelogramS1();
	    	    }
	    	    
	    	    return quadraticParallelogramS1;  
	    	    
	    	case 2:
	    	    // Check if we have already built the matrix
	    	    if(quadraticParallelogramS2 == null) {
	    	        
	    	        // Build the matrix
	    	        buildQuadraticParallelogramS2();
	    	    }
	    	    
	    	    return quadraticParallelogramS2;  
	    	    
	    	case 3:
	    	    // Check if we have already built the matrix
	    	    if(quadraticParallelogramS3 == null) {
	    	        
	    	        // Build the matrix
	    	        buildQuadraticParallelogramS3();
	    	    }
	    	    
	    	    return quadraticParallelogramS3;  
	    	    
	    	case 4:
	    	    // Check if we have already built the matrix
	    	    if(quadraticParallelogramS4 == null) {
	    	        
	    	        // Build the matrix
	    	        buildQuadraticParallelogramS4();
	    	    }
	    	    
	    	    return quadraticParallelogramS4;  
	    	    
	    	default:
	                if(localLogger.isEnabledFor(Level.ERROR)) {
	                    localLogger.log(Level.ERROR, "Requested S-Matrix for unknown S-Index: " + sIndex);
	                }
	                return null;
	    }
	}
	
	/**
	 * Get the S2*-Matrix for the quadratic approach using a parallelogram in the Spanner (FEM6)
	 * @return the S2*-Matrix
	 */
	public static Matrix getQuadraticParallelogramSpannerS2Star() {
	    
	    // Check if we have already built the matrix
	    if(quadraticParallelogramS2Star == null) {
	        
	        // Build the matrix
	        buildQuadraticParallelogramS2Star();
	    }
	    
	    return quadraticParallelogramS2Star;  
	}
	
	/**
	 * Get the S2*-Matrix for the quadratic approach using a triangle in the Spanner (FEM6)
	 * @return the S2*-Matrix
	 */
	public static Matrix getQuadraticTriangleSpannerS2Star() {
	    
	    // Check if we have already built the matrix
	    if(quadraticTriangleS2Star == null) {
	        
	        // Build the matrix
	        buildQuadraticTriangleS2Star();
	    }
	    
	    return quadraticTriangleS2Star;  
	}
	
	/**
	 * Build the S1-Matrix for the linear approach using a triangle
	 */
	private static void buildLinearTriangleS1() {
	    
	    // Init the matrix
	    linearTriangleS1 = MatrixFactory.createUserDefinedMatrix("Linear Triangle S1", 3);
	    
	    // Store all values in the matrix
	    linearTriangleS1.storeLine(1, new double[]{1,-1,0});
	    linearTriangleS1.storeLine(2, new double[]{-1, 1, 0});
	    linearTriangleS1.storeLine(3, new double[]{0, 0, 0});
	    
	    // Factorize
	    linearTriangleS1 = linearTriangleS1.mult(1.0/2.0);
	}
	
	/**
	 * Build the S2-Matrix for the linear approach using a triangle
	 */
	private static void buildLinearTriangleS2() {
	    
	    // Init the matrix
	    linearTriangleS2 = MatrixFactory.createUserDefinedMatrix("Linear Triangle S2", 3);
	    
	    // Store all values in the matrix
	    linearTriangleS2.storeLine(1, new double[]{2,-1,-1});
	    linearTriangleS2.storeLine(2, new double[]{-1, 0, 1});
	    linearTriangleS2.storeLine(3, new double[]{-1, 1, 0});
	    
	    // Factorize
	    linearTriangleS2 = linearTriangleS2.mult(1.0/2.0);
	}

	/**
	 * Build the S3-Matrix for the linear approach using a triangle
	 */
	private static void buildLinearTriangleS3() {
	    
	    // Init the matrix
	    linearTriangleS3 = MatrixFactory.createUserDefinedMatrix("Linear Triangle S3", 3);
	    
	    // Store all values in the matrix
	    linearTriangleS3.storeLine(1, new double[]{1, 0, -1});
	    linearTriangleS3.storeLine(2, new double[]{0, 0, 0});
	    linearTriangleS3.storeLine(3, new double[]{-1, 0, 1});    
	    
	    // Factorize
	    linearTriangleS3 = linearTriangleS3.mult(1.0/2.0);
	}
	
	/**
	 * Build the S4-Matrix for the linear approach using a triangle
	 */
	private static void buildLinearTriangleS4() {
		    
		// Init the matrix
		linearTriangleS4 = MatrixFactory.createUserDefinedMatrix("Linear Triangle S4", 3);
		    
		// Store all values in the matrix
	    linearTriangleS4.storeLine(1, new double[]{2, 1, 1});
	    linearTriangleS4.storeLine(2, new double[]{1, 2, 1});
	    linearTriangleS4.storeLine(3, new double[]{1, 1, 2});	    
	    
	    // Factorize
	    linearTriangleS4 = linearTriangleS4.mult(1.0/24.0);
	}
	
	
	/**
	 * Build the S1-Matrix for the linear approach using a parallelogram
	 */
	private static void buildLinearParallelogramS1() {
	    
	    // Init the matrix
	   linearParallelogramS1 = MatrixFactory.createUserDefinedMatrix("Linear Parallelogram S1", 4);
	    
	    // Store all values in the matrix
	   linearParallelogramS1.storeLine(1, new double[]{2,-2,-1,1});
	   linearParallelogramS1.storeLine(2, new double[]{-2,2,1,-1});
	   linearParallelogramS1.storeLine(3, new double[]{-1,1,2,-2});
	   linearParallelogramS1.storeLine(4, new double[]{1,-1,-2,2});
	   
	    // Factorize
	   linearParallelogramS1 = linearParallelogramS1.mult(1.0/6.0);
	}
	
	/**
	 * Build the S2-Matrix for the linear approach using a parallelogram
	 */
	private static void buildLinearParallelogramS2() {
	    
	    // Init the matrix
	   linearParallelogramS2 = MatrixFactory.createUserDefinedMatrix("Linear Parallelogram S2", 4);
	    
	    // Store all values in the matrix
	   linearParallelogramS2.storeLine(1, new double[]{1,0,-1,0});
	   linearParallelogramS2.storeLine(2, new double[]{0,-1,0,1});
	   linearParallelogramS2.storeLine(3, new double[]{-1,0,1,0});
	   linearParallelogramS2.storeLine(4, new double[]{0,1,0,-1});
	   
	    // Factorize
	   linearParallelogramS2 = linearParallelogramS2.mult(1.0/2.0);
	}
	
	/**
	 * Build the S3-Matrix for the linear approach using a parallelogram
	 */
	private static void buildLinearParallelogramS3() {
	    
	    // Init the matrix
	   linearParallelogramS3 = MatrixFactory.createUserDefinedMatrix("Linear Parallelogram S3", 4);
	    
	    // Store all values in the matrix
	   linearParallelogramS3.storeLine(1, new double[]{2,1,-1,-2});
	   linearParallelogramS3.storeLine(2, new double[]{1,2,-2,-1});
	   linearParallelogramS3.storeLine(3, new double[]{-1,-2,2,1});
	   linearParallelogramS3.storeLine(4, new double[]{-2,-1,1,2});
	   
	    // Factorize
	   linearParallelogramS3 = linearParallelogramS3.mult(1.0/6.0);
	}
	
	/**
	 * Build the S4-Matrix for the linear approach using a parallelogram
	 */
	private static void buildLinearParallelogramS4() {
	    
	    // Init the matrix
	   linearParallelogramS4 = MatrixFactory.createUserDefinedMatrix("Linear Parallelogram S4", 4);
	    
	    // Store all values in the matrix
	   linearParallelogramS4.storeLine(1, new double[]{4,2,1,2});
	   linearParallelogramS4.storeLine(2, new double[]{2,4,2,1});
	   linearParallelogramS4.storeLine(3, new double[]{1,2,4,2});
	   linearParallelogramS4.storeLine(4, new double[]{2,1,2,4});
	   
	    // Factorize
	   linearParallelogramS4 = linearParallelogramS4.mult(1.0/36.0);
	}


	/**
	 * Build the S1-Matrix for the quadratic approach using a triangle
	 */
	private static void buildQuadraticTriangleS1(){
	    
	    // Init the matrix
	    quadraticTriangleS1 = MatrixFactory.createUserDefinedMatrix("Quadratic Triangle S1", 6);
	    
	    // Store all values in the matrix
	    quadraticTriangleS1.storeLine(1, new double[]{3,1,0,-4,0,0});
	    quadraticTriangleS1.storeLine(2, new double[]{1,3,0,-4,0,0});
	    quadraticTriangleS1.storeLine(3, new double[]{0,0,0,0,0,0});
	    quadraticTriangleS1.storeLine(4, new double[]{-4,-4,0,8,0,0});
	    quadraticTriangleS1.storeLine(5, new double[]{0,0,0,0,8,-8});
	    quadraticTriangleS1.storeLine(6, new double[]{0,0,0,0,-8,8});
	    
	    // Factorize
	    quadraticTriangleS1 = quadraticTriangleS1.mult(1.0/6.0);	    
	}
	
	/**
	 * Build the S2-Matrix for the quadratic approach using a triangle
	 */
	private static void buildQuadraticTriangleS2(){
	    
	    // Init the matrix
	    quadraticTriangleS2 = MatrixFactory.createUserDefinedMatrix("Quadratic Triangle S2", 6);
	    
	    // Store all values in the matrix
	    quadraticTriangleS2.storeLine(1, new double[]{6,1,1,-4,0,-4});
	    quadraticTriangleS2.storeLine(2, new double[]{1,0,-1,-4,4,0});
	    quadraticTriangleS2.storeLine(3, new double[]{1,-1,0,0,4,-4});
	    quadraticTriangleS2.storeLine(4, new double[]{-4,-4,0,8,-8,8});
	    quadraticTriangleS2.storeLine(5, new double[]{0,4,4,-8,8,-8});
	    quadraticTriangleS2.storeLine(6, new double[]{-4,0,-4,8,-8,8});
	    
	    // Factorize
	    quadraticTriangleS2 = quadraticTriangleS2.mult(1.0/6.0);	
	}
	
	/**
	 * Build the S3-Matrix for the quadratic approach using a triangle
	 */
	private static void buildQuadraticTriangleS3(){
	    
	    // Init the matrix
	    quadraticTriangleS3 = MatrixFactory.createUserDefinedMatrix("Quadratic Triangle S3", 6);
	    
	    // Store all values in the matrix
	    quadraticTriangleS3.storeLine(1, new double[]{3,0,1,0,0,-4});
	    quadraticTriangleS3.storeLine(2, new double[]{0,0,0,0,0,0});
	    quadraticTriangleS3.storeLine(3, new double[]{1,0,3,0,0,-4});
	    quadraticTriangleS3.storeLine(4, new double[]{0,0,0,8,-8,0});
	    quadraticTriangleS3.storeLine(5, new double[]{0,0,0,-8,8,0});
	    quadraticTriangleS3.storeLine(6, new double[]{-4,0,-4,0,0,8});
	    
	    // Factorize
	    quadraticTriangleS3 = quadraticTriangleS3.mult(1.0/6.0);	
	}
	
	/**
	 * Build the S4-Matrix for the quadratic approach using a triangle
	 */
	private static void buildQuadraticTriangleS4(){
	    
	    // Init the matrix
	    quadraticTriangleS4 = MatrixFactory.createUserDefinedMatrix("Quadratic Triangle S4", 6);
	    
	    // Store all values in the matrix
	    quadraticTriangleS4.storeLine(1, new double[]{6,-1,-1,0,-4,0});
	    quadraticTriangleS4.storeLine(2, new double[]{-1,6,-1,0,0,-4});
	    quadraticTriangleS4.storeLine(3, new double[]{-1,-1,6,-4,0,0});
	    quadraticTriangleS4.storeLine(4, new double[]{0,0,-4,32,16,16});
	    quadraticTriangleS4.storeLine(5, new double[]{-4,0,0,16,32,16});
	    quadraticTriangleS4.storeLine(6, new double[]{0,-4,0,16,16,32});
	    
	    // Factorize
	    quadraticTriangleS4 = quadraticTriangleS4.mult(1.0/360.0);	
	}
	
	/**
	 * Build the S1-Matrix for the quadratic approach using a parallelogram
	 */
	private static void buildQuadraticParallelogramS1() {
	    
	    // Init the matrix
	    quadraticParallelogramS1 = MatrixFactory.createUserDefinedMatrix("Quadratic Parallelogram S1", 8);
	    
	    // Store all values in the matrix
	    quadraticParallelogramS1.storeLine(1, new double[]{52, 28, 23, 17, -80, -6, -40, 6});
	    quadraticParallelogramS1.storeLine(2, new double[]{28, 52, 17, 23, -80, 6, -40, -6});
	    quadraticParallelogramS1.storeLine(3, new double[]{23, 17, 52, 28, -40, 6, -80, -6});
	    quadraticParallelogramS1.storeLine(4, new double[]{17, 23, 28, 52, -40, -6, -80, 6});
	    quadraticParallelogramS1.storeLine(5, new double[]{-80, -80, -40, -40, 160, 0, 80, 0});
	    quadraticParallelogramS1.storeLine(6, new double[]{-6, 6, 6, -6, 0, 48, 0, -48});
	    quadraticParallelogramS1.storeLine(7, new double[]{-40, -40, -80, -80, 80, 0, 160, 0});
	    quadraticParallelogramS1.storeLine(8, new double[]{6, -6, -6, 6, 0, -48, 0, 48});
	    
	    // Factorize
	    quadraticParallelogramS1 = quadraticParallelogramS1.mult(1.0/90.0);	
	}
	
	/**
	 * Build the S2-Matrix for the quadratic approach using a parallelogram
	 */
	private static void buildQuadraticParallelogramS2() {
	    
	    // Init the matrix
	    quadraticParallelogramS2 = MatrixFactory.createUserDefinedMatrix("Quadratic Parallelogram S2", 8);
	    
	    // Store all values in the matrix
	    quadraticParallelogramS2.storeLine(1, new double[]{85, 0, 35, 0, -40, -20, -20, -40});
	    quadraticParallelogramS2.storeLine(2, new double[]{0, -85, 0, -35, 40, 40, 20, 20});
	    quadraticParallelogramS2.storeLine(3, new double[]{35, 0, 85, 0, -20, -40, -40, -20});
	    quadraticParallelogramS2.storeLine(4, new double[]{0, -35, 0, -85, 20, 20, 40, 40});
	    quadraticParallelogramS2.storeLine(5, new double[]{-40, 40, -20, 20, 0, -80, 0, 80});
	    quadraticParallelogramS2.storeLine(6, new double[]{-20, 40, -40, 20, -80, 0, 80, 0});
	    quadraticParallelogramS2.storeLine(7, new double[]{-20, 20, -40, 40, 0, 80, 0, -80});
	    quadraticParallelogramS2.storeLine(8, new double[]{-40, 20, -20, 40, 80, 0, -80, 0});
	    
	    // Factorize
	    quadraticParallelogramS2 = quadraticParallelogramS2.mult(1.0/90.0);	
	}
	
	/**
	 * Build the S3-Matrix for the quadratic approach using a parallelogram
	 */
	private static void buildQuadraticParallelogramS3() {
	    
	    // Init the matrix
	    quadraticParallelogramS3 = MatrixFactory.createUserDefinedMatrix("Quadratic Parallelogram S3", 8);
	    
	    // Store all values in the matrix
	    quadraticParallelogramS3.storeLine(1, new double[]{52, 17, 23, 28, 6, -40, -6, -80});
	    quadraticParallelogramS3.storeLine(2, new double[]{17, 52, 28, 23, 6, -80, -6, -40});
	    quadraticParallelogramS3.storeLine(3, new double[]{23, 28, 52, 17, -6, -80, 6, -40});
	    quadraticParallelogramS3.storeLine(4, new double[]{28, 23, 17, 52, -6, -40, 6, -80});
	    quadraticParallelogramS3.storeLine(5, new double[]{6, 6, -6, -6, 48, 0, -48, 0});
	    quadraticParallelogramS3.storeLine(6, new double[]{-40, -80, -80, -40, 0, 160, 0, 80});
	    quadraticParallelogramS3.storeLine(7, new double[]{-6, -6, 6, 6, -48, 0, 48, 0});
	    quadraticParallelogramS3.storeLine(8, new double[]{-80, -40, -40, -80, 0, 80, 0, 160});
	    
	    // Factorize
	    quadraticParallelogramS3 = quadraticParallelogramS3.mult(1.0/90.0);
	}
	
	/**
	 * Build the S4-Matrix for the quadratic approach using a parallelogram
	 */
	private static void buildQuadraticParallelogramS4() {
	    
	    // Init the matrix
	    quadraticParallelogramS4 = MatrixFactory.createUserDefinedMatrix("Quadratic Parallelogram S4", 8);
	    
	    // Store all values in the matrix
	    quadraticParallelogramS4.storeLine(1, new double[]{6, 2, 3, 2, -6, -8, -8, -6});
	    quadraticParallelogramS4.storeLine(2, new double[]{2, 6, 2, 3, -6, -6, -8, -8});
	    quadraticParallelogramS4.storeLine(3, new double[]{3, 2, 6, 2, -8, -6, -6, -8});
	    quadraticParallelogramS4.storeLine(4, new double[]{2, 3, 2, 6, -8, -8, -6, -6});
	    quadraticParallelogramS4.storeLine(5, new double[]{-6, -6, -8, -8, 32, 20, 16, 20});
	    quadraticParallelogramS4.storeLine(6, new double[]{-8, -6, -6, -8, 20, 32, 20, 16});
	    quadraticParallelogramS4.storeLine(7, new double[]{-8, -8, -6, -6, 16, 20, 32, 20});
	    quadraticParallelogramS4.storeLine(8, new double[]{-6, -8, -8, -6, 20, 16, 20, 32});
	    
	    // Factorize
	    quadraticParallelogramS4 = quadraticParallelogramS4.mult(1.0/180.0);	
	}
	
	/*
	 * 
	 * SPANNER S-MATRICES
	 * 
	 ***************************/
	
	/**
	 * Build the S2Star-Matrix for the quadratic approach using a triangle
	 */
	private static void buildQuadraticTriangleS2Star(){
	    
	    // Init the matrix
	    quadraticTriangleS2Star = MatrixFactory.createUserDefinedMatrix("Quadratic Triangle S2 Star", 6);
	    
	    // Store all values in the matrix
	    quadraticTriangleS2Star.storeLine(1, new double[]{3,0,1,0,0,-4});
	    quadraticTriangleS2Star.storeLine(2, new double[]{1,0,-1,-4,4,0});
	    quadraticTriangleS2Star.storeLine(3, new double[]{0,0,0,0,0,0});
	    quadraticTriangleS2Star.storeLine(4, new double[]{-4,0,0,4,-4,4});
	    quadraticTriangleS2Star.storeLine(5, new double[]{0,0,4,-4,4,-4});
	    quadraticTriangleS2Star.storeLine(6, new double[]{0,0,-4,4,-4,4});
	    
	    // Factorize
	    quadraticTriangleS2Star = quadraticTriangleS2Star.mult(1.0/6.0);	
	}
	
	/**
	 * Build the S2Star-Matrix for the quadratic approach using a triangle
	 */
	private static void buildQuadraticParallelogramS2Star(){
	    
	    // Init the matrix
	    quadraticParallelogramS2Star = MatrixFactory.createUserDefinedMatrix("Quadratic Parallelogram S2 Star", 8);
	    
	    // Store all values in the matrix
	    quadraticParallelogramS2Star.storeLine(1, new double[]{17,-3,7,3,4,-4, -4, -20});
	    quadraticParallelogramS2Star.storeLine(2, new double[]{3,-17,-3,-7,-4,20,4,4});
	    quadraticParallelogramS2Star.storeLine(3, new double[]{7,3,17,-3,-4,-20,4,-4});
	    quadraticParallelogramS2Star.storeLine(4, new double[]{-3,-7,3,-17,4,4,-4,20});
	    quadraticParallelogramS2Star.storeLine(5, new double[]{-20, 20, -4, 4, 0, -16, 0, 16});
	    quadraticParallelogramS2Star.storeLine(6, new double[]{-4, -4, 4, 4, -16, 0, 16, 0});
	    quadraticParallelogramS2Star.storeLine(7, new double[]{-4, 4, -20, 20, 0, 16, 0, -16});
	    quadraticParallelogramS2Star.storeLine(8, new double[]{4, 4, -4, -4, 16, 0, -16, 0});
	    
	    // Factorize
	    quadraticParallelogramS2Star = quadraticParallelogramS2Star.mult(1.0/36.0);	
	}
}




