/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.convert;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import fhm.edu.fem.calc.FEMManager;
import fhm.edu.fem.calc.objects.Border;
import fhm.edu.fem.calc.objects.BorderList;
import fhm.edu.fem.calc.objects.Element;
import fhm.edu.fem.calc.objects.ElementList;
import fhm.edu.fem.calc.objects.Node;
import fhm.edu.fem.calc.objects.NodeList;
import fhm.edu.fem.util.logging.LoggingManager;

/**
 * Class Description:
 *		Used to convert FEMIN files from FEM2 to FEM3
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class FEMINConverterFEM2toFEM3 {
    
	/* Our local logger object */
	protected static Logger localLogger = LoggingManager.getLogger(FEMINConverterFEM2toFEM3.class.getName()); 
    
    /**
     * Construct a new FEMINConverterFEM2toFEM3
     */
    public FEMINConverterFEM2toFEM3 (){}
    
    /**
     * Convert the FEM2 data stored in the FEMManager into FEM3 data
     * @return true if all went well, false otherwise
     */
    public boolean convertFEM2toFEM3(){
        
        // Convert the approach
        if(!this.convertApproachFEM2toFEM3()) {
            
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Conversion failed ! Error while converting the FEM Approach !");
			}
			
			return false;
        }
        
//        // Convert nodes and elements
//        if(!this.convertNodesAndElementsFEM2toFEM3()) {
//            
//			if(localLogger.isEnabledFor(Level.ERROR)) {
//				localLogger.log(Level.ERROR, "Conversion failed ! Error while converting the Nodes and Elements !");
//			}
//			
//			return false;
//        }
//        
//        // Convert borders
//        if(!this.convertBordersFEM2toFEM3()) {
//          
//			if(localLogger.isEnabledFor(Level.ERROR)) {
//				localLogger.log(Level.ERROR, "Conversion failed ! Error while converting the Borders !");
//			}
//			
//			return false;
//        }
        
		if(localLogger.isEnabledFor(Level.INFO)) {
			localLogger.log(Level.INFO, "Conversion complete");
		}
        
        return true;
    }
    
    /**
     * Convert the approach from FEM2 to FEM3
     * @return true if all went well, false otherwise
     */
    protected boolean convertApproachFEM2toFEM3() {
        
        // Grab the fem approach
        int femApproach = FEMManager.getInstance().getFEMApproach();
        
        // Check if we got the proper approach
        if(femApproach != 2) {
            
            // The file has already been transformed            
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Conversion failed ! Only files for FEM2 can be converted. This file is FEM:" + femApproach);
			}
            
            return false;
        }
        
        // Convert the approach
        FEMManager.getInstance().setFemApproach(3);
        return true;
    }
    
    /**
     * Convert the nodes and elements from FEM2 to FEM3
     * @return true if all went well, false otherwise
     */
    protected boolean convertNodesAndElementsFEM2toFEM3() {
        
        // Grab the necessary data
        ElementList elementList = FEMManager.getInstance().getElementList();
        NodeList nodeList = FEMManager.getInstance().getNodeList();
        
        for(int i = 1; i <= elementList.size();i++){
            
            Element curElement = elementList.getElement(i);
            
            // Grab the node list size, can only be 3 for TRIANGLES or 4 for QUADS
            int nodeListSize =  curElement.getNodeList().size();
            
            for(int j = 1; j <= nodeListSize; j++){
                
                Node firstNode = curElement.getNodeList().getNodeInCorrectOrder(j);
                Node secNode = null;
                
                if((j+1) > nodeListSize ) {
                    secNode = curElement.getNodeList().getNodeInCorrectOrder(1); 
                } else {
                    secNode = curElement.getNodeList().getNodeInCorrectOrder(j+1);  
                }                    

                // Check if we already have calculated the pseudo node in some other element
                Node foundNode = elementList.doesNodeExistBetween(firstNode, secNode);
                
                // Check if no node has been found
                if(foundNode == null) {

                    // Calculate the new x/y values
                    double x1 = firstNode.getX();
                    double y1 = firstNode.getY();
                    
                    double x2 = secNode.getX();
                    double y2 = secNode.getY();
                    
                    double newX = ((x1 - x2)/2) + x2;
                    double newY = ((y1 - y2)/2 ) + y2;
                    
                    // Create the new node
                    Node newNode = new Node(nodeList.size()+1, newX, newY, false);

                    // Store the new node in the local element list
                    curElement.getNodeList().setNode(newNode);
                    
                    // Store the new node in the global node list
                    nodeList.setNode(newNode);
                    
                } else {

                    // Store the new found node in the local element list
                	if(!curElement.getNodeList().contains(foundNode)) {
                		curElement.getNodeList().setNode(foundNode);
                	}
                } 
            }
        }   
        
        return true;
    }
    
    /**
     * Convert the borders from FEM2 to FEM3
     * @return true if all went well, false otherwise
     */
    protected boolean convertBordersFEM2toFEM3() {
        
        // Grab the border data
        BorderList borderList = FEMManager.getInstance().getBorderList();
        ElementList elementList = FEMManager.getInstance().getElementList();
        
        // Run through all borders
        for(int i = 1; i <= borderList.size(); i++) {

            Border currentBorder = borderList.getBorder(i);
            
            // Grab the number of existing nodes in the border
            NodeList nodeList = currentBorder.getNodes();
            int nodeListSize = nodeList.size();
            
            // Run through all basic nodes in the border
            for(int j = 1; j <= nodeListSize; j++) {
     
                // Grab first and second current node
                Node firstNode = nodeList.getNodeInCorrectOrder(j);
           
                // Again run through all basic nodes in the border
                for(int o = j; o <= nodeListSize; o++) {
                    
                    Node secondNode = nodeList.getNodeInCorrectOrder(o);

                    // Get the node in between
                    Node midNode = elementList.doesNodeExistBetween(firstNode, secondNode);

                    if(midNode != null) {
                        
                        // Add the found middle node to the node list of the border
                    	if(!nodeList.contains(midNode)) {

                    		nodeList.setNode(midNode); 			
                    	}
                    }
                }
            }
        }
        
        return true;
    }
}
