/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.test;

import fhm.edu.fem.calc.objects.matrix.ENVHullMatrix;
import fhm.edu.fem.calc.objects.matrix.ENVVector;

/**
 * Class Description:
 *		Test Class - Envelope Hull Matrix Chain Style
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class ENVHullChainMatrix extends ENVHullMatrix {

    /**
     * Construct a new ENV for a HullMatrix with the given number of "0.0" elements
     * @param size, the size the ENV should have at startup
     */
    public ENVHullChainMatrix(int size) {
        
        // Initialize the env
        for(int i = 1; i <= size; i++) {

            ENVDataItem newItem = new ENVDataItem(0.0);
            
            // Both pointers are null at the moment because we initialize the new data items with 0.0
            newItem.setPrevious(null);
            newItem.setNext(null);
            
            // Add the item to the data vector
            this.data.add(newItem);
        }
    }
    
    /**
     * Copy Constructor
     * @param env, the env object to incorporate in the new ENVHullMatrix object
     */
    public ENVHullChainMatrix(ENVHullChainMatrix env) {
        this.data = (ENVVector)(env.getData().clone());
    }
    
    /**
     * Add a new element at a specific index of the list, causing all other elements coming after that element to be
     * shifted to the right (including the element currently at the given index)
     * @param index, the index to set the value at (the index starts at 1)
     * @param value, the new value
     */
    public void add(int index, double value) {

        // Create the new item
        ENVDataItem newItem = new ENVDataItem(value);
        
        // Set it in the main data list
        this.data.add(index-1, newItem);
        
        // Include the new item in the list
        this.includeNewItemInList(newItem, index-1);
    }
    
    /**
     * Replace an already existing value in the list with a new value at the specific index
     * @param index, the index of the element to replace (the index starts at 1)
     * @param value, the value to replace the other value with
     * @return the value we replaced
     */
    public double set(int index, double value) {

        ENVDataItem changeItem = (ENVDataItem)(this.data.get(index-1));
        double oldValue = changeItem.getValue();

        // In case the new value is zero we need to change the PREV-NEXT list
        if(value == 0.0) {
            
            if(changeItem.getValue() != 0.0) {
                
                // Remove the item from the PREV-NEXT list
                ENVDataItem prev = changeItem.getPrevious();
                ENVDataItem next = changeItem.getNext();
                
                if(prev != null) {
                    prev.setNext(next);
                }
                
                if(next != null) {
                    next.setPrevious(prev);
                }
                
                // Also set the new value to the item
                changeItem.setValue(value);
            }
            
        } else {
            
            // The new value is not zero, take a look if the old value was zero, in that case we need to include the item in the PREV-NEXT list
            if(changeItem.getValue() == 0.0) {
                
                // The old value was zero, include the newly created item in the list
                changeItem.setValue(value);
                this.includeNewItemInList(changeItem, index-1);
                
            } else {
                
                // The old value was non-zero, so just set the new value to the item
                changeItem.setValue(value);
            }
        }
        
        return changeItem.getValue();
    }
    
    /**
     * Retrieve a value from the list
     * @param index, the index of the value to retrieve (the index starts at 1)
     * @return the value
     */
    public double get(int index) {
        return ((ENVDataItem)(data.get(index-1))).getValue();
    }
    
    /**
     * Retrieve the current size
     * @return size
     */
    public int size() {
        return data.size();
    }
    
    /**
     * Enlarge the ENV by the given difference at the given index
     * @param index, the index we enlarge the ENV at (all other items are shifted to the right)
     * @param difference, the difference to enlarge the ENV
     */
    public void enlargeAt(int index, int difference) {

        // Collect all enlargements in one vector to avoid too many array-copy actions within the Java VM
        this.enlargeVector.clear();
        
        // The difference we need to enlarge between the correct and the previous hull-diagonal-element-index
        for(int i = 1; i <= difference; i++) {
            
            enlargeVector.add(new ENVDataItem(0.0));
        }
        
        // Add the enlarge vector to ENV
        this.data.addAll((index-1), enlargeVector);
    }
    
    /**
     * Clone this object
     * @return the cloned object
     */
    public Object clone() {
        
        // Use the copy constructor for cloning
        return new ENVHullChainMatrix(this);
    }
    
    protected ENVDataItem includeNewItemInList(ENVDataItem newItem, int index) {

        // In case the value is different than zero make sure to include the item in the PREV-NEXT chain
        if(newItem.getValue() != 0.0) {
            
            ENVDataItem tmpItem = null;
            boolean itemFoundAfter = false;
            boolean itemFoundBefore = false;
            
            // Search for the next non-zero item from the item's index position
            for(int i = index+1; i < this.data.size(); i++) {

                tmpItem = (ENVDataItem)(this.data.get(i));
                
                if(tmpItem.getValue() != 0.0) {
                    
                    itemFoundAfter = true;

                    break;
                }
            }
            
            // Check if we did not find the item that way
            if(!itemFoundAfter) {
                
                // Search in the other direction
                for(int i = index-1; i >= 0; i--) {
                    
                    tmpItem = (ENVDataItem)(this.data.get(i));
                    
                    if(tmpItem.getValue() != 0.0) {
                        
                        itemFoundBefore = true;
                        
                        break;
                    }
                }
            }

            // We now should have found an item, if not we add the first non-zero item to the list
            if(!itemFoundAfter && !itemFoundBefore) {
                
                // First item, nothing to do here because both pointers are still zero

            } else {

                // We found an item, include the new item properly into the non-zero-item PREV-NEXT list
                ENVDataItem prev = tmpItem.getPrevious();
                ENVDataItem next = tmpItem.getNext();

                // Depending where we found the item, insert the new item correctly
                if(itemFoundAfter) {

                    newItem.setNext(tmpItem);
                    newItem.setPrevious(prev);
                    
                    if(prev != null) {
                        prev.setNext(newItem);
                    }
                    
                    tmpItem.setPrevious(newItem);
                    
                } else {

                    newItem.setNext(next);
                    newItem.setPrevious(tmpItem);
                    
                    if(next != null) {
                        next.setPrevious(newItem);  
                    }
                    
                    tmpItem.setNext(newItem);
                }
            }
        }

        return newItem;
    }
    
    /**
     * Retrieve the next non-zero item starting from the given index
     * @param index
     * @return the index of the next item or -1 if no item could be found in that row
     */
    public int getNextNonZeroIndex(int index, int rowEndIndex) {
        
        ENVDataItem currentItem = (ENVDataItem)(this.data.get(index-1));

        // Grab the next non-zero item after the current one
        ENVDataItem nextItem = currentItem.getNext();

        // Check if we have a next non-zero item at all
        if(nextItem != null) {
            
            // Check if the next item is still in the same row
            int nextItemIndex = this.data.indexOf(nextItem, index);

            if(nextItemIndex <= rowEndIndex && nextItemIndex >= 0) {

                return (nextItemIndex+1);
            } else {

                return -1;
            }
            
        } else {
            
            // There is no next non-zero item at all, so return -1
            return -1;
        }
    }
    
    public String displayChain() {
        String chainString = "[Chain: \n";
        
        ENVDataItem tmpItem = null;
        
        // Run through the chain, find the first non-zero item
        for(int i = 0; i < this.data.size(); i++) {
            tmpItem = (ENVDataItem)(this.data.get(i));

            if(tmpItem.getValue() != 0.0) {
                break;
            }
        }
        chainString = chainString + tmpItem;
        // Found the first item, displaying chain
        ENVDataItem next = tmpItem.getNext();
        
        do{
            
            chainString = chainString + " <= " + next;
            next = next.getNext();

        } while(next != null);
        
        return chainString + "]";
    }
}
