/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.test;

import java.util.Arrays;
import java.util.Vector;

import fhm.edu.fem.constants.ApplicationConstants;

/**
 * Class Description:
 *		Test Class - Envelope Hull Matrix Speed Up
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class ENVHullMatrixSpeedUp {

    // Where we store our data
    protected Vector<Vector> data = new Vector<Vector>();
    
    // The vector used when enlarging the ENV
    protected Vector<Double> enlargeVector = new Vector<Double>();
    
    // A Vector object we use pretty often
    protected Vector<Double> subVector = null;
    
    // To remember the entry indexes
    protected int[] entryIndexes;
    
    // Common variables allocated here to save allocation time while running
    protected int entryIndex = -1;
    
    /**
     * Construct a new ENV for a HullMatrix with the given number of "0.0" elements
     * @param size, the size the ENV should have at startup
     */
    public ENVHullMatrixSpeedUp(int size) {
        
        // Create the array of entry indexes
        this.entryIndexes = new int[size];
        
        // Initialize the env
        for(int i = 1; i <= size; i++) {
            
            // Create a new subvector
            this.subVector = new Vector<Double>();
            
            // Fill the subvector with the first "0.0" Double element
            this.subVector.add(new Double(0.0));
            
            // Add the subvector to the main vector
            this.data.add(this.subVector);
            
            // Remember the entry index
            this.entryIndexes[i-1] = i;
        }
    }
    
    /**
     * Copy Constructor
     * @param env, the env object to incorporate in the new ENVHullMatrix object
     */
    public ENVHullMatrixSpeedUp(ENVHullMatrixSpeedUp env) {
        
        // Make a clone of the main vector by making a clone of each subvector
        Vector<Vector> oldData = (Vector<Vector>)(env.getData().clone());
        
        for(int i = 0; i < oldData.size(); i++) {
            
            // Create a new subvector
            this.subVector = (Vector<Double>)(oldData.get(i).clone());
            
            // Add the subvector to the main vector
            this.data.add(this.subVector);
        }
        
        // Make a clone of the entry indexes
        this.entryIndexes = (int[])(env.getEntryIndexes().clone());
    }
    
    /**
     * Add a new element at a specific index of the list, causing all other elements coming after that element to be
     * shifted to the right (including the element currently at the given index)
     * @param index, the index to set the value at (the index starts at 1)
     * @param value, the new value
     */
    public void add(int index, double value) {
        
        // Get the correct entry index
        entryIndex = this.getEntryIndexFor(index);
        
        // Grab the correct subvector
        this.subVector = (Vector<Double>)(this.data.get(entryIndex));
        
        // Add the value at the correct index in the subvector
        this.subVector.add(index - this.entryIndexes[entryIndex], new Double(value));
        
        // Adjust the entry indexes affected by the enlargement
        this.adjustEntryIndexesAfterEnlargement(index, 1);
    }
    
    /**
     * Replace an already existing value in the list with a new value at the specific index
     * @param index, the index of the element to replace (the index starts at 1)
     * @param value, the value to replace the other value with
     * @return the value we replaced
     */
    public double set(int index, double value) {
        
        // Get the correct entry index
        entryIndex = this.getEntryIndexFor(index);        

        // Grab the correct subvector
        this.subVector = (Vector<Double>)(this.data.get(entryIndex));

        // Return the value from the correct subindex in the subvector        
        return this.subVector.set(index - this.entryIndexes[entryIndex], new Double(value)).doubleValue();
    }
    
    /**
     * Retrieve a value from the list
     * @param index, the index of the value to retrieve (the index starts at 1)
     * @return the value
     */
    public double get(int index) {

        // Get the correct entry index
        entryIndex = this.getEntryIndexFor(index);
        
        // Return the value from the correct subindex in the subvector
        return ((Double)(this.data.get(entryIndex).get(index - this.entryIndexes[entryIndex]))).doubleValue();
    }
    
    /**
     * Retrieve the number of Double elements in the ENV
     * @return size
     */
    public int size() {
        
        int size = 0;
        
        // Check for the size in all subvectors
        for(int i = 0; i < this.data.size(); i++) {
            
            this.subVector = this.data.get(i);
            size += this.subVector.size();
        }
        
        return size;
    }
    
    /**
     * Enlarge the ENV by the given difference at the given index
     * @param index, the index we enlarge the ENV at (all other items are shifted to the right)
     * @param difference, the difference by which to enlarge the ENV
     */
    public void enlargeAt(int index, int difference) {

        // Collect all enlargements in one vector to avoid too many array-copy actions within the Java VM
        this.enlargeVector.clear();
        
        // The difference we need to enlarge between the correct and the previous hull-diagonal-element-index
        for(int i = 1; i <= difference; i++) {
            
            this.enlargeVector.add(new Double(0.0));
        }
        
        // Get the correct entry index
        this.entryIndex = this.getEntryIndexFor(index);
        
        // Grab the correct subvector
        this.subVector = (Vector<Double>)(this.data.get(entryIndex));

        // Add the enlarge vector to the subvector at the correct subindex
        this.subVector.addAll((index - this.entryIndexes[entryIndex]), enlargeVector);
        
        // Adjust the entry indexes affected by the enlargement
        this.adjustEntryIndexesAfterEnlargement(index, difference);
    }
    
    /**
     * Reduce the ENV by the given difference at the given index
     * @param index, the index we reduce the ENV at (all other items are shifted to the left)
     * @param difference, the difference by which to reduce the ENV
     */
    public void reduceAt(int index, int difference) {
        
        // Get the correct entry index
        this.entryIndex = this.getEntryIndexFor(index);
        
        // Grab the correct subvector
        this.subVector = (Vector<Double>)(this.data.get(entryIndex));
        
        // Reduce the subvector at the correct subindex 
        int reduceAt = index - this.entryIndexes[entryIndex];

        for(int i = 0; i < difference; i++) {
            this.subVector.remove(reduceAt);
        }
        
        // Adjust the entry indexes affected by the reduction
        this.adjustEntryIndexesAfterReduction(index, difference);
    }
    
    /**
     * Retrieve the amount of required memory in this matrix (in byte)
     * @return the amount of required memory in byte
     */
    public long getRequiredMemory() {
        
        long bytes = 0;
        
        // Add the bytes needed by all Double elements to the required amount of memory
        for(int i = 0; i < this.data.size(); i++) {
            
            this.subVector = this.data.get(i);
            bytes += ApplicationConstants.SIZE_DOUBLE * subVector.size();
        }
        
        return bytes;
    }
    
    /**
     * Retrieve the data vector from the ENV
     * @return the data vector holding "Double" objects
     */
    public Vector<Vector> getData() {
        return this.data;
    }
    
    /**
     * Retrieve the entry index array
     * @return the entryIndex array holding all entry indexes used internally in the ENV class
     */
    public int[] getEntryIndexes() {
        return this.entryIndexes;
    }
    
    /**
     * Compute the entry index for the given index and return the correct subvector
     * @param index, the real index
     * @return the entry index for the given index
     */
    protected int getEntryIndexFor(int index) {

    	// Find the entry index for the given "index"
        this.entryIndex = Arrays.binarySearch(this.entryIndexes, index);
        
        // In case the return value is negative we got the possible nearest entry index as (-(insertion point)-1)
        if(this.entryIndex < 0) {
        	
        	// Modify the value to make it fit
            this.entryIndex = (-1*entryIndex-2);
        }
        
        return this.entryIndex;
    }
    
    /**
     * Adjust all entry indexes after an enlargement
     * @param index, the original index we placed the enlargement at
     * @param difference, the amount of elements by which we expanded
     */
    protected void adjustEntryIndexesAfterEnlargement(int index, int difference) {
        
        // Compute the last "still-correct" index in the array
    	int lastCorrectIndex = this.getEntryIndexFor(index);
        
        // Run through all elements in the entry index array following the last correct index and adjust them
        for(int i = (lastCorrectIndex+1); i < this.entryIndexes.length; i++) {
            
            this.entryIndexes[i] = this.entryIndexes[i] + difference;
        }
    }
    
    /**
     * Adjust all entry indexes after a reduction
     * @param index, the original index we placed the reduction at
     * @param difference, the amount of elements by which we reduced
     */
    protected void adjustEntryIndexesAfterReduction(int index, int difference) {
        
        // Compute the last "still-correct" index in the array
        int lastCorrectIndex = this.getEntryIndexFor(index);
        
        // Run through all elements in the entry index array following the last correct index and adjust them
        for(int i = (lastCorrectIndex+1); i < this.entryIndexes.length; i++) {
            
            this.entryIndexes[i] = this.entryIndexes[i] - difference;
        }
    }
    
    /**
     * Clone this object
     * @return the cloned object
     */
    public Object clone() {
        
        // Use the copy constructor for cloning
        return new ENVHullMatrixSpeedUp(this);
    }
}
