/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.util.io;

import java.io.File;
import java.io.FileInputStream;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import fhm.edu.fem.calc.objects.Border;
import fhm.edu.fem.calc.objects.BorderList;
import fhm.edu.fem.calc.objects.CTab;
import fhm.edu.fem.calc.objects.CTabList;
import fhm.edu.fem.calc.objects.Calculation;
import fhm.edu.fem.calc.objects.Condition;
import fhm.edu.fem.calc.objects.Element;
import fhm.edu.fem.calc.objects.ElementList;
import fhm.edu.fem.calc.objects.Force;
import fhm.edu.fem.calc.objects.ForceList;
import fhm.edu.fem.calc.objects.Node;
import fhm.edu.fem.calc.objects.NodeList;
import fhm.edu.fem.constants.InputConstants;
import fhm.edu.fem.constants.SMatrixConstants;
import fhm.edu.fem.util.logging.LoggingManager;

/**
 * Class Description:
 *		This class takes care of loading the FEMIN file
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class FEMINLoader {
    
	/* Our local logger object */
	protected static Logger localLogger = LoggingManager.getLogger(FEMINLoader.class.getName());
	
	/* The approach to use */
	protected int femApproach = -1;
    
    /* The list holding all nodes */
    protected NodeList nodeList = new NodeList();
    
    /* The list holding all elements */
    protected ElementList elementList = new ElementList();
    
    /* The list holding all borders */
    protected BorderList borderList = new BorderList();
    
    /* The list holding all ctabs */
    protected CTabList ctabList = new CTabList();
    
    /* The list holding all calculations */
    protected List<Calculation> calculationList = new ArrayList<Calculation>();
    
    /* For Circle Membrane - The Center Coordinates */
    protected double membraneCircleCenterX = 0.0;
    protected double membraneCircleCenterY = 0.0;
    
    /* For Circle Membrane - The Radiuses */
    protected double membraneCircleRadius1 = 0.0;
    protected double membraneCircleRadius2 = 0.0;
    
    /* For Circle Membrane - The number of Sections */
    protected int membraneCircleSections = 0;
    
    /* For Circle Membrane - The Border Condition */
    protected double membraneCircleCondition = 0.0;
    
    /* For Spanner - The Force List */
    protected ForceList spannerForceList = new ForceList();
    
    /* For Spanner - The Material Constants */
    protected double spannerMaterialConstantE = 0.0;
    protected double spannerMaterialConstantV = 0.0;
    protected double spannerMaterialConstantH = 0.0;
    protected double spannerMaterialConstantHS = 0.0;
    
    /**
     * Construct a new FEMINLoader
     */
    public FEMINLoader() {}
    
    /**
     * Load all settings from the given file
     * @param inFile
     * @return true if all went well, false otherwise
     */
    public boolean load(File inFile) {
    	
    	try {
    		
            // Create the file input stream
            FileInputStream inputStream = new FileInputStream(inFile);
            
            // Create a new bytebuffer to hold the contents on the file based on the file size
            byte buffer[] = new byte[(int)inFile.length()];
            
            // Make sure all lists are empty
            this.nodeList.clear();
            this.elementList.clear();
            this.borderList.clear();
            this.calculationList.clear();
            this.ctabList.clear();
            
            // Read everything from the file and store it in the buffer
            int readBytes = inputStream.read(buffer);
            
            // Convert the byte buffer into one String, easier to work with
            String contents = new String(buffer, 0, readBytes);
            StringReader reader = new StringReader(contents);
            
            // The keyword string we are reading from the contents
            String keyword = "";
            
            // The index pointing to the next character to read
            int charIndex = 0;
            
            // Read from the contents till we reached the end
            while(charIndex < contents.length()) {
            	
            	// Read the next character and append it to the keyword string
            	keyword = keyword + (char)(reader.read());
            	
            	// Raise the char index by one
            	charIndex++;

            	// Take a look if the keyword already matches any of our known keywords
            	if(keyword.contains(InputConstants.KEYWORD_FEM)) {

            		// We found the keyword for the nodes, call the proper method to read all nodes
            		if(!this.loadFEMInfo(reader)) {
            			if(localLogger.isEnabledFor(Level.ERROR)) {
            				localLogger.log(Level.ERROR, "Loading file failed while reading the FEM info!");
            			}
            			
            			return false;
            		}
            		
            		// Clear the found keyword
            		keyword = "";
            		
            	} else if(keyword.contains(InputConstants.KEYWORD_NODES_START)) {

            		// We found the keyword for the nodes, call the proper method to read all nodes
            		if(!this.loadNodes(reader)) {
            			if(localLogger.isEnabledFor(Level.ERROR)) {
            				localLogger.log(Level.ERROR, "Loading file failed while reading nodes!");
            			}
            			
            			return false;
            		}
            		
            		// Clear the found keyword
            		keyword = "";
            		
            	} else if(keyword.contains(InputConstants.KEYWORD_ELEMENTS_START)) {
            		
            		// We found the keyword for the elements, call the proper method to read all elements
            		if(!this.loadElements(reader)) {
            			if(localLogger.isEnabledFor(Level.ERROR)) {
            				localLogger.log(Level.ERROR, "Loading file failed while reading elements!");
            			}
            			
            			return false;
            		}
            		
            		// Clear the found keyword
            		keyword = "";
            		
            	} else if(keyword.contains(InputConstants.KEYWORD_BORDERS_START)) {
            		
            		// We found the keyword for the borders, call the proper method to read all borders
            		if(!this.loadBorders(reader)) {
            			if(localLogger.isEnabledFor(Level.ERROR)) {
            				localLogger.log(Level.ERROR, "Loading file failed while reading borders!");
            			}
            			
            			return false;
            		}
            		
            		// Clear the found keyword
            		keyword = "";
            		
            	} else if(keyword.contains(InputConstants.KEYWORD_CTAB_START)) {
            		
            		// We found the keyword for the calculations, call the proper method to read all the ctabs
            		if(!this.loadCtabs(reader)) {
            			if(localLogger.isEnabledFor(Level.ERROR)) {
            				localLogger.log(Level.ERROR, "Loading file failed while reading a ctab!");
            			}
            			
            			return false;
            		}
            		
            		// Clear the found keyword
            		keyword = "";
            		
            	} else if(keyword.contains(InputConstants.KEYWORD_CALCULATION_START)) {
            		
            		// We found the keyword for the calculations, call the proper method to read all calculations
            		if(!this.loadCalculation(reader)) {
            			if(localLogger.isEnabledFor(Level.ERROR)) {
            				localLogger.log(Level.ERROR, "Loading file failed while reading a calculation!");
            			}
            			
            			return false;
            		}
            		
            		// Clear the found keyword
            		keyword = "";
            		
            	} else if(keyword.contains(InputConstants.KEYWORD_MEMBRANE_CIRCLE_CENTER)) {
            		
            		// We found the keyword for the circle membrane center
            		if(!this.loadMembraneCircleCenter(reader)) {
            			if(localLogger.isEnabledFor(Level.ERROR)) {
            				localLogger.log(Level.ERROR, "Loading file failed while reading Membrane Circle Center!");
            			}
            			
            			return false;
            		}
            		
            		// Clear the found keyword
            		keyword = "";
            		
            	} else if(keyword.contains(InputConstants.KEYWORD_MEMBRANE_CIRCLE_RADIUS1)) {
            		
            		// We found the keyword for the circle membrane radius1
            		if(!this.loadMembraneCircleRadius1(reader)) {
            			if(localLogger.isEnabledFor(Level.ERROR)) {
            				localLogger.log(Level.ERROR, "Loading file failed while reading Membrane Circle Radius 1!");
            			}
            			
            			return false;
            		}
            		
            		// Clear the found keyword
            		keyword = "";
            		
            	} else if(keyword.contains(InputConstants.KEYWORD_MEMBRANE_CIRCLE_RADIUS2)) {
            		
            		// We found the keyword for the circle membrane radius2
            		if(!this.loadMembraneCircleRadius2(reader)) {
            			if(localLogger.isEnabledFor(Level.ERROR)) {
            				localLogger.log(Level.ERROR, "Loading file failed while reading Membrane Circle Radius 2!");
            			}
            			
            			return false;
            		}
            		
            		// Clear the found keyword
            		keyword = "";
            		
            	} else if(keyword.contains(InputConstants.KEYWORD_MEMBRANE_CIRCLE_SECTIONS)) {
            		
            		// We found the keyword for the circle membrane sections
            		if(!this.loadMembraneCircleSections(reader)) {
            			if(localLogger.isEnabledFor(Level.ERROR)) {
            				localLogger.log(Level.ERROR, "Loading file failed while reading Membrane Circle Sections!");
            			}
            			
            			return false;
            		}
            		
            		// Clear the found keyword
            		keyword = "";
            		
            	} else if(keyword.contains(InputConstants.KEYWORD_MEMBRANE_CIRCLE_CONDITION)) {
            		
            		// We found the keyword for the circle membrane border condition
            		if(!this.loadMembraneCircleCondition(reader)) {
            			if(localLogger.isEnabledFor(Level.ERROR)) {
            				localLogger.log(Level.ERROR, "Loading file failed while reading Membrane Circle Border Condition!");
            			}
            			
            			return false;
            		}
            		
            		// Clear the found keyword
            		keyword = "";
            		
            	} else if(keyword.contains(InputConstants.KEYWORD_SPANNER_FORCES_START)) {
            		
            		// We found the keyword for the spanner forces
            		if(!this.loadSpannerForces(reader)) {
            			if(localLogger.isEnabledFor(Level.ERROR)) {
            				localLogger.log(Level.ERROR, "Loading file failed while reading Spanner Forces!");
            			}
            			
            			return false;
            		}
            		
            		// Clear the found keyword
            		keyword = "";
            		
            	} else if(keyword.contains(InputConstants.KEYWORD_SPANNER_MATERIAL_CONSTANT_H)) {
            		
            		// We found the keyword for the spanner material constant H
            		if(!this.loadSpannerMaterialConstantH(reader)) {
            			if(localLogger.isEnabledFor(Level.ERROR)) {
            				localLogger.log(Level.ERROR, "Loading file failed while reading Spanner Material Constant H!");
            			}
            			
            			return false;
            		}
            		
            		// Clear the found keyword
            		keyword = "";
            		
            	} else if(keyword.contains(InputConstants.KEYWORD_SPANNER_MATERIAL_CONSTANT_E)) {
            		
            		// We found the keyword for the spanner material constant e
            		if(!this.loadSpannerMaterialConstantE(reader)) {
            			if(localLogger.isEnabledFor(Level.ERROR)) {
            				localLogger.log(Level.ERROR, "Loading file failed while reading Spanner Material Constant E!");
            			}
            			
            			return false;
            		}
            		
            		// Clear the found keyword
            		keyword = "";
            		
            	} else if(keyword.contains(InputConstants.KEYWORD_SPANNER_MATERIAL_CONSTANT_V)) {
            		
            		// We found the keyword for the spanner material constant v
            		if(!this.loadSpannerMaterialConstantV(reader)) {
            			if(localLogger.isEnabledFor(Level.ERROR)) {
            				localLogger.log(Level.ERROR, "Loading file failed while reading Spanner Material Constant V!");
            			}
            			
            			return false;
            		}
            		
            		// Clear the found keyword
            		keyword = "";
            		
            	} else if(keyword.contains(InputConstants.KEYWORD_SPANNER_MATERIAL_CONSTANT_MAXSPANNUNG)) {
            		
            		// We found the keyword for the spanner material constant e
            		if(!this.loadSpannerMaterialConstantHS(reader)) {
            			if(localLogger.isEnabledFor(Level.ERROR)) {
            				localLogger.log(Level.ERROR, "Loading file failed while reading Spanner Material Constant HS!");
            			}
            			
            			return false;
            		}
            		
            		// Clear the found keyword
            		keyword = "";
            		
            	} else if(keyword.contains(InputConstants.KEYWORD_END)) {
            		
            		// We found the end of the file, let's stop the load-in
            		return true;
            	}
            }
            
            return true;
            
    	} catch (Exception e) {
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Loading file failed: " + inFile.getName(), e);
			}
			
			return false;
    	}
    }
    
    /**
     * Loads the info if the file is in format FEM 2 (linear approach) or FEM 3 (quadratic approach)
     * @param reader, the stringreader to read from
     * @return true if the method was successfull, false otherwise
     */
    protected boolean loadFEMInfo(StringReader reader) {
        
        // Read the integer telling us the approach
        int approach = this.readInteger(reader);
                
        this.femApproach = approach;
        
		if(localLogger.isEnabledFor(Level.INFO)) {
			localLogger.log(Level.INFO, "Loading - FEM Approach: " + this.femApproach + " ...");
		}
        
        return true;
    }
    
    /**
     * Loads all node information from the given string starting at the given index
     * @param reader, the stringreader to read from
     * @return true if the method was successfull, false otherwise
     */
    protected boolean loadNodes(StringReader reader) {
    	
        // Now we are at the position to read the number of nodes in the file
        int fullNumberOfNodes = this.readInteger(reader);

        // Check if we got the number of nodes
        if(fullNumberOfNodes == -1) {
        	
		    // Error, we did not find at least one number
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Syntax Error while loading the amount of nodes from the file!");
			}
			
			return false;
        }

		// Take a look if we are using APPROACH_LINEAR or APPROACH_QUADRATIC
		int realNumberOfNodes = -1;
		
		if(this.femApproach >= 3) {
		    
		    // Quadratic approach, read another number of nodes from the file representing the nodes which coordinates are stored
		    realNumberOfNodes = this.readInteger(reader);
		    
	        // Check if we got the real number of nodes
	        if(realNumberOfNodes == -1) {
	        	
			    // Error, we did not find at least one number
				if(localLogger.isEnabledFor(Level.ERROR)) {
					localLogger.log(Level.ERROR, "Syntax Error while loading the amount of REAL nodes from the file!");
				}
				
				return false;
	        }
	        
			// Show INFO for user
			if(localLogger.isEnabledFor(Level.INFO)) {
				localLogger.log(Level.INFO, "Loading " + fullNumberOfNodes + " nodes from the file. Non-pseudo-Nodes: " + realNumberOfNodes + " ...");
			}
		} else {
			
			// Show INFO for user
			if(localLogger.isEnabledFor(Level.INFO)) {
				localLogger.log(Level.INFO, "Loading " + fullNumberOfNodes + " nodes from the file...");
			}
		}
		
		// Boolean to remember if loading the nodes has been successfull
		boolean successfullyLoadedNodes = false;

		// Call the proper method to read the nodes from the file
		if(this.femApproach == SMatrixConstants.APPROACH_LINEAR) {
		    
		    // Linear approach
		    successfullyLoadedNodes = this.loadNodesLinear(reader, fullNumberOfNodes);
		    
		} else if(this.femApproach >= 3) {
		    
		    // Quadratic approach
		    successfullyLoadedNodes = this.loadNodesQuadratic(reader, fullNumberOfNodes, realNumberOfNodes);
		}
		
		// Show INFO for user
		if(!successfullyLoadedNodes) {
		    
			if(localLogger.isEnabledFor(Level.DEBUG)) {
				localLogger.log(Level.DEBUG, "Failure while loading nodes from the file !");
			}
		}
		
    	return successfullyLoadedNodes;
    }
    
    /**
     * Submethod loading nodes from the file in case of APPROACH_LINEAR
     * @param reader, the stringreader to read from
     * @param numberOfNodes, the number of nodes to read
     * @return true if all went well, false otherwise
     */
    protected boolean loadNodesLinear(StringReader reader, int numberOfNodes) {
    	
		// Start loading the number of nodes we discovered
		for(int nodeIndex = 1; nodeIndex <= numberOfNodes; nodeIndex++) {
		    
		    // Each node consists of a (x, y) value pair, read the x-value, which is a double
	        double xValue = this.readDouble(reader);

//	        // Check if we got a number as x-value
//	        if(xValue == -1) {
//	        	
//			    // Error, we did not find at least one number
//				if(localLogger.isEnabledFor(Level.ERROR)) {
//					localLogger.log(Level.ERROR, "Syntax Error while loading the x-value of node number: " + nodeIndex + " x-value:" + xValue);
//				}
//				
//				return false;
//	        }
	            		
		    // Each node consists of a (x, y) value pair, read the y-value, which is a double
	        double yValue = this.readDouble(reader);
	        
//	        // Check if we got a number as y-value
//	        if(yValue == -1) {
//	        	
//			    // Error, we did not find at least one number
//				if(localLogger.isEnabledFor(Level.ERROR)) {
//					localLogger.log(Level.ERROR, "Syntax Error while loading the y-value of node number: " + nodeIndex);
//				}
//				
//				return false;
//	        }
    		
    		// We now have both the x- and the y-value of the current node
    		Node node = new Node(nodeIndex, xValue, yValue, false);

    		// Store the new node in the node list
    		this.nodeList.setNode(node);			    
		}
		
		return true;
    }
    
    /**
     * Submethod loading nodes from the file in case of APPROACH_QUADRATIC
     * @param reader, the stringreader to read from
     * @param fullNumberOfNodes, the full number of nodes, including PSEUDO nodes
     * @param realNumberOfNodes, the real amount of nodes which coordinates are stored in the file
     * @return true if all went well, false otherwise
     */
    protected boolean loadNodesQuadratic(StringReader reader, int fullNumberOfNodes, int realNumberOfNodes) {

		// Start loading the number of nodes which coordinates are stored in the file
		for(int nodeIndex = 1; nodeIndex <= realNumberOfNodes; nodeIndex++) {
		    
		    // Read the index of the current node
		    int index = this.readInteger(reader);
		    
	        // Check if we got a number as index
	        if(index == -1) {
	        	
			    // Error, we did not find at least one number
				if(localLogger.isEnabledFor(Level.ERROR)) {
					localLogger.log(Level.ERROR, "Syntax Error while loading the index of node number: " + nodeIndex);
				}
				
				return false;
	        }
		    
		    // Each node consists of a (x, y) value pair, read the x-value, which is a double
	        double xValue = this.readDouble(reader);

//	        // Check if we got a number as x-value
//	        if(xValue == -1) {
//	        	
//			    // Error, we did not find at least one number
//				if(localLogger.isEnabledFor(Level.ERROR)) {
//					localLogger.log(Level.ERROR, "Syntax Error while loading the x-value of node number: " + nodeIndex);
//				}
//				
//				return false;
//	        }
	            		
		    // Each node consists of a (x, y) value pair, read the y-value, which is a double
	        double yValue = this.readDouble(reader);
	        
//	        // Check if we got a number as y-value
//	        if(yValue == -1) {
//	        	
//			    // Error, we did not find at least one number
//				if(localLogger.isEnabledFor(Level.ERROR)) {
//					localLogger.log(Level.ERROR, "Syntax Error while loading the y-value of node number: " + nodeIndex);
//				}
//				
//				return false;
//	        }
    		
    		// We now have both the x- and the y-value of the current node
    		Node node = new Node(nodeIndex, xValue, yValue, false);

    		// Store the new node in the node list
    		this.nodeList.setNode(node);			    
		}
		
		// We completely read all nodes which coordinates are stored in the file, now create the PSEUDO nodes
		for(int nodeIndex = (realNumberOfNodes+1); nodeIndex <= fullNumberOfNodes; nodeIndex++) {
		    
		    // Create the new PSEUDO node without x- and y-values
		    Node pseudoNode = new Node(nodeIndex, 0, 0, true);
		    
		    // Store the new node in the node list
		    this.nodeList.setNode(pseudoNode);
		}
		
		return true;
    }
    
    /**
     * Loads all element information from the given string starting at the given index
     * @param reader, the stringreader to read from
     * @return true if the method was successfull, false otherwise
     */
    protected boolean loadElements(StringReader reader) {
    	
    	try {
    	    // Now we are at the position to read the first number of elements in the file
            int numberOfElements = this.readInteger(reader);

            // Check if we got the number of elements
            if(numberOfElements == -1) {
            	
    		    // Error, we did not find at least one number
    			if(localLogger.isEnabledFor(Level.ERROR)) {
    				localLogger.log(Level.ERROR, "Syntax Error while loading the amount of elements from the file!");
    			}
    			
    			return false;
            }
            
            // Now read the second number of elements from the file
            int secondNumberOfElements = this.readInteger(reader);
            
            // Check if we got the second number of elements
            if(secondNumberOfElements == -1) {
            	
    		    // Error, we did not find at least one number
    			if(localLogger.isEnabledFor(Level.ERROR)) {
    				localLogger.log(Level.ERROR, "Syntax Error while loading the amount of elements from the file!");
    			}
    			
    			return false;
            }
            
            // Add both numbers together to get the final number of elements
            numberOfElements = numberOfElements + secondNumberOfElements;
    		
    		// Show INFO for user
    		if(localLogger.isEnabledFor(Level.INFO)) {
    			localLogger.log(Level.INFO, "Loading " + numberOfElements + " elements from the file" + " ...");
    		}
    		
			// The list where we temporarily store the nodes for each element
			NodeList elementNodes = new NodeList();
    		
    		// Start loading the number of nodes we discovered
    		for(int elementIndex = 1; elementIndex <= numberOfElements; elementIndex++) {

    			// To know when one element has been completed
    			boolean elementComplete = false;
    			
    			// Clear the temporary element node list
    			elementNodes.clear();
    			
    			// Read node indexes till the element is complete
    			while(!elementComplete) {
    				
        			// Each element might consist of three, four or six, eight nodes
        			int nodeIndex = this.readInteger(reader);

        	        // Check if we got a proper node index
        	        if(nodeIndex == -1) {
        	        	
        			    // Error, we did not find at least one number
        				if(localLogger.isEnabledFor(Level.ERROR)) {
        					localLogger.log(Level.ERROR, "Syntax Error while loading a node index for element number: " + elementIndex);
        				}
        				
        				return false;
        				
        	        } else {
        	        	
        	        	// We found a proper node index, add the node to the node list for this element
        	        	Node elementNode = this.nodeList.getNode(nodeIndex);
        	        	
        	        	// Check if we found a node with the given node index
        	        	if(elementNode != null) {
        	        		
        	        		// We found the node, add it to the element node list
        	        		elementNodes.setNode(elementNode);
        	        		
        	        	} else {
        	        		
        				    // Error, no node with the specified index found
        					if(localLogger.isEnabledFor(Level.ERROR)) {
        						localLogger.log(Level.ERROR, "Error while loading element number: " + elementIndex + " No node found with the specified index: " + nodeIndex);
        					}
        					
        					return false;
        	        	}
        	        	
        	        }
        	        
        	        // Check if the next char is a space or a tab
        	        char checkWhitespace = (char)(reader.read());
        	        
        	        while(checkWhitespace == ' ') {
        	        	
        	        	// Make a look-ahead to check if the next char is a tab \t or newline \n
        	        	reader.mark(0);
        	        	checkWhitespace = (char)(reader.read());
        	        }

        	        if(checkWhitespace == '\t' || checkWhitespace == '\n' || checkWhitespace == '\r') {

        	            int elementType = -1;

        	            // Determine the elementtype
        	            if(elementNodes.size() == 3 || elementNodes.size() == 6) {
        	                
        	                // TRIANGLE
        	                elementType = SMatrixConstants.FIGURE_TRIANGLE;
        	                
        	            } else if(elementNodes.size() == 4 || elementNodes.size() == 8) {
        	                
        	                // PARALLELOGRAM
        	                elementType = SMatrixConstants.FIGURE_QUADRAT;
        	                
        	            } else {

        	        		if(localLogger.isEnabledFor(Level.ERROR)) {
        	        			localLogger.log(Level.ERROR, "Error while determining element type while loading elements from file !");
        	        		}
        	        		
        	        		return false;
        	            }
        	        	
        	        	// It is a tab, so the element is complete
        	        	Element element = new Element(elementIndex, elementType, (NodeList)(elementNodes.clone()));

        	        	// Set the boolean
        	        	elementComplete = true;

            			// The element is complete, add it to our element list
            			this.elementList.setElement(element);
            			
        	        } else {

    				    // We found a char (number most likely, so go on)
        	            reader.reset();
        	        }
    			}
    		}
    		
    	} catch (Exception e) {
    		if(localLogger.isEnabledFor(Level.ERROR)) {
    			localLogger.log(Level.ERROR, "Error while loading elements from the file !", e);
    		}
    		
    		return false;
    	}
        
    	return true;
    }
    
    /**
     * Loads all border information from the given string starting at the given index
     * @param reader, the stringreader to read from
     * @return true if the method was successfull, false otherwise
     */
    protected boolean loadBorders(StringReader reader) {
    	
    	try {
            // Now we are at the position to read the number of borders in the file
            int numberOfBorders = this.readInteger(reader);

            // Check if we got the number of borders
            if(numberOfBorders == -1) {
            	
    		    // Error, we did not find at least one number
    			if(localLogger.isEnabledFor(Level.ERROR)) {
    				localLogger.log(Level.ERROR, "Syntax Error while loading the amount of borders from the file!");
    			}
    			
    			return false;
            }
    		
    		// Show INFO for user
    		if(localLogger.isEnabledFor(Level.INFO)) {
    			localLogger.log(Level.INFO, "Loading " + numberOfBorders + " borders from the file" + " ...");
    		}
    		
			// The list where we temporarily store the nodes for each element
			NodeList borderNodes = new NodeList();
    		
    		// Start loading the number of borders we discovered
    		for(int borderIndex = 1; borderIndex <= numberOfBorders; borderIndex++) {
    			
    			// Boolean indicating if a border has been completely read
    			boolean borderComplete = false;
    			
    			// Clear the temporary border node list
    			borderNodes.clear();
    			
    			// Find the next border keyword indicating that the information for a certain border is starting here
    			String keyword = "";
    			
    			// Run till the border is complete
    			while(!borderComplete) {
        			
                	// Read the next character and append it to the keyword string
                	keyword = keyword + (char)(reader.read());

                	// Take a look if the keyword already matches any of our known keywords
                	if(keyword.contains(InputConstants.KEYWORD_BORDER)) {

                		// We found the keyword for a certain border, read it's information, the first number to follow is the number of nodes in the border
                		int numberOfBorderNodes = this.readInteger(reader);

                        // Check if we got a proper number of border nodes
                        if(numberOfBorderNodes == -1) {
                        	
                		    // Error, we did not find at least one number
                			if(localLogger.isEnabledFor(Level.ERROR)) {
                				localLogger.log(Level.ERROR, "Syntax Error while loading the amount of nodes in border: " + borderIndex);
                			}
                			
                			return false;
                        }
                        
                        // We got a proper number of border nodes, so read these
                        for(int borderNodeIndex = 0; borderNodeIndex < numberOfBorderNodes; borderNodeIndex++) {

                    		// Read the border node index
                    		int borderNode = this.readInteger(reader);

                            // Check if we got a proper border node index
                            if(borderNode == -1) {
                            	
                    		    // Error, we did not find at least one number
                    			if(localLogger.isEnabledFor(Level.ERROR)) {
                    				localLogger.log(Level.ERROR, "Syntax Error while loading the border node index in border: " + borderIndex);
                    			}
                    			
                    			return false;
                    			
                            } else {
                            	
                	        	// We found a proper node index, add the node to the node list for this border
                	        	Node node = this.nodeList.getNode(borderNode);
 
                	        	// Check if we found a node with the given node index
                	        	if(node != null) {
                	        		
                	        		// We found the node, add it to the element node list
                	        		borderNodes.setNode(node);
                	        		
                	        	} else {
                	        		
                				    // Error, no node with the specified index found
                					if(localLogger.isEnabledFor(Level.ERROR)) {
                						localLogger.log(Level.ERROR, "Error while loading border number: " + borderIndex + " No node found with the specified index: " + borderNodeIndex);
                					}
                					
                					return false;
                	        	}
                            }    
                        }
                        
                        // A border is complete, store it in the main border list
                        Border border = new Border(borderIndex, (NodeList)(borderNodes.clone()));
                        this.borderList.setBorder(border);
                		
                		// Clear the found keyword
                		keyword = "";
                		
                		// Set the boolean to indicate that the border is complete
                		borderComplete = true;
                	}	
    			}	    
    		}
    		
    		if(localLogger.isEnabledFor(Level.INFO)) {
    			localLogger.log(Level.INFO, "Loading borders complete");
    		}
        	
        	return true;    		
    		
    	} catch (Exception e) {
    		if(localLogger.isEnabledFor(Level.ERROR)) {
    			localLogger.log(Level.ERROR, "Error while loading borders from the file !", e);
    		}
    		
    		return false;
    	}
    }
    
    /**
     * Loads all ctab information from the given string starting at the given index
     * @param reader, the stringreader to read from
     * @return true if the method was successfull, false otherwise
     */
    protected boolean loadCtabs(StringReader reader) {
        
        try {
            
            // Now we are at the position to read the number of ctabs in the file
            int numberOfCtabs = this.readInteger(reader);

            // Check if we got the number of ctabs
            if(numberOfCtabs == -1) {
            	
    		    // Error, we did not find at least one number
    			if(localLogger.isEnabledFor(Level.ERROR)) {
    				localLogger.log(Level.ERROR, "Syntax Error while loading the amount of ctabs from the file!");
    			}
    			
    			return false;
            }
    		
    		// Show INFO for user
    		if(localLogger.isEnabledFor(Level.INFO)) {
    			localLogger.log(Level.INFO, "Loading " + numberOfCtabs + " ctabs from the file" + " ...");
    		}
    		
    		// Start loading the number of ctabs we discovered
    		for(int ctabIndex = 1; ctabIndex <= numberOfCtabs; ctabIndex++) {
    			
    		    // Read the left node
    		    int leftNodeIndex = this.readInteger(reader);
    		    int midNodeIndex = this.readInteger(reader);
    		    int rightNodeIndex = this.readInteger(reader);
    		    
    		    // Get the node objects and store them in the new ctab
    		    Node leftNode = this.nodeList.getNode(leftNodeIndex);
    		    Node midNode = this.nodeList.getNode(midNodeIndex);
    		    Node rightNode = this.nodeList.getNode(rightNodeIndex);
    		    
    		    // Create the new ctab object and store it in the main ctab list
    		    CTab newCTab = new CTab(ctabIndex, leftNode, midNode, rightNode);
    		    this.ctabList.setCTab(newCTab);
    		}
        	
        	return true; 
    		
            
    	} catch (Exception e) {
    		if(localLogger.isEnabledFor(Level.ERROR)) {
    			localLogger.log(Level.ERROR, "Error while loading ctabs from the file !", e);
    		}
    		
    		return false;
    	}
    }
    
    /**
     * Loads all calculation information from the given string starting at the given index
     * @param reader, the stringreader to read from
     * @return true if the method was successfull, false otherwise
     */
    protected boolean loadCalculation(StringReader reader) {
       	
    	try {
    		
    		// Show INFO for user
    		if(localLogger.isEnabledFor(Level.INFO)) {
    			localLogger.log(Level.INFO, "Loading calculation from the file ...");
    		}
    		
    		// The list where we store all conditions related to this calculation
    		ArrayList<Condition> conditionList = new ArrayList<Condition>();
    		
    		// Boolean indicating when a calculation has been completely read
    		boolean calculationComplete = false;
    		
    		// Run till the calculation is complete
    		while(!calculationComplete) {
    			
        		// Read the border index
        		int borderIndex = this.readInteger(reader);
        		
                // Check if we got a proper border index
                if(borderIndex == -1) {
                	
        		    // Error, we did not find at least one number
        			if(localLogger.isEnabledFor(Level.ERROR)) {
        				localLogger.log(Level.ERROR, "Syntax Error while loading the border index in a condition");
        			}
        			
        			return false;
        			
                } else {
                	
    	        	// We found a proper border index, add the border to the border list for this condition
    	        	Border conditionBorder = this.borderList.getBorder(borderIndex);

    	        	// Check if we found a border with the given border index
    	        	if(conditionBorder != null) {
    	        		
    	        		// We found the affected border, now read the value the condition needs
    	        		double conditionValue = this.readDouble(reader);
    	        		
    	                // Check if we got a proper value
    	                if(conditionValue == -1) {
    	                	
    	        		    // Error, we did not find at least one number
    	        			if(localLogger.isEnabledFor(Level.ERROR)) {
    	        				localLogger.log(Level.ERROR, "Syntax Error while loading the condition value!");
    	        			}
    	        			
    	        			return false;
    	        			
    	                } else {
    	                	
    	                	// We now have a border affected by the condition and the condition's value
    	                	Condition condition = new Condition(conditionBorder, conditionValue);

    	                	// Add the condition to the condition list for this calculation
    	                	conditionList.add(condition);
    	                }
    	        		
    	        	} else {
    	        		
    				    // Error, no node with the specified index found
    					if(localLogger.isEnabledFor(Level.ERROR)) {
    						localLogger.log(Level.ERROR, "Error while loading conditon. No border found with the specified index: " + borderIndex);
    					}
    					
    					return false;
    	        	}
                }
                
                // Check if there is another condition following which belongs to this calculation
                String keyword = "";
                
                // Boolean indicating when a keyword is complete
                boolean keywordComplete = false;
                
                // Run till the keyword is complete
                while(!keywordComplete) {
                	
                	// Read the next character and append it to the keyword string
                	keyword = keyword + (char)(reader.read());

                	// Take a look if the keyword already matches either the "CONDITION" or the "END OF CALCULATION" keyword
                	if(keyword.contains(InputConstants.KEYWORD_CALCULATION_START)) {

                		// We found another condition for this calculation, do nothing
                		
                		// Set the boolean to indicate that the keyword is complete
                		keywordComplete = true;
                		
                	} else if(keyword.contains(InputConstants.KEYWORD_CALCULATION_END)) {
                		
                		// We found the end of the calculation, set the boolean to indicate that the calculation is complete
                		calculationComplete = true;
                		
                		// Set the boolean to indicate that the keyword is complete
                		keywordComplete = true;
                	}
                }
    		}   
    		
    		// The calculation is complete, store it
    		int calculationId = this.calculationList.size() + 1;
    		this.calculationList.add(new Calculation(calculationId, (List<Condition>)(conditionList.clone())));
    		
    		// Show INFO for user
    		if(localLogger.isEnabledFor(Level.INFO)) {
    			localLogger.log(Level.INFO, "Done loading: " + conditionList.size() + " conditions for calculation: " + calculationId);
    		}
        	
        	return true;    		
    		
    	} catch (Exception e) {
    		if(localLogger.isEnabledFor(Level.ERROR)) {
    			localLogger.log(Level.ERROR, "Error while loading conditions for a calculation from the file !", e);
    		}
    		
    		return false;
    	}
    }
    
    /**
     * Loads the center point of a mebrane circle
     * @param reader, the stringreader to read from
     * @return true if the method was successfull, false otherwise
     */
    protected boolean loadMembraneCircleCenter(StringReader reader) {
        
        try {
            
        	// Now we are at the position to read the center point of the membrane
        	double centerX = this.readDouble(reader);
        	double centerY = this.readDouble(reader);
    		
    		// Show INFO for user
    		if(localLogger.isEnabledFor(Level.INFO)) {
    			localLogger.log(Level.INFO, "Loaded Membrane Circle Center from the file: (x/y) " + centerX + " / " + centerY + " ...");
    		}
    		
    		this.membraneCircleCenterX = centerX;
    		this.membraneCircleCenterY = centerY;
        	
        	return true; 
    		
            
    	} catch (Exception e) {
    		if(localLogger.isEnabledFor(Level.ERROR)) {
    			localLogger.log(Level.ERROR, "Error while loading membrane circle center from the file !", e);
    		}
    		
    		return false;
    	}
    }
    
    /**
     * Loads the radius1 of a mebrane circle
     * @param reader, the stringreader to read from
     * @return true if the method was successfull, false otherwise
     */
    protected boolean loadMembraneCircleRadius1(StringReader reader) {
        
        try {
            
        	// Now we are at the position to read the first radius
        	double radius1 = this.readDouble(reader);
    		
    		// Show INFO for user
    		if(localLogger.isEnabledFor(Level.INFO)) {
    			localLogger.log(Level.INFO, "Loaded Membrane Circle Radius1 from the file: " + radius1 + " ...");
    		}
    		
    		this.membraneCircleRadius1 = radius1;
        	
        	return true; 
            
    	} catch (Exception e) {
    		if(localLogger.isEnabledFor(Level.ERROR)) {
    			localLogger.log(Level.ERROR, "Error while loading membrane circle radius1 from the file !", e);
    		}
    		
    		return false;
    	}
    }
    
    /**
     * Loads the radius2 of a mebrane circle
     * @param reader, the stringreader to read from
     * @return true if the method was successfull, false otherwise
     */
    protected boolean loadMembraneCircleRadius2(StringReader reader) {
        
        try {
            
        	// Now we are at the position to read the second radius
        	double radius2 = this.readDouble(reader);
    		
    		// Show INFO for user
    		if(localLogger.isEnabledFor(Level.INFO)) {
    			localLogger.log(Level.INFO, "Loaded Membrane Circle Radius2 from the file: " + radius2 + " ...");
    		}
    		
    		this.membraneCircleRadius2 = radius2;
        	
        	return true; 
            
    	} catch (Exception e) {
    		if(localLogger.isEnabledFor(Level.ERROR)) {
    			localLogger.log(Level.ERROR, "Error while loading membrane circle radius2 from the file !", e);
    		}
    		
    		return false;
    	}
    }
    
    /**
     * Loads the number of sections for the membrane circle
     * @param reader, the stringreader to read from
     * @return true if the method was successfull, false otherwise
     */
    protected boolean loadMembraneCircleSections(StringReader reader) {
        
        try {
            
        	// Now we are at the position to read the number of sections
        	int sections = this.readInteger(reader);
    		
    		// Show INFO for user
    		if(localLogger.isEnabledFor(Level.INFO)) {
    			localLogger.log(Level.INFO, "Loaded Membrane Circle Sections from the file: " + sections + " ...");
    		}
    		
    		this.membraneCircleSections = sections;
        	
        	return true; 
            
    	} catch (Exception e) {
    		if(localLogger.isEnabledFor(Level.ERROR)) {
    			localLogger.log(Level.ERROR, "Error while loading membrane circle sections from the file !", e);
    		}
    		
    		return false;
    	}
    }
    
    /**
     * Loads the border condition of a mebrane circle
     * @param reader, the stringreader to read from
     * @return true if the method was successfull, false otherwise
     */
    protected boolean loadMembraneCircleCondition(StringReader reader) {
        
        try {
            
        	// Now we are at the position to read the border condition of the membrane
        	double borderCondition = this.readDouble(reader);
    		
    		// Show INFO for user
    		if(localLogger.isEnabledFor(Level.INFO)) {
    			localLogger.log(Level.INFO, "Loaded Membrane Circle Border Condition from the file: " + borderCondition + " ...");
    		}
    		
    		this.membraneCircleCondition = borderCondition;
        	
        	return true; 
    		
    	} catch (Exception e) {
    		if(localLogger.isEnabledFor(Level.ERROR)) {
    			localLogger.log(Level.ERROR, "Error while loading membrane circle condition from the file !", e);
    		}
    		
    		return false;
    	}
    }

    /**
     * Loads all spanner forces information from the given reader
     * @param reader, the stringreader to read from
     * @return true if the method was successfull, false otherwise
     */
    protected boolean loadSpannerForces(StringReader reader) {
    	
    	try {
            // Now we are at the position to read the number of forces from the file
            int numberOfForces = this.readInteger(reader);

            // Check if we got the number of forces
            if(numberOfForces == -1) {
            	
    		    // Error, we did not find at least one number
    			if(localLogger.isEnabledFor(Level.ERROR)) {
    				localLogger.log(Level.ERROR, "Syntax Error while loading the amount of forces from the file!");
    			}
    			
    			return false;
            }
    		
    		// Show INFO for user
    		if(localLogger.isEnabledFor(Level.INFO)) {
    			localLogger.log(Level.INFO, "Loading " + numberOfForces + " forces from the file" + " ...");
    		}
    		
    		// Start loading the amount of forces we discovered
    		for(int forceIndex = 1; forceIndex <= numberOfForces; forceIndex++) {
    			
    			// Boolean indicating if a force has been completely read
    			boolean forceComplete = false;
    			
    			// Find the next border keyword indicating that the information for a certain border is starting here
    			String keyword = "";
    			
    			// Run till the border is complete
    			while(!forceComplete) {
        			
                	// Read the next character and append it to the keyword string
                	keyword = keyword + (char)(reader.read());

                	// Take a look if the keyword already matches any of our known keywords
                	if(keyword.contains(InputConstants.KEYWORD_SPANNER_FORCE)) {

                		// We found the keyword for a certain force, read it's information, the first number to follow is the node id
                		int nodeId = this.readInteger(reader);

                        // Check if we got a proper node id
                        if(nodeId == -1) {
                        	
                		    // Error, we did not find at least one number
                			if(localLogger.isEnabledFor(Level.ERROR)) {
                				localLogger.log(Level.ERROR, "Syntax Error while loading the node ID of the force: " + nodeId);
                			}
                			
                			return false;
                        }
                        
                        // The next number is the x-value of the force
                        double xValue = this.readDouble(reader);
                        
                        // Check if we got a proper x-value
                        if(xValue == -1) {
                        	
                		    // Error, we did not find at least one number
                			if(localLogger.isEnabledFor(Level.ERROR)) {
                				localLogger.log(Level.ERROR, "Syntax Error while loading the x-value of the force: " + xValue);
                			}
                			
                			return false;
                        }
                        
                        // The next number is the y-value of the force
                        double yValue = this.readDouble(reader);
                        
                        // Check if we got a proper y-value
                        if(yValue == -1) {
                        	
                		    // Error, we did not find at least one number
                			if(localLogger.isEnabledFor(Level.ERROR)) {
                				localLogger.log(Level.ERROR, "Syntax Error while loading the y-value of the force: " + yValue);
                			}
                			
                			return false;
                        }

                        // The current force is complete, store it in the main force list
                        Node forceNode = this.nodeList.getNode(nodeId);
                        Force force = new Force(forceIndex, xValue, yValue, forceNode);
                        this.spannerForceList.setForce(force);
                		
                		// Clear the found keyword
                		keyword = "";
                		
                		// Set the boolean to indicate that the force is complete
                		forceComplete = true;
                	}	
    			}	    
    		}
    		
    		if(localLogger.isEnabledFor(Level.INFO)) {
    			localLogger.log(Level.INFO, "Loading Forces complete");
    		}
        	
        	return true;    		
    		
    	} catch (Exception e) {
    		if(localLogger.isEnabledFor(Level.ERROR)) {
    			localLogger.log(Level.ERROR, "Error while loading forces from the file !", e);
    		}
    		
    		return false;
    	}
    }
    
    /**
     * Loads the spanner material constant E from the file
     * @param reader, the stringreader to read from
     * @return true if the method was successfull, false otherwise
     */
    protected boolean loadSpannerMaterialConstantE(StringReader reader) {
        
        try {
            
        	// Now we are at the position to read the spanner material constant E
        	double materialConstantE = this.readDouble(reader);
    		
    		// Show INFO for user
    		if(localLogger.isEnabledFor(Level.INFO)) {
    			localLogger.log(Level.INFO, "Loaded Spanner Material Constant E from the file: " + materialConstantE + " ...");
    		}
    		
    		this.spannerMaterialConstantE = materialConstantE;
        	
        	return true; 
    		
    	} catch (Exception e) {
    		if(localLogger.isEnabledFor(Level.ERROR)) {
    			localLogger.log(Level.ERROR, "Error while loading the Spanner Material Constant E from the file !", e);
    		}
    		
    		return false;
    	}
    }
    
    /**
     * Loads the spanner material constant V from the file
     * @param reader, the stringreader to read from
     * @return true if the method was successfull, false otherwise
     */
    protected boolean loadSpannerMaterialConstantV(StringReader reader) {
        
        try {
            
        	// Now we are at the position to read the spanner material constant V
        	double materialConstantV = this.readDouble(reader);
    		
    		// Show INFO for user
    		if(localLogger.isEnabledFor(Level.INFO)) {
    			localLogger.log(Level.INFO, "Loaded Spanner Material Constant V from the file: " + materialConstantV + " ...");
    		}
    		
    		this.spannerMaterialConstantV = materialConstantV;
        	
        	return true; 
    		
    	} catch (Exception e) {
    		if(localLogger.isEnabledFor(Level.ERROR)) {
    			localLogger.log(Level.ERROR, "Error while loading the Spanner Material Constant V from the file !", e);
    		}
    		
    		return false;
    	}
    }
    
    /**
     * Loads the spanner material constant "Hoechstspannung" from the file
     * @param reader, the stringreader to read from
     * @return true if the method was successfull, false otherwise
     */
    protected boolean loadSpannerMaterialConstantHS(StringReader reader) {
        
        try {
            
        	// Now we are at the position to read the spanner material constant HS
        	double materialConstantHS = this.readDouble(reader);
    		
    		// Show INFO for user
    		if(localLogger.isEnabledFor(Level.INFO)) {
    			localLogger.log(Level.INFO, "Loaded Spanner Material Constant HS from the file: " + materialConstantHS + " ...");
    		}
    		
    		this.spannerMaterialConstantHS = materialConstantHS;
        	
        	return true; 
    		
    	} catch (Exception e) {
    		if(localLogger.isEnabledFor(Level.ERROR)) {
    			localLogger.log(Level.ERROR, "Error while loading the Spanner Material Constant HS from the file !", e);
    		}
    		
    		return false;
    	}
    }
    
    /**
     * Loads the spanner material constant h from the file
     * @param reader, the stringreader to read from
     * @return true if the method was successfull, false otherwise
     */
    protected boolean loadSpannerMaterialConstantH(StringReader reader) {
        
        try {
            
        	// Now we are at the position to read the spanner material constant H
        	double materialConstantH = this.readDouble(reader);
    		
    		// Show INFO for user
    		if(localLogger.isEnabledFor(Level.INFO)) {
    			localLogger.log(Level.INFO, "Loaded Spanner Material Constant H from the file: " + materialConstantH + " ...");
    		}
    		
    		this.spannerMaterialConstantH = materialConstantH;
        	
        	return true; 
    		
    	} catch (Exception e) {
    		if(localLogger.isEnabledFor(Level.ERROR)) {
    			localLogger.log(Level.ERROR, "Error while loading the Spanner Material Constant H from the file !", e);
    		}
    		
    		return false;
    	}
    }
    
    /**
     * Read the next number from that string, no matter if it's an integer or double
     * @param value, the number object where our value is stored
     * @return the read integer
     */
    protected int readInteger(StringReader reader) {
    	try {
    	    // Read a value from the string
    	    String valueString = "";

    	    // Create the booleans we work with
    	    boolean valueStarted = false;
    	    boolean valueComplete = false;
    	    boolean isError = false;
    	    
    	    // Read till we found the next whitespace or the character we read is not a number anymore
    	    while(!valueComplete && !isError) {
    	    	
    	    	// Mark the stream, so we can return to this position in case it is a whitespace
    	    	reader.mark(0);
    	    	
    	    	// Read the next char from the reader
    	    	char nextChar = (char)(reader.read());

    	    	// Check if the char is a digit, whitespace or letter
    	    	if(Character.isDigit(nextChar)) {
    	    		
    	    		// The char is a digit, add it to our value string
    	    		valueString = valueString + nextChar;
    	    		
    	    		// The value has started
    	    		valueStarted = true;
    	    		
    	    	} else if(Character.isWhitespace(nextChar)) {
    	    		
    	    		// The char is a whitespace, check if the value already started
    	    		if(valueStarted) {
    	    			valueComplete = true;
    	    			
    	    			// The value is complete, reset the string by one char
    	    			reader.reset();
    	    		}
    	    		
    	    	} else {
    	    		
    	    		// The char is neither a digit, nor a whitespace, ERROR!
    	    		isError = true;
    	    	}
    	    }

            // Let's check how we stopped the while
    	    if(isError) {
    	    	
    	    	// Error took place, return -1
    	    	return -1;
    	    	
    	    } else if(valueComplete) {
    	    	
    	    	// Complete value read, return it
    	    	return Integer.parseInt(valueString);
    	    } else {
    	    	
    	    	return -1;
    	    }
    	    
    	} catch (Exception e) {
    		if(localLogger.isEnabledFor(Level.ERROR)) {
    			localLogger.log(Level.ERROR, "Error while reading Integer from StringReader !", e);
    		}
    	}
    	
    	return -1;
    }
    
    /**
     * Read the next number from that string, no matter if it's an integer or double
     * @param value, the number object where our value is stored
     * @return the read double
     */
    protected double readDouble(StringReader reader) {
    	
    	try {
    	    // Read a value from the string
    	    String valueString = "";

    	    // Create the booleans we work with
    	    boolean valueStarted = false;
    	    boolean valueComplete = false;
    	    boolean isError = false;
    	    
    	    // Read till we found the next whitespace or the character we read is not a number anymore
    	    while(!valueComplete && !isError) {
    	    	
    	    	// Mark the stream, so we can return to this position in case it is a whitespace
    	    	reader.mark(0);
    	    	
    	    	// Read the next char from the reader
    	    	char nextChar = (char)(reader.read());

    	    	// Check if the char is a digit, whitespace, dot or something else
    	    	if(Character.isDigit(nextChar) || nextChar == ',' || nextChar == '.' || nextChar == '-') {
    	    		
    	    		// The char is a digit, add it to our value string
    	    		valueString = valueString + nextChar;
    	    		
    	    		// The value has started
    	    		valueStarted = true;
    	    		
    	    	} else if(Character.isWhitespace(nextChar)) {
    	    		
    	    		// The char is a whitespace, check if the value already started
    	    		if(valueStarted) {
    	    			valueComplete = true;
    	    			
    	    			// The value is complete, reset the string by one char
    	    			reader.reset();
    	    		}
    	    		
    	    	} else {
    	    		
    	    		// The char is neither a digit, nor a whitespace, ERROR!
    	    		isError = true;
    	    	}
    	    }

            // Let's check how we stopped the while
    	    if(isError) {
    	    	
    	    	// Error took place, return -1
    	    	return -1;
    	    	
    	    } else if(valueComplete) {
    	        
    	        // Check if we got a ',' (comma) in the valueString, replace it with a '.' (dot)
    	        valueString = valueString.replace(',', '.');
    	    	
    	    	// Complete value read, return it
    	    	return Double.parseDouble(valueString);
    	    } else {
    	    	
    	    	return -1;
    	    }
    	    
    	} catch (Exception e) {
    		if(localLogger.isEnabledFor(Level.ERROR)) {
    			localLogger.log(Level.ERROR, "Error while reading Integer from StringReader !", e);
    		}
    	}
    	
    	return -1;
    }
    
    /**
     * Retrieve all nodes
     * @return the list of all nodes
     */
    public NodeList getNodes() {
        return this.nodeList;
    }
    
    /**
     * Retrieve all elements
     * @return the list of all elements
     */
    public ElementList getElements() {
        return this.elementList;
    }
    
    /**
     * Retrieve all borders
     * @return the list of all borders
     */
    public BorderList getBorders() {
        return this.borderList;
    }
    
    /**
     * Retrieve all ctabs
     * @return the list of all ctabs
     */
    public CTabList getCTabs() {
        return this.ctabList;
    }
    
    /**
     * Retrieve all calculations
     * @return the list of all calculations
     */
    public List<Calculation> getCalculations() {
        return this.calculationList;
    }
    
    /**
     * Retrieve the FEM approach to use
     * @return the FEM approach
     */
    public int getFEMApproach() {
        return this.femApproach;
    }
    
	/**
	 * @return Returns the membraneCircleCenterX.
	 */
	public double getMembraneCircleCenterX() {
		return membraneCircleCenterX;
	}
	
	/**
	 * @return Returns the membraneCircleCenterY.
	 */
	public double getMembraneCircleCenterY() {
		return membraneCircleCenterY;
	}
	
	/**
	 * @return Returns the membraneCircleRadius1.
	 */
	public double getMembraneCircleRadius1() {
		return membraneCircleRadius1;
	}
	
	/**
	 * @return Returns the membraneCircleRadius2.
	 */
	public double getMembraneCircleRadius2() {
		return membraneCircleRadius2;
	}
	
	/**
	 * @return Returns the membraneCircleSections.
	 */
	public int getMembraneCircleSections() {
		return membraneCircleSections;
	}
	
	/**
	 * @return Returns the membraneCircleCondition.
	 */
	public double getMembraneCircleCondition() {
		return membraneCircleCondition;
	}
	
	/**
	 * @return Returns the spannerForceList.
	 */
	public ForceList getSpannerForceList() {
		return spannerForceList;
	}
	
	/**
	 * @return Returns the spannerMaterialConstant.
	 */
	public double getSpannerMaterialConstantE() {
		return spannerMaterialConstantE;
	}
	
	/**
	 * @return Returns the spannerMaterialConstant.
	 */
	public double getSpannerMaterialConstantV() {
		return spannerMaterialConstantV;
	}
	
	/**
	 * @return Returns the spannerMaterialConstant.
	 */
	public double getSpannerMaterialConstantHS() {
		return spannerMaterialConstantHS;
	}
	
	/**
	 * @return Returns the spannerMaterialConstant.
	 */
	public double getSpannerMaterialConstantH() {
		return spannerMaterialConstantH;
	}
}