/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * DV-Anwendungen in der Technik - Prof.Dr.Gleich - 8.Semester
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 8T
 */

/**
 * USABILITY:
 *  Copyright (c) 2005 
 *  Thomas Woellert (thomas@woellert.net)
 *  Stefan Weissbach (weissbst@gmx.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package fhm.edu.fem.util.io;

import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.io.StringWriter;
import java.util.List;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import fhm.edu.fem.calc.FEMManager;
import fhm.edu.fem.calc.objects.Border;
import fhm.edu.fem.calc.objects.BorderList;
import fhm.edu.fem.calc.objects.Calculation;
import fhm.edu.fem.calc.objects.Condition;
import fhm.edu.fem.calc.objects.Element;
import fhm.edu.fem.calc.objects.ElementList;
import fhm.edu.fem.calc.objects.Node;
import fhm.edu.fem.calc.objects.NodeList;
import fhm.edu.fem.constants.InputConstants;
import fhm.edu.fem.constants.SMatrixConstants;
import fhm.edu.fem.util.logging.LoggingManager;

/**
 * Class Description:
 *		Used to save new FEMIN files based on the data contained by the FEMManager
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 * @author Stefan Weissbach (weissbst@gmx.de)
 */
public class FEMINSaver {

    /* Our local logger object */
	protected static Logger localLogger = LoggingManager.getLogger(FEMINSaver.class.getName());
    
    /**
     * Create a new FEMINSaver
     */
    public FEMINSaver() {}
    
    /**
     * Save all data (which is stored in the FEMManager) to the given file in FEMIN style
     * @param outFile, the file to save to
     * @param femPlotterCompatible, true if the file should be saved FEM_PLOTTER compatible
     * @return true if all went well, false otherwise
     */
    public boolean save(File outFile, boolean femPlotterCompatible) {
        
        try {
            
            // Grab the FEMManager
            FEMManager femManager = FEMManager.getInstance();
            
            // Open the file output stream
            FileOutputStream outputStream = new FileOutputStream(outFile, false);
            
            // Grab all data from the FEMManager
            int femApproach = femManager.getFEMApproach();
            NodeList nodeList = femManager.getNodeList();
            ElementList elementList = femManager.getElementList();
            BorderList borderList = femManager.getBorderList();
            List<Calculation> calculationList = femManager.getCalculationList();
            
            // Create a new StringWriter
            StringWriter writer = new StringWriter();
            
            // Write the FEM Approach to the file
            if(!this.writeFEMApproach(writer, femPlotterCompatible, femApproach)) {
                
                if(localLogger.isEnabledFor(Level.ERROR)) {
    				localLogger.log(Level.ERROR, "Saving FEM Approach to FEMIN file failed! File: " + outFile.getName());
    			}
                
                return false;
            }
            
            // Write the Nodes to the file
            if(!this.writeNodes(writer, femPlotterCompatible, femApproach, nodeList)) {
                
                if(localLogger.isEnabledFor(Level.ERROR)) {
    				localLogger.log(Level.ERROR, "Saving Nodes to FEMIN file failed! File: " + outFile.getName());
    			}
                
                return false;               
            }
            
            // Write the Elements to the file
            if(!this.writeElements(writer, femPlotterCompatible, elementList)) {
                
                if(localLogger.isEnabledFor(Level.ERROR)) {
    				localLogger.log(Level.ERROR, "Saving Elements to FEMIN file failed! File: " + outFile.getName());
    			}
                
                return false;                
            }
            
            // Write the Borders to the file
            if(!this.writeBorders(writer, borderList)) {
                
                if(localLogger.isEnabledFor(Level.ERROR)) {
    				localLogger.log(Level.ERROR, "Saving Borders to FEMIN file failed! File: " + outFile.getName());
    			}
                
                return false;
            }
            
            // Write the Conditions and Calculations to the file
            if(!this.writeCalculations(writer, calculationList)) {
                
                if(localLogger.isEnabledFor(Level.ERROR)) {
    				localLogger.log(Level.ERROR, "Saving Calculations and Conditions to FEMIN file failed! File: " + outFile.getName());
    			}
                
                return false;
            }
            
            // Write the closing footer of the FEMIN file
            if(!this.writeCloseFooter(writer)) {
                
                if(localLogger.isEnabledFor(Level.ERROR)) {
                    localLogger.log(Level.ERROR, "Saving closing footer to FEMIN file failed! File: " + outFile.getName());
                }
                
                return false;
            }
            
            // Grab the buffer from the writer
            StringBuffer buffer = writer.getBuffer();
            String string = buffer.toString();
            
            // Create an OutputStreamWriter
            OutputStreamWriter outputStreamWriter = new OutputStreamWriter(outputStream);
            
            // Write the String to the output stream and close it
            outputStreamWriter.write(string);
            outputStreamWriter.close();
            
            return true;
            
        } catch (Exception e) {
            
			if(localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Saving FEMIN file failed: " + outFile.getName(), e);
			}
			
			return false;
        }
    }
    
    /**
     * Write the FEM Approach to the file
     * @param writer, the writer to use
     * @param femPlotterCompatible, true if FEM_PLOTTER compatible
     * @param femApproach, the approach to write out
     * @return true if all went well, false otherwise
     */
    protected boolean writeFEMApproach(StringWriter writer, boolean femPlotterCompatible, int femApproach) {
                
        // Approach is known, write it to the writer
    	if(!femPlotterCompatible) {
    		
    		writer.write(InputConstants.KEYWORD_FEM + " " + femApproach + "\n");
    	}
       
        return true;
    }
    
    /**
     * Write the Nodes to the file
     * @param writer, the writer to use
     * @param femPlotterCompatible, true if FEM_PLOTTER compatible
     * @param femApproach, the approach to determine how to write certain things to the file
     * @param nodeList, the node list to write to the file
     * @return true if all went well, false otherwise
     */
    protected boolean writeNodes(StringWriter writer, boolean femPlotterCompatible, int femApproach, NodeList nodeList) {
    
        // Determine how many nodes and pseudo nodes we have in that node list
        int nodeCounter = 0;
        int pseudoNodeCounter = 0;
        
        for(int i = 1; i <= nodeList.size(); i++) {
            
            if(nodeList.getNodeInCorrectOrder(i).isPseudoNode()) {
                
                // We found a pseudo node
                pseudoNodeCounter++;
                
            } else {
                
                // We found a normal node
                nodeCounter++;
            }
        }
        
        // Depending on the approach, either write only the normal node count or both counts
        if(femApproach == SMatrixConstants.APPROACH_LINEAR || femPlotterCompatible) {
            
            // APPROACH_LINEAR
            // Write the first line telling how many nodes
        	if(femPlotterCompatible) {
        		writer.write("KOORD" + " " + nodeCounter + "\n");
        	} else {
        		writer.write(InputConstants.KEYWORD_NODES_START_1 + " " + nodeCounter + "\n");
        	}
            
        } else {
            
            // APPROACH_QUADRATIC
            // Write the first line telling how many nodes and pseudo nodes
            writer.write(InputConstants.KEYWORD_NODES_START_1 + " " + (pseudoNodeCounter+nodeCounter) + " " + nodeCounter + "\n");
            
        }
        
        // Write all the other coordinate information to the writer
        for(int i = 1; i <= nodeList.size(); i++) {
            
            // Grab the actual node from the list
            Node currentNode = nodeList.getNodeInCorrectOrder(i);
            
            // Only save the coordinates of non-pseudo nodes
            //if(!currentNode.isPseudoNode()) {
            if(true) {    
                // We found a proper node, write it's info to the file based on the used approach
                if(femApproach == SMatrixConstants.APPROACH_LINEAR || femPlotterCompatible) {
                  
                    // Write both coordinates to the file
                    writer.write(currentNode.getX() + " " + currentNode.getY() + " ");
                    
                } else {
                  
                    // Write the node's ID number and both coordinates to the file
                    writer.write(currentNode.getId() + " " + currentNode.getX() + " " + currentNode.getY() + "     ");
                }
                
                // Make a line break every six nodes to make it better read-able
                if(i%6 == 0) {
                    writer.write("\n");
                }
            }
        }
        
        writer.write("\n");
        
        return true;
    }

    /**
     * Write the Elements to the file
     * @param writer, the writer to use
     * @param femPlotterCompatible, true if FEM_PLOTTER compatible
     * @param elementList, the element list to write out
     * @return true if all went well, false otherwise
     */
    protected boolean writeElements(StringWriter writer, boolean femPlotterCompatible, ElementList elementList) {
        
        // Determine how many triangles and how many quadratics we used
        int triCounter = 0;
        int quadCounter = 0;
        
        // Run through all elements
        for(int i = 1; i <= elementList.size(); i++) {
            
            // Grab the current element
            Element currentElement = elementList.getElement(i);
            
            // Do the check
            if(currentElement.getNodeList().size() == 3 || currentElement.getNodeList().size() == 6) {
                
                // Triangle, count it
                triCounter++;
                
            } else if(currentElement.getNodeList().size() == 4 || currentElement.getNodeList().size() == 8) {
                
                // Parallelogram, count it
                quadCounter++;
                
            } else {
            
                // Unknown element type
                if(localLogger.isEnabledFor(Level.ERROR)) {
    				localLogger.log(Level.ERROR, "Unknown Element Type: " + currentElement.getElementType() + " for ElementIndex: " + i);
    			}
            }
        }
        
        // Write the first line of the elements to the file, telling us how many TRIANGLES and PARALLELOGRAMS we have
        if(femPlotterCompatible) {
        	
        	writer.write(InputConstants.KEYWORD_ELEMENTS_START + " " + (triCounter + quadCounter) + "\n");
        	
        } else {
        	
        	writer.write(InputConstants.KEYWORD_ELEMENTS_START + " " + triCounter + " " + quadCounter + "\n");
        }
        
        // Run through all TRIANGLE elements and write out their node lists
        for(int i = 1; i <= elementList.size(); i++) {
            
            Element currentElement = elementList.getElement(i);
            
            if(currentElement.getNodeList().size() == 3 || currentElement.getNodeList().size() == 6) {
                
                NodeList nodeList = currentElement.getNodeList();
                
                for(int j = 1; j <= nodeList.size(); j++) {
                    
                    // Write the ID number of the node
                    writer.write(nodeList.getNodeInCorrectOrder(j).getId() + " ");
                }
                
                // Make a line break
                writer.write("\n");
            }
        }
        
        // Run through all PARALLELOGRAM elements and write out their node lists
        for(int i = 1; i <= elementList.size(); i++) {
            
            Element currentElement = elementList.getElement(i);
            
            if(currentElement.getNodeList().size() == 4 || currentElement.getNodeList().size() == 8) {
                
                NodeList nodeList = currentElement.getNodeList();
                
                for(int j = 1; j <= nodeList.size(); j++) {
                    
                    // Write the ID number of the node
                    writer.write(nodeList.getNodeInCorrectOrder(j).getId() + " ");
                }
                
                // Make a line break
                writer.write("\n");
            }
        }
        
        return true;
    }

    /**
     * Write the Borders to the file
     * @param writer, the writer to use
     * @param borderList, the border list to write out
     * @return true if all went well, false otherwise
     */
    protected boolean writeBorders(StringWriter writer, BorderList borderList) {
        
        // Write the Keyword indicating that the borders are starting here and the actual number of borders
        writer.write(InputConstants.KEYWORD_BORDERS_START + " " + borderList.size() + "\n");
        
        // Write the specific borders together with the nodes they consist of
        for(int i = 1; i <= borderList.size(); i++) {
            
            // Grab the current border and it's nodes
            Border border = borderList.getBorder(i);
            NodeList nodeList = border.getNodes();
            
            // Print out the number of nodes in the border and the node indexes themselves
            writer.write(InputConstants.KEYWORD_BORDER + " " + nodeList.size());
            
            for(int j = 1; j <= nodeList.size(); j++) {
                
                writer.write(" " + nodeList.getNodeInCorrectOrder(j).getId());
            }
            
            // Nodes complete, make a line break
            writer.write("\n");
        }
        
        return true;
    }    
    

    /**
     * Write the Calculations to the file
     * @param writer, the writer to use
     * @param calculationList, the calculation list to write out
     * @return true if all went well, false otherwise
     */
    protected boolean writeCalculations(StringWriter writer, List<Calculation> calculationList) {
        
        // Go through all calculations in the list
        for(int i = 0; i < calculationList.size(); i++) {
            
            Calculation calc = calculationList.get(i);
            List<Condition> conditionList = calc.getConditions();
            
            // Run through all conditions in the calculation
            for(int j = 0; j < conditionList.size(); j++) {
                
                Condition cond = conditionList.get(j);
                
                // Write the keyword of the condition, the border index it affects and the condition value
                writer.write(InputConstants.KEYWORD_CALCULATION_START + " " + cond.getBorder().getId() + " " + cond.getValue() + "\n");
            }
            
            // Write the calculation END keyword at the end of the calculation
            writer.write(InputConstants.KEYWORD_CALCULATION_END + "\n");
        }
        
        // Always write "RECHNEN" at the end of the file, even in case no condition is present
        if(calculationList.size() == 0) {
        
        	writer.write(InputConstants.KEYWORD_CALCULATION_END + "\n");
        }
        
        return true;
    }
    
    /**
     * Write the closing footer of the file
     * @param writer, the writer to use
     * @return true if all went well, false otherwise
     */
    protected boolean writeCloseFooter(StringWriter writer) {
        
        // Write the closing keyword to the file
        writer.write(InputConstants.KEYWORD_END);
        
        return true;
    }
}
